

#ifndef Asagent_includeD
#define Asagent_includeD Yes

#include "../s_tools/sfile.h"

typedef char Asagent_helplinE[81];

/*
*/

#define Asagent_use_search_distancE 256                   /* set to 0 or 256 */
                                                  /*  256 only if command
                                                      -search_distance is known
                                                      to similar */

#define Sagent_random_seed_leN  16  /* number of strong random bytes to be
                                       used for high quality random numbers */


struct AsagenT {

 /* Name of the program which is using this agent */
 char progname[SfileadrL];

 /* interface to supervising application */
 char *app;                                            /* should be (void *) */
 int (*App_interpreter) ();                              /* (*app, int flag) */
 int (*App_eval) ();               /* (*app, char *in, char **out, int flag) */
 int (*App_determine_current) ();          /* (*app, char **adrpt,int *idx,  */
                                           /*        int *count,   int flag) */
 int (*App_cursor_advance) ();                /* (*app, int *step, int flag) */
 int (*App_seems_known) (); /* (*app, char *adr, int *filec, char ***filev,
                                                                   int flag) */

 int service_mode;           /* indicates the type of service provided
                                 0= server mode, working for application
                                 1= client mode, finally connected to a server
                              */
 int data_mode;
 struct SconN *client_sconn;
 int server_crucial;
 int preclient_disable_internal;
 int loop_return_on_idle;

 int encapsulate_views;
 int input_fd;
 int override_event_fd;
 int request_mode;
 struct LstrinG *current_users;

 struct LstrinG *markstack;   /* keeps track of server's open marks */
 char *notified_mark_id;      /* stores id of outermost mark notification
                                 received by client */
 int notified_mark_fd;
 char *notified_reply_id;     /* id of outermost reply notification */

 /* state informations about the remote application */
 int remote_idx;
 int remote_count;
 char *remote_adr;

 int silent_end;             /* 1= do not print messages on end */

 int end_on_flag;            /* end program on several triggers
                                 bit0= end of list reached
                                 bit1= prompt encountered in client mode
                                 bit2= timeout
                                 bit3= idle
                                 bit4= mark
                              */
 int end_on_timeout;
 int end_on_idle;
 char *end_on_mark;

 int recursion_level;        /* counts recursions of Asagent_interpreter() */
 int recursion_max;          /* to prevent endless recursion */

 char start_dir[SfileadrL];
 int arguments_to_queue;
 int permission_check;

 struct SaccesS  *access;
 struct AsqueuE *queue;
 int no_inconn_interrupt;
 struct AsinconN *interrupting_inconn;
 struct AsevenT *event;
 int stdout_pipe[2];
 int stderr_pipe[2];
 int preserve_orphan_events;

 struct AsviewprotO *view_prototype;
 struct AsvieW *views;
 int view_counter;
 int ask_counter;

 int disable_external; /*
                                bit0= current_file
                                bit1= double_files
                        */
 int disable_internal; /*
                                bit0= current_file
                                bit1= double_files
                                bit2= stdout
                                bit3= stderr
                                bit16= stdin/termios from terminal
                              */
 struct AsvieW *internal_current_view;
 int repair_icv;
 int repair_icv_idx;         /* index of last repaired file */

 int look_for_doubles; /* 0=never , 1=if not already guarded , 2=always */
 int doubles_shown;

 char *maindir;
 char *trashdir;
                              /* Lifetime is only a single pass of _loop() : */
 int keep_views;              /* tell _loop() that the view was not changed */
 int no_keep_views_when_idle; /* tell _loop() to reset keep_views on idle */
 int view_is_updated;         /* tell _loop() that the view was refreshed */
 int doubles_view_is_updated; /* tell _loop() the doubles view was refreshed */
 struct AsrM *deferred_rm_jobs;
 int loop_wait_after_hide;  /* how long to wait with a new show after a hide */
 int loop_wait_after_cmd;           /* how long to wait between two commands */
 int loop_wait_queue_limit;

 struct AsviewmsG *viewmsg;

 int smem_stderr; /* debugging:  messages about development of memory usage */
};


#define Asagent_viewer_cmD "display"

/*

                      Application models of AsagenT

 Currently there are two application models which also may be combined.

 1) AsagenT controls the components of the application by calling
    the application interpreter or by calling input connection handlers
    which both are provided by the application.

                         External functions

    The interpreter has to be a function like Asagent_interpreter() 
    It gets an application handle rather than a pointer to an AsagenT
    but the flag parameters and possible return values should have
    the same meaning. It should call Asagent_interpreter() with any
    command that it cannot recognize by itself. In that case, the
    return value of Asagent_interpreter() should be forwarded unchanged.

    Application handle and interpreter are attached by a call like
      Asagent_attach_app(agent,application_handle,Application_interpreter_func,
                         NULL,NULL,NULL,NULL,0);
    The other parameters which are NULL in this example could be used to
    set further external functions which all may use the application handle :
      _eval ................ text substitution before command execution
      _determine_current ... obtain index and file address of current item
                                (an image to be displayed by views)
      _cursor_advance ...... set current item within item list
      _seems_known ......... obtain a list of possibly similar files
    For examples and parameter definitions see imvapp.c , Imvapp_*()

                         Handled connections

    Input connection handlers are installed by creating a SconN object
    and attaching it to the list of input connections. 
    Obtain a SconN object by :
      Sconn_tcpclient_new(&sconn,hostname,portnumber,NULL,0);
    or from an abritrary file descriptor fd (e.g. result of open() or socket()
    which later has to be closed by the application) :
      Sconn_new(&sconn,"my_app",NULL,0);
      Sconn_attach_fd(sconn,5,fd,"my_app in",0);
    Use it to create a new entry in the agent's input list
      Asagent_new_handled_inconn(agent,sconn,handle,Handler_function,0);

    The handler function is eventually called as
      Handler_function(handle,sconn,flag);
    flag==0 means that pending data or EOF were detected at sconn.
    flag==1 is a notification that the sconn has been detached from the
    list of active input connections. The SconN object is still available
    then and eventually has to be destroyed by the application.

    In most cases of external disconnection the handler function will be
    called with flag==0 and will detect EOF by Sconn_read() returning 0.
    To avoid being notified again and again, it should then return 0 to
    indicate EOF to its caller which will detach the connection and call
    again with flag==1. 

    If you provided a file descriptor of your own and have to close it by
    yourself, do not close it before you destroy the input list entry
    and your own reference to the SconN object. SconN counts its referers
    and only gets destroyed if all referers have released it.
    Use Sconn_get_referers() to lookup that count. See also sconn.h
    To remove the input list item (and decrement reference count) use:
       Asagent_destroy_inconn(agent,sconn,0);
    Input list items also get automatically removed if the agent is destroyed.

    The handler function should return 1 on success (even if no data were
    actually available), 0 on EOF and -1 in case of a fatal error.

    
 2) Asagent returns control to the application when data become
    available or EOF occurs at a special kind of input connection.

                       Interrupting connections

    An interrupting connection may be created from a SconN object by calling  
      Asagent_new_interrupting_inconn(agent,sconn,NULL,NULL,0);

    Interrupting connections may be temporarily disabled by :
      Asagent_set_no_inconn_interrupt(agent,1,0);
    and re-enabled by
      Asagent_set_no_inconn_interrupt(agent,0,0);

    After an interrupt of Asagent_loop() a pointer to the interrupting
    SconN may be obtained by
      Asagent_get_interrupting_sconn(agent,&sconn,&fd,0);

    If EOF occurs i(e.g. Sconn_read() returns 0) the application should call 
      Asagent_destroy_inconn(agent,sconn,0);
    to remove the connection from the list of active inputs. Otherwise
    the EOF would interrupt the event loop again and again.
    
    It is possible to combine interrupting and handled connections by
    providing a handle and a handler function :
      Asagent_new_interrupting_inconn(agent,sconn,handle,Handler_function,0);
    The handler function then is called whenever data are available
    and when the connection gets abandoned by the agent. Nevertheless
    the agent's event loop gets interrupted after the handler was called
    with flag==0 . In most applications the handler function will just return 1
    if flag==0 and only become active when the connection got detached.
    Be aware that Asagent_destroy_inconn() calls the handler with flag==1 .

    The rules for unrefering SconN apply (see also at Handled Connections):
      Release the input list item as well as the SconN which was used to
      create it.
      Do not close the fd you created the SconN from before all references
      to that Sconn have been released.
*/

#endif /* ! Asagent_includeD */

