
/*
 cc -g -c ../as/asview.c
*/

#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>


#include "../s_tools/smem.h"
#include "../s_tools/sconn.h"

#include "../as/asvmsg.h"

#include "../as/asview.h"



/* ----------------------------- AsviewsubscR ----------------------------- */


int Asviewsubscr_new(subscr,type,representation,next,flag)
struct AsviewsubscR **subscr;
char *type,*representation;
struct AsviewsubscR *next;
int flag;
{
 struct AsviewsubscR *s;
 
 if(strlen(type)>=Asview_id_strL || strlen(representation)>=Asview_id_strL)
   return(0);

 *subscr= s= TSOB_FELD(struct AsviewsubscR,1);
 if(s==NULL)
   return(-1);
 strcpy(s->type,type);
 
 Asviewsubscr_class_from_type(s->type,s->class,0);
 strcpy(s->representation,representation);
 s->next= next;
 return(1);
}


int Asviewsubscr_destroy(subscr,flag)
struct AsviewsubscR **subscr;
int flag;
/*
 bit0= recursively delete all objects linked by (*subscr)->next
*/
{
 if((*subscr)==NULL)
   return(0);
 if(flag&1)
   Asviewsubscr_destroy(&((*subscr)->next),flag);
 Smem_freE((char *) *subscr);
 (*subscr)= NULL;
 return(1);
}


int Asviewsubscr_class_from_type(type,class,flag)
char *type;
char *class;
int flag;
{
 int ret;

 ret= Asviewmsg_class_from_type(type,class,0);
 return(ret);
}


int Asviewsubscr_rep_from_type(type,representation,flag)
char *type;
char *representation;
int flag;
{
 strcpy(representation,"data");
 if(strcmp(type,"current_file")==0 || strcmp(type,"current_file_copy")==0)
   strcpy(representation,"address");
 else if(strcmp(type,"double_files")==0 || strcmp(type,"double_files_copy")==0)
   strcpy(representation,"address");
 else if(strcmp(type,"list_change")==0)
   strcpy(representation,"address");
 return(1);
}


int Asviewsubscr_get_rep(subscr,type,rep,flag)
struct AsviewsubscR *subscr;
char *type;
char *rep;
int flag;
{
 strcpy(rep,subscr->representation);
 if(strcmp(rep,"default")==0)
   Asviewsubscr_rep_from_type(type,rep,0);
 return(1);
}

/* ----------------------------- AsviewprotO ----------------------------- */


int Asviewproto_new(proto,cmd,flag)
struct AsviewprotO **proto;
char *cmd;
int flag;
{
 struct AsviewprotO *p;

 *proto= p= TSOB_FELD(struct AsviewprotO,1);
 if(p==NULL)
   return(-1);
 p->cmd= p->geometry_option= p->geometry_left= p->geometry_right= NULL;
 if(Sregex_string(&(p->cmd),cmd,0)<=0) 
   goto failed;
 if(Sregex_string(&(p->geometry_option),"-geometry",0)<=0) 
   goto failed;
 if(Sregex_string(&(p->geometry_left),"+0+0",0)<=0) 
   goto failed;
 if(Sregex_string(&(p->geometry_right),"-0+0",0)<=0) 
   goto failed;
 return(1);
failed:;
  Asviewproto_destroy(proto,0);
  return(-1);
}


int Asviewproto_destroy(proto,flag)
struct AsviewprotO **proto;
int flag;
{
 if(*proto==NULL)
   return(0);
 if((*proto)->cmd!=NULL)
   Smem_freE((char *) (*proto)->cmd);
 if((*proto)->geometry_option!=NULL)
   Smem_freE((char *) (*proto)->geometry_option);
 if((*proto)->geometry_left!=NULL)
   Smem_freE((char *) (*proto)->geometry_left);
 if((*proto)->geometry_right!=NULL)
   Smem_freE((char *) (*proto)->geometry_right);
 Smem_freE((char *) *proto);
 *proto= NULL;
 return(1);
}


/* ------------------------------- AsvieW ------------------------------- */


int Asview_new(view,agent,id,type,proto,pos,flag)
struct AsvieW **view;
struct AsagenT *agent;
char *id;
char *type;
struct AsviewprotO *proto;
int pos;
int flag;
{
 struct AsvieW *v;
 int i;

 *view= v= TSOB_FELD(struct AsvieW,1);
 if(v==NULL)
   return(-1);
 v->agent= agent;
 strcpy(v->id,id);
 strcpy(v->type,type);
 v->broken= 0;
 v->encapsulated= 0;
 v->keep_copy_time= 0;
 v->files= NULL;
 v->filecount= 0;
 v->index= -1;
 v->shown_purposes= NULL;
 v->subscriptions= NULL;

 v->proto= proto;
 v->pos= pos;
 v->pid= 0;
 v->pid_valid= 0;

 v->sconn= NULL;

 v->copy_target= NULL;
 v->copy_limit= 1;
 v->copy_counter= (time(0)%2000000)*1000;

 v->prev= NULL;
 v->next= NULL;

 if(Lstring_new(&(v->shown_purposes),"init",NULL,0)<=0)
   goto failed;

 return(1);
failed:;
 Asview_destroy(view,0);
 return(-1);
}


int Asview_files_destroy(view,flag)
struct AsvieW *view;
int flag;
{
 int i;

 if(view->files==NULL) 
   return(0);
 for(i=0;i<view->filecount;i++)
   if(view->files[i]!=NULL)
     Smem_freE((char *) view->files[i]);
 Smem_freE((char *) view->files);
 view->files= NULL;
 view->filecount= 0;
 return(1);
}


int Asview_destroy(view,flag)
struct AsvieW **view;
int flag;
/*
 bit0= connection broken. do not send notifications
*/
{
 int i;
 char line[Asview_id_strL+80];

 if(*view==NULL)
   return(0);
 if(!((flag&1) || (*view)->broken)) {
   Asview_hide(*view,"",1);
   sprintf(line,"end 1 :id:data:%ld:%s\n",strlen((*view)->id),(*view)->id);
   Asview_send((*view),line,0);
 }
 Asview_files_destroy(*view,0);
 Lstring_destroy_all(&((*view)->shown_purposes),0);

 Sconn_unrefer(&((*view)->sconn),0);
 Asviewsubscr_destroy(&((*view)->subscriptions),1);

 Sregex_string(&((*view)->copy_target),NULL,0);

 if((*view)->prev!=NULL)
   (*view)->prev->next= (*view)->next;
 if((*view)->next!=NULL)
   (*view)->next->prev= (*view)->prev;
 Smem_freE((char *) (*view));
 *view= NULL;
 return(1);
}


int Asview_set_links(view,prev,next,flag)
struct AsvieW *view;
struct AsvieW *prev,*next;
int flag;
{
 view->prev= prev;
 view->next= next;
 return(1);
}


int Asview_get_links(view,prev,next,flag)
struct AsvieW *view;
struct AsvieW **prev,**next;
int flag;
{
 *prev= view->prev;
 *next= view->next;
 return(1);
}


int Asview_get_subscr(view,type,result,flag)
struct AsvieW *view;
char *type;
struct AsviewsubscR **result;
int flag;
{
 struct AsviewsubscR *subscr= NULL,*all_subscr= NULL;

 *result= NULL;
 for(subscr= view->subscriptions; subscr!=NULL; subscr= subscr->next) {
   if(strcmp(subscr->type,type)==0)
     *result= subscr;
   if(strcmp(subscr->type,"all")==0)
     all_subscr= subscr;
 }
 if((*result)!=NULL)
   return(1);
 *result= all_subscr;
 if((*result)!=NULL)
   return(1);
 return(0);
}


int Asview_is_subscribed(view,type,flag)
struct AsvieW *view;
char *type;
int flag;
{
 struct AsviewsubscR *subscr;
 int ret;

 ret= Asview_get_subscr(view,type,&subscr,0);
 return(ret);
}


int Asview_add_subscr(view,type,representation,flag)
struct AsvieW *view;
char *type,*representation;
int flag;
{
 int ret;
 struct AsviewsubscR *subscr= NULL;

 ret= Asview_get_subscr(view,type,&subscr,0);
 if(ret<0)
   goto ex;
 if(ret>0) {
   if(strcmp(subscr->representation,representation)!=0 
             && strlen(representation)<Asview_id_strL)
     strcpy(subscr->representation,representation);
   {ret= 2; goto ex;}                                  /* already subscribed */
 }
 ret= Asviewsubscr_new(&subscr,type,representation,view->subscriptions,0);
 if(ret<=0)
   goto ex;
 view->subscriptions= subscr;
 ret= 1;
ex:;
 return(ret);
}


int Asview_set_fd(view,fd,flag)
struct AsvieW *view;
int fd;
int flag;
{
 int ret;

 if(fd<0){
   Sconn_detach_cfd(view->sconn,1);
 } else {
   ret= Sconn_attach_fd(view->sconn,5,"",fd,1);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Asview_set_copy_par(view,copy_target,copy_limit,keep_time,flag)
struct AsvieW *view;
char *copy_target;
int copy_limit;
int flag;
{
 int ret;

 if(Sregex_string(&(view->copy_target),copy_target,0)<=0)
   return(-1);
 view->copy_limit= copy_limit;
 view->keep_copy_time= keep_time;
 return(1);
}


int Asview_attach_sconn(view,sconn,type,filename,fd,flag)
struct AsvieW *view;
struct SconN *sconn;
int type;
char *filename;
int fd;
int flag;
{
 int ret;
 struct SaccesS *access;

 if(sconn==NULL) {
   Asagent_get_access(view->agent,&access,0);
   ret= Sconn_new(&sconn,filename,access,0);
   if(ret<=0)
     goto ex;
   ret= Sconn_attach_fd(sconn,type,filename,fd,1);
   if(ret<=0)
     goto ex;
   view->sconn= sconn;
 } else
   Sconn_refer(sconn,&(view->sconn),0);

 ret= 1;
ex:
 return(ret);
}


int Asview_get_fd(view,fd,flag)
struct AsvieW *view;
int *fd;
int flag;
/*
 bit0= get input fd of connection
 bit1= get fd even if not valid
*/
{
 *fd= Sconn_get_fd(view->sconn,(flag&2) | !(flag&1));
 return(1);
}


int Asview_set_proto(view,proto,flag)
struct AsvieW *view;
struct AsviewprotO *proto;
int flag;
{
 if(view->proto!=NULL)
   view->proto= proto;
 return(1);
}


int Asview_sendbin(view,text,count,flag)
struct AsvieW *view;
char *text;
int count;
int flag;
/*
 bit0= send no data, only try to flush output buffer
 bit1= return 3 if buffer is not empty
*/
{
 int ret;
 struct AsvieW *current;

 if(view->broken)
   return(0);
 if(view->sconn==NULL)
   return(2);
 if(flag&1)
   count= 0;
 ret= Sconn_write(view->sconn,text,count,(flag&1)<<1);
 if(ret==-1 || ret==0)
   goto failed;
 if(flag&2)
   if(Sconn_wbuff_count(view->sconn,0)>0)
     return(3);
 return(1);
failed:;
 /*  mark as broken all views using this sconn */
 view->broken= 1;
 for(current= view; current->prev!=NULL; current= current->prev);
 for(;current!=NULL;current= current->next)
   if(view->sconn==current->sconn) 
     current->broken= 1;
 return(0);
}


int Asview_send(view,text,flag)
struct AsvieW *view;
char *text;
int flag;
{
 int ret;

 ret= Asview_sendbin(view,text,strlen(text),flag);
 return(ret);
}


int Asview_send_start(view,flag)
struct AsvieW *view;
int flag;
{
 int ret;
 char line[2*Asview_id_strL];

 sprintf(line,"init 1 :id:data:%ld:%s\n",strlen(view->id),view->id);
 ret= Asview_send(view,line,0);
 return(ret);
}


int Asview_send_filedata(view,name,file_bytes,flag)
struct AsvieW *view;
char *name;
off_t file_bytes;
int flag;
/*
 bit4= use field 'name' of current viewmsg of agent as source
*/
{
 FILE *fp;
 int ret,index,bytecount;
 off_t i;
 char buff[1024],*bytept= NULL;
 struct AsviewmsG *msg;

 if(flag&16)
   goto use_viewmsg;

 /* send file content , truncate or pad exactly to file_bytes */
 fp= fopen(name,"r");
 if(fp==NULL)
   return(0);
 for(i=0;i<file_bytes;i++){
   ret= fread(buff,1,sizeof(buff),fp);
   if(ret>1) {
     if(i+ret-1>=file_bytes)
       ret= file_bytes-i;
     i+= ret-1;
   } else if(ret<1){
     buff[0]= 0;
     ret= 1;
   }
   ret= Asview_sendbin(view,buff,ret,0);
   if(ret==-1)
     {ret= 0; goto ex;}
 } 
 ret= 1;
 goto ex;

use_viewmsg:;
 ret= Asagent_get_viewmsg(view->agent,&msg,0);
 if(ret<=0)
   goto ex;
 ret= Asviewmsg_fetch_field(msg,name,&bytept,&bytecount,0);
 if(ret<=0)
   goto ex;
 if(bytecount<file_bytes)
   return(-1);
 ret= Asview_sendbin(view,bytept,file_bytes,0);
 if(ret==-1)
   {ret= 0; goto ex;}

 ret= 1;
ex:;
 if(fp!=NULL)
   fclose(fp);
 return(ret);
}


int Asview_send_files(view,purpose,flag)
struct AsvieW *view;
char *purpose;
int flag;
/*
 bit4= use current viewmsg of agent as source of file content
*/
{
 struct AsviewsubscR *subscr;
 off_t file_bytes;
 int ret,i,fieldcount,send_index= 0;
 char line[2*SfileadrL],representation[Asview_id_strL],numbertext[80];

 ret= Asview_get_subscr(view,purpose,&subscr,0);
 if(ret<=0)
   goto ex; 
 Asviewsubscr_get_rep(subscr,purpose,representation,0);
 fieldcount= view->filecount;
 if(strcmp(representation,"data")==0) 
   fieldcount*= 2;
 if(view->filecount==1 && view->index>=0) {
   send_index= 1;
   fieldcount++;
 }
 sprintf(line,"%s %ld ",purpose,fieldcount);
 ret= Asview_send(view,line,0);
 if(ret<=0)
   goto ex; 
 if(send_index) {
   sprintf(numbertext,"%ld",view->index+1);
   sprintf(line,":idx:data:%ld:%s",strlen(numbertext),numbertext);
   ret= Asview_send(view,line,0);
   if(ret<=0)
     goto ex;
 }
 for(i=0;i<view->filecount;i++) {
   sprintf(line,":file%ld:address:%ld:%s",
                i+1,strlen(view->files[i]),view->files[i]);
   ret= Asview_send(view,line,0);
   if(ret<=0)
     goto ex;
   if(strcmp(representation,"data")==0) {
     file_bytes= Sfile_size(view->files[i],0);
     if(file_bytes<=0)
       {ret= 0; goto ex;}
     sprintf(line,":content%ld:%s:%ld:",
                  i+1,representation,file_bytes);
     ret= Asview_send(view,line,0);
     if(ret<=0)
       goto ex;
     if(flag&16) {
       sprintf(numbertext,"content%ld",i+1);
       ret= Asview_send_filedata(view,numbertext,file_bytes,16);
     } else
       ret= Asview_send_filedata(view,view->files[i],file_bytes,0);
     if(ret<=0)
       goto ex;
   }
 }
 ret= Asview_send(view,"\n",0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Asview_try_outbuffers(view,flag)
struct AsvieW *view;
int flag;
/*
 return:
  <=0 : error
  1 : all buffers empty
  2 : still data pending
*/
{
 int ret,buffer_not_empty= 0;
 struct AsvieW *current,*next;

 if(view==NULL)
   return(0);
 for(current= view; current->prev!=NULL; current= current->prev);
 for(;current!=NULL;current= next) {
   next= current->next;
   if(current->sconn==NULL)  
 continue;
   ret= Asview_sendbin(current,"",0,1|2);   /* send remnant data if possible */
   if(ret<0)
     goto ex;
   if(ret==3)
     buffer_not_empty= 1;
 }
 ret= 1+buffer_not_empty;
ex:;
 return(ret);
}


int Asview_broadcast_files(view,purpose,flag)
struct AsvieW *view;
char *purpose;
int flag;
/*
 bit0= hide rather than show
 bit1= do not serve parameter view itself
 bit6= do not restrict broadcasting to views which are enabled by agent
*/
{
 int ret;
 struct AsvieW *current,*next;

 for(current= view; current->prev!=NULL; current= current->prev);
 for(;current!=NULL;current= next) {
   next= current->next;
  if(!(flag&64)) {
     if(!Asagent_view_is_enabled(current->agent,current,0))
 continue;
   }
   if(strcmp(current->type,"copy")==0)             /* these serve themselves */
 continue;
   if(strcmp(current->type,"child")==0)          /* these are not interested */
 continue;
   if(!Asview_is_subscribed(current,purpose,0))
 continue;
   if(current==view) {
     if(!(flag&2))
       ret= Asview_send_files(current,purpose,0);/* below would be very bad */
   } else
     ret= Asview_notify_file(current,purpose,view->files,view->filecount,
                              view->index,(flag&(1|64))|8);
   if(ret<0)
     goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Asview_obtain_copyfiles(view,purpose,files,count,flag)
struct AsvieW *view;
char *purpose;
char ***files;
int *count;
int flag;
{
 int ret,i;
 struct AsvieW *current,*next;

 if((*files)!=NULL)
   Smem_freE((char *) *files);
 *count= 0;
 if(view==0)
   return(0);

 for(current= view; current->prev!=NULL; current= current->prev);
 for(;current!=NULL;current= next) {
   next= current->next;
   if(strcmp(current->type,"copy")!=0)
 continue;
   if(!Asview_is_subscribed(current,purpose,0))
 continue;
   if(current->filecount<=0)
 continue;
   *files= TSOB_FELD(char *,current->filecount);
   if((*files)==NULL)
     return(-1);
   for(i=0;i<current->filecount;i++)
     (*files)[i]= current->files[i];
   *count= current->filecount;
   return(1);
 }
 return(0);
}


int Asview_show(view,purpose,files,count,index,flag)
struct AsvieW *view;
char *purpose;
char **files;
int count;
int index;
int flag;
/*
 bit1= do not show 'child' type
 bit2= report files via ls -l if view is 'child' type
 bit3= show empty file lists too
 bit4= use current viewmsg of agent as source of file content
       (does not apply to 'child' type)
 bit5= do not show non 'child' views
 bit6= do not restrict broadcasting to views which are enabled by agent
 bit8= file addresses are preceeded by distance number text and blank, skip it 
*/
{
 int i,ret,errn;
 char line[8*SfileadrL],**arglist,*current_dir,*fnpt,*dir;
 char errmsg[SfileadrL],adr[SfileadrL],*expt,*spt,name[80];
 int actual_count,bytecount,wpt;
 char *actual_purpose= NULL,*bytept= NULL,*cpt;
 pid_t child_pid;
 struct AsviewmsG *msg= NULL;
 struct LstrinG *shown= NULL;
 struct stat stbuf;

 static int show_child_arglist= 0;

 if(strcmp(view->type,"child")==0 && (flag&2))
   return(2);
 if(strcmp(view->type,"child")!=0 && (flag&32))
   return(2);

 if(view->pid_valid || Lstring_is_listed(view->shown_purposes,purpose)) {
   ret= Asview_hide(view,purpose,flag&(2|64));
   if(ret<0)
     return(ret);
 }

 Asview_files_destroy(view,0);
 view->files= TSOB_FELD(char *,count);
 if(view->files==NULL){
failed:;
   return(-1);
 }
 for(i=0;i<count;i++)
   view->files[i]= NULL;
 for(i=0;i<count;i++) {
   cpt= NULL;
   if(flag&256)
     cpt= strchr(files[i],' ');
   if(cpt==NULL)
     cpt= files[i];
   else
     cpt++;
   if(Sregex_string(&(view->files[i]),cpt,0)<=0)
     goto failed;
 }
 view->filecount= count;
 if(strncmp(purpose,"current_file",12)==0 && count==1)
   view->index= index;
 else
   view->index= -1;

 if(strcmp(view->type,"pipe")==0)
   goto pipe_message;

 if(strcmp(view->type,"copy")==0)
   goto copy_files;

child_fork:;

 wait3(NULL,WNOHANG,NULL); /* burry any pending dead children */
 if(count<0)
   return(1);
 if(flag&4) {
   for(i=0;i<view->filecount;i++) 
     Asagent_print_fileinfo(view->agent,view->files[i],0);
 }

 child_pid= fork();
 if(child_pid==-1)
   return(0);
 if(child_pid==0) {
   /* this is the child */

   /* redirect stdout and stderr to the agent's input system */
   /* never call Asagent_stdXXX() after this */
   ret= Asagent_pipe_stdtxt(view->agent,0);
   if(ret<=0)
     return(ret);

   arglist= TSOB_FELD(char *,view->filecount+4);
   if(arglist==NULL)
     return(-1);
   arglist[0]= view->proto->cmd;
   arglist[1]= view->proto->geometry_option;
   if(view->pos==0)
     arglist[2]= view->proto->geometry_left;
   else
     arglist[2]= view->proto->geometry_right;
   wpt= 3;
   for(i=0;i<view->filecount;i++) {
     if(stat(view->files[i],&stbuf)==-1)
   continue;
     arglist[wpt++]= view->files[i];
   }
   arglist[wpt]= NULL;

   if(show_child_arglist)
     for(i=0;i<wpt;i++) 
       fprintf(stderr,"arglist[%ld]=\"%s\"\n",i,arglist[i]);

   if(wpt<=3) {
     /* ??? show a placeholder pic ??? */;
   }

   execvp(view->proto->cmd,arglist); /* the process should be replaced now */

   fprintf(stderr,"execvp() failed\n");
   exit(0);

 } else {
   /* this is the parent */
   view->pid= child_pid;
   view->pid_valid= 1; 
 }
 Asagent_check_pipes(view->agent,1);
 return(1);

pipe_message:;
 if(!Lstring_is_listed(view->shown_purposes,purpose))
   if(Lstring_new(&shown,purpose,view->shown_purposes,0)<=0)
     return(-1);
 ret= Asview_send_files(view,purpose,0);
 if(ret<=0)
   return(ret);
 return(1);

copy_files:;
 if(view->copy_target==NULL) {
   Asview_files_destroy(view,0);
   return(0);
 }
 if(!Lstring_is_listed(view->shown_purposes,purpose))
   if(Lstring_new(&shown,purpose,view->shown_purposes,0)<=0)
     return(-1);
 if(Sregex_string(&actual_purpose,purpose,0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&actual_purpose,"_copy",1)<=0)
   {ret= -1; goto ex;}
 for(i=0;i<view->filecount && i<view->copy_limit;i++) {
   /* transplant ending and fiddle in index */
   expt= strrchr(view->files[i],'.');
   spt= strrchr(view->files[i],'/');
   if(expt==NULL || (spt!=NULL && expt<spt))
     expt= view->files[i]+strlen(view->files[i]);
   view->copy_counter++;
   sprintf(adr,"%s_%lu%s",view->copy_target,view->copy_counter,expt);
   /* copy file content */
   if(flag&16) {
     ret= Asagent_get_viewmsg(view->agent,&msg,0);
     if(ret<=0)
       goto ex;
     sprintf(name,"content%ld",i+1);
     ret= Asviewmsg_fetch_field(msg,name,&bytept,&bytecount,0);
     if(ret<=0)
        goto ex;
     ret= Sfile_bin_to_file(adr,bytept,bytecount,1);
   } else {
     ret= Sfile_copy_file(view->files[i],adr,&errn,&errmsg,0);
   }
   if(ret<=0){
     Asagent_stderr(view->agent,"--- view of type 'copy' failed",1);
     Asagent_stderr(view->agent,errmsg,1);
 break;
   }
   /* replace list entry */
   if(Sregex_string(&(view->files[i]),adr,0)<=0)
     {ret= -1; goto ex;}
 }
 actual_count= i;
 /* release remaining filenames */
 for(i=actual_count;i<view->filecount;i++)
   Sregex_string(&(view->files[i]),NULL,0);
 view->filecount= actual_count;
 /* send notification about actual filenames */
 if(view->filecount>0 || (flag&8)){
   if(view->sconn==NULL) {
     ret= Asview_broadcast_files(view,actual_purpose,flag&64);
   }else{
     ret= Asview_send_files(view,actual_purpose,0);
   }
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 if(ret<=0)
   view->filecount= 0; /* to protect the original files from unlink */
 Sregex_string(&actual_purpose,NULL,0);
 return(ret);
}


int Asview_hide(view,purpose,flag)
struct AsvieW *view;
char *purpose;
int flag;
/*
 bit0= types "pipe","copy" : hide all subscribed purposes of class "imgfile"
 bit1= do not hide "child" type
 bit5= do not hide non "child" views
 bit6= do not restrict broadcasting to views which are enabled by agent
 bit7= hide type "pipe" even if the purpose is not recorded in shown_purposes
*/
{
 int ret,i;
 char *actual_purpose= NULL;
 struct AsviewsubscR *subscr;
 struct LstrinG *shown= NULL;

 if(strcmp(view->type,"child")==0 && (flag&2))
   return(2);
 if(strcmp(view->type,"pipe")==0)
   goto pipe_message;
 if(strcmp(view->type,"copy")==0)
   goto copy_files;

 if(!view->pid_valid)
   return(2);
 kill(view->pid,SIGTERM);
 view->pid_valid= 0;
 wait3(NULL,WNOHANG,NULL);
 Asview_files_destroy(view,0);
 return(1);

pipe_message:;
 if(flag&32)
   return(2);
 if(Lstring_find_text(view->shown_purposes,purpose,&shown,0)>0)
   Lstring_destroy(&shown,0);
 else if(!(flag&128))
   return(2);
 ret= 1;
 if(flag&1){
   for(subscr= view->subscriptions; subscr!=NULL; subscr= subscr->next)
     if(strcmp(subscr->class,"imgfile")==0) {
       Asview_files_destroy(view,0);
       ret= Asview_send_files(view,subscr->type,0);
     }
 }else{
   Asview_files_destroy(view,0);
   ret= Asview_send_files(view,purpose,0);
 }
 return(ret);

copy_files:;
 if(flag&32)
   return(2);
 if(flag&1){
   for(subscr= view->subscriptions; subscr!=NULL; subscr= subscr->next)
     if(strcmp(subscr->class,"imgfile")==0) {
       ret= Asview_hide(view,subscr->type,flag&~1);
       if(ret<0)
         goto ex;
     }
   return(1);
 }
 if(Lstring_find_text(view->shown_purposes,purpose,&shown,0)>0)
   Lstring_destroy(&shown,0);
 else
   {ret= 1; goto ex;}
 if(Sregex_string(&actual_purpose,purpose,0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&actual_purpose,"_copy",1)<=0)
   {ret= -1; goto ex;}

 /* remove files */;
 for(i=0;i<view->filecount;i++) {
   if(view->files[i]!=NULL)
     if(strlen(view->files[i])>0) {
       if(view->keep_copy_time>0) {
         ret= Asagent_add_rm_job(view->agent,
                                  Sfile_microtime(0)+view->keep_copy_time,
                                  view->files[i],0);
         if(ret<0)
           goto ex;
       } else {
         unlink(view->files[i]);
       }
     }
 }
 /* send notification about canceling of files */
 if(view->sconn==NULL) {
   ret= Asview_broadcast_files(view,actual_purpose,1|2|(flag&64));
   if(ret<=0)
     goto ex;
 }
 Asview_files_destroy(view,0);
 ret= Asview_send_files(view,actual_purpose,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 Sregex_string(&actual_purpose,NULL,0);
 return(ret);
}


int Asview_has_died(view,exitvalue,flag)
struct AsvieW *view;
int *exitvalue;
int flag;
/*
 return:
  0= no info available
  1= process died from exit (see *exitvalue)
  2= process died in another way
*/
{
 int status,ret;

 if(view==NULL)
   return(0);
 if(!view->pid_valid)
   return(0);
 ret= waitpid(view->pid,&status,WNOHANG);
 if(ret!=view->pid)
   return(0);
 if(WIFSIGNALED(status) || !WIFEXITED(status))
   return(2);
 *exitvalue= WEXITSTATUS(status);
 return(1);
}


int Asview_notify_text(view,purpose,text,flag)
struct AsvieW *view;
char *purpose;
char *text;
int flag;
/*
 bit0= append newline to text
*/
{
 char line[2*SfileadrL],representation[Asview_id_strL];
 struct AsviewsubscR *subscr;
 int ret;

 if(!Asview_is_subscribed(view,purpose,0))
   return(2);
 ret= Asview_get_subscr(view,purpose,&subscr,0);
 if(ret<=0)
   goto ex;
 Asviewsubscr_get_rep(subscr,purpose,representation,0);
 sprintf(line,"%s %ld :text:%s:%ld:",
              purpose,1,representation,strlen(text)+(flag&1));
 if(strlen(text)+10>=sizeof(line)-strlen(line))
   return(-2);
 strcat(line,text);
 if(flag&1)
   strcat(line,"\n");
 strcat(line,"\n");
 ret= Asview_send(view,line,0);
ex:;
 return(ret);
}


int Asview_notify_ask(view,type,id,text,constraint,flag)
struct AsvieW *view;
char *type;
char *id;
char *text;
char *constraint;
int flag;
{
 char line[2*SfileadrL],representation[Asview_id_strL];
 struct AsviewsubscR *subscr;
 int ret;

 if(!Asview_is_subscribed(view,"ask",0))
   return(2);
 ret= Asview_get_subscr(view,"ask",&subscr,0);
 if(ret<=0)
   goto ex;
 Asviewsubscr_get_rep(subscr,"ask",representation,0);
 sprintf(line,"ask 4 :id:%s:%ld:%s:type:%s:%ld:%s",
              representation,strlen(id),id,
              representation,strlen(type),type);
 sprintf(line+strlen(line),":text:%s:%ld:%s:constraint:%s:%ld:%s\n",
              representation,strlen(text),text,
              representation,strlen(constraint),constraint);
 ret= Asview_send(view,line,0);
ex:;
 return(ret);
}


int Asview_notify_list_change(view,idx,action,item,flag)
struct AsvieW *view;
int idx;
char *action;
char *item;
int flag;
{
 char line[2*SfileadrL],representation[Asview_id_strL],numstring[81];
 struct AsviewsubscR *subscr;
 int ret;

 if(!Asview_is_subscribed(view,"list_change",0))
   return(2);
 ret= Asview_get_subscr(view,"list_change",&subscr,0);
 if(ret<=0)
   goto ex;
 Asviewsubscr_get_rep(subscr,"list_change",representation,0);
 if(strcmp(representation,"data")==0)
   strcpy(representation,"address");
 sprintf(numstring,"%ld",idx+1);
 sprintf(line,"list_change 3 :idx:data:%ld:%s:action:data:%ld:%s",
              strlen(numstring),numstring,strlen(action),action);
 sprintf(line+strlen(line),":item:%s:%ld:%s\n",
              representation,strlen(item),item);
 ret= Asview_send(view,line,0);
ex:;
 return(ret);
}


int Asview_notify_file(view,purpose,files,count,index,flag)
struct AsvieW *view;
char *purpose;
char **files;
int count;
int index;
int flag;
/*
 bit0= hide, i.e : cancel previously sent file list
 bit1= do not show 'child' type
 bit2= report files via ls -l
 bit3= show empty file lists too
 bit4= use current viewmsg of agent as source of file content
 bit5= do not show non 'child' views
 bit6= do not restrict broadcasting to views which are enabled by agent
 bit7= hide type "pipe" even if the purpose is not recorded in shown_purposes
 bit8= file addresses are preceeded by distance number text and blank, skip it 
*/
{
 int ret;

 if(!Asview_is_subscribed(view,purpose,0))
   return(2);
 if(flag&1)
   ret= Asview_hide(view,purpose,flag&(2|32|64|128));
 else 
   ret= Asview_show(view,purpose,files,count,index,flag&(2|4|8|16|32|64|256));
 return(ret);
}


int Asview_notify_prompt(view,idx,count,flag)
struct AsvieW *view;
int idx,count;
int flag;
/*
 bit0= unprompt
*/
{
 int ret;
 char line[2*SfileadrL],idxtext[80],counttext[80];

 if(!Asview_is_subscribed(view,"prompt",0))
   return(2);
 if(flag&1) {
   sprintf(line,"prompt 0 \n");
 } else {
   sprintf(idxtext,"%ld",idx);
   sprintf(counttext,"%ld",count);
   sprintf(line,"prompt 2 :idx:data:%ld:%s:count:data:%ld:%s\n",
                strlen(idxtext),idxtext,strlen(counttext),counttext);
 }
 ret= Asview_send(view,line,0);
 return(ret); 
}


int Asview_notify_mark(view,mark_id,flag)
struct AsvieW *view;
char *mark_id;
int flag;
/*
 bit0= mark end
 bit8= reply rather than mark
*/
{
 int ret;
 char line[2*SfileadrL],*mode,*subscription;

 if(flag&256)
   subscription= "reply";
 else 
   subscription= "mark";
 if(flag&1) 
   mode= "end";
 else
   mode= "start";
 sprintf(line,"%s 2 :state:data:%ld:%s:id:data:%ld:%s\n",
               subscription,strlen(mode),mode,strlen(mark_id),mark_id);
 ret= Asview_send(view,line,0);
 return(ret); 
}


