
/*
 cc -g -c ../as/asview.c
*/

#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>


#include "../s_tools/smem.h"

#include "../as/asvmsg.h"



/* -------------------------- AsviewmsgfielD----------------------------- */


int Asviewmsgfield_new(field,name,representation,number_of_bytes,bytes,flag)
struct AsviewmsgfielD **field;
char *name;
char *representation;
int number_of_bytes;
char *bytes;
int flag;
{
 struct AsviewmsgfielD *f;
 int i;
 
 *field= f= TSOB_FELD(struct AsviewmsgfielD,1);
 if(f==NULL)
   return(-1);
 f->name= NULL;
 f->representation= NULL;
 f->number_of_bytes= 0;
 f->bytecount= 0;
 f->bytes= NULL;
 f->buffer= NULL;
 f->stage= 0;

 if(Sregex_string(&(f->name),name,0)<=0)
   goto failed;
 if(Sregex_string(&(f->representation),representation,0)<=0)
   goto failed;
 if(number_of_bytes>0) {
   f->bytes= TSOB_FELD(char,number_of_bytes);
   if(f->bytes==NULL)
     goto failed;
   for(i=0;i<number_of_bytes;i++)
     f->bytes[i]= bytes[i];
 }
 f->number_of_bytes= number_of_bytes;
 return(1);

failed:;
 Asviewmsgfield_destroy(field,0); 
 return(-1);
}


int Asviewmsgfield_destroy(field,flag)
struct AsviewmsgfielD **field;
int flag;
{
 int i;
 
 if((*field)==NULL)
   return(0);
 Sregex_string(&((*field)->name),NULL,0);
 Sregex_string(&((*field)->representation),NULL,0);
 if((*field)->bytes!=NULL)
   Smem_freE((char *) (*field)->bytes);
 Sregex_string(&((*field)->buffer),NULL,0);
 Smem_freE((char *) *field);
 *field= NULL;
 return(1);
}


int Asviewmsgfield_parse_new(field,text,textcount,rptpt,flag)
struct AsviewmsgfielD **field;
char *text;
int textcount;
char **rptpt;
int flag;
{
 int ret,number_of_bytes,i,tail;
 struct AsviewmsgfielD *f;
 char *rpt;

 if(Asviewmsgfield_is_done(*field,0))
   Asviewmsgfield_destroy(field,0);
 if(*field==NULL) {
   ret= Asviewmsgfield_new(field,"","",0,NULL,0);
   if(ret<=0)
     goto ex;
 }
 f= *field;

 /* :name:representation:number_of_bytes:bytes */
 rpt= text;
 if(f->stage==0) {
   for(i=0;i<textcount;i++)
     if(*(rpt++)==':')
   break;
   if(i<textcount)
     f->stage= 1;
 }
 if(f->stage==1) {
   ret= Sregex_cut_neck(rpt,':',&(f->name),&rpt,1);
   if(ret!=1)
     goto ex;
   f->stage= 2;
 }
 if(f->stage==2) {
   ret= Sregex_cut_neck(rpt,':',&(f->representation),&rpt,1);
   if(ret!=1)
     goto ex;
   f->stage= 3;
 }
 if(f->stage==3) {
   ret= Sregex_cut_neck(rpt,':',&(f->buffer),&rpt,1);
   if(ret!=1)
     goto ex;
   sscanf(f->buffer,"%ld",&(f->number_of_bytes));
   Sregex_string(&(f->buffer),NULL,0);
   if(f->bytes!=NULL)
     Smem_freE((char *) f->bytes);
   f->bytes= TSOB_FELD(char,f->number_of_bytes+1);
   if(f->bytes==NULL)
     {ret= -1; goto ex;}
   f->stage= 4;
 }
 if(f->stage==4) {
   tail= textcount-(rpt-text);
   for(i=0;i<tail&&f->bytecount<f->number_of_bytes;i++) 
     f->bytes[f->bytecount++]= *(rpt++);
   if(f->bytecount>=f->number_of_bytes) {
     f->bytes[f->number_of_bytes]= 0;
     f->stage= -1;
   }
 }

 if(f->stage>=0)
   ret= 2;
 else
   ret= 1;
ex:;
 *rptpt= rpt;
 return(ret);
}


int Asviewmsgfield_is_done(field,flag)
struct AsviewmsgfielD *field;
int flag;
{
 if(field==NULL)
   return(0);
 return(field->stage==-1);
}



/* ----------------------------- AsviewmsG ------------------------------ */



int Asviewmsg_new(msg,type,class,number_of_fields,flag)
struct AsviewmsG **msg;
char *type;
char *class;
int number_of_fields;
int flag;
{
 struct AsviewmsG *m;
 int i,ret;
 
 *msg= m= TSOB_FELD(struct AsviewmsG,1);
 if(m==NULL)
   return(-1);
 m->type= NULL;
 m->class= NULL;
 m->number_of_fields= 0;
 m->fields= NULL;
 m->fieldcount= 0;
 m->buffer= NULL;
 m->stage= 0;

 if(Sregex_string(&(m->type),type,0)<=0)
   goto failed;
 if(Sregex_string(&(m->class),class,0)<=0)
   goto failed;
 ret= Asviewmsg_new_fieldlist(m,number_of_fields,0);
 if(ret<=0)
   goto failed;
 return(1);

failed:;
 Asviewmsg_destroy(msg,0); 
 return(-1);
}


int Asviewmsg_new_fieldlist(msg,number_of_fields,flag)
struct AsviewmsG *msg;
int number_of_fields;
int flag;
{
 int i;

 Asviewmsg_destroy_fieldlist(msg,0);

 msg->fields= TSOB_FELD(struct AsviewmsgfielD *,number_of_fields+1);
 if(msg->fields==NULL)
   return(-1);
 for(i=0;i<number_of_fields+1;i++)
   msg->fields[i]= NULL;
 msg->number_of_fields= number_of_fields;
 msg->fieldcount= 0;
 return(1);
} 


int Asviewmsg_destroy_fieldlist(msg,flag)
struct AsviewmsG *msg;
int flag;
{
 int i;

 if(msg->fields!=NULL) {
   for(i=0;i<msg->number_of_fields;i++)
     Asviewmsgfield_destroy(&(msg->fields[i]),0);
   Smem_freE((char *) msg->fields);
 }
 msg->number_of_fields= 0;
 msg->fieldcount= 0;
 return(1);
}


int Asviewmsg_destroy(msg,flag)
struct AsviewmsG **msg;
int flag;
{
 if((*msg)==NULL)
   return(0);
 Sregex_string(&((*msg)->type),NULL,0);
 Sregex_string(&((*msg)->class),NULL,0);
 Asviewmsg_destroy_fieldlist(*msg,0);
 Sregex_string(&((*msg)->buffer),NULL,0);
 Smem_freE((char *) *msg);
 *msg= NULL;
 return(1);
}


int Asviewmsg_parse_head(msg,text,textcount,rptpt,flag)
struct AsviewmsG **msg;
char *text;
int textcount;
char **rptpt;
int flag;
/*
 return:
   <=0 : error
     1 : head read, msg object created
     2 : head is still too short, nothing read or created
*/
{
 char *type= NULL,class[81],*rpt,*npt,*bpt;
 int number_of_fields= 0,i,ret;
 struct AsviewmsG *m= NULL;


 if(*msg==NULL) {
   ret= Asviewmsg_new(msg,"","",0,0);
   if(ret<=0)
     goto ex;
   (*msg)->stage= 1;
 }
 m= *msg;

 /* read : type number_of_fields */
 rpt= text;
 if(m->stage==1) {
   ret= Sregex_cut_neck(rpt,' ',&(m->type),&rpt,1);
   if(ret<=0)
     goto ex;
   if(ret==2) {
     if(strlen(m->type)>Imvvmsg_type_maxleN) {
       m->stage= -1;
       ret= 0;
     }
     goto ex;
   }
   Asviewmsg_class_from_type(m->type,class,0);
   Sregex_string(&(m->class),class,0);
   m->stage= 2;
 }
 if(m->stage==2) {
   npt= strchr(rpt,'\n');
   bpt= strchr(rpt,' ');
   if(npt!=NULL && (bpt==NULL || npt<bpt)) {
     ret= Sregex_cut_neck(rpt,'\n',&(m->buffer),&rpt,1);
     rpt--; /* do not consume newline */
   } else
     ret= Sregex_cut_neck(rpt,' ',&(m->buffer),&rpt,1);
   if(ret<=0)
     goto ex;
   if(ret==2) {
     if(m->buffer[0]==0) 
       goto ex;                               /* do not touch m->buffer[1] ! */
     if(m->buffer[0]!='0' || m->buffer[1]!=0)
       goto ex;
   }  
   sscanf(m->buffer,"%ld",&number_of_fields);
   Sregex_string(&(m->buffer),NULL,0);
   ret= Asviewmsg_new_fieldlist(m,number_of_fields,0);
   if(ret<=0)
     goto ex;
   m->stage= 3;
 }

 if(m->stage>=0 && m->stage<3)
   ret= 2;
 else
   ret= 1;
ex:;
 Sregex_string(&type,NULL,0);
 *rptpt= rpt;
 return(ret);
}


int Asviewmsg_parse_fields(msg,text,textcount,rptpt,flag)
struct AsviewmsG *msg;
char *text;
int textcount;
char **rptpt;
int flag;
{
 char *type= NULL,*rpt;
 int number_of_fields= 0,i,ret,unfinished= 0;

 if(msg->stage<3)
   return(3);

 rpt= text;
 
 if(msg->stage==3) {
   for(i=msg->fieldcount;i<msg->number_of_fields;i++){
     msg->fieldcount= i;
     ret= Asviewmsgfield_parse_new(&(msg->fields[i]),rpt,textcount-(rpt-text),
                                    &rpt,0);
     if(ret<=0)
       goto ex;
     if(!Asviewmsgfield_is_done(msg->fields[i],0)) {
       unfinished= 1;
   break;
     }
   } 
   if(unfinished || i<msg->number_of_fields) {
     ret= 2;
     goto ex;
   } else {
     msg->fieldcount= msg->number_of_fields;
     msg->stage= 4;
   }
 }
 while(msg->stage==4 && textcount-(rpt-text)>0){
   rpt++;
   if(*(rpt-1)=='\n') 
     msg->stage= -1;
 }
 if(msg->stage==-1)
   ret= 1;
 else
   ret= 2;
ex:;
 *rptpt= rpt;
 return(ret);
}


int Asviewmsg_is_complete(msg,flag)
struct AsviewmsG *msg;
int flag;
{
 if(msg==NULL)
   return(-1);
 return(msg->stage==-1);
}


int Asviewmsg_fetch_type(msg,msgtype,flag)
struct AsviewmsG *msg;
char **msgtype;
int flag;
{
 *msgtype= msg->type;
 return(1);
}


int Asviewmsg_get_fieldcount(msg,flag)
struct AsviewmsG *msg;
int flag;
{
 return(msg->fieldcount);
}


int Asviewmsg_fetch_field_idx(msg,idx,fieldname,bytept,bytecount,flag)
struct AsviewmsG *msg;
int idx;
/* be aware, these are only references and no copied values */
char **fieldname;
char **bytept;
int *bytecount;
int flag;
{
 int ret,i;
 static char default_text[2]={""};

 *bytept= default_text;
 *bytecount= 0;
 if(idx<0 || idx>=msg->fieldcount)
   return(0);
 *fieldname= msg->fields[idx]->name;
 *bytept= msg->fields[idx]->bytes;
 *bytecount= msg->fields[idx]->bytecount;
 return(1);
}


int Asviewmsg_fetch_field(msg,fieldname,bytept,bytecount,flag)
struct AsviewmsG *msg;
char *fieldname;
/* be aware, this is only  a reference and no copied value */
char **bytept;
int *bytecount;
int flag;
{
 int ret,i;
 static char default_text[2]={""};

 *bytept= default_text;
 *bytecount= 0;
 for(i=0;i<msg->fieldcount;i++){
   if(strcmp(fieldname,msg->fields[i]->name)==0) {
     *bytept= msg->fields[i]->bytes;
     *bytecount= msg->fields[i]->bytecount;
     return(1);
   }
 }
 return(0);
}


int Asviewmsg_fetch_frep_idx(msg,idx,representation,flag)
struct AsviewmsG *msg;
int idx;
/* be aware, this is only  a reference and no copied value */
char **representation;
int flag;
{
 int ret,i;
 static char default_text[2]={""};

 if(idx<0 || idx>=msg->fieldcount)
   return(0);
 *representation= msg->fields[idx]->representation;
 return(1);
}


int Asviewmsg_fetch_frep(msg,fieldname,representation,flag)
struct AsviewmsG *msg;
char *fieldname;
/* be aware, this is only  a reference and no copied value */
char **representation;
int flag;
{
 int ret,i;
 static char default_text[2]={""};

 for(i=0;i<msg->fieldcount;i++){
   if(strcmp(fieldname,msg->fields[i]->name)==0) {
     *representation= msg->fields[i]->representation;
     return(1);
   }
 }
 return(0);
}


int Asviewmsg_get_filev(msg,filev,filec,flag)
struct AsviewmsG *msg;
char ***filev;
int *filec;
int flag;
{
 int ret,i;

 if((*filev)!=NULL)
   Smem_freE((char *) (*filev));
 *filec= 0;
 if(strcmp(msg->type,"current_file")!=0 &&
    strcmp(msg->type,"current_file_copy")!=0 &&
    strcmp(msg->type,"double_files")!=0 &&
    strcmp(msg->type,"double_files_copy")!=0)
   return(0);
 if(msg->fieldcount<=0)
   return(1);
 *filev= TSOB_FELD(char *,msg->fieldcount);
 if((*filev)==NULL)
   return(-1);
 for(i=0;i<msg->fieldcount;i++)
   if(strcmp(msg->fields[i]->representation,"address")==0)
     (*filev)[(*filec)++]= msg->fields[i]->bytes;
 return(1);
}


int Asviewmsg_class_from_type(type,class,flag)
char *type;
char *class;
int flag;
{

 strcpy(class,"text");
 if(strcmp(type,"init")==0)
   strcpy(class,"ctrl");
 else if(strcmp(type,"end")==0)
   strcpy(class,"ctrl");
 else if(strcmp(type,"current_file")==0 || strcmp(type,"current_file_copy")==0)
   strcpy(class,"imgfile");
 else if(strcmp(type,"double_files")==0 || strcmp(type,"double_files_copy")==0)
   strcpy(class,"imgfile");
 else if(strcmp(type,"ask")==0)
   strcpy(class,"input");
 return(1);
}

