
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>


#include "../s_tools/smem.h"

#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"


#include "ctyp.h"

#include "cgen.h"



/* ----------------------------- CgeN ------------------------- */

int Cgen_new(cgen,flag)
struct CgeN **cgen;
int flag;
{
 int ret;
 struct CgeN *c;

 *cgen= c= TSOB_FELD(struct CgeN,1);
 if(c==NULL) {
   fprintf(stderr,"Cannot create cgen object : %s\n",strerror(errno));
   return(-1);
 }
 c->classname= NULL;
 c->structname= NULL;
 c->functname= NULL;
 c->is_managed_list= 0;
 c->elements= NULL;
 c->last_element= NULL;
 c->fp= NULL;
 c->filename[0]= 0;
 c->msg[0]= 0;
 return(1);
}


int Cgen_destroy(cgen,flag)
struct CgeN **cgen;
int flag;
{
 struct CgeN *c;
 struct CtyP *ct,*next_ct;
 
 c= *cgen;
 if(c==NULL)
   return(0);

 Sregex_string(&(c->classname),NULL,0);
 Sregex_string(&(c->structname),NULL,0);
 Sregex_string(&(c->functname),NULL,0);
 for(ct= c->elements; ct!=NULL; ct= next_ct) {
   next_ct= ct->next;
   Ctyp_destroy(&ct,0);
 }

 free((char *) c);
 *cgen= NULL; 
 return(1);
}


int Cgen_make_names(cgen,flag)
struct CgeN *cgen;
int flag;
{
 int l;

 if(Sregex_string(&(cgen->structname),cgen->classname,0)<=0)
   return(-1);
 cgen->structname[0]= toupper(cgen->structname[0]);
 l= strlen(cgen->structname);
 cgen->structname[l-1]= toupper(cgen->structname[l-1]);
 if(Sregex_string(&(cgen->functname),cgen->classname,0)<=0)
   return(-1);
 cgen->functname[0]= toupper(cgen->functname[0]);
 return(1);
}


int Cgen_read_fp(cgen,fp,flag)
struct CgeN *cgen;
FILE *fp;
int flag;
/*
 bit0= return 0 if eof at classname
*/
{
 char line[4096],*cpt,*bpt;
 int ret;

 line[0]= 0;
 while(line[0]==0 || line[0]=='#') {
   fprintf(stderr,"[-list] classname ?\n");
   if(Sfile_fgets(line,sizeof(line)-1,fp)==NULL) {
     if(!(flag&1))
       return(2);
no_name:;
     sprintf(cgen->msg,"No classname given.");
     return(0);
   }
   printf("%s\n",line);
 }
 cpt= line;
 if(cpt[0]=='-') {
    /* look for management specifiers:
       -l*   listable by prev-next chain
    */
   if(cpt[1]=='l' || cpt[1]=='L') {
     cgen->is_managed_list= 1;
   }
   while(*cpt!=0 && !isspace(*cpt)) cpt++;
   while(*cpt!=0 && isspace(*cpt)) cpt++;
   if(*cpt==0)
     goto no_name; 
 }
 if(Sregex_string(&(cgen->classname),cpt,0)<=0)
   return(-1);
 ret= Cgen_make_names(cgen,0);
 if(ret<=0)
   return(ret);

 while(1) {
   ret= Ctyp_read_fp(&(cgen->last_element),fp,cgen->msg,0);
   if(ret<=0)
     return(ret);
   if(ret==2)
 break;
   if(cgen->elements==NULL)
     cgen->elements= cgen->last_element;
 }
 if(cgen->is_managed_list) {
   sprintf(line,"-c struct %s *prev",cgen->structname);
   ret= Ctyp_new_from_line(&(cgen->last_element),cgen->last_element,
                           line,cgen->msg,0);
   if(ret<=0)
     return(ret);
   if(cgen->elements==NULL)
     cgen->elements= cgen->last_element;
   sprintf(line,"-c struct %s *next",cgen->structname);
   ret= Ctyp_new_from_line(&(cgen->last_element),cgen->last_element,
                           line,cgen->msg,0);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Cgen_open_wfile(cgen,flag)
struct CgeN *cgen;
int flag;
{
 struct stat stbuf;
 int ret;

 if(stat(cgen->filename,&stbuf)!=-1) {
   sprintf(cgen->msg,"File '%s' already existing.",cgen->filename);
   ret= 0; goto ex;
 }
 if(cgen->fp!=NULL)
   {fclose(cgen->fp); cgen->fp= NULL;}
 cgen->fp= fopen(cgen->filename,"w");
 if(cgen->fp==NULL) {
   sprintf(cgen->msg,"Cannot open file '%s' in w-mode. %s",
                     cgen->filename,strerror(errno));
   ret= 0; goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Cgen_write_datestr(cgen,flag)
struct CgeN *cgen;
int flag;
{
 time_t t0;
 char timetext[81];

 t0= time(0);
 strftime(timetext,sizeof(timetext),"%a, %d %b %Y %H:%M:%S GMT",
                                    gmtime(&t0));
 fprintf(cgen->fp,"/* ( generated by CgeN on  %s ) */\n",timetext);
 return(1);
}


int Cgen_write_h(cgen,flag)
struct CgeN *cgen;
int flag;
{
 int ret,i,pointer_level;
 FILE *fp= NULL;
 struct CtyP *ct;

 sprintf(cgen->filename,"%s.h",cgen->classname);
 ret= Cgen_open_wfile(cgen,0);
 if(ret<=0)
   goto ex;
 fp= cgen->fp;
 
 fprintf(fp,"\n");
 fprintf(fp,"#ifndef %s_includeD\n",cgen->functname);
 fprintf(fp,"#define %s_includeD\n",cgen->functname);
 fprintf(fp,"\n");
 fprintf(fp,"\n");
 fprintf(fp,"\n");
 fprintf(fp,"struct %s {\n",cgen->structname);
 fprintf(fp,"\n");
 ct= cgen->elements;
 for(ct= cgen->elements;ct!=NULL;ct= ct->next) {
   if(Ctyp_is_struct(ct,0))
     fprintf(fp," struct");
   fprintf(fp," %s ",ct->dtype);
   pointer_level= Ctyp_get_pointer_level(ct,0);
   for(i=0;i<pointer_level;i++)
     fprintf(fp,"*");
   fprintf(fp,"%s;\n",ct->name);
 }
 fprintf(fp,"\n");
 fprintf(fp,"};\n");
 fprintf(fp,"\n");
 fprintf(fp,"\n");
 fprintf(fp,"#endif /* %s_includeD */\n",cgen->functname);
 fprintf(fp,"\n");
 Cgen_write_datestr(cgen,0);
 ret= 1;
ex:;
 if(cgen->fp!=NULL) 
   {fclose(cgen->fp); cgen->fp= NULL;}
 return(ret);
}


int Cgen_write_c_head(cgen,flag)
struct CgeN *cgen;
int flag;
{
 int ret;
 FILE *fp= NULL;

 fp= cgen->fp;
 fprintf(fp,"\n");
 fprintf(fp,"/*\n");
 fprintf(fp,"  cc -g -o %s.c\n",cgen->classname);
 fprintf(fp,"*/\n");
 Cgen_write_datestr(cgen,0);
 fprintf(fp,"\n");
 fprintf(fp,"#include <sys/types.h>\n");
 fprintf(fp,"#include <stdlib.h>\n");
 fprintf(fp,"#include <stdio.h>\n");
 fprintf(fp,"#include <string.h>\n");
 fprintf(fp,"#include <errno.h>\n");
 fprintf(fp,"\n");
 fprintf(fp,"#include \"../s_tools/smem.h\"\n");
 fprintf(fp,"#include \"../s_tools/sfile.h\"\n");
 fprintf(fp,"#include \"../s_tools/sregex.h\"\n");
 fprintf(fp,"\n");
 fprintf(fp,"#include \"%s.h\"\n",cgen->classname);
 fprintf(fp,"\n");
 fprintf(fp,"\n");
 fprintf(fp,"/* --------------------------  %s  ----------------------- */\n",
            cgen->structname);
 fprintf(fp,"\n");
 return(1);
}


int Cgen_write_c_new(cgen,flag)
struct CgeN *cgen;
int flag;
{
 int ret,pointer_level,management;
 unsigned long array_size;
 FILE *fp= NULL;
 struct CtyP *ct;

 fp= cgen->fp;
 fprintf(fp,"\n");
 fprintf(fp,"int %s_new(objpt,flag)\n",cgen->functname);
 fprintf(fp,"struct %s **objpt;\n",cgen->structname);
 fprintf(fp,"int flag;\n");
 fprintf(fp,"{\n");
 fprintf(fp," struct %s *o;\n",cgen->structname);
 fprintf(fp," int ret,i;\n");
 fprintf(fp,"\n");
 fprintf(fp," *objpt= o= TSOB_FELD(struct %s,1);\n",cgen->structname);
 fprintf(fp," if(o==NULL)\n");
 fprintf(fp,"   return(-1);\n");
 fprintf(fp,"\n");
 for(ct= cgen->elements; ct!=NULL; ct= ct->next) {
   array_size= Ctyp_get_array_size(ct,0);
   pointer_level= Ctyp_get_pointer_level(ct,0);
   if(array_size>0) {
     if(strcmp(ct->dtype,"char")==0) {
       fprintf(fp," o->%s[0]= 0;\n;",ct->name);
     } else if(pointer_level>0) {
       fprintf(fp," for(i=0;i<array_size;i++)\n;");
       fprintf(fp,"   o->%s[i]= NULL;\n",ct->name);
     } else {
       fprintf(fp," for(i=0;i<array_size;i++)\n;");
       fprintf(fp,"   o->%s[i]= 0;\n",ct->name);
     }  
   } else if(pointer_level>0) {
     fprintf(fp," o->%s= NULL;\n",ct->name);
   } else
     fprintf(fp," o->%s= 0;\n",ct->name);
 }
 fprintf(fp,"\n");
 fprintf(fp," return(1);\n");
 fprintf(fp,"failed:;\n");
 fprintf(fp," %s_destroy(objpt,0);\n",cgen->functname);
 fprintf(fp," return(-1);\n");
 fprintf(fp,"}\n");
 fprintf(fp,"\n");
 for(ct= cgen->elements; ct!=NULL; ct= ct->next) {
   management= Ctyp_get_management(ct,0);
   if(management==4) {
     if(ct->next==NULL) {
no_last_pt:;
       sprintf(cgen->msg,
               "Lonely -l found. A -v of same type must follow.\nName is : %s",
               ct->name);
       return(0);
     }
     if(strcmp(ct->next->dtype,ct->dtype)!=0 
               || ct->next->is_pointer!=ct->is_pointer)
       goto no_last_pt;
     ret= Cgen_write_c_new_type(cgen,ct,ct->next,0);
     if(ret<=0)
       return(ret);
   }
 }
 return(1);
}


int Cgen_write_c_new_type(cgen,ct_first,ct_last,flag)
struct CgeN *cgen;
struct CtyP *ct_first,*ct_last;
int flag;
{
 int ret,l,management,pointer_level,i;
 FILE *fp= NULL;
 char funct[4096],classname[4096],*npt;

 strcpy(funct,ct_first->dtype);
 strcpy(classname,funct);
 l= strlen(funct);
 if(l>0) {
   funct[0]= toupper(funct[0]);
   funct[l-1]= tolower(funct[l-1]);
   classname[0]= tolower(funct[0]);
   classname[l-1]= funct[l-1];
 }
 fp= cgen->fp;
 fprintf(fp,"\n");
 fprintf(fp,"int %s_new_%s(o,flag)\n",cgen->functname,ct_first->name);
 fprintf(fp,"struct %s *o;\n",cgen->structname);
 fprintf(fp,"int flag;\n");
 fprintf(fp,"{\n");
 fprintf(fp," int ret;\n");
 fprintf(fp," struct %s *c= NULL;\n",ct_first->dtype);
 fprintf(fp,"\n");
 fprintf(fp," ret= %s_new(&c,0);\n",funct);
 fprintf(fp," if(ret<=0)\n");
 fprintf(fp,"   return(ret);\n");
 fprintf(fp," %s_link(c,o->%s,0);\n",funct,ct_last->name);
 fprintf(fp," o->%s= c;\n",ct_last->name);
 fprintf(fp," if(o->%s==NULL)\n",ct_first->name);
 fprintf(fp,"   o->%s= c;\n",ct_first->name);
 fprintf(fp," return(1);\n");
 fprintf(fp,"}\n");
 fprintf(fp,"\n");
 ret= 1;
ex:;
 return(ret);
}


int Cgen_write_c_destroy(cgen,flag)
struct CgeN *cgen;
int flag;
{
 int ret,l,management,pointer_level,i;
 FILE *fp= NULL;
 struct CtyP *ct,*next;
 char funct[4096],*npt;

 fp= cgen->fp;
 fprintf(fp,"\n");
 fprintf(fp,"int %s_destroy(objpt,flag)\n",cgen->functname);
 fprintf(fp,"struct %s **objpt;\n",cgen->structname);
 fprintf(fp,"int flag;\n");
 fprintf(fp,"{\n");
 fprintf(fp," struct %s *o;\n",cgen->structname);
 fprintf(fp,"\n");
 fprintf(fp," o= *objpt;\n");
 fprintf(fp," if(o==NULL)\n");
 fprintf(fp,"   return(0);\n");
 fprintf(fp,"\n");
 for(ct= cgen->elements; ct!=NULL; ct= ct->next) {
   management= Ctyp_get_management(ct,0);
   if(management==1 || management==4) {
     strcpy(funct,ct->dtype);
     l= strlen(funct);
     if(l>0) {
       funct[0]= toupper(funct[0]);
       funct[l-1]= tolower(funct[l-1]);
     }
     if(strcmp(ct->dtype,"char")==0)
       fprintf(fp," Sregex_string(");
     else if(strcmp(ct->dtype,"LstrinG")==0 || management==4)
       fprintf(fp," %s_destroy_all(",funct);
     else
       fprintf(fp," %s_destroy(",funct);
     
     pointer_level= Ctyp_get_pointer_level(ct,0)-2;
     for(i=0; i>pointer_level; i--)
       fprintf(fp,"&");
     for(i=0; i<pointer_level; i++)
       fprintf(fp,"*");
     fprintf(fp,"(o->%s)",ct->name);
     if(strcmp(ct->dtype,"char")==0)
       fprintf(fp,",NULL");
     fprintf(fp,",0);\n");
   } else if(management==2) {
     next= ct->next;
     if(next==NULL) {
broken_chain:;
       sprintf(cgen->msg,
               "Lonely -c found. They have to appear in pairs.\nName is : %s",
               ct->name);
       ret= 0; goto ex;
     }
     if(next->management!=3)
       goto broken_chain;
     fprintf(fp," if(o->%s!=NULL)\n",ct->name);
     fprintf(fp,"   o->%s->%s= o->%s;\n",ct->name,next->name,next->name);
     fprintf(fp," if(o->%s!=NULL)\n",next->name);
     fprintf(fp,"   o->%s->%s= o->%s;\n",next->name,ct->name,ct->name);
     ct= next;
   }
 }
 fprintf(fp,"\n");
 fprintf(fp," Smem_freE((char *) o);\n");
 fprintf(fp," *objpt= NULL;\n");
 fprintf(fp," return(1);\n");
 fprintf(fp,"}\n");
 fprintf(fp,"\n");
 if(cgen->is_managed_list){
   ret= Cgen_write_c_destroy_all(cgen,0);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Cgen_write_c_destroy_all(cgen,flag)
struct CgeN *cgen;
int flag;
{
 int ret,l,management,pointer_level,i;
 FILE *fp= NULL;
 struct CtyP *ct;

 fp= cgen->fp;
 fprintf(fp,"\n");
 fprintf(fp,"int %s_destroy_all(objpt,flag)\n",cgen->functname);
 fprintf(fp,"struct %s **objpt;\n",cgen->structname);
 fprintf(fp,"int flag;\n");
 fprintf(fp,"{\n");
 fprintf(fp," struct %s *o,*n;\n",cgen->structname);
 fprintf(fp,"\n");
 fprintf(fp," o= *objpt;\n");
 fprintf(fp," if(o==NULL)\n");
 fprintf(fp,"   return(0);\n");
 fprintf(fp," for(;o->prev!=NULL;o= o->prev);\n");
 fprintf(fp," for(;o!=NULL;o= n) {\n");
 fprintf(fp,"   n= o->next;\n");
 fprintf(fp,"   %s_destroy(&o,0);\n",cgen->functname);
 fprintf(fp," }\n");
 fprintf(fp," *objpt= NULL;\n");
 fprintf(fp," return(1);\n");
 fprintf(fp,"}\n");
 fprintf(fp,"\n");
 ret= 1;
ex:;
 return(ret);
}


int Cgen_write_c_access(cgen,flag)
struct CgeN *cgen;
int flag;
{
 int ret,l,management,pointer_level,i;
 FILE *fp= NULL;
 struct CtyP *ct;
 char funct[4096],*npt;

 fp= cgen->fp;
 for(ct= cgen->elements; ct!=NULL; ct= ct->next) {
   pointer_level= Ctyp_get_pointer_level(ct,0);
   fprintf(fp,"\n");
   fprintf(fp,"int %s_get_%s(o,pt,flag)\n",cgen->functname,ct->name);
   fprintf(fp,"struct %s *o;\n",cgen->structname);
   if(Ctyp_is_struct(ct,0))
     fprintf(fp,"struct ");
   fprintf(fp,"%s ",ct->dtype);
   for(i=0; i<pointer_level+1; i++)
     fprintf(fp,"*");
   fprintf(fp,"pt;\n");
   fprintf(fp,"int flag;\n");
   fprintf(fp,"{\n");
   fprintf(fp," *pt= o->%s;\n",ct->name);
   fprintf(fp," return(1);\n");
   fprintf(fp,"}\n");
   fprintf(fp,"\n");
 }
 if(cgen->is_managed_list) {
   fprintf(fp,"\n");
   fprintf(fp,"int %s_link(o,link,flag)\n",cgen->functname);
   fprintf(fp,"struct %s *o;\n",cgen->structname);
   fprintf(fp,"struct %s *link;\n",cgen->structname);
   fprintf(fp,"int flag;\n");
   fprintf(fp,"/*\n");
   fprintf(fp,"  bit0= insert as link->prev rather than as link->next\n");
   fprintf(fp,"*/\n");
   fprintf(fp,"{\n");
   fprintf(fp," if(o->prev!=NULL)\n");
   fprintf(fp,"   o->prev->next= o->next;\n");
   fprintf(fp," if(o->next!=NULL)\n");
   fprintf(fp,"   o->next->prev= o->prev;\n");
   fprintf(fp," o->prev= o->next= NULL;\n");
   fprintf(fp," if(link==NULL)\n");
   fprintf(fp,"   return(1);\n");
   fprintf(fp," if(flag&1) {\n");
   fprintf(fp,"   o->next= link;\n");
   fprintf(fp,"   o->prev= link->prev;\n");
   fprintf(fp,"   if(o->prev!=NULL)\n");
   fprintf(fp,"     o->prev->next= o;\n");
   fprintf(fp,"   link->prev= o;\n");
   fprintf(fp," } else {\n");
   fprintf(fp,"   o->prev= link;\n");
   fprintf(fp,"   o->next= link->next;\n");
   fprintf(fp,"   if(o->next!=NULL)\n");
   fprintf(fp,"     o->next->prev= o;\n");
   fprintf(fp,"   link->next= o;\n");
   fprintf(fp," }\n");
   fprintf(fp," return(1);\n");
   fprintf(fp,"}\n");
   fprintf(fp,"\n");
 }
 return(1);
}


int Cgen_write_c(cgen,flag)
struct CgeN *cgen;
int flag;
/*
 bit0= also write access functions *_set_* *_get_* [*_link_*]
*/
{
 int ret;

 sprintf(cgen->filename,"%s.c",cgen->classname);
 ret= Cgen_open_wfile(cgen,0);
 if(ret<=0)
   goto ex;
 ret= Cgen_write_c_head(cgen,0);
 if(ret<=0)
   goto ex;
 ret= Cgen_write_c_new(cgen,0);
 if(ret<=0)
   goto ex;
 ret= Cgen_write_c_destroy(cgen,0);
 if(ret<=0)
   goto ex;
 if(flag&1) {
   ret= Cgen_write_c_access(cgen,0);
   if(ret<=0)
     goto ex;
 }
 
 ret= 1;
ex:;
 if(cgen->fp!=NULL)
   {fclose(cgen->fp); cgen->fp= NULL;}
 return(ret);
}


/* -------------------------------------------------------------- */

main(argc,argv)
int argc;
char **argv;
{
 struct CgeN *cgen= NULL;
 int ret, msg_printed= 0,first=1;

 while(!feof(stdin)) {
   ret= Cgen_new(&cgen,0);
   if(ret<=0)
     goto ex;
   ret= Cgen_read_fp(cgen,stdin,first);
   if(ret<=0)
     goto ex;
   if(ret==2)
 break;
   first= 0;
   ret= Cgen_write_h(cgen,0);
   if(ret<=0)
     goto ex;
   ret= Cgen_write_c(cgen,1);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:
 if(cgen!=NULL)
   if(cgen->msg[0]!=0) {
     fprintf(stderr,"%s\n",cgen->msg);
     msg_printed= 1;
   }
 if(ret<=0 &&!msg_printed) {
   if(errno>0)
     fprintf(stderr,"Error : %s\n",strerror(errno));
   else if(ret==-1)
     fprintf(stderr,"Program run failed (probably due to lack of memory)\n");
   else
     fprintf(stderr,"Program run failed\n");
 }
 Cgen_destroy(&cgen,0);
 exit(1-ret);
}
