
/*
  cc -g -o httpproxy.c
*/

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "httprequest.h"

#include "httpproxy.h"


/* --------------------------  HttpproxY  ----------------------- */


int Httpproxy_new(objpt,agent,msg,flag)
struct HttpproxY **objpt;
struct AsagenT *agent;
char *msg;                             /* must point to at least 4096 bytes */
int flag;
{
 struct HttpproxY *o;
 int ret,i;

 *objpt= o= TSOB_FELD(struct HttpproxY,1);
 if(o==NULL)
   return(-1);

 o->agent= agent;
 o->logfile= NULL;
 o->log_client_side= 0;
 o->parse_request= 0;
 o->log_tee= 0;
 o->request_counter= 0;
 o->client_byte_count= 0;
 o->server_byte_count= 0;
 o->byte_count_limit= Httpproxy_byte_count_limiT;
 o->from_client= NULL;
 o->to_server= NULL;
 o->microsleep= 0;
 o->requests= NULL;
 o->last_request= NULL;
 o->cookies= NULL;
 o->last_cooky= NULL;
 o->docs= NULL;
 o->last_doc= NULL;
 o->msg= msg;

 return(1);
failed:;
 Httpproxy_destroy(objpt,0);
 return(-1);
}


int Httpproxy_destroy_lists(o,flag)
struct HttpproxY *o;
int flag;
{
 Httprequest_destroy_all(&(o->requests),0);
 o->last_request= NULL;
 Httpcooky_destroy_all(&(o->cookies),0);
 o->last_cooky= NULL;
 Httpdoc_destroy_all(&(o->docs),0);
 o->last_doc= NULL;
 return(1);
}
 

int Httpproxy_destroy(objpt,flag)
struct HttpproxY **objpt;
int flag;
{
 struct HttpproxY *o;

 o= *objpt;
 if(o==NULL)
   return(0);

 Sregex_string(&(o->logfile),NULL,0);
 Httpproxy_destroy_lists(o,0);
 Sconn_destroy(&(o->from_client),0);
 Sconn_destroy(&(o->to_server),0);

 Smem_freE((char *) o);
 *objpt= NULL;
 return(1);
}


int Httpproxy_set_logfile(o,name,flag)
struct HttpproxY *o;
char *name;
int flag;
{
 if(Sregex_string(&(o->logfile),name,0)<=0)
   return(-1);
 return(1);
}


int Httpproxy_get_logfile(o,pt,flag)
struct HttpproxY *o;
char **pt;
int flag;
{
 *pt= o->logfile;
 return(1);
}


int Httpproxy_ret_byte_count_limit(prx,flag)
struct HttpproxY *prx;
int flag;
{
 return(prx->byte_count_limit);
}


int Httpproxy_set_byte_count_limit(prx,value,flag)
struct HttpproxY *prx;
int value;
int flag;
{
 prx->byte_count_limit= value;
 return(1);
}


int Httpproxy_ret_log_tee(prx,flag)
struct HttpproxY *prx;
int flag;
{
 return(prx->log_tee);
}


int Httpproxy_set_log_tee(prx,value,flag)
struct HttpproxY *prx;
int value;
int flag;
{
 prx->log_tee= value;
 return(1);
}


int Httpproxy_ret_microsleep(prx,flag)
struct HttpproxY *prx;
int flag;
{
 return(prx->microsleep);
}


int Httpproxy_set_microsleep(prx,value,flag)
int value;
struct HttpproxY *prx;
int flag;
{
 prx->microsleep= value;
 return(1);
}


int Httpproxy_get_from_client(o,pt,flag)
struct HttpproxY *o;
struct SconN **pt;
int flag;
{
 *pt= o->from_client;
 return(1);
}


int Httpproxy_get_to_server(o,pt,flag)
struct HttpproxY *o;
struct SconN **pt;
int flag;
{
 *pt= o->to_server;
 return(1);
}


int Httpproxy_get_requests(o,pt,flag)
struct HttpproxY *o;
struct HttprequesT **pt;
int flag;
{
 *pt= o->requests;
 return(1);
}


int Httpproxy_get_cookies(o,pt,flag)
struct HttpproxY *o;
struct HttpcookY **pt;
int flag;
{
 *pt= o->cookies;
 return(1);
}


int Httpproxy_get_docs(o,pt,flag)
struct HttpproxY *o;
struct HttpdoC **pt;
int flag;
{
 *pt= o->docs;
 return(1);
}


int Httpproxy_send_error(prx,statustext,flag)
struct HttpproxY *prx;
char *statustext;
int flag;
{
 int ret;
 time_t t0;
 char buff[2*8192],*hv,timetext[81];

 if(prx->from_client==NULL)
   {ret= 0; goto ex;}

 t0= time(0);
 strftime(timetext,sizeof(timetext),"%a, %d %b %Y %H:%M:%S GMT",
                                    gmtime(&t0));
 buff[0]= 0;
 sprintf(buff+strlen(buff),"HTTP/1.0 %s\r\n",statustext);
 sprintf(buff+strlen(buff),"Server: struct HttpproxY\r\n");
 sprintf(buff+strlen(buff),"Date: %s\r\n",timetext);
 sprintf(buff+strlen(buff),"Connection: close\r\n\r\n");
 sprintf(buff+strlen(buff),"<HTML><HEAD><TITLE>%s</TITLE></HEAD>",statustext);
 sprintf(buff+strlen(buff),"<BODY><H1>%s</H1><PRE>%s</PRE>",
                           statustext,prx->msg);
 sprintf(buff+strlen(buff),"<HR>( sent by struct HttpproxY )</BODY></HTML>\n");
 ret= Sconn_write(prx->from_client,buff,strlen(buff),4);
 if(ret<=0)
   {ret= 0;goto ex;}
 ret= 1;
ex:;
 return(ret);
}


int Httpproxy_log_write(prx,text,len,flag)
struct HttpproxY *prx;
char *text;
int len;
int flag;
/*
 bit0= log to stdout in any case
*/
{
 int ret,to_stdout= 0;
 char *stdout_text= NULL;

 if(strcmp(prx->logfile,"-")==0 || prx->logfile[0]==0) 
   to_stdout= 1;
 if(!to_stdout) {
   ret= Sfile_bin_to_file(prx->logfile,text,len,2);
   if(ret<=0)
     goto ex;
 }
 if(to_stdout || prx->log_tee || (flag&1)) {
   if(Sregex_string_cut(&stdout_text,text,len,0)<=0)
     goto ex;
   Asagent_stdout(prx->agent,stdout_text,4);
/*
   fwrite(text,len,1,stdout); fflush(stdout);
*/
 }
 ret= 1;
ex:;
 Sregex_string(&stdout_text,NULL,0);
 return(ret);
}


int Httpproxy_log(prx,line,len,flag)
struct HttpproxY *prx;
char *line;
int len;
int flag;
/*
 bit0= this is the Request-Line (i.e. first line of request) without CRLF
       or the first reply bytes of the server
 bit1= this comes from the server (else from client)
 bit2= report end of connection
 bit3= print also to stdout if log file is active
*/
{
 int ret,to_stdout= 0,i;
 char *buff= NULL,*wpt,text[81];

 if(prx->logfile==NULL) 
   {ret= 2; goto ex;}

 if(flag&4) {
   Httpproxy_log_write(prx,"\n\n",2,0);
   sprintf(text,"++++++ #%ld Connection closed ( %ld + %ld bytes)\n\n",
                prx->request_counter,prx->client_byte_count,
                prx->server_byte_count);
   ret= Httpproxy_log_write(prx,text,strlen(text),1);
   goto ex;
 }

 if(flag&1) {
   if(flag&2) {
     Httpproxy_log_write(prx,"\n\n",2,0);
     sprintf(text,"++++++ #%ld Server reply :\n",prx->request_counter);
     Httpproxy_log_write(prx,text,strlen(text),1);
     ret= Httpproxy_log_write(prx,"\n",1,0);
   } else {
     prx->request_counter++;
     Httpproxy_log_write(prx,"\n\n",2,0);
     sprintf(text,"++++++ #%ld Client request :\n",prx->request_counter);
     Httpproxy_log_write(prx,text,strlen(text),1);
     ret= Httpproxy_log_write(prx,"\n",1,0);
   }
   if(ret<=0)
     goto ex;
 }
 wpt= buff= TSOB_FELD(unsigned char,3*len+10);
 if(buff==NULL)
   {ret= -1; goto ex;}
 for(i=0;i<len;i++)
   if(line[i]!=10 && (line[i]<32 || line[i]>126 || line[i]=='\\')) {
     sprintf(wpt,"\\%-2.2X",(unsigned int) line[i]);
     wpt+= 3;
   } else
     *(wpt++)= line[i];
 *wpt= 0;
 ret= Httpproxy_log_write(prx,buff,wpt-buff,!!(flag&8));
 if(ret<=0)
   goto ex;
 if((flag&1) && !(flag&2))
   Httpproxy_log_write(prx,"\\0D\n",4,!!(flag&8));
 ret= 1;
ex:;
 if(buff!=NULL)
   Smem_freE(buff);
 return(ret);
}


int Httpproxy_new_request(prx,flag)
struct HttpproxY *prx;
int flag;
{
 int ret= 0;
 struct HttprequesT *req= NULL;

 ret= Httprequest_new(&req,0);
 if(ret<=0)
   return(ret);
 Httprequest_link(req,prx->last_request,0);
 prx->last_request= req;
 if(prx->requests==NULL)
   prx->requests= req;
 return(1);
}


int Httpproxy_handle_firstline(prx,flag)
struct HttpproxY *prx;
int flag;
{
 int ret= 0,portnumber= -1;
 char *line= NULL,*request_uri,*hostname= NULL,*path= NULL;
 struct HttprequesT *req;

 ret= Sconn_readline(prx->from_client,&line,1);
 if(ret<=0) 
   {ret= 0; goto ex;}
 if(prx->log_client_side)
   Httpproxy_log(prx,line,strlen(line),1|8);
 prx->client_byte_count+= strlen(line)+2;

 /* obtain target address */
 ret= Httpproxy_new_request(prx,0);
 if(ret<=0)
   goto ex;
 req= prx->last_request;
 ret= Httprequest_first_line(req,line,0);
 if(ret<=0)
   goto ex;
 ret= Httprequest_split_uri(req,&hostname,&portnumber,&path,0);
 if(ret<=0)
   goto ex;
 if(hostname==NULL) {
   /* >>> ??? buffer input until 'Host:' header was found */
   sprintf(prx->msg,"Client is not aware of proxy and sends no absoluteURI.");
   Httpproxy_send_error(prx,"404 Not found.",0);
   ret= 0; goto ex;
 }
 
 ret= Sconn_tcpclient_new(&(prx->to_server),hostname,portnumber,NULL,0);
 if(ret<=0) {
   sprintf(prx->msg,"Failed to connect to origin server.");
   if(ret==0)
     Httpproxy_send_error(prx,"404 Not found.",0);
   goto ex;
 }

 /* compose new first line and send */
 if(Sregex_string(&line,req->method,0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&line," ",1)<=0)
   {ret= -1; goto ex;}
 if(path==NULL) {
   if(Sregex_string(&line,"*",1)<=0)
     {ret= -1; goto ex;}
 } else {
   if(Sregex_string(&line,path,1)<=0)
     {ret= -1; goto ex;}
 }
 if(Sregex_string(&line," ",1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&line,req->http_version,1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&line,"\r\n",1)<=0)
   {ret= -1; goto ex;}
 ret= Sconn_write(prx->to_server,line,strlen(line),4);
 if(ret<=0) {
   sprintf(prx->msg,"Cannot write to http server.");
   if(ret==0)
     Httpproxy_send_error(prx,"404 Not found.",0);
   ret= 0; goto ex;
 }
 if(!prx->log_client_side) {
   line[strlen(line)-2]= 0;
   Httpproxy_log(prx,line,strlen(line),1|8);
 }
 ret= 1;
ex:;
 if(ret<=0 && prx->msg[0]==0)
   sprintf(prx->msg,"No complete first line of request readable");
 Sregex_string(&line,NULL,0);
 Sregex_string(&hostname,NULL,0);
 Sregex_string(&path,NULL,0);
 return(ret);
}


int Httpproxy_exchange_data(prx,flag)
struct HttpproxY *prx;
int flag;
{
 int ret= 0,ready_count,bytes,server_silent= 1,log_flag,byte_count= 0;
 int truncated= 0,byte_count_limit= 8192,bytes_to_log;
 int server_first_line= 1,log_len,i;
 struct SconN *conns[2],*ready[2],*write_conn;
 char buf[4096],*conn_id= NULL,*lpt;

 /* forward all bytes of request while listening for reply
    and forwarding reply bytes unchanged to client connection */
 /* detect EOF and return */

 conns[0]= prx->from_client;
 conns[1]= prx->to_server;
 while(1) {
   ret= Sconn_select(conns,2,ready,&ready_count,1000000,0);
   if(ret<0)
     goto ex;
   if(ready_count<=0)
 continue;
   ret= Sconn_read(ready[0],buf,sizeof(buf),&bytes,0);
   if(ret<0) {
     if(Sconn_get_id(ready[0],&conn_id,0)<=0)
       conn_id= "-";
     sprintf(prx->msg,"Error when reading connection %s.",conn_id);
     ret= 0;
     goto ex;
   }
   if(ret==0) /* EOF */
 break;
   if(bytes==0)
 continue;
   log_flag= 0;
   if(ready[0]==prx->from_client) {
     if(prx->parse_request) {
       ret= Httprequest_parse(prx->last_request,buf,bytes,0);
       if(ret<=0)
         goto ex;
     }
     write_conn= prx->to_server;
     byte_count= prx->client_byte_count;
     prx->client_byte_count+= bytes;
   } else {
     log_flag= 2;
     if(server_silent) {
       log_flag|= 1;
       server_silent= 0;
       server_first_line= 1;
       truncated= 0;
     }
     write_conn= prx->from_client;
     byte_count= prx->server_byte_count;
     prx->server_byte_count+= bytes;
   }
   bytes_to_log= bytes;
   if(byte_count+bytes_to_log > prx->byte_count_limit) {
     bytes_to_log= prx->byte_count_limit - byte_count;
     if(truncated==0)
       truncated= 1;
   }
   lpt= buf;
   if(ready[0]==prx->to_server && server_first_line) {
     for(i=0;i<bytes;i++)
       if(buf[i]=='\n')
     break;
     if(i>=bytes)
       log_len= bytes;
     else {
       log_len= i+1;
       server_first_line= 0;
     }
     Httpproxy_log(prx,buf,log_len,log_flag|8); 
     lpt+= log_len;
     bytes_to_log-= log_len;
     log_flag&= ~1;
   }
   if(bytes_to_log>0)
     Httpproxy_log(prx,lpt,bytes_to_log,log_flag);
   if(truncated==1) {
     truncated= 2;
     Httpproxy_log(prx,"\n++++++ truncated\n",18,log_flag&2);
   }
   ret= Sconn_write(write_conn,buf,bytes,4);
   if(ret<=0) {
     if(Sconn_get_id(write_conn,&conn_id,0)<=0)
       conn_id= "-";
     sprintf(prx->msg,"Error when writing connection %s.",conn_id);
     ret= 0; goto ex;
   }
   byte_count+= bytes;
   if(prx->microsleep>0)
     Sfile_microsleep(prx->microsleep,0);
 }

 ret= 1;
ex:;
 Sregex_string(&conn_id,NULL,0);
 return(ret);
}


int Httpproxy_handle_request(prx,fd,logfile,flag)
struct HttpproxY *prx;
int fd;
char *logfile;
int flag;
{
 int ret;

 Sconn_destroy(&(prx->from_client),0);
 Sconn_destroy(&(prx->to_server),0);
 prx->client_byte_count= 0;
 prx->server_byte_count= 0;
 if(Sregex_string(&(prx->logfile),logfile,0)<=0)
   {ret= -1; goto ex;}
 ret= Sconn_new_from_tcpfd(&(prx->from_client),fd,"from_client",NULL,0);
 if(ret<=0)
   goto ex;
 ret= Httpproxy_handle_firstline(prx,0);
 if(ret<=0)
   goto ex;
 ret= Httpproxy_exchange_data(prx,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 Httpproxy_log(prx,"",0,4);
 Sconn_destroy(&(prx->from_client),0);
 Sconn_destroy(&(prx->to_server),0);
 return(ret);
}

