
/*
  cc -g -o sproxy.c
*/

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "../as/asagent.h"

#include "sproxy.h"

/* --------------------------  SproxY  ----------------------- */


int Sproxy_new(objpt,agent,hostname,portnumber,logfile,flag)
struct SproxY **objpt;
struct AsagenT *agent;
char *hostname;
int portnumber;
char *logfile;
int flag;
{
 struct SproxY *o;
 int ret,i;

 *objpt= o= TSOB_FELD(struct SproxY,1);
 if(o==NULL)
   return(-1);

 o->hostnames= NULL;
 o->portnumber= portnumber;
 o->external_s_mask= 0;
 o->logfile= NULL;
 o->agent= agent;
 o->serv= NULL;
 o->serv_sconn= NULL;
 o->httpproxy= NULL;
 o->msg[0]= 0;

 if(Sregex_string(&(o->logfile),logfile,0)<=0)
   goto failed;
 ret= Lstring_new(&(o->hostnames),hostname,NULL,0);
 if(ret<=0)
   goto failed;

 ret= Httpproxy_new(&(o->httpproxy),agent,o->msg,0);
 if(ret<=0)
   goto failed;
 return(1);
failed:;
 Sproxy_destroy(objpt,0);
 return(-1);
}


int Sproxy_destroy(objpt,flag)
struct SproxY **objpt;
int flag;
{
 struct SproxY *o;

 o= *objpt;
 if(o==NULL)
   return(0);

 Sproxy_stop(o,0);
 Lstring_destroy_all(&(o->hostnames),0);
 Sregex_string(&(o->logfile),NULL,0);
 Httpproxy_destroy(&(o->httpproxy),0);

 Smem_freE((char *) o);
 *objpt= NULL;
 return(1);
}


int Sproxy_start(o,flag)
struct SproxY *o;
int flag;
{
 int ret;

 ret= Sproxy_stop(o,0); 
 ret= Stcpserv_new(&(o->serv),o->hostnames->text,o->portnumber,0xFFFFFFFF,0);
 if(ret<=0)
   goto ex;
 o->portnumber= Stcpserv_get_portnumber(o->serv,0);
 /* attach to agent as interrupting connection */;
 ret= Sconn_new(&(o->serv_sconn),"sproxy server face",NULL,0);
 if(ret<=0)
   goto ex;
 ret= Sconn_attach_fd(o->serv_sconn,5,"sproxy server face in",
                      Stcpserv_get_server_socket(o->serv),0);
 if(ret<=0)
   goto ex;
 ret= Asagent_new_interrupting_inconn(o->agent,o->serv_sconn,NULL,NULL,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Sproxy_stop(o,flag)
struct SproxY *o;
int flag;
{
 int ret;

 if(o==NULL)
   return(0);
 if(o->serv==NULL)
   return(0);

 if(o->serv_sconn!=NULL)
   Asagent_destroy_inconn(o->agent,o->serv_sconn,0);
 Sconn_destroy(&(o->serv_sconn),0);
 Stcpserv_destroy(&(o->serv),0);
 return(1);
}


int Sproxy_get_hostnames(o,pt,flag)
struct SproxY *o;
struct LstrinG **pt;
int flag;
{
 *pt= o->hostnames;
 return(1);
}


int Sproxy_set_portnumber(o,value,flag)
struct SproxY *o;
int value;
int flag;
{
 o->portnumber= value;
 return(1);
}


int Sproxy_get_portnumber(o,pt,flag)
struct SproxY *o;
int *pt;
int flag;
{
 *pt= o->portnumber;
 return(1);
}


int Sproxy_get_external_s_mask(o,pt,flag)
struct SproxY *o;
int *pt;
int flag;
{
 *pt= o->external_s_mask;
 return(1);
}


int Sproxy_get_logfile(o,pt,flag)
struct SproxY *o;
char **pt;
int flag;
{
 *pt= o->logfile;
 return(1);
}


int Sproxy_get_agent(o,pt,flag)
struct SproxY *o;
struct AsagenT **pt;
int flag;
{
 *pt= o->agent;
 return(1);
}


int Sproxy_get_serv(o,pt,flag)
struct SproxY *o;
struct StcpserV **pt;
int flag;
{
 *pt= o->serv;
 return(1);
}


int Sproxy_get_httpproxy(o,pt,flag)
struct SproxY *o;
struct HttpproxY **pt;
int flag;
{
 *pt= o->httpproxy;
 return(1);
}


int Sproxy_get_msg(o,pt,flag)
struct SproxY *o;
char **pt;
int flag;
{
 *pt= o->msg;
 return(1);
}


int Sproxy_helptext(prx,reply,flag)
struct SproxY *prx;
char **reply;
int flag;
/*
 bit0-3 = mode : 0=intro , 1=command reference , 2=appendice ,
                 not here: 3=external helpers , 4=key map help text
*/
{
 static Asagent_helplinE intro_text[]= {
"",
"",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
};
 static Asagent_helplinE reference_text[]= {
"",
"                Command reference of application commands",
"",
"-help",
"    Print this text.",
"",
"-log_file:[[+]fileaddress | '-']",
"    Direct loging to the given file address. If the address is preceded by",
"    a '+' sign then logging is done to both file and the agent's stdout.",
"    If the address is a single '-' sign then logging is done to stdout only.",
"    Since this command can impose a severe security problem it is a member",
"    of the permission class -security.",
"",
"-log_limit:number_of_bytes",
"    Truncation limit for logging of a single message",
"",
"-proxy_port:portnumber",
"    Set service port number. This is the port where the client may reach",
"    the proxy. The proxy service at the formerly used port ends and is",
"    resumed at the new port.",
"    If no portnumber has been set during processing of program arguments",
"    has been set, then the service will start at an arbitrary portnumber.",
"    Watch out for the start message to get this number.",
"",
"-read_delay:microseconds",
"    Set the time to be idle after a successful read-write cycle.",
"    On systems which deliver small portions of data but also buffer at",
"    least 1 KByte this delay can decrease processor load without reducing",
"    throughput if the delay time is less than 1/R seconds. R is the",
"    maximum throughput in KBytes per second.",
"",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
};
 static Asagent_helplinE appendice_text[]= {
"",
"",
"                      Appendix application P :  Permission Classes",
"",
"This is a list of application permission classes and their commands.",
"see \"Appendix agent P\" for more information about permission classes.",
"",
"  Class       Command (resp. Action)",
"------------------------------------------------------------------------",
"",
"  security    -log_file",
"  service     -proxy_port , -read_delay",
"  log         -log_limit",
"  help        -help",
"",
"------------------------------------------------------------------------",
"",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
};

 Asagent_helplinE *helptext;
 int mode,ret,i;
 char line[SfileadrL];

 if(Sregex_string(reply,"",0)<=0)
   {ret= -1; goto ex;};
 mode= flag&15;
 if(mode==0)
   helptext= intro_text;
 else if(mode==1)
   helptext= reference_text;
 else if(mode==2)
   helptext= appendice_text;
 else
   {ret= 0; goto ex;}
 for(i=0;strcmp(helptext[i],
                "@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
                )!=0;i++) {
   sprintf(line,"%s\n",helptext[i]);
   if(Sregex_string(reply,line,1)<=0)
     {ret= -1; goto ex;}
 }
 ret= 1;
ex:;
 return(ret);
}



int Sproxy_cmd_help(prx,flag)
struct SproxY *prx;
int flag;
{
 int ret,i;
 char *helptext= NULL;
 struct AsagenT *agent;

 agent= prx->agent;
 ret= Asagent_get_permission(agent,"help","help",0);
 if(ret<=0)
   goto ex;
 Asagent_stdout(agent,"",1);
 Asagent_stdout(agent,"",1);
 Asagent_stdout(agent,"                Features and Options of ",0);
 Asagent_stdout(agent,agent->progname,1);
 Asagent_stdout(agent,"",1);
 for(i=0;i<3;i++) {
   ret= Sproxy_helptext(prx,&helptext,i);
   if(ret<=0)
     goto ex;
   Asagent_stdout(agent,helptext,4);
 }
 Asagent_stdout(agent,"See also:  sagent -help",1);

/*
 for(i=0;i<3;i++) {
   ret= Asagent_helptext(agent,&helptext,i);
   if(ret<=0)
     goto ex;
   Asagent_stdout(agent,helptext,4);
 }
*/
 Asagent_stdout(agent,"",1);
 ret= 1;
ex:
 Sregex_string(&helptext,NULL,0);
 return(ret);
}


Sproxy_cmd_log_file(prx,text,flag)
struct SproxY *prx;
char *text;
int flag;
{
 int ret,i,value;
 struct AsagenT *agent;
 char line[161],*cpt;

 agent= prx->agent;
 ret= Asagent_get_permission(agent,"file","log_file",0);
 if(ret<=0)
   return(ret);
 cpt= text;
 Httpproxy_set_log_tee(prx->httpproxy,0,0);
 if(*cpt=='+') {
   Httpproxy_set_log_tee(prx->httpproxy,1,0);
   cpt++;
 }
 if(Sregex_string(&(prx->logfile),cpt,0)<=0)
   return(-1);
 if(Httpproxy_set_logfile(prx->httpproxy,cpt,0)<=0)
   return(-1);
 sprintf(line,"Log file now : ");
 Asagent_stderr(agent,line,0);
 Asagent_stderr(agent,cpt,1);
 if(Httpproxy_ret_log_tee(prx->httpproxy,0))
   Asagent_stderr(agent,"(also mirrored to stdout)\n",1);
 return(1);
}


Sproxy_cmd_log_limit(prx,text,flag)
struct SproxY *prx;
char *text;
int flag;
{
 int ret,i,value;
 struct AsagenT *agent;
 char line[161];

 agent= prx->agent;
 ret= Asagent_get_permission(agent,"log","log_limit",0);
 if(ret<=0)
   return(ret);
 value= Httpproxy_ret_byte_count_limit(prx->httpproxy,0);
 sscanf(text,"%ld",&value);
 if(value<0)
   return(0);
 Httpproxy_set_byte_count_limit(prx->httpproxy,value,0);
 sprintf(line,"Log limit now %ld bytes\n",value);
 Asagent_stderr(agent,line,1);
 return(1);
}


Sproxy_cmd_proxy_port(prx,text,flag)
struct SproxY *prx;
char *text;
int flag;
{
 int ret,i,value;
 struct AsagenT *agent;
 char line[161];

 agent= prx->agent;
 ret= Asagent_get_permission(agent,"service","proxy_port",0);
 if(ret<=0)
   return(ret);
 value= prx->portnumber;
 sscanf(text,"%ld",&value);
 if(value<0 || value>65535) {
   sprintf(line,"This value not allowed [0..65535]: %ld",value);
   Asagent_stderr(agent,line,1);
   return(0);
 }
 prx->portnumber= value;
 /* Sproxy_stop(prx,0); */
 ret= Sproxy_start(prx,0);
 if(ret<=0) {
   sprintf(line,"sproxy START FAILED at port %ld",prx->portnumber);
   Asagent_stderr(agent,line,1);
   return(ret);
 }
 sprintf(line,"sproxy ready for service at port %ld\n",prx->portnumber);
 Asagent_stderr(agent,line,1);
 return(1);
}


Sproxy_cmd_read_delay(prx,text,flag)
struct SproxY *prx;
char *text;
int flag;
{
 int ret,i,value;
 struct AsagenT *agent;
 char line[161];

 agent= prx->agent;
 ret= Asagent_get_permission(agent,"service","read_delay",0);
 if(ret<=0)
   return(ret);
 value= prx->portnumber;
 sscanf(text,"%ld",&value);
 if(value<0 || value>1000000) {
   sprintf(line,"This value not allowed [0..1000000]: %ld",value);
   Asagent_stderr(agent,line,1);
   return(0);
 }
 value= Httpproxy_ret_microsleep(prx->httpproxy,0);
 sscanf(text,"%ld",&value);
 if(value<0)
   return(0);
 Httpproxy_set_microsleep(prx->httpproxy,value,0);
 sprintf(line,"Read delay now %ld microseconds\n",value);
 Asagent_stderr(agent,line,1);
 return(1);
}


int Sproxy_interpreter(prx,flag)
struct SproxY *prx;
int flag;
/*
 bit0= quiet : do not report command
 bit1= with bit0 : do not report user
 bit2= client mode (call Asagent_interpreter() and return)
 bit3= forced local mode
*/
/*
 return values:
  <=0 = error
  1 = all went normal
  2 = redo user interaction
  3 = end agent service
  4 = command not recognized
*/
{
 int ret,i;
 char *cmd= NULL,line[SfileadrL];
 struct AsagenT *agent;

 agent= prx->agent;
 if(agent==NULL)
   {ret= 0; goto ex;}
 if(((flag&4) || Asagent_get_service_mode(agent,0)==1) && !(flag&8)) {
   /* client-mode: there are no non-remote commands in SproxY */
   ret= Asagent_interpreter(agent,4|(flag&3));
   goto ex;
 }
 Asagent_get_event_connection(agent,1);          /* also set agent->input_fd */
 ret= Asagent_get_event_cmd(agent,&cmd,0);
 if(ret<=0)
   goto ex;
 if(strncmp(cmd,"--sproxy",8)==0) {
   /* remove own module prefix if there */
   for(i=8;cmd[i]!=0;i++)
     cmd[i-8]= cmd[i];
   cmd[i]= 0;
 }
 if(!(flag&1)) {
   sprintf(line,"%ld> ",Asagent_get_event_connection(agent,0));
   Asagent_stdout(agent,line,0);
   Asagent_stdout(agent,cmd,1);
 }
 if(cmd[0]==0||cmd[0]=='#'||cmd[0]=='!') {
   ;
 } else if(strncmp(cmd,"-help",5)==0) {
   ret= Sproxy_cmd_help(prx,0);
   if(ret<=0)
     goto ex;
 } else if (strncmp(cmd,"-log_file",9)==0 && cmd[9]!=0 && cmd[9]!='_') { 
   ret= Sproxy_cmd_log_file(prx,cmd+10,0);
   if(ret<=0)
     goto ex;
 } else if (strncmp(cmd,"-log_limit",10)==0 && cmd[10]!=0 && cmd[10]!='_') { 
   ret= Sproxy_cmd_log_limit(prx,cmd+11,0);
   if(ret<=0)
     goto ex;
 } else if (strncmp(cmd,"-proxy_port",11)==0 && cmd[11]!=0 && cmd[11]!='_') { 
   ret= Sproxy_cmd_proxy_port(prx,cmd+12,0);
   if(ret<=0)
     goto ex;
 } else if (strncmp(cmd,"-read_delay",11)==0 && cmd[11]!=0 && cmd[11]!='_') { 
   ret= Sproxy_cmd_read_delay(prx,cmd+12,0);
   if(ret<=0)
     goto ex;

 /* >>> } else if (strncmp(cmd,"-xyz",4)==0 && cmd[4]!=0 && cmd[4]!='_') { */

 } else {
   ret= Asagent_interpreter(agent,1|2|(flag&(4|8)));
   if(ret<=0 || ret==3)
     goto ex;
   if(ret==4) {
     Asagent_stderr(agent,"--- command not recognized : '",0);
     Asagent_stderr(agent,cmd,0);
     Asagent_stderr(agent,"'",1);
     goto ex;
   }
 }
 if(ret==0)
   ret= 1;

 ret= 1;
ex:;
 Sregex_string(&cmd,NULL,0);
 return(ret);
}


int Sproxy_handle_request(prx,flag)
struct SproxY *prx;
int flag;
{
 int ret,fd;

 ret= Stcpserv_accept_fd(prx->serv,&fd,0);
 if(ret<0)
   return(ret);
 if(ret==0)
   return(2);
 ret= Httpproxy_handle_request(prx->httpproxy,fd,prx->logfile,0);
 if(ret<=0)
   goto ex;
 Httpproxy_destroy_lists(prx->httpproxy,0);

 ret= 1;
ex:;
 return(ret);
}


int Sproxy_loop(prx,flag)
struct SproxY *prx;
int flag;
{
 int ret;

 prx->msg[0]= 0;
 while(1) {
   ret= Asagent_loop(prx->agent,1);
   if(ret<=0) 
     goto ex;
   if(ret==3)
     goto ex;
   if(ret!=5) 
 continue;
   ret= Sproxy_handle_request(prx,0);
   if(ret<0) 
     goto ex;
   if(ret==0 && prx->msg[0]!=0)
     Asagent_stderr(prx->agent,prx->msg,1);
 }

 ret= 1;
ex:
 if(ret<=0 && prx->msg[0]!=0)
   Asagent_stderr(prx->agent,prx->msg,1);
 return(ret);
}



/* --------------------------------------------------------------- */

static struct AsagenT *agent= NULL;


void Cleanup_handler_pipe()
{
 return;
}


void Cleanup_handler_exit()
{
 alarm(0);
 Asagent_stderr(agent,"\n\nTIMEOUT ! Program aborted\n\n",1);
 Asagent_destroy(&agent,0);
 exit(4);
}


main(argc,argv)
int argc;
char **argv;
{
 int ret,ard= 1;
 struct SproxY *prx= NULL;
 char line[1024];
 
 signal(SIGPIPE,Cleanup_handler_pipe);
 signal(SIGALRM,Cleanup_handler_exit);

 fprintf(stdout,"sproxy %s starting up\n\n",Sproxy_versioN);

 /* full memory supervision */
 Smem_set_record_items(1);

 ret= Asagent_new(&agent,argv[0],1,0);
 if(ret<=0) {
   fprintf(stderr,"%s : creation of agent object failed\n",argv[0]);
   ret= 1; goto ex;
 }
 ret= Sproxy_new(&prx,agent,"localhost",0,"-",0);
 if(ret<=0) {
   fprintf(stderr,"%s : creation of sproxy object failed\n",argv[0]);
   ret= 4; goto ex;
 }
 Asagent_attach_app(agent,prx,Sproxy_interpreter,NULL,NULL,NULL,NULL,0);

 ard= 1;
 ret= Asagent_read_options(agent,argc,argv,&ard,0);
 if(ret<=0) {
   ret= 2; goto ex;
 }
 if(ret==3)
   goto ok;

 if(prx->serv==NULL) {
   ret= Sproxy_start(prx,0);
   if(ret<=0)
     {ret= 4; goto ex;}
 }

 sprintf(line,"sproxy ready for service at port %ld\n",prx->portnumber);
 Asagent_stderr(agent,line,1);
 ret= Sproxy_loop(prx,0);
 if(ret<=0) {
   fprintf(stderr,"%s : event loop aborted\n",argv[0]);
   ret= 3; goto ex;
 }
 if(ret==3)
   goto ok;

ok:;
 ret= 0;
ex:
 Sproxy_destroy(&prx,0);
 Asagent_destroy(&agent,0);

 Smem_stderr(1|2);
 exit(ret);
}

