
/*
 cc -g -c imgdbf.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "imgsamp.h"
#include "imgexempt.h"

#include "imgdb.h"
#include "imgdbf.h"


/* -------------------------------- ImgdbF ------------------------------ */


int Imgdbf_new(imdbf,imdb,flag)
struct ImgdbF **imdbf;
struct ImgdB *imdb;
int flag;
{
 int ret;
 struct ImgdbF *m;

 *imdbf= m= TSOB_FELD(struct ImgdbF,1);
 if(*imdbf==NULL)
   return(-1);
 m->imdb= imdb;
 m->file_nam[0]= 0;
 m->fp_nam= NULL;
 m->file_samp[0]= 0;
 m->fp_samp= NULL;
 m->searchpos= 0;
 m->file_exempt[0]= 0;
 m->fp_exempt= NULL;
 return(1);
failed:;
 Imgdbf_destroy(imdbf,0);
 return(-1);
}


int Imgdbf_destroy(imdbf,flag)
struct ImgdbF **imdbf;
int flag;
{
 struct ImgdbF *m;
 
 m= *imdbf;
 if(m==NULL)
   return(0);
 Imgdbf_fclose(m,1);
 Imgdbf_fclose(m,2);
 if(m->fp_exempt!=NULL)
   fclose(m->fp_exempt);
 Smem_freE((char *) m);
 *imdbf= NULL;
 return(1);
}


int Imgdbf_copy(source,target,flag)
struct ImgdbF *source,*target;
int flag;
{
 strcpy(target->file_nam,source->file_nam);
 Imgdbf_fclose(target,1);
 if(source->fp_nam!=NULL) {
   /* >>> open target->fp_nam in same state as source->fp_nam */; 
 }
 strcpy(target->file_samp,source->file_samp);
 Imgdbf_fclose(target,2);
 if(source->fp_samp!=NULL) {
   /* >>> open target->fp_samp in same state as source->fp_samp */; 
 }
 target->searchpos= source->searchpos;
 strcpy(target->file_exempt,source->file_exempt);
 if(target->fp_exempt!=NULL)
   {fclose(target->fp_exempt); target->fp_exempt= NULL;}
 if(source->fp_samp!=NULL) {
   /* >>> open target->fp_exempt in same state as source->fp_exempt */; 
 }
 return(1);
}


int Imgdbf_fopen(imdbf,mode,flag)
struct ImgdbF *imdbf;
char *mode;
int flag;
/*
 bit0-3: mode for Imgdb_get_mapname()
         1= name of list of file names
         2= name of list of samples
 bit14= create in any case
 bit15= close only, do not open new file pointer
*/
{
 char filename[SfileadrL],testname[SfileadrL],*name_pt;
 int typ;
 FILE **fp_pt;

 typ= flag&15;
 if(typ!=1 && typ!=2)
   typ= 1;
 
 Imgdb_get_mapname(imdbf->imdb,filename,typ);
 if(typ==1) {
   fp_pt= &(imdbf->fp_nam);
   name_pt= imdbf->file_nam;
 } else {
   fp_pt= &(imdbf->fp_samp);
   name_pt= imdbf->file_samp;
 }
 if(*fp_pt!=NULL){
   fclose(*fp_pt); 
   *fp_pt= NULL;
   name_pt[0]= 0;
 }
 if(!(flag&(1<<15))){
   if(!Imgmap_get_guarding(imdbf->imdb->imp,0)) {
     Imgdb_get_mapname(imdbf->imdb,testname,5);
     if(Sfile_size(testname,1)!=-1) {
       Imgdb_get_mapname(imdbf->imdb,testname,0);
       sprintf(imdbf->imdb->errmsg,
               "Access to guarded map file '%s' is disabled.",testname);
       goto ex;
     }
   }
   if(flag&(1<<14))
     if((*fp_pt= fopen(filename,"ab"))!=NULL)
       {fclose(*fp_pt); *fp_pt= NULL;}
   *fp_pt= fopen(filename,mode);
   if(*fp_pt==NULL)
     sprintf(imdbf->imdb->errmsg,
             "cannot open map file '%s' errno=%ld",filename,errno);
   else
     strcpy(name_pt,filename);
 }
ex:;
 return(*fp_pt!=NULL);
}


int Imgdbf_fclose(imdbf,flag)
struct ImgdbF *imdbf;
int flag;
/*
 bit0-3: mode for Imgdb_get_mapname()
         1= name of list of file names
         2= name of list of samples
*/
{
 Imgdbf_fopen(imdbf,"",(flag&15)|(1<<15));
 return(1);
}


int Imgdbf_is_open(imdbf,flag)
struct ImgdbF *imdbf;
int flag;
/*
 bit0-3: mode like for Imgdb_get_mapname()
         0= both filepointers
         1= list of file names
         2= list of samples
*/
{
 int mode,ret;

 mode= (flag&15);
 if(mode==2)
   return(imdbf->fp_samp!=NULL);
 else if(mode==1)
   return(imdbf->fp_nam!=NULL);
 ret= Imgdbf_is_open(imdbf,1) && Imgdbf_is_open(imdbf,2);
 return(ret);
}


int Imgdbf_db_exists(imdbf,flag)
struct ImgdbF *imdbf;
int flag;
{
 char filename[SfileadrL];

 Imgdb_get_mapname(imdbf->imdb,filename,1);
 if(Sfile_size(filename,1)==-1)
   return(0);
 return(1);
}


int Imgdbf_fseek(imdbf,pos,flag)
struct ImgdbF *imdbf;
unsigned long pos;
int flag;
/*
 bit0-3: mode for Imgdb_get_mapname()
         1= name of list of file names
         2= name of list of samples
*/
{
 if((flag&15)==2) {
   fseek(imdbf->fp_samp,pos,0);
 } else {
   fseek(imdbf->fp_nam,pos,0);
 }
 return(1);
}


int Imgdbf_ftell(imdbf,pos,flag)
struct ImgdbF *imdbf;
unsigned long *pos;
int flag;
/*
 bit0-3: mode for Imgdb_get_mapname()
         1= name of list of file names
         2= name of list of samples
*/
{
 if((flag&15)==2) {
   *pos= ftell(imdbf->fp_samp);
 } else {
   *pos= ftell(imdbf->fp_nam);
 }
 return(1);
}


int Imgdbf_activate_searchpos(imdbf,flag)
struct ImgdbF *imdbf;
int flag;
{
 int ret;

 ret= Imgdbf_fseek(imdbf,imdbf->searchpos,2);
 return(ret);
}


int Imgdbf_preserve_searchpos(imdbf,flag)
struct ImgdbF *imdbf;
int flag;
/*
 bit15= reset
*/
{
 if(flag&(1<<15)) {
   imdbf->searchpos= 0;
   return(1);
 }
 Imgdbf_ftell(imdbf,&(imdbf->searchpos),2);
 return(1);
}


int Imgdbf_address_samp(imdbf,samp,sadr,flag)
struct ImgdbF *imdbf;
struct ImgsamplE *samp;
int flag;
{
 int ret;

 if(imdbf->fp_samp==NULL) {
   ret= Imgdbf_fopen(imdbf,"r+b",2);
   if(ret<=0)
     return(ret);
 }
 fseek(imdbf->fp_samp,samp->samp_adr,0);
 return(1);
}


int Imgdbf_write_nam(imdbf,filename,nam_adr,flag)
struct ImgdbF *imdbf;
char *filename;
unsigned long *nam_adr;
int flag;
/*
 bit0= address with  *nam_adr 
 bit15= make entry at nam_adr invalid (write a zero)
*/
{
 int ret;
 char null=0;

 if(flag&1) 
   fseek(imdbf->fp_nam,*nam_adr,0);
 if(nam_adr!=NULL)
   *nam_adr= ftell(imdbf->fp_nam);
 if(flag&(1<<15) || filename==NULL) {
   ret= fwrite(&null,1,1,imdbf->fp_nam);
   if(ret!=1) 
     return(0);
 }else{
   ret= fwrite(filename,strlen(filename),1,imdbf->fp_nam);
   if(ret!=1)
     return(0);
   fputc('\n',imdbf->fp_nam);
 }
 return(1);
}


int Imgdbf_write_samp(imdbf,writesample,sadr,flag)
struct ImgdbF *imdbf;
struct ImgsamplE *writesample;
unsigned long sadr;
int flag;
/*
 bit0= address with writesample->samp_adr
 bit1= sadr addressing is already done

 bit15= mark written sample as completely invalid
*/
{
 int i,ret,head,sp;

 if(!(flag&2)) {
   if(flag&1) {
     if((writesample->invalid&1) && !(flag&(1<<15))) {
       sprintf(imdbf->imdb->errmsg,
               "No valid sample address set in Imgdbf_write_samp().");
       return(0);
     }
     fseek(imdbf->fp_samp,writesample->samp_adr,0);
   }
   sadr= ftell(imdbf->fp_samp);
 }
 if(flag&(1<<15)) 
   writesample->invalid= -1;
 Imgsample_set_samp_adr(writesample,sadr,0);

 head= writesample->invalid;
 if(fwrite(&head,sizeof(int),1,imdbf->fp_samp)!=1)
   return(0);
 if(fwrite(&(writesample->id),sizeof(writesample->id),1,imdbf->fp_samp)!=1) 
   return(0);
 if(fwrite(&(writesample->nam_adr),sizeof(writesample->nam_adr),1,
           imdbf->fp_samp)!=1) 
   return(0);
 if(fwrite(&(writesample->crc),sizeof(writesample->crc),1,imdbf->fp_samp)!=1) 
   return(0);
 sp= imdbf->imdb->spoints;
 if(fwrite(writesample->red,sp*sizeof(Imgsamp_coorD),1,imdbf->fp_samp)!=1)
   return(0);
 if(fwrite(writesample->green,sp*sizeof(Imgsamp_coorD),1,imdbf->fp_samp)!=1)
   return(0);
 if(fwrite(writesample->blue,sp*sizeof(Imgsamp_coorD),1,imdbf->fp_samp)!=1)
   return(0);
 ret= 1;
ex:;
 return(ret);
}


int Imgdbf_record(imdbf,filename,writesample,flag)
struct ImgdbF *imdbf;
char *filename;
struct ImgsamplE *writesample;
int flag;
{
 int ret,i;
 char name[SfileadrL];
 unsigned long nam_adr,samp_adr;

 /* filename */
 ret= Imgdbf_fopen(imdbf,"ab",1);
 if(ret<=0)
   return(ret);
 if(Imgdbf_write_nam(imdbf,filename,&nam_adr,0)<=0)
   goto write_err;
 Imgdbf_fclose(imdbf,1);
 Imgsample_set_nam_adr(writesample,nam_adr,0);

 /* sample */
 ret= Imgdbf_fopen(imdbf,"ab",2);
 if(ret<=0)
   return(ret);
 if(Imgdbf_write_samp(imdbf,writesample,0,0)<=0)
   goto write_err;
 Imgdbf_fclose(imdbf,2);
 return(1);

write_err:;
 Imgdbf_fclose(imdbf,1);
 Imgdbf_fclose(imdbf,2);
 sprintf(imdbf->imdb->errmsg,"cannot write name to map file '%s' errno=%ld",
                             name,errno);
 return(0);
}


int Imgdbf_read_samp(imdbf,readsample,flag)
struct ImgdbF *imdbf;
struct ImgsamplE *readsample;
int flag;
/*
  bit0= address with readsample->samp_adr
*/
{
 int i,sp,ret;
 unsigned long samp_adr;

 Imgsample_set_filename(readsample,NULL,0);
 if(flag&1)
   fseek(imdbf->fp_samp,readsample->samp_adr,0);
 readsample->samp_adr= ftell(imdbf->fp_samp);
 if(fread(&(readsample->invalid),sizeof(readsample->invalid),1,
          imdbf->fp_samp)!=1)
   return(0);
 if(fread(&(readsample->id),sizeof(readsample->id),1,imdbf->fp_samp)!=1)
   return(0);
 if(fread(&(readsample->nam_adr),sizeof(readsample->nam_adr),1,
          imdbf->fp_samp)!=1)
   return(0);
 if(fread(&(readsample->crc),sizeof(readsample->crc),1,imdbf->fp_samp)!=1)
   return(0);
 sp=readsample->spoints= imdbf->imdb->spoints;
 if(fread(readsample->red,sp*sizeof(Imgsamp_coorD),1,imdbf->fp_samp)!=1)
   return(0);
 if(fread(readsample->green,sp*sizeof(Imgsamp_coorD),1,imdbf->fp_samp)!=1)
   return(0);
 if(fread(readsample->blue,sp*sizeof(Imgsamp_coorD),1,imdbf->fp_samp)!=1)
   return(0);
 return(1);
}


int Imgdbf_read_nam(imdbf,readsample,flag)
struct ImgdbF *imdbf;
struct ImgsamplE *readsample;
int flag;
/*
  bit0= address with readsample->samp_adr
*/
{
 int i,sp,ret,c;
 unsigned long samp_adr;
 char filename[SfileadrL];

 if(!Imgdbf_is_open(imdbf,1)) {
   ret= Imgdbf_fopen(imdbf,"rb",1);
   if(ret<=0)
     goto ex;
 }
 if(flag&1)
   fseek(imdbf->fp_nam,readsample->nam_adr,0);
 i=0;
 while(i<SfileadrL-1){
   c= fgetc(imdbf->fp_nam);
   if(c==EOF) {
     Imgdb_get_mapname(imdbf->imdb,filename,1);
     sprintf(imdbf->imdb->errmsg,
             "incomplete filename in map file '%s' errno=%ld",filename,errno);
     filename[i++]= 0;
     ret= 0; goto ex;
   }
   if(c=='\r')
 continue;
   if(c=='\n') c= 0;
   filename[i++]= c;
 if(c==0) break;
 }
 filename[SfileadrL-1]= 0;
 ret= Imgsample_set_filename(readsample,filename,0);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 return(ret);
}


int Imgdbf_truncate(imdbf,name,length,flag)
struct ImgdbF *imdbf;
char *name;
unsigned long length;
int flag;
{
 if(truncate(name,length)==-1){
   sprintf(imdbf->imdb->errmsg,
           "on deleting last entry: truncate(,%ld) failed, errno=%ld",
           length,errno);
   sprintf(imdbf->imdb->errmsg+strlen(imdbf->imdb->errmsg),
           "\nin imgdbf file '%s'\n",name);
   return(0);
 }
 return(1);
}


int Imgdbf_del_back(imdbf,samp,flag)
/* delete sample and everything that has a higher address */
struct ImgdbF *imdbf;
struct ImgsamplE *samp;
int flag;
{
 int ret;
 char name[SfileadrL];

 Imgdb_get_mapname(imdbf->imdb,name,2);
 if(Imgdbf_truncate(imdbf,name,samp->samp_adr,0)<=0)
   {ret= 0; goto ex;}
 Imgdb_get_mapname(imdbf->imdb,name,1);
 if(Imgdbf_truncate(imdbf,name,samp->nam_adr,0)<=0)
   {ret= 0; goto ex;}
 ret= 1;
ex:;
 return(ret);
}


int Imgdbf_rm_sample(imdbf,samp,flag)
struct ImgdbF *imdbf;
struct ImgsamplE *samp;
int flag;
/*
 bit0= keep files open if operation is successful
 bit1= no need to reopen files
*/
{
 int ret;

 if(imdbf->fp_nam!=NULL || imdbf->fp_samp!=NULL || !(flag&2)) { 
   ret= Imgdbf_fopen(imdbf,"r+b",1);
   if(ret<=0)
     goto ex;
   ret= Imgdbf_fopen(imdbf,"r+b",2);
   if(ret<=0)
     goto ex;
 }
 if(Imgdbf_write_samp(imdbf,samp,0,(1<<15)|1)<=0)
   {ret= 0; goto ex;}
 if(Imgdbf_write_nam(imdbf,"",&(samp->nam_adr),(1<<15)|1)<=0)
   {ret= 0; goto ex;}
 ret= 1;
ex:
 if(!(flag&1)) {
   Imgdbf_fclose(imdbf,1);
   Imgdbf_fclose(imdbf,2);
 }
 return(ret);
}


int Imgdbf_rename(imdbf,samp,new_name,flag)
struct ImgdbF *imdbf;
char *new_name;
struct ImgsamplE *samp;
int flag;
{
 int ret;
 struct ImgsamplE *delete_samp;

 ret= Imgmap_temp_sample(imdbf->imdb->imp,&delete_samp,0);
 if(ret<=0)
   goto ex;

 /* remove old sample */;
 ret= Imgsample_copy(samp,delete_samp,0);
 if(ret<=0)
   goto ex;
 ret= Imgdbf_rm_sample(imdbf,delete_samp,1);
 if(ret<=0)
   goto ex;
 /* record new one */;
 ret= Imgdbf_record(imdbf,new_name,samp,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 Imgsample_destroy(&delete_samp,0);
 return(ret);
}


int Imgdbf_add_exempt(imdbf,filename,lookalikes,flag)
struct ImgdbF *imdbf;
char *filename;
struct LstrinG **lookalikes;
int flag;
{
 int ret;
 struct LstrinG *ls;
 char listname[SfileadrL];

 ret= Imgdb_get_mapname(imdbf->imdb,listname,6);
 if(ret<=0)
   goto ex;
 ret= Sfile_string_to_file(listname,filename,2|4);
 if(ret<=0)
   goto ex;
 for(ls= *lookalikes;ls!=NULL;ls=ls->next) {
   ret= Sfile_string_to_file(listname,ls->text,2|4);
   if(ret<=0)
     goto ex;
 }
 ret= Sfile_string_to_file(listname,"",2|4);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 return(ret);
}


int Imgdbf_exemptlist_is_changed(imdbf,list_seal,seal,flag)
struct ImgdbF *imdbf;
unsigned long list_seal;
unsigned long *seal;
int flag;
{
 int ret;
 struct stat stbuf;
 char listname[SfileadrL];

 *seal= 0;
 ret= Imgdb_get_mapname(imdbf->imdb,listname,6);
 if(ret<=0)
   return(ret);
 if(stat(listname,&stbuf)==-1)
   return(2);
 *seal= stbuf.st_mtime&0xFFFFFFFF;
 if(list_seal == *seal)
   return(0);
 return(1);
}


int Imgdbf_count_exempts(imdbf,count,flag)
struct ImgdbF *imdbf;
int *count;
int flag;
/*
 bit0= do not add 1 for security
*/
{
 int ret;
 char listname[SfileadrL];

 ret= Imgdb_get_mapname(imdbf->imdb,listname,6);
 if(ret<=0)
   return(ret);
 *count= Sfile_line_count(listname,2);
 if(*count<0)
   return(0);
 if(!(flag&1))
   (*count)++;
 return(1);
}


int Imgdbf_read_exempt(imdbf,listname,filename,lstring,flag)
struct ImgdbF *imdbf;
char *listname;
char filename[SfileadrL];
struct LstrinG **lstring;
int flag;
/*
 bit0= resume reading
*/
/*
 return:  1=success      
          0=everything is read (or cannot open file)
         -1=fatal failure
*/
{
 char line[SfileadrL];
 int ret,state;
 
 if(imdbf->fp_exempt==NULL || !(flag&1)) {
   if(imdbf->fp_exempt!=NULL) {
     fclose(imdbf->fp_exempt);
     imdbf->fp_exempt= NULL;
   }
   imdbf->fp_exempt= fopen(listname,"r");
   if(imdbf->fp_exempt==NULL)
     return(0);
   strcpy(imdbf->file_exempt,listname);
 }
 state= 0;
 while(Sfile_fgets(line,sizeof(line),imdbf->fp_exempt)!=NULL) {
   if(line[0]==0) {
     if(state!=0) 
       {ret= 1; goto ex;}
   } else if(state==0) {                            /* first line of a group */
     strcpy(filename,line);
     state= 1;
   } else {                                                /* following line */
     ret= Lstring_new(lstring,line,*lstring,1);
     if(ret<=0)
       {ret= -1; goto ex;}
   }
 }
 ret= 0;
ex:;
 if(ret<=0 && imdbf->fp_exempt!=NULL) {
   fclose(imdbf->fp_exempt);
   imdbf->fp_exempt= NULL;
   imdbf->file_exempt[0]= 0;
 }
 return(ret); 
}


int Imgdbf_garbage_coll(imdbf,flag)
struct ImgdbF *imdbf;
int flag;
{
 struct ImgcmP *icmp;
 struct ImgmaP *imp,*target= NULL,*buff= NULL;
 struct ImgdB *imdb;
 char mapname[SfileadrL],filename[SfileadrL],newname[SfileadrL];
 int i,file_exists=0,ret,highcode,write_dbs_mem;
 struct stat stbuf;

 imdb= imdbf->imdb;
 imp= imdbf->imdb->imp;
 Imgmap_get_icmp(imp,&icmp,0);
 ret= Imgmap_new(&target,icmp,"",1);
 if(ret<=0)
   goto ex;
 ret= Imgmap_new(&buff,icmp,"",1);
 if(ret<=0)
   goto ex;

 Imgmap_delete_cache(imp,0);
 Imgmap_end_query(imp,0);
 Imgmap_copy(imp,target,2|4);
 Imgmap_get_mapname(imp,mapname,0);
 strcat(mapname,"_garb");
 Imgmap_set_mapname(target,mapname,1);

 highcode= Imgdb__mapname_highcode(0);
 for(i=1;i<=highcode;i++){
   if(i==5 || i==7 || i==8)
 continue;
   Imgmap_get_mapname(target,filename,i);
   if(stat(filename,&stbuf)!=-1) {
     if(!file_exists)
       sprintf(imdb->errmsg,"cannot create temporary data base %s\n",mapname);
     sprintf(imdb->errmsg+strlen(imdb->errmsg),"file already existing: %s",
                                             filename);
     file_exists= 1;
   }
 }
 if(file_exists)
   {ret= -1; goto ex;}
                                         /* eventually disable SQL mirroring */
 write_dbs_mem= imdb->write_dbs;
 imdb->write_dbs= 1;
 ret= Imgmap_copy_all(imp,target,0);
                                       /* eventually re-enable SQL mirroring */
 imdb->write_dbs= write_dbs_mem;
 if(ret<=0){
   /* on failure remove partially written files */
   for(i=1;i<5;i++){
     Imgmap_get_mapname(target,filename,i);
     if(stat(filename,&stbuf)!=-1)
       unlink(filename);
   }
   ret= -1; goto ex;
 }
 /* rename old files */
 Imgmap_copy(imp,buff,2|4);
 Imgmap_get_mapname(imp,mapname,0);
 strcat(mapname,"_");
 strcat(mapname,Sfile_datestr(time(0),3));
 Imgmap_set_mapname(buff,mapname,0);
 for(i=1;i<=highcode;i++){
   if(i==5 || i==7 || i==8)
 continue;
   Imgmap_get_mapname(imp,filename,i);
   Imgmap_get_mapname(buff,newname,i);
   if(stat(filename,&stbuf)!=-1) {
     if(rename(filename,newname)==-1){
       sprintf(imdb->errmsg,"cannot rename old file %s \n",mapname);
       for(i=i-1;i>=0;i--){
         Imgmap_get_mapname(imp,filename,i);
         Imgmap_get_mapname(buff,newname,i);
         if(stat(filename,&stbuf)!=-1)
           rename(newname,filename);
       }
       Imgmap_get_mapname(target,mapname,0);
       sprintf(imdb->errmsg+strlen(imdb->errmsg),
               "please rename new data base %s* manually",mapname);
       ret= 0; goto ex;
     }
   }
 }
 /* rename new files */
 for(i=1;i<=highcode;i++){
   if(i==5 || i==7 || i==8)
 continue;
   Imgmap_get_mapname(imp,newname,i);
   Imgmap_get_mapname(target,filename,i);
   if(stat(filename,&stbuf)!=-1)
     rename(filename,newname);
 }
 if(Imgmap_get_caching(imp,0)) {
   Imgmap_close(imp,0);
   ret= Imgmap_load_cache(imp,0);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 Imgmap_destroy(&target,0);
 Imgmap_destroy(&buff,0);
 return(ret);
}


