
/*
 cc -g -c imgmap.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"

#include "imgsamp.h"

#include "imgmatch.h"


/* -------------------------------- ImgmatcH ------------------------------ */


int Imgmatch_new(matcher,flag)
struct ImgmatcH **matcher;
int flag;
{
 int ret;
 struct ImgmatcH *m;

 *matcher= m= TSOB_FELD(struct ImgmatcH,1);
 if(m==NULL)
   return(-1);
 m->search_method= Imgmatch_search_methoD;
 m->color_eps= Imgmatch_color_epS;
 m->bw_eps= Imgmatch_bw_epS;
 m->color_max_misses= Imgmatch_color_max_misseS;
 m->bw_max_misses= Imgmatch_bw_max_misseS;
 m->check_distance= 0;
 m->max_distance_factor= 1.0;
 m->check_ratio= 0; 
 m->ratio_tolerance= 0.0;
 m->cm[0]= m->cm[1]= m->cm[2]= 0.0;
 m->eps= 0;
 m->max_misses= 0;
 
 return(1);
failed:;
 Imgmatch_destroy(matcher,0);
 return(-1);
}


int Imgmatch_destroy(matcher,flag)
struct ImgmatcH **matcher;
int flag;
{
 struct ImgmaP *m;
 
 if(*matcher==NULL)
   return(0);
 Smem_freE((char *) *matcher);
 *matcher= NULL;
 return(1);
}


int Imgmatch_copy(source,target,flag)
struct ImgmatcH *source,*target;
int flag;
{
 int i;

 target->search_method= source->search_method;
 target->color_eps= source->color_eps;
 target->bw_eps= source->bw_eps;
 target->color_max_misses= source->color_max_misses;
 target->bw_max_misses= source->bw_max_misses;
 target->check_distance= source->check_distance;
 target->max_distance_factor= source->max_distance_factor;
 target->check_ratio= source->check_ratio;
 target->ratio_tolerance= source->ratio_tolerance;
 for(i=0;i<3;i++)
   target->cm[i]= source->cm[i];
 target->eps= source->eps;
 target->max_misses= source->max_misses;
 return(1);
}


int Imgmatch_set_eps(matcher,color_eps,bw_eps,flag)
struct ImgmatcH *matcher;
double color_eps,bw_eps;
int flag;
{
 matcher->color_eps= color_eps;
 matcher->bw_eps= bw_eps;
 return(1);
}


int Imgmatch_get_searchpar(matcher,search_method,eps,max_misses,flag)
struct ImgmatcH *matcher;
int *search_method;
double *eps;
int *max_misses;
int flag;
{
 *search_method= matcher->search_method;
 *eps= matcher->eps;
 *max_misses= matcher->max_misses;
 return(1);
}


int Imgmatch_set_match_par(matcher,search_method,
                           color_eps,bw_eps,color_max_misses,bw_max_misses,
                           check_distance,max_distance_factor,
                           flag)
struct ImgmatcH *matcher;
int search_method;
double color_eps,bw_eps;
int color_max_misses,bw_max_misses;
int check_distance;
double max_distance_factor;
int flag;
{
 if(search_method>=0)
   matcher->search_method= search_method;
 if(color_eps>=0.0)
   matcher->color_eps= color_eps;
 if(bw_eps>=0.0)
   matcher->bw_eps= bw_eps;
 if(color_max_misses>=0)
   matcher->color_max_misses= color_max_misses;
 if(bw_max_misses>=0)
   matcher->bw_max_misses= bw_max_misses;
 if(check_distance>=0) {
   matcher->check_distance= check_distance;
   matcher->max_distance_factor= max_distance_factor;
 }
 return(1);
}


int Imgmatch_get_match_par(matcher,search_method,
                           color_eps,bw_eps,color_max_misses,bw_max_misses,
                           check_distance,max_distance_factor,
                           flag)
struct ImgmatcH *matcher;
int *search_method;
double *color_eps,*bw_eps;
int *color_max_misses,*bw_max_misses;
int *check_distance;
double *max_distance_factor;
int flag;
{
 *search_method= matcher->search_method;
 *color_eps= matcher->color_eps;
 *bw_eps= matcher->bw_eps;
 *color_max_misses= matcher->color_max_misses;
 *bw_max_misses= matcher->bw_max_misses;
 *check_distance= matcher->check_distance;
 *max_distance_factor= matcher->max_distance_factor;
 return(1);
}


int Imgmatch_get_cm(matcher,cm,flag)
struct ImgmatcH *matcher;
double cm[3];
int flag;
{
 int i;
 
 for(i=0;i<3;i++)
   cm[i]= matcher->cm[i];
 return(1);
}


int Imgmatch__cmp_pdist(p1,p2)
double *p1,*p2;
{
 if(*p1<*p2)
   return(-1);
 if(*p1>*p2)
   return(1);
 return(0);
}


int Imgmatch__sum_pdist(pdist,spoints,max_misses,dist,flag)
double pdist[];
int spoints,max_misses;
double *dist;
int flag;
{
 int i;
 double d;
 double Sregex__sqrt();

 /* calculate overall distance omitting max_misses of the highest */
 qsort(pdist,spoints,sizeof(double),Imgmatch__cmp_pdist);
 d= 0;
 for(i=0;i<spoints-max_misses;i++)
   d+= pdist[i];
 if(spoints-max_misses>0)
   *dist= Sregex__sqrt(d/((double) (spoints-max_misses)),1e-6,0);
 else
   *dist= 0.0;
 return(1);
}
     

int Imgmatch_test_diff(matcher,insample,samp,max_misses,in_eps,dist,flag)
struct ImgmatcH *matcher;
struct ImgsamplE *insample;
struct ImgsamplE *samp;
int max_misses;
double in_eps;
double *dist;
int flag;
/*
 bit0= determine *dist if match
*/
{
 int i,misses,eps;

 eps= in_eps*Imgsamp_scalE;
 misses= 0;
 for(i=0;i<samp->spoints && misses<=max_misses;i++) {
   if(abs(insample->red[i]-samp->red[i])>eps) {
     misses++;
 continue;
   }
   if(abs(insample->green[i]-samp->green[i])>eps) {
     misses++;
 continue;
   }
   if(abs(insample->blue[i]-samp->blue[i])>eps) {
     misses++;
 continue;
   }
 }
 if(i==samp->spoints && misses<=max_misses) {
   if(flag&1) {
     double absd[3],pdist[Imgsamp_max_sizE];

     /* calculate point distances */
     for(i=0;i<samp->spoints;i++) {
       absd[0]= abs(insample->red[i]-samp->red[i]);
       absd[1]= abs(insample->green[i]-samp->green[i]);
       absd[2]= abs(insample->blue[i]-samp->blue[i]);
       pdist[i]= (absd[0]*absd[0]+absd[1]*absd[1]+absd[2]*absd[2])
                 / ((double) Imgsamp_scalE * (double) Imgsamp_scalE) ;
     }
     Imgmatch__sum_pdist(pdist,samp->spoints,max_misses,dist,0);
   }
   return(1);
 }
 return(0);
}


int Imgmatch_test_double_diff(matcher,insample,samp,max_misses,in_eps,dist,flag)
/* relative to center of mass : invariant against movement in color space */
struct ImgmatcH *matcher;
struct ImgsamplE *insample;
struct ImgsamplE *samp;
int max_misses;
double in_eps;
double *dist;
int flag;
/*
 bit0= determine *dist if match
*/
{
 int i,misses;
 double eps;
 double v[Imgsamp_max_sizE][3],cm[3];
 double Sregex__sqrt();

 misses= 0;
 cm[0]= cm[1]= cm[2]= 0.0;
 eps= in_eps*Imgsamp_scalE;
 for(i=0;i<insample->spoints;i++) {
   cm[0]+= 
   v[i][0]= (double)(insample->red[i]-samp->red[i]);
   cm[1]+= 
   v[i][1]= (double)(insample->green[i]-samp->green[i]);
   cm[2]+= 
   v[i][2]= (double)(insample->blue[i]-samp->blue[i]);
 }
 cm[0]/= (double) insample->spoints;
 cm[1]/= (double) insample->spoints;
 cm[2]/= (double) insample->spoints;
 for(i=0;i<samp->spoints && misses<=max_misses;i++) {
   if(fabs(v[i][0]-cm[0])>eps) {
     misses++;
 continue;
   }
   if(fabs(v[i][1]-cm[1])>eps) {
     misses++;
 continue;
   }
   if(fabs(v[i][2]-cm[2])>eps) {
     misses++;
 continue;
   }
 }
 if(i==insample->spoints && misses<=max_misses) {
   if(flag&1) {
     double absd[3],pdist[Imgsamp_max_sizE],d;

     /* calculate point distances */
     for(i=0;i<samp->spoints;i++) {
       absd[0]= fabs(v[i][0]-cm[0]);
       absd[1]= fabs(v[i][1]-cm[1]);
       absd[2]= fabs(v[i][2]-cm[2]);
       pdist[i]= (absd[0]*absd[0]+absd[1]*absd[1]+absd[2]*absd[2])
                 / ((double) Imgsamp_scalE * (double) Imgsamp_scalE) ;
     }
     Imgmatch__sum_pdist(pdist,samp->spoints,max_misses,dist,0);
   }
   return(1);
 }
 return(0);
}


int Imgmatch_test_scaled_diff(matcher,insample,incm,samp,
                              max_misses,in_eps,dist,flag)
/* center of mass matched by three independend scale factors : 
   invariant 0-centered scaling in color space */
struct ImgmatcH *matcher;
struct ImgsamplE *insample;
double incm[3];
struct ImgsamplE *samp;
int max_misses;
double in_eps;
double *dist;
int flag;
/*
 bit0= determine *dist if match
*/
{
 int i,misses,ret;
 double eps;
 double v[Imgsamp_max_sizE][3],cm[3],scale[3],movement[3];
 static double dark_limit= 16*Imgsamp_scalE;
 static double fixpoint[3]={ 64.0*Imgsamp_scalE,
                             64.0*Imgsamp_scalE, 64.0*Imgsamp_scalE};
 double Sregex__sqrt();

 if(incm[0]<dark_limit && incm[1]<dark_limit && incm[2]<dark_limit) {
   ret= Imgmatch_test_diff(matcher,insample,samp,max_misses,in_eps,0);
   return(ret);
 }
 misses= 0;
 cm[0]= cm[1]= cm[2]= 0.0;
 eps= in_eps*Imgsamp_scalE;
 for(i=0;i<samp->spoints;i++) {
   cm[0]+= samp->red[i];
   cm[1]+= samp->green[i];
   cm[2]+= samp->blue[i];
 }
 cm[0]/= (double) samp->spoints;
 cm[1]/= (double) samp->spoints;
 cm[2]/= (double) samp->spoints;
 if((cm[0]-fixpoint[0])>=dark_limit)
   scale[0]= (incm[0]-fixpoint[0])/(cm[0]-fixpoint[0]);
 else
   scale[0]= 1.0;
 if((cm[1]-fixpoint[1])>=dark_limit)
   scale[1]= (incm[1]-fixpoint[1])/(cm[1]-fixpoint[1]);
 else
   scale[1]= 1.0;
 if((cm[2]-fixpoint[2])>=dark_limit)
   scale[2]= (incm[2]-fixpoint[2])/(cm[2]-fixpoint[2]);
 else
   scale[2]= 1.0;
 movement[0]= (1.0-scale[0])*fixpoint[0];
 movement[1]= (1.0-scale[1])*fixpoint[1];
 movement[2]= (1.0-scale[2])*fixpoint[2];
 for(i=0;i<samp->spoints && misses<=max_misses;i++) {
   if(fabs(((double) samp->red[i])*scale[0]+movement[0]
           -((double) insample->red[i]))>eps) {
     misses++;
 continue;
   }
   if(fabs(((double) samp->green[i])*scale[1]+movement[1]
      -((double) insample->green[i]))>eps) {
     misses++;
 continue;
   }
   if(fabs(((double) samp->blue[i])*scale[2]+movement[2]
      -((double) insample->blue[i]))>eps) {
     misses++;
 continue;
   }
 }
 if(i==samp->spoints && misses<=max_misses) {
   if(flag&1) {
     double absd[3],pdist[Imgsamp_max_sizE],d;

     /* calculate point distances */
     for(i=0;i<samp->spoints;i++) {
       absd[0]= fabs(((double) samp->red[i])*scale[0]+movement[0]
                -((double) insample->red[i]));
       absd[1]= fabs(((double) samp->green[i])*scale[1]+movement[1]
                -((double) insample->green[i]));
       absd[2]= fabs(((double) samp->blue[i])*scale[2]+movement[2]
                -((double) insample->blue[i]));
       pdist[i]= (absd[0]*absd[0]+absd[1]*absd[1]+absd[2]*absd[2])
                 / ((double) Imgsamp_scalE * (double) Imgsamp_scalE) ;
     }
     Imgmatch__sum_pdist(pdist,samp->spoints,max_misses,dist,0);
   }
   return(1);
 }
 return(0);
}


int Imgmatch_is_bw(matcher,insample,flag)
struct ImgmatcH *matcher;
struct ImgsamplE *insample;
int flag;
/*
flag:
 bit0= set eps from *_eps , use color_eps or bw_eps according to bit1
 bit1=  0=color_eps  1=bw_eps
return:
-1= error
 0= no
 1= yes
*/
{
 int ret,i,j;
 int colorful= 0,boring,nearby;
 double d,dd,*cm,eps;

 if(flag&1) {
   if(flag&2)
     matcher->eps= matcher->bw_eps;
   else
     matcher->eps= matcher->color_eps;
 }
 for(i=0;i<3;i++)
   matcher->cm[i]= 0.0;
 cm= matcher->cm;
 colorful= 0;
 boring= 0;
 if(insample==NULL)
   return(0);

 /* overall color average of the sample */
 for(i=0;i<insample->spoints;i++) {
   cm[0]+= insample->red[i];
   cm[1]+= insample->green[i];
   cm[2]+= insample->blue[i];
 }
 for(j=0;j<3;j++)
   cm[j]/= ((double) insample->spoints);

 eps= matcher->eps;
 if(eps>matcher->bw_eps)
   eps+= matcher->bw_eps;
 for(i=0;i<insample->spoints;i++) {

   /* maximum intensity difference within a single pixel */
   if(matcher->search_method==1) {
     d= abs(insample->red[i]-cm[0]-insample->green[i]+cm[1]);
     if(abs(insample->red[i]-cm[0]-insample->blue[i]+cm[2])>d) 
        d= abs(insample->red[i]-cm[0]-insample->blue[i]+cm[2]);
   }else{
     d= abs(insample->red[i]-insample->green[i]);
     if(abs(insample->red[i]-insample->blue[i])>d) 
        d= abs(insample->red[i]-insample->blue[i]);
   }

   /* eventually count as colorful pixel */ 
   if(d/Imgsamp_scalE>eps)
     colorful++;

   /* look for heaps of similar colored pixels */
   nearby= 0;
   for(j=0;j<insample->spoints;j++) {
     if(j==i)
   continue;
     d= abs(insample->red[i]-insample->red[j]);
     dd= abs(insample->green[i]-insample->green[j]);
     if(dd>d)
       d= dd;
     dd= abs(insample->blue[i]-insample->blue[j]);
     if(dd>d)
       d= dd;
     if(d/Imgsamp_scalE<eps)
       nearby++;
   }
   if(nearby >= insample->spoints/2 ) {
     boring= 1;
 break;
   }

 }

 /* if sample is greyish or boring : it is considered bw */
 if(colorful<insample->spoints/2 || boring) 
   return(1);
 return(0);
}


int Imgmatch_choose_tolerances(matcher,insample,flag)
struct ImgmatcH *matcher;
struct ImgsamplE *insample;
int flag;
/*
 bit0= set eps to 256.0
*/
{
 int ret,i;
 
 /* use higher accuracy if insample is boring (e.g. black+white) */
 matcher->eps= matcher->color_eps;
 matcher->max_misses= matcher->color_max_misses;
 if(insample==NULL)
   {ret= 2; goto ex;}
 for(i=0;i<Imgmatch_color_eps_maxtrY;i++) {
   /* the definition of greyness and boringness depends on eps
      therefore it is attempted to approach bw tolerances smoothly */
   ret= Imgmatch_is_bw(matcher,insample,0);
   if(ret<0)
     {ret= -1; goto ex;}
   if(ret==0) 
     {ret= 1; goto ex;}
   matcher->eps*= Imgmatch_color_eps_downgradE;
   if(matcher->eps<matcher->bw_eps)
 break;
 }
 matcher->eps= matcher->bw_eps;
 matcher->max_misses= matcher->bw_max_misses;
 ret= 1;
ex:;
 if(flag&1)
   matcher->eps= 256.0;
 return(ret);
}


int Imgmatch_test(matcher,insample,samp,dist,flag)
struct ImgmatcH *matcher;
struct ImgsamplE *insample;
struct ImgsamplE *samp;
double *dist;
int flag;
/*
 bit0= determine *dist if match
*/
{
 int ret= 0;
 double eps;

 if(matcher->check_distance)
   flag|= 1;
 if(matcher->search_method==1) {
   ret= Imgmatch_test_double_diff(matcher,insample,samp,
                                 matcher->max_misses,matcher->eps,dist,flag&1);
 } else if(matcher->search_method==2) {
   ret= Imgmatch_test_scaled_diff(matcher,insample,matcher->cm,
                                  samp,matcher->max_misses,matcher->eps,dist,
                                  flag&1);
 } else {
   ret= Imgmatch_test_diff(matcher,insample,samp,
                           matcher->max_misses,matcher->eps,dist,flag&1);
 }
 if(ret>0 && matcher->check_distance)
   if(*dist > matcher->eps * matcher->max_distance_factor)
     ret= 0;
 return(ret);
}


