
#include <stdio.h>
#include <string.h>

#include <mysql/mysql.h>

/*

UDF compliant definitions of custom MySQL functions :

  SIMILAR_IMG(sample1,sample2[...,constant_options,...])
     Image sample comparison function. Samples may be provided by program
     similar. Constant_options are the parameters of similar -match_par:
       color_tolerance,bw_tolerance,color_ignore,bw_ignore,method
     The first four are numbers, method is a string 'color_diff' or
     'color_double_diff'.
  SIMILAR_FMT(sampletext,format_name)
     Converts sampletext into the format depicted by format_name.
     Defined names are:
       smp17  ... standard binary format for storage in SQL databases
       smp16  ... image description part of standard ascii format 
                  to become complete for data interchange it must be
                  ammended by a '+' and the image file address. 
                  For the purposes of SIMILAR_FMT this ammendment is
                  not necessary, thoug.
       (native) . (Database server specific high performance format)
       name   ... If sampletext is in standard ascii format, then the
                  fileaddress (part after '+') is returned. Else, an
                  empty text is returned.

To compile:
  gcc -shared -DSmem_no_own_functionS -o similar_udf.so similar_udf.c imgsamp.c

To test open links:
  gcc -DSmem_no_own_functionS -DSimilar_udf_maiN similar_udf.c imgsamp.c

To activate in MySQL, copy similar_udf.so to /usr/lib .
Then execute at the mysql client :
  CREATE FUNCTION similar_img RETURNS INTEGER SONAME "similar_udf.so";
  CREATE FUNCTION similar_fmt RETURNS STRING SONAME "similar_udf.so";

To deactivate (activation survives server restart) :
  DROP FUNCTION similar_img;
  DROP FUNCTION similar_fmt;

*/

#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"

#include "imgsamp.h"

#include "similar_udf.h"


/* ------------------- abducted helpers ---------------- */


int Sregex_string_cut(handle,text,len,flag)
char **handle;
char *text;
int len;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int l=0;
 char *old_handle;

 if((flag&1)&&*handle!=NULL)
   l+= strlen(*handle);
 old_handle= *handle;
 if(text!=NULL) {
   l+= len;
   *handle= TSOB_FELD(char,l+1);
   if(*handle==NULL) {
     *handle= old_handle;
     return(0);
   }
   if((flag&1) && old_handle!=NULL)
     strcpy(*handle,old_handle);
   else
     (*handle)[0]= 0;
   strncat(*handle,text,len);
 } else {
   *handle= NULL;
 }
 if(old_handle!=NULL)
   Smem_freE(old_handle);
 return(1);
}


int Sregex_string(handle,text,flag)
char **handle;
char *text;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int ret,l=0;

 if(text!=NULL)
   l= strlen(text);
 
/* #define Sregex_looking_for_contenT 1 */
#ifdef Sregex_looking_for_contenT
 /* a debugging point if a certain text content has to be caught */
 if(text!=NULL)
   if(strcmp(text,"clear")==0)
     ret= 0;
#endif

 ret= Sregex_string_cut(handle,text,l,flag&1);
 return(ret);
}


int Sregex_hex_value(text,value,length,flag)
char *text;
unsigned long *value;
int length;
int flag;
{
 int single,i;

 *value= 0;
 if(length>8)
   return(-1);
 if(length>strlen(text))
   return(0);
 for(i=0;i<length;i++) {
   (*value)<<= 4;
   if(text[i]>='0' && text[i]<='9')
     single= text[i]-'0';
   else if (text[i]>='A' && text[i]<='F')
     single= text[i]-'A'+10;
   else if (text[i]>='a' && text[i]<='f')
     single= text[i]-'a'+10;
   else
     return(0);
   (*value)|= single;
 }
 return(1); 
}



/* ------------------- internal functions ---------------- */

static int Similar_udf_destroy();


static int Similar_udf_new(udf,func_type,flag)
struct Similar_udF **udf;
int func_type;
int flag;
{
 int i;
 struct Similar_udF *m;

 *udf= m= TSOB_FELD(struct Similar_udF,1);
 if(m==NULL)
   return(-1);
 m->func_type= func_type;
 m->matcher= NULL;
 m->sample[0]= m->sample[1]= NULL;
 m->sample_is_constant[0]= m->sample_is_constant[1]= 0;
 m->result_buffer= NULL;
 for(i=0;i<sizeof(m->payload);i++)
   m->payload[0]= 0;
 return(1);
failed:;
 Similar_udf_destroy(udf,0);
 return(-1);
}


static int Similar_udf_destroy(udf,flag)
struct Similar_udF **udf;
int flag;
{
 int i;

 if(*udf==NULL)
   return(0);
 Imgmatch_destroy(&((*udf)->matcher),0);
 for(i=0;i<2;i++)
   Imgsample_destroy(&((*udf)->sample[i]),0);
 Sregex_string(&((*udf)->result_buffer),NULL,0);
 free((char *) *udf);
 *udf= NULL;
 return(1);
}


static int Similar_udf_decode_sample(udf,data,datalen,idx,flag)
struct Similar_udF *udf;
char *data;
int datalen;
int idx;
int flag;
/*
 bit0= permission to refer to data without copying if native format
       (this must be revoked before the data get freed)
*/
{
 int ret,status,i;
 char text[2*SfileadrL];

 if(udf->sample[idx]==NULL) {
   ret= Imgsample_new(&(udf->sample[idx]),0);
   if(ret<=0)
     {ret= -1; goto ex;}
 }

 /* look for header of native with desirable bytesex */;
 if( datalen==106 && *((unsigned short *) data)==2) {
#ifdef Imgsample_pointer_layeR
   if(flag&1) { /* refer directly to SQL data */
     Imgsample_point_rgb(udf->sample[idx],data+4,0);
     ret= 1; goto ex;
   }
#endif /* Imgsample_pointer_layeR */
   ret= Imgsample_smp17_deformat(udf->sample[idx],data,datalen,0);
 } else {
   ret= Imgsample_decode(udf->sample[idx],data,datalen,udf->payload,0);
 }
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
} 


int Similar_udf_check_format(udf,data,datalen,fmt,message,flag)
struct Similar_udF *udf;
char *data;
int datalen;
char fmt[16];
char message[81];
int flag;
/*
 return:
  0= unknown
  1= binary smp17
  2= ascii smp16
  3= native smp17
  4= name
  5,6,7=  average values of red,green,blue
  8= swapped native smp17
*/
{
 int l,i,ret;

 l= 15;
 if(l>datalen)
   l= datalen;
 strncpy(fmt,data,l);
 fmt[l]= 0;
 if(datalen>=16)
   goto unknown_target_format;
 for(i=0;i<l;i++)
   if(isalpha(fmt[i]))
     fmt[i]= tolower(fmt[i]);
 if(strcmp(fmt,"smp17")==0 || strcmp(fmt,"binary")==0) {
   /* convert to standard binary database sample format */;
   ret= 1; goto ex;
 } else if(strcmp(fmt,"smp16")==0 || strcmp(fmt,"ascii")==0) {
   /* convert to standard ascii sample format */;
   ret= 2; goto ex;
 } else if(strcmp(fmt,"native")==0) {
   /* convert to server specific native sample format */;
   ret= 3; goto ex;
 } else if(strcmp(fmt,"name")==0) {
   /* return name part (empty string if there is none) */;
   ret= 4; goto ex;
 } else if(strcmp(fmt,"red")==0) {
   ret= 5; goto ex;
 } else if(strcmp(fmt,"green")==0) {
   ret= 6; goto ex;
 } else if(strcmp(fmt,"blue")==0) {
   ret= 7; goto ex;
 } else if(strcmp(fmt,"swapped")==0) {
   /* emulate incompatible server's specific native sample format */;
   ret= 8; goto ex;
 }
unknown_target_format:;
 sprintf(message,"SIMILAR_FMT : unknown target format : %s",fmt);
 ret= 0; 
ex:;
 return(ret);
}


/* ------------------- UDF functions ---------------- */


my_bool similar_img_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
 int i,ret,l,check_distance= 0;
 struct Similar_udF *udf= NULL;
 double num= 0.0;
 double color_eps,bw_eps,max_distance_factor;
 int color_max_misses,bw_max_misses,search_method,chosen= 0;
 char numtext[81],method_text[80];
 
 if(args->arg_count<2) {
usage:;
   strcpy(message,
          "usage : SIMILAR_IMG(sample1,sample2[,...,constant_options,...])");
   return(1);
 }
 for(i=0;i<2;i++) {
   if(args->arg_type[i] != STRING_RESULT)
     goto usage;
 }
 if(args->arg_count>2) {
   if(args->args[2]==NULL)
     goto usage;                                  /* argument not a constant */
 }

 ret= Similar_udf_new(&udf,0,0);
 if(ret<=0) {
nomem:;
   strcpy(message,"SIMILAR_IMG : cannot allocate memory.");
   return(1);
 }
 initid->ptr= (char *) udf;

 ret= Imgmatch_new(&(udf->matcher),0);
 if(ret<=0) 
   goto nomem;
 Imgmatch_get_match_par(udf->matcher,&search_method,
                        &color_eps,&bw_eps,&color_max_misses,&bw_max_misses,0);
 for(i=2;i<args->arg_count;i++) {
   /* read option string and modify defaults in matcher object */;
   if(args->args[i]==NULL)
     goto usage;
   num= 0.0;
   if(args->arg_type[i] == STRING_RESULT) {
     if(args->lengths[i]<80)
       strncpy(numtext,args->args[i],args->lengths[i]);
     else
       strcpy(numtext,"-1");
     sscanf(numtext,"%lf",&num);
   } else if(args->arg_type[i] == INT_RESULT)
     num= *((long long *) args->args[i]);
   else
     num= *((double *) args->args[i]);
   if(i==2) 
     color_eps= num;
   else if(i==3)
     bw_eps= num;
   else if(i==4)
     color_max_misses= num;
   else if(i==5)
     bw_max_misses= num;
   else if(i==6) {
     if(args->arg_type[i] == STRING_RESULT) {
       l= sizeof(method_text)-1;
       if(args->lengths[i]<l)
         l= args->lengths[i];
       strncpy(method_text,args->args[i],l);
       method_text[l]= 0;
       if(strcmp(method_text,"color_diff")==0 || 
          strcmp(method_text,"0")==0)
           search_method= 0;
       if(strcmp(method_text,"color_double_diff")==0 || 
          strcmp(method_text,"1")==0)
           search_method= 1;
       if(strcmp(method_text,"color_scaled_diff")==0||
          strcmp(method_text,"2")==0)
           search_method= 2;
     } else {
       if(num>=0 && num<=2)
         search_method= num;
     }
   } else if(i==7) {
     max_distance_factor= num;
     if(max_distance_factor>=0.0)
       check_distance= 1;
     else
       check_distance= 0;
   }
 }
 Imgmatch_set_match_par(udf->matcher,search_method,
                        color_eps,bw_eps,color_max_misses,bw_max_misses,
                        check_distance,max_distance_factor,0);
 for(i=0;i<2;i++) {
   udf->sample_is_constant[i]= 0;
   Imgsample_destroy(&(udf->sample[i]),0);
   if(args->args[i] == NULL)
 continue;
   ret= Similar_udf_decode_sample(udf,args->args[i],args->lengths[i],i,0);
   if(ret<=0) {
unreadable:;
     sprintf(message,"SIMILAR_FMT : unreadable sample format in argument %ld",
                     i+1);
     Similar_udf_destroy((struct Similar_udF **) &(initid->ptr),0);
     return(1);
   }
   Imgmatch_choose_tolerances(udf->matcher,udf->sample[i],0);
   udf->sample_is_constant[i]= 1;
   chosen= 1;
 }
 if(!chosen)
   Imgmatch_choose_tolerances(udf->matcher,NULL,0);
 return(0);
}


void similar_img_deinit(UDF_INIT *initid)
{
 if(initid->ptr!=NULL)
   Similar_udf_destroy((struct Similar_udF **) &(initid->ptr),0);
}


long long similar_img(UDF_INIT *initid, UDF_ARGS *args,
                      char *is_null, char *error)
{
 long long ret;
 int i;
 struct Similar_udF *udf= NULL;
 static int suppress_error= 0;
 double dist;
 
 if(initid->ptr==NULL) {
   *error= 1;
   *is_null= 1;
   return(0);
 }
 udf= (struct Similar_udF *) initid->ptr;
 *error= 0;
 *is_null= 0;

 /* decode both samples */
 for(i=0;i<2;i++) {
   if(udf->sample_is_constant[i] && udf->sample[i]!=NULL)
 continue;
   if(args->args[i]==NULL || args->lengths[i]==0)
     goto unreadable;
   ret= Similar_udf_decode_sample(udf,args->args[i],args->lengths[i],i,1);
   if(ret<=0) {
unreadable:;
     sprintf(udf->payload,"SIMILAR_FMT : unreadable sample format");
     *error= 1;
     goto ex;
   }
 }
 ret= Imgmatch_test(udf->matcher,udf->sample[0],udf->sample[1],&dist,0);
ex:;
 if(suppress_error) 
   *error= 0;
 if(*error!=0) 
   *is_null= 1;
#ifdef Imgsample_pointer_layeR
 /* reset possibly altered rgb pointer layer */
 if(udf->sample[0]->rgb!=udf->sample[0]->red)
   Imgsample_point_rgb(udf->sample[0],udf->sample[0]->rgb,0); 
 if(udf->sample[1]->rgb!=udf->sample[1]->red)
   Imgsample_point_rgb(udf->sample[1],udf->sample[1]->rgb,0); 
#endif /* Imgsample_pointer_layeR */
 return(ret);
}


my_bool similar_fmt_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
 int i,ret;
 struct Similar_udF *udf= NULL;
 char fmt[16];
 
 if(args->arg_count<2 || args->arg_count>2) {
usage:;
   strcpy(message,
          "usage : SIMILAR_FMT(sampletext,format_name)");
   return(1);
 }
 for(i=0;i<args->arg_count;i++) {
   if(args->arg_type[i] != STRING_RESULT)
     goto usage;
 }
 ret= Similar_udf_new(&udf,1,0);
 if(ret<=0) {
   strcpy(message,"SIMILAR_FMT : cannot allocate memory.");
   return(1);
 }
 initid->ptr= (char *) udf;
 if(args->args[1]!=NULL && args->lengths[1]!=0) {
   ret= Similar_udf_check_format(udf,args->args[1],args->lengths[1],
                                 fmt,message,0);
   if(ret<=0) {
     Similar_udf_destroy(&udf,1,0);
     return(1);
   }
 }
 return(0);
}


void similar_fmt_deinit(UDF_INIT *initid)
{
 if(initid->ptr!=NULL)
   Similar_udf_destroy((struct Similar_udF **) &(initid->ptr),0);
}


char *similar_fmt(UDF_INIT *initid, UDF_ARGS *args,
            char *result, unsigned long *length,
            char *is_null, char *error)
{
 long long ret;
 char *result_pt,*text= NULL,fmt[16],tr;
 int i,l,result_length= 0,avg[3],fmt_code;
 struct Similar_udF *udf;
 static int suppress_error= 0;
 
 *error= 0;
 *is_null= 0;
 result_pt= result;
 udf= (struct Similar_udF *) initid->ptr;

 if(args->args[0]==NULL || args->lengths[0]==0)
   goto unreadable;
 ret= Similar_udf_decode_sample(udf,args->args[0],args->lengths[0],0,0);
 if(ret<=0) {
unreadable:;
   /* sprintf(result,"SIMILAR_FMT : %s",udf->payload); */
   sprintf(result,"SIMILAR_FMT : unreadable sample format");
   result_length= strlen(result);
   *error= 1;
   goto ex;
 }
 ret= Similar_udf_check_format(udf,args->args[1],args->lengths[1],
                               fmt,result,0);
 if(ret<=0) {
   result_length= strlen(result);
   *error= 1;
   goto ex;
 }
 fmt_code= ret;
 if(fmt_code==1) {
   /* convert to standard binary database sample format */;
   ret= Imgsample_smp17_format(udf->sample[0],result,&result_length,avg,0);
   if(ret<=0) {
conversion_failed:;
     sprintf(result,"SIMILAR_FMT : conversion to format %s failed.",fmt);
     result_length= strlen(result);
     *error= 1;
     goto ex;
   }
 } else if(fmt_code==2) {
   /* convert to standard ascii sample format */;
   ret= Imgsample_serialize(udf->sample[0],&text,1);
   if(ret<=0) 
     goto conversion_failed;
   strcpy(result,text);
   result_length= strlen(result);
 } else if(fmt_code==3 || fmt_code==8) {
   /* convert to server specific native sample format */;
   ret= Imgsample_smp17_format(udf->sample[0],result,&result_length,avg,
                              2|((fmt_code==8)<<2));
   if(ret<=0) 
     goto conversion_failed;
 } else if(fmt_code==4) {
   /* return name part (empty string if there is none) */;
   if(udf->sample[0]->filename==NULL) {
     result[0]= 0;
     result_length= 0;
   } else {
     if(Sregex_string(&(udf->result_buffer),udf->sample[0]->filename,0)<=0)
       goto conversion_failed;
     result_length= strlen(udf->result_buffer);
     result_pt= udf->result_buffer;
   }
 } else if(fmt_code>=5 && fmt_code<=7) {
   /* average values */
   ret= Imgsample_smp17_format(udf->sample[0],result,&result_length,avg,0);
   if(ret<=0)
     goto conversion_failed;
   sprintf(result,"%ld",avg[fmt_code-5]);
   result_length= strlen(result);
 } else {
unknown_target_format:;
   sprintf(result,
           "SIMILAR_FMT : uncoordinated : unknown target format %s",fmt);
   result_length= strlen(result);
   *error= 1;
   goto ex;
 }
 
ex:;
 if(suppress_error) 
   *error= 0;
 if(*error!=0) 
   *is_null= 1;
 *length= result_length;
 Sregex_string(&text,NULL,0);
 return(result_pt);
}


/* just to test final linkability : the shortest main i ever made
     gcc -DSimilar_udf_maiN  similar_udf.c ; rm a.out
*/

#ifdef Similar_udf_maiN
main() {exit(0);}
#endif /* Similar_udf_maiN */


