
/*
in src/stic_build :
 cc -g -o ../../bin/notify_ip_adr ../misc/notify_ip_adr.c ../smtp/smtpsend.c \
    stcp.o sfile.o sregex.o smem.o
*/


#include <unistd.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"

#define Notify_ip_adr_versioN "0.4"


main(argc,argv)
int argc;
char **argv;
{
 int ret,i,first_try= 1,invalid_read= 0,max_delivery_tries= 4;
 int retry_of_pending_interval= 300; /* set to 0 to disable */
 unsigned int cycle_count= 1, pending_cycle= 0;
 char *result= NULL,*ifmark,*addrmark,*ipt,*ept;
 char address[16],old_address[16],subject[81],sent_by[81],*errmsg;
 struct SmtpsenD *sender= NULL;

 static int test_pending= 0;
 
 if(test_pending)
   retry_of_pending_interval= 30;

 strcpy(old_address,"init");
 if(argc<4) {
   fprintf(stderr,"\n");
   fprintf(stderr,"usage: %s ifconfig_command ifmark addrmark \\\n",argv[0]);
   fprintf(stderr,"                          [smtphost from to [to ...]]\n");
   fprintf(stderr,"Program version: %s\n",Notify_ip_adr_versioN);
   fprintf(stderr,"Watches the IP address of an interface and notifies a\n");
   fprintf(stderr,"list of recipients by e-mail about changes.\n");
   fprintf(stderr,"The ifconfig_command is executed once per second.\n");
   fprintf(stderr,"Within its output the text given by ifmark is searched.\n");
   fprintf(stderr,"The first occurence of addrmark after ifmark indicates\n");
   fprintf(stderr,"the beginning of the IP address.\n");
   fprintf(stderr,"Mail is sent via smtphost using the SMTP protocol.\n");
   fprintf(stderr,"Example (works with SuSE 7.2, english):\n");
   fprintf(stderr,"  %s /sbin/ifconfig 'ppp0      Link' 'inet addr:' \\\n",
                  argv[0]);
   fprintf(stderr,"      mail.gmx.net scdbackup@gmx.net stic-source@gmx.net\n");
   fprintf(stderr,"For connecting to the IP address with ssh, tell it to\n");
   fprintf(stderr,"use the known key for the host (e.g. hostname ts2)\n");
   fprintf(stderr,"  ssh 217.88.170.25 -o 'HostKeyAlias ts2'\n");
   fprintf(stderr,"\n");
   exit(1);
 }

 while(1) {
   ret= Sfile_popen_resultstring(argv[1],&result,0);
   if(ret<=0) {
     if(!first_try)
       goto failed_read;
     fprintf(stderr,"command failed : %s\n",argv[1]);
     fprintf(stderr,"%s\n",strerror(errno));
     exit(2);
   }
   ifmark= strstr(result,argv[2]);
   if(ifmark==NULL) {
     if(!first_try)
       goto failed_read;
     fprintf(stderr,"cannot find ifmark : %s\n",argv[2]);
     fprintf(stderr,"in output text :\n %s\n",result);
     exit(3);
   }
   addrmark= strstr(ifmark,argv[3]);
   if(addrmark==NULL) {
     if(!first_try)
       goto failed_read;
     fprintf(stderr,"cannot find addrmark : %s\n",argv[3]);
     fprintf(stderr,"in output text :\n %s\n",ifmark);
     exit(4);
   }
   ipt= addrmark+strlen(argv[3]);
   for(ept= ipt; (*ept)=='.' || ((*ept)>='0' && (*ept)<='9'); ept++);
   *ept= 0;
   if(ept-ipt>15) {
     if(!first_try)
       goto failed_read;
     fprintf(stderr,"alleged ip address to long : %s\n",ipt);
     exit(5);
   }
   strcpy(address,ipt);

   if(strcmp(old_address,"init")==0) {
     printf("-------------   %s\n",Sfile_datestr(time(0),1|2|256));
     printf("notify_ip_adr %s : startup with ip_address='%s'\n",
            Notify_ip_adr_versioN,address);
     fflush(stdout);
     goto next_cycle;
   }
   if(strcmp(address,old_address)!=0) {
     Smtpsend_destroy(&sender,0); /* any pending notification is void now */
     printf("-------------   %s\n",Sfile_datestr(time(0),1|2|256));
     printf("new ip_address='%s'\n",address);
     fflush(stdout);
     if(invalid_read) {
       invalid_read= 0;
       if(strncmp(address,"192.168",7)==0) {
         printf(
          "looks like a newly started ppp interface. notification omitted.\n");
         fflush(stdout);
         goto next_cycle;
       }
     } else {
       if(strncmp(address,"192.168",7)==0) {
         printf("looks like we are offline now. notification omitted.\n");
         fflush(stdout);
         goto next_cycle;
       }
     }
     if(argc<7) {
       printf("not a single mail address given. notification omitted.\n");
       fflush(stdout);
       goto next_cycle;
     }

     sprintf(subject,"notify_ip_adr : %s",address);
     sprintf(sent_by,"sent by notify_ip_adr %s",Notify_ip_adr_versioN);
     ret= Smtpsend_new(&sender,argv[4],argv[5],argv[6],subject,sent_by,0);
     if(ret<=0) {
err_creation:;
       fprintf(stderr,"%s : creation of mailer object failed\n",argv[0]);
       exit(6);
     }
     for(i=7;i<argc;i++) {
       ret= Smtpsend_add_to(sender,argv[i],0);
       if(ret<=0)
         goto err_creation;
     }
     /* This repetition is for initial difficulties of the connection */
     /* For temporarily unavailable servers : retry_of_pending_interval */
     for(i=0;i<max_delivery_tries;i++) {
       printf("waiting for 15 seconds\n",address);
       fflush(stdout);
       sleep(15); /* wait until the connection is really up */

       if(test_pending)
         ret= 0;
       else
         ret= Smtpsend_deliver(sender,0);
         fflush(stdout);
       if(ret<=0) {
         fprintf(stderr,"%s : delivery of mail failed\n",argv[0]);
         Smtpsend_get_errmsg(sender,&errmsg,0);
         if(errmsg[0])
           fprintf(stderr,"( %s )\n",errmsg);
       } else {
     break;
       }
     }
     fflush(stdout);
     if(i>=max_delivery_tries) {
       if(retry_of_pending_interval>0)
         fprintf(stderr,
           "%s : delivery of pending mail will be retried every %ld seconds\n",
           argv[0],retry_of_pending_interval);
       pending_cycle= cycle_count;
     } else {
       Smtpsend_destroy(&sender,0);
     }
   } else if(invalid_read) {
     printf("-------------   %s\n",Sfile_datestr(time(0),1|2|256));
     printf(
   "missing text marks reappeared with same ip_address. service up again ?\n");
     fflush(stdout);
     invalid_read= 0;
   } else {
     if(sender!=NULL && retry_of_pending_interval>0) 
     if(((cycle_count-pending_cycle+1)%retry_of_pending_interval)==0) {
       printf("-------------   %s\n",Sfile_datestr(time(0),1|2|256));
       printf("trying to deliver pending mail\n");
       ret= Smtpsend_deliver(sender,0);
       fflush(stdout);
       if(ret<=0) {
         fprintf(stderr,"%s : delivery of pending mail failed\n",argv[0]);
         Smtpsend_get_errmsg(sender,&errmsg,0);
         if(errmsg[0])
           fprintf(stderr,"( %s )\n",errmsg);
       } else {
         printf("pending mail sucessfully delivered\n");
         fflush(stdout);
         Smtpsend_destroy(&sender,0);
       }
     }
   }

   if(0) {
failed_read:;
     if(!invalid_read) {
       printf("-------------   %s\n",Sfile_datestr(time(0),1|2|256));
       printf("previously existing text marks vanished. service down ?\n"); 
       fflush(stdout);
     }
     invalid_read= 1;
   }
next_cycle:;
   cycle_count++;
   first_try= 0;
   strcpy(old_address,address);
   if(result!=NULL)
     free(result);
   result= NULL;

   sleep(1);
 }

 exit(0);
}

