
#ifndef Nntpconn_defineD

#define Nntpconn_defineD 

#define Nntpconn_strL 4096

struct NntpconN {

 /* Server */
 char servername[Nntpconn_strL];
 int portnumber;
 int allows_posting;

 /* Dummy server mode */
 int dummy;                           /* server is actually a list of files */
 struct LstrinG *sources;                           /* names of these files */
 struct LstrinG *curr_source;                   /* name of the current file */
 int endseq_mode;                       /* 0=NNTP:'.'  1=Mailfolder:"From " */
 long seek0_offset;
 int eof_encountered;

 /* User */
 char login[Nntpconn_strL];
 char password[Nntpconn_strL];

 /* Socket */
 int sock;
 int active;
 int timeout;
 double slow_server_limit;

 /* For primitive I/O */
 char sent_cmd[Nntpconn_strL];
 char status_line[Nntpconn_strL];
 int status_code;
 char line[Nntpconn_strL];
 int line_valid;
 int endseq_pt;     /* memorizes position in end sequence */
 int zero_data_count;
 int zero_data_max;

 /* status display */
 int show_io_stat;
 double t_bytecount;
 double t_start;
 double t_shown;
 double t_shown_bytecount;
 double t_show_interval;
 int show_io_blink;

 /* bandwidth limitation */
 char baudlimitfile[Nntpconn_strL];
 double r_baudlimit;
 double start_time;
 double r_bytecount;
 double w_bytecount;

 /* For line-mode I/O */
 char buffer[4*1024];
 char *buffrpt;
 int buffer_status; /* 0=nothing read , 1=finished , 
                       2=data pending in NntpconN */

 /* Error Handling */
 char errmsg[Nntpconn_strL];
 int errn; /* >0 = UNIX errno
              <0 = internal errors
                   -1 = result buffer overflow
                   -2 = connection not active (open)
                   -3 = cannot initialize regexp [NntploaD] 
                   -4 = name of output file too long [NntploaD]
                   -5 = malformed status reply [NntpheaD]
                   -6 = read timeout
                   -7 = cannot address group [NntpclienT]
                   -8 = cannot access first message in group [NntpclienT]
                   -9 = message/xover required but no group opened [NntpclienT]
                  -10 = download required but no message opened [NntpclienT]
                  -11 = message already downloaded (no overwrite) [NntpclienT]
                  -12 = bad message range [NntpclienT]
                  -13 = unknown option [NntpclienT]
                  -14 = interrupted by user [NntpclienT]
                  -15 = cannot set regular expression [NntpclienT]
                  -16 = bad value for timeout [NntpclienT]
                  -17 = not enough memory [NntpclienT,NntpposT]
                  -18 = malformed filter
                  -19 = premature end of binary data
                  -20 = cannot get list of groups
                  -21 = creation of memory object failed
                  -22 = auto removing not permitted or max agelimit violated
                  -23 = invalid time code
                  -24 = cannot obtain maintainance lock
                  -25 = server does not permit posting
                  -26 = posting required but no group opened [NntpclienT]
                  -27 = negative reply on ARTICLE HEAD or BODY request
                  -28 = overflow of References: buffer
                  -29 = socket returns empty data
                  -30 = attempt to post empty control message
                  -31 = XOVER not enabled for this server
                  -32 = extension command not supported by server
                  -33 = filesize changed during read_from
                  -34 = buffer overrun prevented
                  -35 = attempt to authenticate failed
                  -36 = non recoverable timeout (made from -6)
                  -37 = server returns much less article bytes than announced
                  -38 = connection broken
                  -39 = unknown host
                  -40 = programming mistake detected
                  -41 = not a hashed directory
                  -42 = already a hashed directory
                  -43 = not a UNIX style mail folder
                  -44 = may not feed mailfolder group via NNTP
            */
 
 /* output redirection functionality */
 FILE *outlet;
 char outlet_name[Nntpconn_strL];

 /* externally managed info about download and indexing files */
 struct NntpfileR  *filer;

 /* >>> this should be obsolete since NntpfileR contains a groupname too */
 /* >>> but NntpgrouP sets and NntpheaD reads this even if there is no filer */
 /* Optional Info: possibly set and read by clients of nntpconn */
 /* >>> one could make an automat which scans transactions for "GROUP" */
 char groupname[Nntpconn_strL];

};

#define Nntpconn_keepalivE /* setsockopt() SO_KEEPALIVE */


static char Nntpconn_endseQ[]= 
              {13, 10, '.', 13, 10, 0, 0, 0, 0, 0, 0 };
static int Nntpconn_endseq_fidX[]= 
              { 0,  0,   0,  0,  0, 0, 0, 0, 0, 0, 0 };
#define Nntpconn_endseqL 11

static char Nntpconn_mailfolder_endseQ[]= 
              {10, 10, 'F', 'r', 'o', 'm', ' ', 0, 0, 0, 0, 0, 0 };
/* where to resume comparison in case of failure at particular position */
/* here : to properly detect the endiseq in   10 10 10 "From "
          a failure at idx=2 'F' has to resume at idx=1 and test for 10
*/
static int Nntpconn_mailfolder_endseq_fidX[]= 
              { 0,  0,   1,   0,   0,   0,   0, 0, 0, 0, 0, 0, 0 };
#define Nntpconn_mailfolder_endseqL 13

#endif /* Nntpconn_defineD */


 
#ifdef Nntpconn_never_defineD


                             Function Overview 


int Nntpconn_new(conn,servername,portnumber,flag)
struct NntpconN **conn;
char *servername;
int portnumber;
int flag;
/*
 bit0= dummy (read from file servername)
 bit1= do not open connection
*/
   Creates a connection object and usually establishes the connection.


int Nntpconn_destroy(conn,flag)
struct NntpconN **conn;
int flag;
   Closes connection and removes connection object form memory.


int Nntpconn_open(conn,flag)
struct NntpconN *conn;
int flag;
   Opens the connection to the preset server and port.


int Nntpconn_close(conn,flag)
struct NntpconN *conn;
int flag;
/*
 bit15= Emergency close : no QUIT, no waiting for answer
*/
   Closes the connection.


int Nntpconn_set_auth(conn,login,password,flag)
struct NntpconN *conn;
char *login,*password;
int flag;
   Set login and password. They are used if the server requires authentication.
   

int Nntpconn_set_outlet(conn,fp,name,flag)
struct NntpconN *conn;
FILE *fp;
char *name;
int flag;
   Set file target for message text output.


int Nntpconn_set_timeout(conn,timeout,flag)
struct NntpconN *conn;
int timeout;
int flag;
   Set connection timeout in seconds.


int Nntpconn_set_slow_server(conn,limit,flag)
struct NntpconN *conn;
double limit;
int flag;
   Set a minimum baudrate which has to be outdone after 2*timeout .


int Nntpconn_set_r_baudlimit(conn,limit,flag)
struct NntpconN *conn;
double limit;
int flag;
   Set a maximum baudrate which will be obeyed when reading data from socket.


int Nntpconn_get_t_count(conn,t_start,t_bytecount,flag)
struct NntpconN *conn;
double *t_start;
double *t_bytecount;
int flag;
   Query the start time and the transferred bytes (read+write).


int Nntpconn_set_t_count(conn,t_start,t_bytecount,flag)
struct NntpconN *conn;
double t_start;
double t_bytecount;
int flag;
   Set the start time and the transferred bytes (read+write).


int Nntpconn_reset_t_count(conn,flag)
struct NntpconN *conn;
int flag;
   Set the start time and the transferred bytes (read+write) to 0.


int Nntpconn_reset_rw_count(conn,flag)
struct NntpconN *conn;
int flag;
   Set the counters for read and written bytes to 0.


int Nntpconn_set_groupname(conn,groupname,flag)
struct NntpconN *conn;
char *groupname;
int flag;
   Set the groupname to be used in further activities.


int Nntpconn_get_groupname(conn,groupname,flag)
struct NntpconN *conn;
char *groupname;
int flag;
   Query goupname.


int Nntpconn_set_filer(conn,filer,flag)
struct NntpconN *conn;
struct NntpfileR *filer;
int flag;
   Attach a NnntpfileR object to the connection object.
   It determines the file system target for downloaded data and adjusts
   HTML linkage.


nt Nntpconn_err_not_active(conn,flag)
struct NntpconN *conn;
int flag;
   Called if data are requested from a connection that is not open.


int Nntpconn_result(conn,line,recv_text,recv_siz,save_name,flag)
struct NntpconN *conn;
char *line;
char *recv_text;
unsigned long recv_siz;
char *save_name;
int flag;
   Dispatches read text to various output channels.


int Nntpconn_check_slow_server(conn,flag)
struct NntpconN *conn;
int flag;
   Checks for a very slow service and eventually causes an error.


int Nntpconn_show_stat(conn,flag)
struct NntpconN *conn;
int flag;
/*
 bit0= clear full line and exit
 bit1= always show cmd
 bit3= display "speed" rather than sent_cmd
 bit4= display with newline
*/
   Displays the connection status line which consists of the current
   NNTP command, the transfer statistics and eventually the idle time.


int Nntpconn_get_stat(conn,t_time,t_bytes,flag)
struct NntpconN *conn;
double *t_time;
double *t_bytes;
int flag;
   Query the transferred bytes and the time which elapsed since start
   of the counting. (Usually start of transaction)


int Nntpconn_select(conn,flag)
struct NntpconN *conn;
int flag;
/*
 bit0+1= mode:  0=read ; 1= write ; 2= exception
 bit2= timeout only 0.1 seconds on first try
 bit3= only display statistics (bit0+1 have to be 0, conn->show_io_stat!=0)
 bit4= display with newline
 bit5= do not check for slow service
*/
   Waits for data to arrive at the connection socket. Eventually displays 
   idle messages and may cause timeout error.


int Nntpconn_send(conn,send_text,flag)
struct NntpconN *conn;
char *send_text;
int flag;
/*
 bit0= raw: do not append crlf
*/
   Sends data to the server. Data usually are complete text lines.


int Nntpconn_recv(conn,line,size,flag)
struct NntpconN *conn;
char *line;
int size;
int flag;
   Receives data from the server.


int Nntpconn_recv_status(conn,flag)
struct NntpconN *conn;
int flag;
/*
 bit0= preliminarily show the incoming characters
*/
   Receives data byte by byte and handles them as part of a status reply.


int Nntpconn_recv_text(conn,recv_text,recv_siz,save_name,flag)
struct NntpconN *conn;
char *recv_text;
unsigned long recv_siz;
char *save_name;
int flag;
/*
 bit0= write reply to stdout
 bit1= write reply to recv_text
 bit2= append reply to file save_name
 bit8= only check wether there are still data expected
*/
   Receives data in larger pieces up to recv_siz and eventually recognizes
   the end of article dot.


int Nntpconn_enforce_limits(conn,filename,flag)
struct NntpconN *conn;
char *filename;
int flag;
/*
 bit0= read configuration files
 bit1= do not wait (just read)
 bit2= do nothing if conn->baudlimit>0
*/
   Enforces the defined upper baud limits by waiting eventually. 
   (called by Nntpconn_transaction)


int Nntpconn_transaction_reset(conn,flag)
struct NntpconN *conn;
int flag;
/*
 bit3= do not reset transaction count
*/
   Prepares for a new transaction.


int Nntpconn_errmsg_reset(conn,flag)
struct NntpconN *conn;
int flag;
   Reset error message.


int Nntpconn_transaction(conn,send_text,recv_text,recv_siz,save_name,flag)
struct NntpconN *conn;
char *send_text;
char *recv_text;
unsigned long recv_siz;
char *save_name;
int flag;
/*
 bit0= write reply to stdout
 bit1= write reply to recv_text
 bit2= append reply to file save_name
 bit3= do not reset transaction count
 bit4= do not try to authenticate (set error)

 bit8= do not send send_text
 bit9= do not read status reply
bit10= do not read reply data after status line
*/
/*
 return:
   <0 fatal error
   =0 retryable error
    1 OK Transaction completed
    2 OK Transaction needs to be continued
*/
   Performs a NNTP transaction. A transaction consists of a command to
   be sent, a status reply to be received and (depending on the status
   reply) a text body to be received.
   This function should be called again with bit8|bit9 as long as it returns 2.


int Nntpconn_authenticate(conn,flag)
struct NntpconN *conn;
int flag;
    Performs the authentication handshake by AUTHINFO USER and AUTHINFO PASS.
    (the protocol demands the passwords to be transferred in clear)
   

nt Nntpconn_getline(conn,line,linsiz,flag)
struct NntpconN *conn;
char *line;
int linsiz;
int flag;
    Read a complete body text line from server.


#endif /* Nntpconn_never_defineD */

