
/*
  cc -g -c nntpfiler.c
*/

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>


/* Macro for allocating object arrays */
#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));

#include "../s_tools/sfile.h"

#include "nntpconn.h"
#include "nntphead.h"
#include "nntpfiler.h"


/* --------------------- Nntpfiler -------------------- */


int Nntpfiler_new(filer,conn,basedir,servername,port,groupname,flag)
struct NntpfileR **filer;
struct NntpconN *conn;
char *basedir;
char *servername;
int port;
char *groupname;
int flag;
/*
 bit0..3= initial value of filer->html 
*/
{
 struct NntpfileR *f;

 *filer= f= TSOB_FELD(struct NntpfileR,1);
 if(f==NULL)
   return(-1);
 f->conn= conn;
 strcpy(f->basedir,basedir);
 strcpy(f->servername,servername);
 f->port= port;
 strcpy(f->groupname,groupname);
 f->html= (flag&15);
 f->lvalid= 0;
 f->lfilev= NULL;
 f->lfilec= 0;
 f->lmarks= NULL;
 f->lmsgidx= NULL;
 return(1);
}


int Nntpfiler_destroy(filer,flag)
struct NntpfileR **filer;
int flag;
{
 if((*filer)==NULL)
   return(0);
 free((char *) *filer);
 *filer= NULL;
 return(1);
}


int Nntpfiler_del_index_cache(filer,flag)
struct NntpfileR *filer;
int flag;
{
 if(filer==NULL)
   return(0);
 filer->lvalid= 0;
 if(filer->lmarks!=NULL)
   free(filer->lmarks);
 filer->lmarks= NULL;
 if(filer->lmsgidx!=NULL)
   free(filer->lmsgidx);
 filer->lmsgidx= NULL;
 if(filer->lfilev!=NULL) /* release contentfile list */
   Sfile_ls("",&(filer->lfilec),&(filer->lfilev),2);
 return(1);
}


int Nntpfiler_cpy(target,source,flag)
struct NntpfileR *target;
struct NntpfileR *source;
int flag;
/*
 bit0= do not copy conn
*/
{
 if(!(flag&1))
   target->conn= source->conn;
 strcpy(target->basedir,source->basedir);
 strcpy(target->groupname,source->groupname);
 target->html= source->html;
 return(1);
}


int Nntpfiler_set_conn(filer,conn,flag)
struct NntpfileR *filer;
struct NntpconN *conn;
int flag;
{
 filer->conn= conn;
 return(1);
}


int Nntpfiler_set_basedir(filer,basedir,flag)
struct NntpfileR *filer;
char *basedir;
int flag;
{
 strcpy(filer->basedir,basedir);
 return(1);
}


int Nntpfiler_set_servername(filer,servername,port,flag)
struct NntpfileR *filer;
char *servername;
int port;
int flag;
{
 strcpy(filer->servername,servername);
 filer->port= port;
 return(1);
}


int Nntpfiler_set_groupname(filer,groupname,flag)
struct NntpfileR *filer;
char *groupname;
int flag;
{
 strcpy(filer->groupname,groupname);
 return(1);
}


int Nntpfiler_set_html(filer,html,flag)
struct NntpfileR *filer;
int html;
int flag;
{
 filer->html= html;
 return(1);
}


int Nntpfiler_chdir_primitive(dir,errn,errmsg,flag)
char *dir;
int *errn;
char *errmsg;
int flag;
/*
 bit0= print to stderr, do not use errn,errmsg
*/
{
 /* set working directory */
 if(chdir(dir)==-1) {
   if(flag&1) {
     fprintf(stderr,
             "Error %ld on changing to target directory %s\n%s",
             errno,dir,strerror(errno));
   } else {
     *errn= errno;
     sprintf(errmsg,
             "Error %ld on changing to target directory %s\n%s",
             errno,dir,strerror(errno));
   }
   return(-1);
 }
 return(1);
}


int Nntpfiler_std_dir(filer,dir,flag)
struct NntpfileR *filer;
char *dir;
int flag;
/*
 bit0-7= mode: 0= group dir
               1= server+port dir
               2= bin dir
               3= message idlist dir
*/
{
 int mode;

 mode= flag&255;
 if(mode==0){
   sprintf(dir,"%s/%s.port%ld/%s",
               filer->basedir,filer->conn->servername,filer->conn->portnumber,
               filer->groupname);
 }else if(mode==1){
   sprintf(dir,"%s/%s.port%ld",
               filer->basedir,filer->conn->servername,filer->conn->portnumber);
 }else if(mode==2){
   sprintf(dir,"%s/%s.port%ld/%s/bin",
               filer->basedir,filer->conn->servername,filer->conn->portnumber,
               filer->groupname);
 }else if(mode==3){
   sprintf(dir,"%s/msg_idlist",filer->basedir);
 }else{
   dir[0]= 0;
   return(-1);
 }
 return(1);
}


int Nntpfiler_create_dir(filer,dir,purpose,flag)
struct NntpfileR *filer;
char *dir;
char *purpose;
int flag;
/*
 bit0= allow automatic removing of old files
*/
{
 int ret,was_there= 0;
 char name[SfileadrL];
 struct stat stbuf; 

 if(stat(dir,&stbuf)==-1){
mkserverdir:;
   ret= mkdir(dir,0x1FF);
   if(ret==-1){
     filer->conn->errn= errno;
     sprintf(filer->conn->errmsg,
             "Error %ld on creating %s directory %s\n%s",
             errno,purpose,dir,strerror(errno));
     return(-1);
   }
   was_there= 0;
   if(flag&1){
     sprintf(name,"%s/%s",dir,Sfile_autoremove_olD);
     Sfile_string_to_file(name,"-3d",0);
   }
 } else {
   if(!S_ISDIR(stbuf.st_mode)){
     filer->conn->errn= EEXIST;
     sprintf(filer->conn->errmsg,"Cannot create %s directory %s\n%s",
                                 purpose,dir,
                                 "A non-directory object already exists.");
     return(-1);
   }
   was_there= 1;
 }
 return(!was_there);
}


int Nntpfiler_create_std_dir(filer,progname,flag)
struct NntpfileR *filer;
char *progname;
int flag;
{
 char dir[4096],hdir[4096];
 int ret,ret_group,i;

 ret= Nntpfiler_create_dir(filer,filer->basedir,"Base-Directory",0);
 if(ret<0)
   return(0);
 Nntpfiler_std_dir(filer,dir,3);
 ret= Nntpfiler_create_dir(filer,dir,"Message-ID-List",1);
 if(ret<0)
   return(0);
 for(i=0;i<Nntpfiler_idlist_hasH;i++){
   strcpy(hdir,dir);
   strcat(hdir,"/");
   sprintf(hdir+strlen(hdir),Nntpfiler_idlist_formaT,i);
   ret= Nntpfiler_create_dir(filer,hdir,"Message-ID-List",0);
   if(ret<0)
     return(0);
 }
 if(strcmp(filer->conn->servername,"-")==0 || filer->conn->servername[0]==0)
   return(3); /* just the general files have been created */
 Nntpfiler_std_dir(filer,dir,1);
 ret= Nntpfiler_create_dir(filer,dir,"Server+Port",0);
 if(ret<0)
   return(0);
 Nntpfiler_std_dir(filer,dir,0);
 if(strcmp(filer->groupname,"-")==0 || filer->groupname[0]==0)
   return(4);
 ret= ret_group= Nntpfiler_create_dir(filer,dir,"Group",1);
 if(ret<0)
   return(0);
 if(filer->html==2) {
   ret= Nntpfiler_php3group(filer,dir,0);
   if(ret<0)
     return(0);
 }
 Nntpfiler_std_dir(filer,dir,2);
 ret= Nntpfiler_create_dir(filer,dir,"Binary Files",0);
 if(ret<0)
   return(0);
 return(1+(ret_group==0)); /* 2= group directory was already there */
}


int Nntpfiler_msg_target(targetname,targetadr,dir,msgno,flag)
char *targetname,*targetadr;
char *dir;
int msgno;
int flag;
/*
 bit0= do not check for existing message
 bit1= do not copy result to targetname
 bit2= do not copy result to targetadr
 bit3= return(3) if message is older than 3600 seconds
*/
{
 char filename[4096],fileadr[4096];
 struct stat stbuf;

 sprintf(filename,"msg%-8.8ld.html",msgno);
 sprintf(fileadr,"%s/%s",dir,filename);
 if(!(flag&2))
   strcpy(targetname,filename);
 if(!(flag&4))
   strcpy(targetadr,fileadr);
 if(!(flag&1)){
   /* check wether the message already exists */;
   if(lstat(fileadr,&stbuf)!=-1) {
     if(flag&8)
       if(time(0)-stbuf.st_mtime>3600)
         return(3);
     return(2);
   }
 }
 return(1);
}


int Nntpfiler_escape_html(buffer,buffer_size,line,flag)
char *buffer,*line;
int buffer_size;
int flag;
{
 int ret;
 char *cpt,*wpt;

 wpt= buffer;
 for(cpt= line;*cpt!=0;cpt++){
   *wpt= 0;
   if(wpt-buffer>buffer_size-10) {
     return(0);
   }
   if(*cpt=='<') {
     strcat(wpt,"&lt;");
     wpt+= strlen(wpt);
   }else if(*cpt=='>') {
     strcat(wpt,"&gt;");
     wpt+= strlen(wpt);
   }else if(*cpt=='"') {
     strcat(wpt,"&quot;");
     wpt+= strlen(wpt);
   }else if(*cpt=='&') {
     strcat(wpt,"&amp;");
     wpt+= strlen(wpt);
   } else
     *(wpt++)= *cpt;
 }
 *wpt= 0;
 return(1);
}


int Nntpfiler_esc_filename(name,namsiz,flag)
unsigned char *name;
int namsiz;
int flag;
/*
 bit0= do not check for existing files
 bit1= assume a pseudo collision with original name
*/
{
 unsigned char *buff,*wpt,*cpt,infix[80];
 int i,k,bad,l,ret= 1,testcount;
 struct stat stbuf;

 if(namsiz<=0) 
   return(-1);
 buff= TSOB_FELD(unsigned char,namsiz+10);
 if(buff==NULL)
   return(-1);

 wpt= buff;
 l= strlen(name);

 strcpy(buff,name);
 ret= Sfile_esc_filename(buff,namsiz-10,8|4);
 if(ret<=0)
   goto ex;
 wpt= buff+strlen(buff);
 
 if(!(flag&1)){
   /* check for existing files and choose other name */
   Sfile_unique_adr(buff,!!(flag&2));
 }

 strcpy(name,buff);
ex:;
 free((char *) buff);
 return(ret);
}


int Nntpfiler_format_linkline(filer,msgno,
                              links,listname,labelname,prevname,nextname,flag)
struct NntpfileR *filer;
int msgno;
char *links,*listname,*labelname,*prevname,*nextname;
int flag;
{
 sprintf(links,"</PRE><CENTER><KBD>");
 if(filer->html==2) {
   sprintf(links+strlen(links),"<A HREF=\"fetch.php3?msg=%ld\">%ld</A>",
                               msgno-1,msgno-1);
   strcat(links," | ");
 }
 sprintf(links+strlen(links),"<A HREF=\"%s\">previous</A>",prevname);
 strcat(links," | ");
 sprintf(links+strlen(links),"<A HREF=\"%s#%s\">up</A>",listname,labelname);
 strcat(links," | ");
 sprintf(links+strlen(links),"<A HREF=\"%s\">next</A>",nextname);
 if(filer->html==2) {
   strcat(links," | ");
   sprintf(links+strlen(links),"<A HREF=\"fetch.php3?msg=%ld\">%ld</A>",
                               msgno+1,msgno+1);
 }
 sprintf(links+strlen(links),"</KBD></CENTER><PRE>\r\n");
 return(1);
}


int Nntpfiler_adjust_mlinks(filer,msgno,listname,prevname,nextname,flag)
struct NntpfileR *filer;
int msgno;
char *listname,*prevname,*nextname;
int flag;
{
 int ret= 0,cmplen;
 char links[321],fileadr[SfileadrL],line[1024],filename[SfileadrL],*cpt;
 char dir[SfileadrL];

 if(filer==NULL)
   return(2);

 Nntpfiler_std_dir(filer,dir,0);
 ret= Nntpfiler_msg_target(filename,fileadr,dir,msgno,1);
 if(ret<=0)
   return(0);
 ret= Nntpfiler_format_linkline(filer,msgno,
                                links,listname,filename,prevname,nextname,0);
 if(ret<=0)
   return(ret);
 cpt= strchr(links,'=');
 if(cpt==NULL)
   cmplen= strlen(links);
 else
   cmplen= cpt-links+1;
 ret= Sfile_change_file_lines(fileadr,links,cmplen,links,
                              &(filer->conn->errn),filer->conn->errmsg,0);
 return(ret);
}


int Nntpfiler_adjust_llinks(filer,listadr,msgname,prevname,nextname,flag)
struct NntpfileR *filer;
char *listadr;
char *msgname,*prevname,*nextname;
int flag;
{
 int ret= 0,cmplen;
 char links[SfileadrL],search_text[81];

 if(filer==NULL)
   return(2);

 strcpy(search_text,"<HR><BR><A HREF=");
 sprintf(links,"%s\"%s\">",search_text,prevname);
 sprintf(links+strlen(links),"previous index</A> | ");
 sprintf(links+strlen(links),"<A HREF=\"%s\">",msgname);
 sprintf(links+strlen(links),"first downloaded message</A> | ");
 sprintf(links+strlen(links),"<A HREF=\"%s\">",nextname);
 sprintf(links+strlen(links),"next index</A>\n");
 ret= Sfile_change_file_lines(listadr,search_text,strlen(search_text),links,
                              &(filer->conn->errn),filer->conn->errmsg,0);
 return(ret);
}


int Nntpfiler_message_id_key(filer,msgid,adr,adrsiz,flag)
struct NntpfileR *filer;
char *msgid;
char *adr;
int adrsiz;
int flag;
/*
 bit0= do not strip first and last character
*/
{
 int l,hc;
 char line[Nntpfiler_strmaX];

 strcpy(adr,"");
 if(filer==NULL)
   return(2);
 Nntpfiler_std_dir(filer,adr,3);
 strcpy(line,msgid);
 Sregex_trimline(line,0);
 if((l=strlen(line))<0)
   return(2);
 if(!(flag&1))
   line[--l]= 0;
 Sfile_esc_filename(line+!(flag&1),sizeof(line)-2,1);
 strcat(adr,"/");
 hc= Sfile_hashcode(line+1,Nntpfiler_idlist_hasH,1);
 sprintf(adr+strlen(adr),Nntpfiler_idlist_formaT,hc);
 strcat(adr,"/");
 strcat(adr,line+1);
 return(1);
}


int Nntpfiler_mntc_lock(filer,description,flag)
struct NntpfileR *filer;
char *description;
int flag;
/*
 bit0= wait until lock can be obtained
 bit1= release lock
*/
{
 char adr[SfileadrL],dir[SfileadrL];
 int ret,count= 0,temp_adr_handle= 0;

 Nntpfiler_std_dir(filer,dir,0);
 sprintf(adr,"%s/%s",dir,Nntpfiler_mntc_namE);
 if(flag&2){
   unlink(adr);
   /* de-register as temporary file object */
   Cleanup_handler_temp_file(adr,0,1);
   return(1);
 }
 /* register as temporary file object */
 temp_adr_handle= Cleanup_handler_temp_file(adr,0,0);
 do {
   ret= Sfile_get_lock(adr,(char *) filer,Nntpfiler_lock_waiT,description,0);
   if(ret<=0){
     if(flag&1){
       Sfile_microsleep(1000000,0);
       count++;
       fprintf(stderr,"\r  waiting for release of maintenance lock : %ld s\r",
                      count);
     }else{
       filer->conn->errn= -24;
       sprintf(filer->conn->errmsg,"cannot obtain maintenance lock for : %s",
                                   description);
       /* de-register */
       Cleanup_handler_temp_file("",temp_adr_handle,1|2);
       return(0);
     }
   }
 } while(ret<=0);
 if(count>0)
   fprintf(stderr,
           "\r                                                            \r");
 return(1);
}


int Nntpfiler_del_mntc_locks(filer,flag)
struct NntpfileR *filer;
int flag;
{
 int ret=0,filec= 0,i;
 char **filev= NULL,mntc_expr[SfileadrL];

 ret= Nntpfiler_chdir_primitive(filer->basedir,&(filer->conn->errn),
                                filer->conn->errmsg,0);
 if(ret<=0)
   return(2);
 sprintf(mntc_expr,"*/*/%s",Nntpfiler_mntc_namE);
 ret= Sfile_ls(mntc_expr,&filec,&filev,0);
 if(ret<=0)
   filec= 0;
 if(filec==0){
   ret= 2;
   goto ex;
 }
 printf("removing all maintenance locks:\n");
 for(i=0;i<filec;i++){
   printf("%s\n",filev[i]);
   unlink(filev[i]);
 }
 printf("done\n");
 ret= 1;
ex:;
 Sfile_ls("",&filec,&filev,2);
 return(ret);
}


int Nntpfiler_id_to_msgadr(filer,msgid,msgadr,msgname,flag)
struct NntpfileR *filer;
char *msgid;
char *msgadr,*msgname;
int flag;
/*
 bit0= do not strip first and last character of msgid
*/
{
 char *mpt,id_adr[SfileadrL],line[SfileadrL],dir[SfileadrL];
 int ret;

 Nntpfiler_std_dir(filer,dir,0);
 /* get msgadr */
 ret= Nntpfiler_message_id_key(filer,msgid,id_adr,sizeof(id_adr),flag&1);
 if(ret<=0)
   return(ret);
 ret= Sfile_file_to_text(id_adr,line,sizeof(line),1|2);
 if(ret<0)
   return(ret);
 if(ret==2 || ret==0)
   return(2);
 sprintf(msgadr,"%s/%s",filer->basedir,line);
 /* check wether same directory */
 mpt= strrchr(msgadr,'/')+1;
 strcpy(msgname,mpt);
 if(strncmp(dir,msgadr,(mpt-msgadr)-1)!=0)
   return(3);
 return(1);
}


int Nntpfiler_add_reference(filer,ref_id,msgno,subject,from,flag)
struct NntpfileR *filer;
char *ref_id;
int msgno;
char *subject,*from;
int flag;
{
 int ret;
 char msgname[SfileadrL],dir[SfileadrL];

 /* get own msgname */
 Nntpfiler_std_dir(filer,dir,0);
 ret= Nntpfiler_msg_target(msgname,msgname,dir,msgno,1|4);
 if(ret<=0)
   return(ret);
 ret= Nntpfiler_add_reference_n(filer,ref_id,msgname,subject,from,0);
 return(ret);
}


int Nntpfiler_add_reference_n(filer,ref_id,msgname,subject,from,flag)
struct NntpfileR *filer;
char *ref_id;
char *msgname;
char *subject,*from;
int flag;
{
 int ret;
 char msgadr[SfileadrL],ref_link[SfileadrL];
 char esc_subject[SfileadrL],esc_from[SfileadrL];
 static char search[]={"</PRE></BODY></HTML>"};

 /* get referred msgadr */
 ret= Nntpfiler_id_to_msgadr(filer,ref_id,msgadr,esc_from,0);
 if(ret!=1)
   return(ret);

 /* build ref_link string (+ search string) */
 ret= Nntpfiler_escape_html(esc_subject,sizeof(esc_subject),subject,0);
 if(ret<=0)
   return(ret);
 ret= Nntpfiler_escape_html(esc_from,sizeof(esc_from),from,0);
 if(ret<=0)
   return(ret);
 sprintf(ref_link," <A HREF=%s> %s </A> from: <B>%s</B>\r\n",
                  msgname,esc_subject,esc_from);
 strcat(ref_link,search);
 strcat(ref_link,"\r\n");
 ret= Sfile_change_file_lines(msgadr,search,strlen(search),ref_link,
                              &(filer->conn->errn),filer->conn->errmsg,0);
 return(ret);
}


int Nntpfiler_list_result(filer,fp,text,flag)
struct NntpfileR *filer;
FILE *fp;
char *text;
int flag;
{
 int ret;
 char *buffertext,buffertext_data[4096];

 if(filer->html && !(flag&1)) {
   ret= Nntpfiler_escape_html(buffertext_data,sizeof(buffertext_data),
                              text,0);
   if(ret<=0)
     return(ret);
   buffertext= buffertext_data;
 } else {
   if((flag&1)&&!filer->html)
     return(2);
   buffertext= text;
 }

 if(!(flag&2))
   fprintf(stderr,"%s",buffertext);
 fprintf(fp,"%s",buffertext);
 return(1);
}


int Nntpfiler_list_index(filer,flag)
struct NntpfileR *filer;
int flag;
{
 int ret= 0,lfilec=0,i;
 char **lfilev= NULL,line[SfileadrL],link[SfileadrL];
 FILE *fp= NULL;

 /* create index of listfiles */
 fp= fopen("list_index.html","w");
 if(fp==NULL){
   filer->conn->errn= errno;
   sprintf(filer->conn->errmsg,"Error %ld on opening output file : %s\n%s",
                                errno,line,strerror(errno));
   ret= 0;
   goto ex;
 }

 sprintf(line,"%s.port%ld-%s",filer->servername,filer->port,filer->groupname);
 Sfile_esc_filename(line,sizeof(line),2);
 sprintf(link,"</PRE><CENTER><KBD><A HREF=\"../../main_index.html#%s\">",line);
 strcat(link,"main index</A></KBD></CENTER><BR><PRE>\n");

 sprintf(line,"<HTML>\n");
 Nntpfiler_list_result(filer,fp,line,1|2);
 sprintf(line,"<HEAD><TITLE>list of available indexfiles</TITLE></HEAD>\n");
 Nntpfiler_list_result(filer,fp,line,1|2);
 sprintf(line,"<BODY><KBD><BR>\n\n");
 Nntpfiler_list_result(filer,fp,line,1|2);
 Nntpfiler_list_result(filer,fp,link,1|2);
 sprintf(line,"<HR><BR>\n");
 Nntpfiler_list_result(filer,fp,line,1|2);
 sprintf(line,"</PRE><CENTER><KBD>Server <B>%s</B>",filer->conn->servername);
 Nntpfiler_list_result(filer,fp,line,1|2);
 sprintf(line," , Port %ld<BR>\n",filer->conn->portnumber);
 Nntpfiler_list_result(filer,fp,line,1|2);
 sprintf(line,"Group <B>%s</B>\n",filer->groupname);
 Nntpfiler_list_result(filer,fp,line,1|2);
 if(filer->html==2) {
   Nntpfiler_list_result(filer,fp,"<BR>",1|2);
   sprintf(line,"%s | %s","<A HREF=\"list_interval.php3\">update group</A>",
                          "<A HREF=\"remove_old.php3\">remove old files</A>");
   Nntpfiler_list_result(filer,fp,line,1|2);
 }
 sprintf(line,"</KBD></CENTER><PRE><BR>\n");
 Nntpfiler_list_result(filer,fp,line,1|2);
 sprintf(line,"<HR><BR>\n");
 Nntpfiler_list_result(filer,fp,line,1|2);

 /* get a list of contentlist files */
 ret= Sfile_ls("^list_[0-9]......[0-9]_[0-9]......[0-9]\\.html$",
               &lfilec,&lfilev,4|8);
 if(ret<=0)
   lfilec= 0;
 if(lfilec<=0){
   sprintf(line,"<HR><B>no index files found</B><HR>\n\n");
   Nntpfiler_list_result(filer,fp,line,1|2);
 } else {
   for(i=0;i<lfilec;i++){
     sprintf(line,"<A HREF=\"%s\" NAME=\"%s\">%s</A>",
                  lfilev[i],lfilev[i],lfilev[i]);
     Nntpfiler_list_result(filer,fp,line,1|2);
     if(strcmp(lfilev[i],"list_00000000_00000000.html")==0) {
       sprintf(line," (copy of most recent list , ");
       Nntpfiler_list_result(filer,fp,line,1|2);
       sprintf(line,"<A HREF=\"msg_new.html\">first message there</A>)\n");
       Nntpfiler_list_result(filer,fp,line,1|2);
       sprintf(line,"<BR><HR><BR>\n");
     } else {
       sprintf(line,"<BR><BR>\n");  
     }
     Nntpfiler_list_result(filer,fp,line,1|2);
   }
 }

 Nntpfiler_list_result(filer,fp,"<HR>\n",1|2);
 Nntpfiler_list_result(filer,fp,link,1|2);
 sprintf(line,"</KBD></BODY></HTML>\n");
 Nntpfiler_list_result(filer,fp,line,1|2);
 ret= 1;
ex:
 if(lfilev!=NULL) /* release contentfile list */
   Sfile_ls("",&lfilec,&lfilev,2);
 if(fp!=NULL) fclose(fp);
 return(ret);
}


int Nntpfiler_clean_all_reflists(filer,agelimit,flag)
struct NntpfileR *filer;
time_t agelimit;
int flag;
{
 int filec= 0,i,ret;
 char **filev= NULL,dir[SfileadrL];
 FILE *fp;
 struct stat stbuf;

 /* cd group directory */
 Nntpfiler_std_dir(filer,dir,0);
 ret= Nntpfiler_chdir_primitive(dir,&(filer->conn->errn),
                                filer->conn->errmsg,0);
 if(ret<=0)
   goto ex;
 ret= Sfile_ls("^msg[0-9]......[0-9]\\.html$",&filec,&filev,8);
 if(ret<=0)
   goto ex;
 for(i= 0;i<filec;i++){
   /* check wether young enough */
   if(Sfile_older_than(filev[i],agelimit,1))
 continue;
   /* remove the list of referring messages */
   Sfile_del_file_lines(filev[i],"Reply list :",0,"</PRE></BODY></HTML>",
                        0,&(filer->conn->errn),filer->conn->errmsg,0);
 }
ex:
 if(filev!=NULL) /* release contentfile list */
   Sfile_ls("",&filec,&filev,2);
 if(fp!=NULL) fclose(fp);
 return(ret);
}


int Nntpfiler_adjust_refs(filer,agelimit,flag)
struct NntpfileR *filer;
time_t agelimit;
int flag;
{
 int ret,mid,filec= 0,hc,i,lg;
 char dir[SfileadrL],bdir[SfileadrL],idir[SfileadrL],**filev= NULL;
 char ref_id[Nntphead_refL],msgadr[SfileadrL],msgid[Nntpfiler_strmaX];
 char adr[SfileadrL],ref_msgadr[SfileadrL],*cpt,*spt;
 char from[Nntphead_strL],subject[Nntphead_strL],refers[Nntphead_strL];
 char *ref_adr,ref_esc[Nntphead_strL];

 ret= Nntpfiler_clean_all_reflists(filer,agelimit,0);
 if(ret<=0)
   return(ret);

 Nntpfiler_std_dir(filer,bdir,3);
 ret= Nntpfiler_chdir_primitive(bdir,&(filer->conn->errn),
                                filer->conn->errmsg,0);
 if(ret<=0)
   goto ex;

 lg= strlen(filer->groupname);

 /* loop over msg_idlist directories */
 for(hc=0;hc<Nntpfiler_idlist_hasH;hc++){
   sprintf(idir,Nntpfiler_idlist_formaT,hc);
   sprintf(dir,"%s/*",idir);
   /* get list of id files */;
   ret= Sfile_ls(dir,&filec,&filev,0);
   if(ret<=0)
     goto ex;
   /* loop over id files in single directory */
   for(mid=0;mid<filec;mid++){
     if(Sfile_older_than(filev[mid],agelimit,1))
   continue;
     cpt= strrchr(filev[mid],'/');
     if(cpt==NULL) cpt= filev[mid];
     else          cpt++;
     strcpy(msgid,cpt);
     Sfile_unescape_filename(msgid,0);
     /* read msgadr and reference msgid */
     strcpy(adr,filev[mid]);
     ret= Sfile_file_to_text(adr,msgadr,sizeof(msgadr),1);
     if(ret<=0)
   continue;
     ret= Sfile_read_file_line(adr,"References:",0,ref_id,sizeof(ref_id),
                               &(filer->conn->errn),filer->conn->errmsg,2);
     if(ret<=0)
   continue;
     Sregex_trimline(ref_id,0);
     if(ref_id[0]==0)
   continue;
     ret= Sfile_read_file_line(adr,"Subject: ",0,subject,sizeof(subject),
                               &(filer->conn->errn),filer->conn->errmsg,2);
     if(ret<=0)
   continue;
     ret= Sfile_read_file_line(adr,"From: ",0,from,sizeof(from),
                               &(filer->conn->errn),filer->conn->errmsg,2);
     if(ret<=0)
   continue;
     /* check wether msg is in group */
     cpt= strchr(msgadr,'/');
     if(cpt==NULL)
   continue;
     cpt++;
     if(cpt[lg]!='/')
   continue;
     if(strncmp(cpt,filer->groupname,lg)!=0)
   continue;
     /* remove all path components but the message file name */
     cpt= strrchr(msgadr,'/')+1;
     for(i=cpt-msgadr;msgadr[i]!=0;i++)
       msgadr[i-(cpt-msgadr)]= msgadr[i];
     msgadr[i-(cpt-msgadr)]= 0;
     /* get ref_msgadr of referenced id */
     ret= Nntpfiler_message_id_key(filer,ref_id,adr,sizeof(adr),0);
     if(ret<=0)
       goto ex;
     ret= Sfile_file_to_text(adr,ref_msgadr,sizeof(ref_msgadr),1|2);
     if(ret<=0)
   continue;
     /* check wether msg is in group */
     ref_adr= strchr(ref_msgadr,'/');
     if(ref_adr==NULL)
   continue;
     ref_adr++;
     if(ref_adr[lg]!='/')
   continue;
     if(strncmp(ref_adr,filer->groupname,lg)!=0)
   continue;
     ref_adr= strrchr(ref_msgadr,'/')+1;
     ret= Nntpfiler_add_reference_n(filer,ref_id,msgadr,subject,from,0);
     if(ret<=0)
       goto ex;
     /* change "refers :" line in message */
     if((ret= Nntpfiler_escape_html(ref_esc,sizeof(ref_esc),ref_id,0))<=0)
       goto ex;
     sprintf(refers,"refers: <A HREF=%s> %s </A>\n",ref_adr,ref_esc);
     ret= Sfile_change_file_lines(msgadr,"refers :",0,refers,
                                 &(filer->conn->errn),filer->conn->errmsg,0);
   }
 }
 ret= 1;
ex:;
 Sfile_ls("",&filec,&filev,2);
 return(ret);
}


int Nntpfiler_read_lintervals(filer,filec,filev,marks,msgidx,flag)
struct NntpfileR *filer;
int *filec;
char ***filev;
int **marks,**msgidx;
int flag;
{
 int i,ret;
 int lfilec= 0,*lmarks= NULL,*lmsgidx= NULL;
 char **lfilev= NULL,dir[SfileadrL];

 Nntpfiler_std_dir(filer,dir,0);
 ret= Nntpfiler_chdir_primitive(dir,&(filer->conn->errn),
                                filer->conn->errmsg,0);
 if(ret<=0)
   goto ex;
 /* get a list of contentlist files */
 Sfile_ls("^list_[0-9]......[0-9]_[0-9]......[0-9]\\.html$",
          &lfilec,&lfilev,4|8);
 *filec= lfilec;
 *filev= lfilev;
 /* read number intervals from list names */
 lmarks= TSOB_FELD(int,lfilec+1);
 *marks= lmarks;
 lmsgidx= TSOB_FELD(int,lfilec+1);
 *msgidx= lmsgidx;
 if(lmarks==NULL || lmsgidx==NULL){
   filer->conn->errn= -21;
   sprintf(filer->conn->errmsg,
           "Not enough memory for %ld list_*_* files",lfilec);
   ret= -1;
   goto ex;
 }
 for(i=0;i<lfilec;i++){
   lfilev[i][13]= 0;
   sscanf(lfilev[i]+5,"%ld",&lmarks[i]);
   lfilev[i][13]= '_';
   lmsgidx[i]= 0;
 }
 if(lfilec>0){
   sscanf(lfilev[lfilec-1]+14,"%ld",&lmarks[lfilec]);
   lmarks[lfilec]++;
 } else
   lmarks[lfilec]= 0;
 ret= 1;
ex:;
 return(ret);
}


int Nntpfiler_det_indexfile(filer,msgno,filename,flag)
struct NntpfileR *filer;
int msgno;
char *filename;
int flag;
/*
 bit0= permission to reuse cached lmarks and lmsgidx 
*/
{
 int ret,i;

 strcpy(filename,"list_00000000_00000000.html");
 if((flag&1) && filer->lvalid) {
   ;
 } else {
   Nntpfiler_del_index_cache(filer,0);
   ret= Nntpfiler_read_lintervals(filer,&(filer->lfilec),
                                        &(filer->lfilev),
                                        &(filer->lmarks),
                                        &(filer->lmsgidx),0);
   if(ret<=0)
     goto ex;
   filer->lvalid= 1;
 }
 for(i=0;i<filer->lfilec;i++) 
   if(filer->lmarks[i]<=msgno && msgno<filer->lmarks[i+1])
     break;
 ret= 1;
 if(i<filer->lfilec){
   ret= 2;
   strcpy(filename,filer->lfilev[i]);
 }
ex:;
 return(ret);
}


int Nntpfiler_merge_tee(filer,boss,flag)
struct NntpfileR *filer;
struct NntpclienT *boss;
int flag;
/*
 bit0= do not aquire maintenance lock
*/
{
 int i,ret,locked=0,filec= 0;
 struct NntpmergE *merger= NULL;
 char **filev= NULL,dir[SfileadrL];

 if(!(flag&1)){
   ret= Nntpfiler_mntc_lock(filer,"merging tee files",1);
   if(ret<=0)
     goto ex;
   locked= 1;
 }

 Nntpfiler_std_dir(filer,dir,2);
 ret= Nntpfiler_chdir_primitive(dir,NULL,NULL,1);
 if(ret<=0)
   goto ex;
 ret= Sfile_ls("^_.*\\.tee$",&filec,&filev,8);
 if(ret<0)
   goto ex;
 if(filec<=0 || ret==0)
   {ret= 2; goto ex;}

 ret= Nntpmerge_new(&merger,boss,filev,filec,0);
 if(ret<=0)
   goto ex;
 ret= Nntpmerge_all(merger,0);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 if(locked)
   Nntpfiler_mntc_lock(filer,"merging tee files",2);
 Nntpmerge_destroy(&merger,0);
 if(filev!=NULL)
   Sfile_ls(dir,&filec,&filev,2);
 return(ret);
}


int Nntpfiler_adjust_links(filer,flag)
struct NntpfileR *filer;
int flag;
/*
 bit0= do not aquire maintenance lock
*/
{
 int i,ret,mfilec=0,msgno,lfilec=0,lidx,curr_list,locked=0;
 int *lmsgidx=NULL,*lmarks=NULL;
 char **mfilev= NULL,**lfilev= NULL;
 char *prev,*next,*nmsg,listname[SfileadrL],msg_new[81],dir[SfileadrL];
 char *sname,*pname,*gname,in_portname[16],default_lname[81];
 char list_index[81];

 strcpy(list_index,"list_index.html");
 strcpy(default_lname,"list_00000000_00000000.html");
 msg_new[0]= 0;

 if(!(flag&1)){
   ret= Nntpfiler_mntc_lock(filer,"adjusting links",1);
   if(ret<=0)
     goto ex;
   locked= 1;
 }

 ret= Nntpfiler_read_lintervals(filer,&lfilec,&lfilev,&lmarks,&lmsgidx,0);
 if(ret<=0)
   goto ex;

 lidx= 1;

 /* get list of message files */
 ret= Sfile_ls("^msg[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]\\.html$",
               &mfilec,&mfilev,4|8);
 if(ret<=0)
   mfilec= 0;
 /* connect links to neighbours */
 curr_list= -1;
 for(i=0;i<mfilec;i++){
   sscanf(mfilev[i]+3,"%ld",&msgno);
   /* set contentlist name. (default is 0_0) */
   strcpy(listname,default_lname);
   /* find matching listname */
   if(msgno>=lmarks[0]){
     while(msgno>=lmarks[lidx] & lidx<lfilec+1)
       lidx++;
     if(lidx<=lfilec){
       if(lidx!=curr_list)
         lmsgidx[lidx-1]= i+1;
       strcpy(listname,lfilev[lidx-1]);
       if(lidx==lfilec && msg_new[0]==0)
         strcpy(msg_new,mfilev[i]);
       curr_list= lidx;
     }
   }
   /* set neighbour names */
   if(i-1<0)
     prev= listname;
   else
     prev= mfilev[i-1];
   if(i+1>=mfilec)
     next= listname;
   else
     next= mfilev[i+1];
   ret= Nntpfiler_adjust_mlinks(filer,msgno,listname,prev,next,0);
   if(ret<=0)
     goto ex;
 }

 /* adjust links in contentlist files */
 for(lidx=1;lidx<=lfilec;lidx++){
   if(strcmp(lfilev[lidx-1],default_lname)==0)
 continue;

   if(lidx-2<0)
     prev= list_index;
   else if(strcmp(lfilev[lidx-2],default_lname)==0)
     prev= list_index;
   else
     prev= lfilev[lidx-2];

   if(lidx>=lfilec)
     next= list_index;
   else
     next= lfilev[lidx];

   if(lmsgidx[lidx-1]-1<0)
     nmsg= next;
   else
     nmsg= mfilev[lmsgidx[lidx-1]-1];

   ret= Nntpfiler_adjust_llinks(filer,lfilev[lidx-1],nmsg,prev,next,0);
   if(ret<=0)
     goto ex;
 }

 /* make an index of listfiles */
 ret= Nntpfiler_list_index(filer,0);
 if(ret<=0)
   goto ex;

 /* Copy certain messages and the newest list to standard names */
 if(mfilec>0){
   unlink("msg_first.html");
   Sfile_copy_file(mfilev[0],"msg_first.html",
                   &(filer->conn->errn),filer->conn->errmsg,0);
   unlink("msg_new.html");
   if(msg_new[0]!=0)
     Sfile_copy_file(msg_new,"msg_new.html",
                   &(filer->conn->errn),filer->conn->errmsg,0);
   unlink("msg_last.html");
   Sfile_copy_file(mfilev[mfilec-1],"msg_last.html",
                   &(filer->conn->errn),filer->conn->errmsg,0);
 }
 if(lfilec>0){
   unlink(default_lname);
   Sfile_copy_file(lfilev[lfilec-1],default_lname,
                   &(filer->conn->errn),filer->conn->errmsg,0);
 }

 ret= 1;
ex:;
 if(locked)
   Nntpfiler_mntc_lock(filer,"adjusting links",2);
 if(lmarks!=NULL)
   free(lmarks);
 if(lmsgidx!=NULL)
   free(lmsgidx);
 if(lfilev!=NULL) /* release contentfile list */
   Sfile_ls(dir,&lfilec,&lfilev,2);
 if(mfilev!=NULL) /* release filelist */
   Sfile_ls(dir,&mfilec,&mfilev,2);
 return(ret);
}


int Nntpfiler_check_maxlimit(filer,agelimit,flag)
struct NntpfileR *filer;
time_t agelimit;
int flag;
{
 int ret,valid;
 char line[SfileadrL];
 time_t maxlimit;

 ret= Sfile_file_to_text(Sfile_autoremove_olD,line,sizeof(line),1);
 if(ret<=0){
   filer->conn->errn= -22;
   sprintf(filer->conn->errmsg,"File '%s' not found in working directory",
           Sfile_autoremove_olD);
   return(0);
 }
 if(ret==2) /* empty file, no limit */
   return(1);
 ret= Sfile_decode_time(line,&valid,&maxlimit,0);
 if(ret<=0) {
   filer->conn->errn= -23;
   sprintf(filer->conn->errmsg,"Invalid time code : '%s' (read from file '%s)",
           line,Sfile_autoremove_olD);
   return(0);
 }
 if(maxlimit<agelimit) {
   filer->conn->errn= -22;
   sprintf(filer->conn->errmsg,
           "Agelimit '%s' violated. (imposed by file '%s')",
           line,Sfile_autoremove_olD);
   return(0);
 }
 return(1);
}


int Nntpfiler_remove_old_files(filer,timecode,flag)
struct NntpfileR *filer;
char *timecode;
int flag;
/*
 bit0-7= mode : 0=group files , 1=message id list
*/
{
 FILE *fp= NULL;
 int i,ret,filec=0,hashdirc=0,mode,valid,firstno,lastno,verbous= 1,locked= 0;
 int removed= 0;
 char **filev=NULL,**hashdirv=NULL,line[SfileadrL],portname[81],dir[SfileadrL];
 char wd[SfileadrL];
 time_t agelimit,xoverlimit;

 ret= Sfile_decode_time(timecode,&valid,&agelimit,0);
 if(ret<=0)
   goto ex;

 mode= flag&255;
 if(mode==0){
   Nntpfiler_std_dir(filer,dir,0);
   ret= Nntpfiler_chdir_primitive(dir,&(filer->conn->errn),
                                  filer->conn->errmsg,0);
   if(ret<=0)
     goto ex;
   ret= Nntpfiler_check_maxlimit(filer,agelimit,0);
   if(ret<=0)
     goto ex;

   ret= Nntpfiler_mntc_lock(filer,"removing old files",1);
   if(ret<=0)
     goto ex;
   locked= 1;

   /* There may be very many files to process ,
      therefore we use  ls | grep 'regexpr' rather than  ls parsexpr  */

   /* XOVER caches (with special maximum lifetime) */
   ret= Sfile_ls("^xover_[0-9]......[0-9]_[0-9]......[0-9]$",&filec,&filev,4|8);
   if(ret<=0)
      filec= 0;
   if(time(0)-agelimit>Nntpfiler_xover_maxlifE)
     xoverlimit= time(0)-Nntpfiler_xover_maxlifE;
   else
     xoverlimit= agelimit;
   ret= Sfile_remove_old_files(filec,filev,xoverlimit,verbous);
   if(ret<=0)
     goto ex;
   /* binary files (possibly too large for shell argument list) */
   getcwd(wd,sizeof(wd));
   if(chdir("bin")!=-1) {
     ret= Sfile_ls("",&filec,&filev,4);
     if(ret<=0)
        filec= 0;
     ret= Sfile_remove_old_files(filec,filev,agelimit,verbous|(1<<15));
     if(ret<=0)
       goto ex;
     chdir(wd);
   }
   /* message files */
   ret= Sfile_ls("^msg[0-9]......[0-9]\\.html$",&filec,&filev,4|8);
   if(ret<=0)
      filec= 0;
   ret= Sfile_remove_old_files(filec,filev,agelimit,verbous);
   if(ret<=0)
     goto ex;



   /* list files : depending on which messages survived */;
   /* qui vive ? */
   ret= Sfile_ls("^msg[0-9]......[0-9]\\.html$",&filec,&filev,4|8);
   if(ret<=0)
      filec= 0;
   if(filec>0)
     sscanf(filev[0]+3,"%ld",&firstno);
   else
     firstno= 0x7FFFFFFF;
   ret= Sfile_ls("^list_[0-9]......[0-9]_[0-9]......[0-9]\\.html$",
                 &filec,&filev,4|8);
   if(ret<=0)
      filec= 0;

#ifdef Nntpfiler_disableD

   for(i=0;i<filec;i++){
     sscanf(filev[i]+14,"%ld",&lastno);
     if(lastno>0 && firstno>lastno){
       unlink(filev[i]);
       removed++;
       if(verbous){
         if(removed==1)
           printf("removed (listing no valid message any more) :\n");
         printf(" %s",filev[i]);
         if((removed%3)==0)
            printf("\n");
       }
     }
   }

#else /* Nntpfiler_disableD */

   for(i=0;i<filec;i++){
     lastno= 0;
     if(strlen(filev[i])>14)
       sscanf(filev[i]+14,"%ld",&lastno);
     if(lastno<=0 || firstno<=lastno){
       filev[i][0]= 0;                        /* make invisible for remover */
     }
   }
   ret= Sfile_remove_old_files(filec,filev,agelimit,verbous);
   if(ret<=0)
     goto ex;

#endif /* ! Nntpfiler_disableD */

   if((removed%3)!=0)
     printf("\n");
   ret= Nntpfiler_adjust_links(filer,1);
   if(ret<=0)
     goto ex;
 } else if(mode==1) {
   Nntpfiler_std_dir(filer,dir,3);
   Nntpfiler_chdir_primitive(dir,&(filer->conn->errn),
                                 filer->conn->errmsg,0);
   if(ret<=0)
     goto ex;
   ret= Nntpfiler_check_maxlimit(filer,agelimit,0);
   if(ret<=0)
     goto ex;
   /* hash directories */;
   ret= Sfile_ls("[0-9A-F]*",&hashdirc,&hashdirv,1);
   if(ret<=0)
      hashdirc= 0;
   /* message ids */;
   for(i=0;i<hashdirc;i++){
     sprintf(line,"%s/*",hashdirv[i]);
     ret= Sfile_ls(line,&filec,&filev,4);
     if(ret<=0)
        filec= 0;
     ret= Sfile_remove_old_files(filec,filev,agelimit,verbous);
     if(ret<=0)
       goto ex;
   }
 }

 ret= 1;
ex:; 
 if(locked)
   Nntpfiler_mntc_lock(filer,"removing old files",2);
 if(fp!=NULL)
   fclose(fp);
 Sfile_ls("",&filec,&filev,2);
 Sfile_ls("",&hashdirc,&hashdirv,2);
 return(ret);
}


int Nntpfiler_main_index(filer,progname,flag)
struct NntpfileR *filer;
char *progname;
int flag;
{
 char rc_expr[81],**rfilev= NULL,*serverdir,*groupname,line[SfileadrL];
 char dir[SfileadrL],old_server[SfileadrL],text[SfileadrL],*dpt;
 int ret,rfilec= 0,i,htmlrec;
 FILE *fp= NULL;

 htmlrec= filer->html;
 if(filer->html<1)
   filer->html= 1;
 ret= Nntpfiler_chdir_primitive(filer->basedir,&(filer->conn->errn),
                                filer->conn->errmsg,0);
 if(ret<=0)
   {ret= 0; goto ex;}
 sprintf(rc_expr,"*/*/.%src",progname);
 ret= Sfile_ls(rc_expr,&rfilec,&rfilev,4);
 if(ret<=0)
   rfilec= 0;

 sprintf(line,"%s/main_index.html",filer->basedir);
 fp= fopen(line,"w");
 if(fp==NULL){
   filer->conn->errn= errno;
   sprintf(filer->conn->errmsg,"Error %ld on opening output file : %s\n%s",
                                errno,line,strerror(errno));
   ret= 0;
   goto ex;
 }

 sprintf(line,"<HTML>\n");
 Nntpfiler_list_result(filer,fp,line,1|2);
 sprintf(line,"<HEAD><TITLE>list of snntpbatch newsgroups</TITLE></HEAD>\n");
 Nntpfiler_list_result(filer,fp,line,1|2);
 sprintf(line,"<BODY><KBD><BR>\n\n");
 Nntpfiler_list_result(filer,fp,line,1|2);
 if(filer->html==2) {
   ret= Nntpfiler_php3main(filer,0);
   if(ret<0)
     return(0);
   sprintf(line,
           "<CENTER><KBD><A HREF=\"subscribe.php3\">subscribe anywhere</A>");
   Nntpfiler_list_result(filer,fp,line,1|2);
   Nntpfiler_list_result(filer,fp,"</KBD></CENTER><HR>",1|2);
 }
 if(rfilec<=0){
   sprintf(line,"<HR><B>no groups subscribed</B><HR>\n\n");
   Nntpfiler_list_result(filer,fp,line,1|2);
 }

 old_server[0]= 0;
 for(i=0;i<rfilec;i++){
   serverdir= rfilev[i];
   groupname= strchr(serverdir,'/');
   *groupname= 0;
   groupname++;
   *strchr(groupname,'/')= 0;
   if(Sfile_type(serverdir,0)==3) {
     /* symbolic link : test wether leading to a sibling file in basedir */
     if(Sfile_readlink(serverdir,dir,sizeof(dir),0)<=0)
 continue;
     if(strchr(dir,'/')==NULL)
 continue;
   }
   if(strcmp(old_server,serverdir)!=0){
     Nntpfiler_list_result(filer,fp,"</PRE><CENTER><KBD>",1|2);
     sprintf(line,"<HR>Serverdirectory <B>%s</B><BR>\n",serverdir);
     Nntpfiler_list_result(filer,fp,line,1|2);
     if(filer->html==2) {
       dpt= strrchr(serverdir,'.');
       if(dpt!=NULL)
         *dpt= 0;
       sprintf(line,
        "<A HREF=\"subscribe.php3?server=%s&login=off\">subscribe group</A>\n",
               serverdir);
       if(dpt!=NULL)
         *dpt= '.';
       Nntpfiler_list_result(filer,fp,line,1|2);
     }
     Nntpfiler_list_result(filer,fp,"<HR></KBD></CENTER><PRE>\n",1|2);
   }
   sprintf(text,"%s-%s",serverdir,groupname);
   Sfile_esc_filename(text,sizeof(text),2);
   sprintf(line,"<A NAME=\"%s\">",text);
   Nntpfiler_list_result(filer,fp,line,1|2);
   sprintf(line,"Group <B>%s</B><BR>\n",groupname);
   Nntpfiler_list_result(filer,fp,line,1|2);
   sprintf(dir,"%s/%s",serverdir,groupname);

   sprintf(line,"<A HREF=\"%s/msg_first.html\">first message</A> |",dir);
   Nntpfiler_list_result(filer,fp,line,1|2);
   sprintf(line,"<A HREF=\"%s/list_index.html\">group index</A> |",dir);
   Nntpfiler_list_result(filer,fp,line,1|2);
   sprintf(line,
         "<A HREF=\"%s/list_00000000_00000000.html\">most recent index</A> ",
           dir);
   Nntpfiler_list_result(filer,fp,line,1|2);
   sprintf(line,"<A HREF=\"%s/msg_new.html\">(start message)</A> |",dir);
   Nntpfiler_list_result(filer,fp,line,1|2);
   sprintf(line,"<A HREF=\"%s/msg_last.html\">last message</A>\n",dir);
   Nntpfiler_list_result(filer,fp,line,1|2);
   sprintf(line,"<BR><BR>\n<HR><BR>\n");
   Nntpfiler_list_result(filer,fp,line,1|2);
   strcpy(old_server,serverdir);
 }

 sprintf(line,"</KBD></BODY></HTML>\n");
 Nntpfiler_list_result(filer,fp,line,1|2);

 ret= 1;
ex:; 
 if(fp!=NULL)
   fclose(fp);
 Sfile_ls("",&rfilec,&rfilev,2);
 filer->html= htmlrec;
 return(ret);
}


int Nntpfiler_install_file(filer,master_dir,target_dir,filename,content,flag)
struct NntpfileR *filer;
char *master_dir;
char *target_dir;
char *filename;
char *content;
int flag;
/*
 bit0= do not try to copy from master_dir
 bit2= do not overwrite target but return 2 if it already exists
*/
{
 char source[SfileadrL],target[SfileadrL],errmsg[SfileadrL];
 int ret,errn=0;

 sprintf(source,"%s/%s",master_dir,filename);
 sprintf(target,"%s/%s",target_dir,filename);
 if(flag&2)
   if(Sfile_size(target,1)!=-1)
     {ret= 2; goto ex;}
 if(!(flag&1) && Sfile_size(source,1)!=-1) {
   ret= Sfile_copy_file(source,target,&errn,errmsg,0);
   if(ret>0)
     goto ex;
 }
 ret= Sfile_string_to_file(target,content,1);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Nntpfiler_php3main(filer,flag)
struct NntpfileR *filer;
int flag;
{
 char *reply= NULL,adr[SfileadrL];
 int ret;

 ret= Nntpfiler_php3code_stic_dir(&reply,0);
 if(ret<=0)
   goto ex;
 ret= Nntpfiler_install_file(filer,"",filer->basedir,"stic_dir.php3",reply,3);
 if(ret<=0)
   goto ex;
 ret= Nntpfiler_php3code_subscribe(&reply,0);
 if(ret<=0)
   goto ex;
 ret= Nntpfiler_install_file(filer,"",filer->basedir,"subscribe.php3",reply,3);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 Sregex_string(&reply,NULL,0);
 return(ret);
}


int Nntpfiler_php3group(filer,dir,flag)
struct NntpfileR *filer;
char dir[];
int flag;
{
 char *reply= NULL,adr[SfileadrL];
 int ret;

 ret= Nntpfiler_php3code_snntp_doorman(&reply,0);
 if(ret<=0)
   goto ex;
 ret= Nntpfiler_install_file(filer,filer->basedir,dir,"snntp_doorman.php3",
                             reply,0);
 if(ret<=0)
   goto ex;

 ret= Nntpfiler_php3code_fetch(&reply,0);
 if(ret<=0)
   goto ex;
 ret= Nntpfiler_install_file(filer,filer->basedir,dir,"fetch.php3",reply,0);
 if(ret<=0)
   goto ex;

 ret= Nntpfiler_php3code_remove_old(&reply,0);
 if(ret<=0)
   goto ex;
 ret= Nntpfiler_install_file(filer,filer->basedir,dir,"remove_old.php3",
                             reply,0);
 if(ret<=0)
   goto ex;

 ret= Nntpfiler_php3code_list_interval(&reply,0);
 if(ret<=0)
   goto ex;
 ret= Nntpfiler_install_file(filer,filer->basedir,dir,"list_interval.php3",
                             reply,0);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 Sregex_string(&reply,NULL,0);
 return(ret);
}


/* --- automatically created functions which reproduce some PHP3 scripts --- */


int Nntpfiler_php3code_subscribe(reply,flag)
char **reply;
int flag;
{
 static char text[][120] = {
"<?php",
"",
"# automatically created and prone to overwriting by snntpbatch -subscribe",
"",
"# This script must be in a snntpbatch base directory above any server",
"# directory address (i.e.  $servername . \".port\" . $portnumber )",
"# In this base directory, there has to be an executable snntpbatch with",
"# the same semantics as the stic program.",
"",
"# Parameters :",
"# server        address of the news server",
"# group         name of the newsgroup",
"# user_id       AUTHINFO user id news for server",
"# password      AUTHINFO password for news server",
"# login         off = do not produce user_id and password entry fields",
"# force_rc      on = re-install $serverdir/.snntpbatchrc and -login",
"",
"$auth_script = \"authenticate.php3\";",
"if (file_exists($auth_script)!=false) {",
"  include $auth_script;",
"}",
"",
"function file_open ($fileadr,$mode) {",
"  $fp = fopen($fileadr,$mode);",
"  if ($fp == false) {",
"    echo \"<HTML><BODY>failed to fopen(\\\"$fileadr\\\",\\\"$mode\\\")</BODY></HTML>\\n\";",
"    exit();",
"  }",
"  return $fp;",
"}",
"",
"include \"stic_dir.php3\";",
"",
"if ($group == \"\") {",
"  echo \"<HTML><HEAD><TITLE>snntp subscribe</TITLE></HEAD><BODY>\\n\";",
"  echo \"<PRE>\\n\";",
"  echo \"<FORM method=\\\"get\\\" action=\\\"$SCRIPT_NAME\\\">\\n\";",
"  echo \"Subscribe to a newsgroup <HR>\\n\";",
"  if ($server == \"\") {",
"    echo \"server<BR><INPUT name=\\\"server\\\" size=60 maxlength=256><BR>\\n\";",
"  } else {",
"    echo \"on server <B>$server</B>\";",
"    echo \"<INPUT type=hidden name=\\\"server\\\" value=\\\"$server\\\"><BR>\\n\";",
"  }",
"  echo \"group<BR><INPUT name=\\\"group\\\" size=60 maxlength=256><BR>\";",
"  if ($server == \"\") {",
"    echo \"(groupname \\\"-\\\" subscribes to the server only)\\n\";",
"  }",
"  echo \"<HR>\\n\";",
"  if($login != \"off\") {",
"    echo \"A user id with password is only required if the news server\\n\";",
"    echo \"demands authentication and if this is the first time that\\n\";",
"    echo \"one of its newsgroups is subscribed. It will be ignored else\\n\";",
"    echo \"unless you activate \\\"force_rc\\\" .<BR>\\n\";",
"    echo \"user_id  <INPUT name=\\\"user_id\\\" size=16 maxlength=256>\\n\";",
"    echo \"password <INPUT type=password name=\\\"password\\\" size=16 maxlength=256>\\n\";",
"    echo \"force_rc <INPUT type=checkbox name=\\\"force_rc\\\" value=\\\"on\\\">\\n\";",
"    echo \"<HR>\\n\";",
"  } else {",
"    echo \"<INPUT type=hidden name=\\\"login\\\" value=\\\"off\\\">\";",
"  }",
"  echo \"<input type=submit value=\\\"subscribe\\\">\";",
"  echo \"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\";",
"  echo \"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\";",
"  echo \"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\";",
"  echo \"<input type=reset value=\\\"clear\\\">\\n\";",
"  echo \"</FORM>\\n\";",
"  echo \"</PRE>\\n\";",
"  echo \"</BODY></HTML>\\n\";",
"  exit();",
"}",
"",
"$cmd  = \"./snntpbatch \";",
"$cmd .= \"-basedir \\$(pwd) \";",
"$cmd .= \"-html php \" ;",
"$cmd .= \"-subscribe $server - $group \" ;",
"$cmd .= \" 2>&1\" ;",
"",
"# execute snntpbatch command",
"set_time_limit(3600);",
"$fd = popen($cmd,\"r\");",
"while (!feof($fd)) {",
"  $buffer .= fgets($fd, 4096);",
"}",
"$ret = pclose($fd);",
"if ($ret != 0) {",
"  echo \"<HTML><HEAD><TITLE>FAILURE</TITLE>\";",
"  echo \"</HEAD><BODY><PRE>\";",
"  echo \"$cmd\";",
"  echo \"<HR>\";",
"  echo $buffer;",
"  echo \"</PRE></BODY></HTML>\";",
"  exit();",
"}",
"sleep(1);",
"",
"$serverdir = $server . \".port119\";",
"$serverrc = $serverdir . \"/.snntpbatchrc\";",
"if( ($server != \"-\" && $server != \"\") && ",
"    (file_exists($serverrc) == false || $force_rc == \"on\")) {",
"  $fp = file_open($serverrc,\"w\");",
"  if($user_id != \"\") {",
"    $server_login = \"-login \\\"$user_id\\\" \\\"$password\\\"\";",
"    $server_hideout = \"$stic_dir/httpd_snntpbatch_server_dir/$server\";",
"    $fsh = file_open($server_hideout,\"w\");",
"    chmod($server_hideout,0600);",
"    fputs($fsh,\"$server_login\\n\");",
"    fclose($fsh);",
"    fputs($fp,\"-read_from \\\"$server_hideout\\\"\\n\");",
"  }",
"  fputs($fp,\"-html php\\n\");",
"  fputs($fp,\"-group_quick on\\n\");",
"  fputs($fp,\"-pacifier off\\n\");",
"  fputs($fp,\"-list_on_stdout off\\n\");",
"  fclose($fp);",
"}",
"",
"echo \"<HTML><HEAD><TITLE>DONE</HEAD></TITLE><BODY><PRE>\";",
"if($group != \"-\") {",
"  echo \"The group has been subscribed.\\n\";",
"  echo \"<HR>\";",
"  echo \"<A HREF=\\\"$serverdir/$group/list_index.html\\\">to group index (list_index.html)</A>\\n\";",
"}else if ($server != \"-\") {",
"  echo \"The server has been subscribed.\\n\";",
"  echo \"<HR>\";",
"  echo \"<A HREF=\\\"main_index.html\\\">to main index</A>\\n\";",
"} else {",
"  echo \"Done : -subscribe $server - $group\\n\";",
"  echo \"<HR>\";",
"  echo \"<A HREF=\\\"main_index.html\\\">to main index</A>\\n\";",
"}",
"echo \"</PRE></BODY></HTML>\";",
"",
"?>",
"@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
 };
 char line[120+4];
 int i,ret;

 if(Sregex_string(reply,"",0)<=0)
   {ret= -1; goto ex;};
 for(i=0;strcmp(text[i],
             "@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@")!=0;i++){
   sprintf(line,"%s\n",text[i]);
   if(Sregex_string(reply,line,1)<=0)
     {ret= -1; goto ex;};
 }
 ret= 1;
ex:;
 return(ret);
}


int Nntpfiler_php3code_stic_dir(reply,flag)
char **reply;
int flag;
{
 static char text[][120] = {
"<?php",
"# $stic_dir = \"/home/thomas/stic_dir\";",
"",
"function stic_file_open ($fileadr,$mode) {",
"  $fp = fopen($fileadr,$mode);",
"  if ($fp == false) {",
"    echo \"<HTML><BODY>failed to fopen(\\\"$fileadr\\\",\\\"$mode\\\")</BODY></HTML>\\n\";",
"    exit();",
"  }",
"  return $fp;",
"}",
"function read_stic_dir () {",
"  global $stic_dir;",
"",
"  $fp = stic_file_open(\".stic_main_dir\",\"r\");",
"  $stic_dir = fgets($fp,1024);",
"  $pos = strpos($stic_dir,\"\\n\");",
"  if($pos != false) {",
"    $stic_dir= substr($stic_dir,0,$pos);",
"  }",
"  fclose($fp);",
"}",
"read_stic_dir();",
"?>",
"@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
 };
 char line[120+4];
 int i,ret;

 if(Sregex_string(reply,"",0)<=0)
   {ret= -1; goto ex;};
 for(i=0;strcmp(text[i],
             "@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@")!=0;i++){
   sprintf(line,"%s\n",text[i]);
   if(Sregex_string(reply,line,1)<=0)
     {ret= -1; goto ex;};
 }
 ret= 1;
ex:;
 return(ret);
}


int Nntpfiler_php3code_snntp_doorman(reply,flag)
char **reply;
int flag;
{
 static char text[][120] = {
"<?php",
"",
"# automatically created and prone to overwriting by snntpbatch -subscribe",
"",
"# This script normally must be in a snntpbatch group directory below a ",
"# regular server directory address (i.e.  $servername . \".port\" . $portnumber )",
"",
"# Parameters see global list of function doorman",
"",
"function doorman( $door_layers, $anywhere, $check_maintenance) {",
"#   $door_layers int        how far to look up for authenticate scripts",
"#   $anywhere    bool       do not demand to be in a group directory",
"#                           and do not determine server parameters",
"#   $check_maintenance      exit() if file ./maintenance_lock is found",
"",
"  global $PHP_AUTH_USER;",
"  global $PHP_AUTH_PW;",
"  # output parameters if anywhere == false",
"  global $wd;",
"  global $basedir;",
"  global $server;",
"  global $port;",
"  global $group;",
"",
"  # Try bottom up wether there are authentication scripts to include",
"  $prefix = \"\";",
"  for ($i = 0 ; $i < $door_layers ; $i++) {",
"    if (file_exists($prefix . \"authenticate.php3\") == true) {",
"      include ($prefix . \"authenticate.php3\");",
"  break;",
"    }",
"    $prefix .= \"../\";",
"  }",
"",
"  if ($anywhere == false) {",
"    # learn basedir servername, portnumber and groupname from current directory",
"    unset($answer);",
"    exec(\"pwd\",$answer);",
"    $wd = $answer[0];",
"    $server = basename(dirname($wd));",
"    $dtp = strrpos($server,\".\");",
"    if ($dtp == false) {",
"      echo \"<HTML><HEAD><TITLE>FAILURE NOTICE</TITLE></HEAD><BODY><PRE>\";",
"      echo \"False installation detected.\";",
"      echo \"Parent directory of this script is not a snntpbatch server directory :\";",
"      echo \"$server\";",
"      echo \"</PRE></BODY></HTML>\";",
"      exit();",
"    } else {",
"      $port = substr($server, $dtp + 5);",
"      $server = substr($server,0,$dtp);",
"    }",
"    $group = basename($wd);",
"    $basedir = dirname(dirname($wd));",
"  }",
"",
"  if ($check_maintenance == true) {",
"    if (file_exists(\"maintenance_lock\") == true) {",
"      echo \"<HTML><HEAD><TITLE>FAILURE NOTICE</TITLE></HEAD><BODY><PRE>\";",
"      echo \"Found file \\\"maintenance_lock\\\" . Please try again later.\";",
"      echo \"</PRE></BODY></HTML>\";",
"      exit();",
"    }",
"  }",
"}",
"?>",
"@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
 };
 char line[120+4];
 int i,ret;

 if(Sregex_string(reply,"",0)<=0)
   {ret= -1; goto ex;};
 for(i=0;strcmp(text[i],
             "@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@")!=0;i++){
   sprintf(line,"%s\n",text[i]);
   if(Sregex_string(reply,line,1)<=0)
     {ret= -1; goto ex;};
 }
 ret= 1;
ex:;
 return(ret);
}


int Nntpfiler_php3code_fetch(reply,flag)
char **reply;
int flag;
{
 static char text[][120] = {
"<?php",
"",
"# automatically created and prone to overwriting by snntpbatch -subscribe",
"",
"# this script must be in a snntpbatch group directory",
"",
"# Parameters :",
"#   $msg    message number",
"#   $load   wether to fetch the message via snntpbatch :",
"#           1= load message if not already present",
"#          >1= load message in any case",
"#       else   eventually show intermediate page and call with load=1",
"#              If empty($with_intermediate_page) , it works like $load == 1",
"#",
"",
"require \"snntp_doorman.php3\";",
"doorman(3,false,true);",
"",
"# If set display an annoying intermediate page while loading ",
"# $with_intermediate_page = 1;",
"",
"# force $msgno to be a pure number",
"$msgno = 1 + $msg;",
"$msgno = $msgno - 1;",
"if ($msgno > 0) {",
"  $msgname = sprintf(\"msg%08d.html\",$msgno);",
"} ",
"$cmd  = \"$basedir/snntpbatch \";",
"$cmd .= \"-basedir $basedir \";",
"$cmd .= \"-html php \" ;",
"$cmd .= \"-pacifier off \" ;",
"$cmd .= \"-overwrite on \";",
"if (!(empty($server) || empty($group))) {",
"  $cmd .= \"-server $server \";",
"  $cmd .= \"-group $group \";",
"}",
"$cmd .= \"-fetch \" . $msg ;",
"$cmd .= \" 2>&1\" ;",
"",
"if (!empty($with_intermediate_page)) {",
"  if (file_exists($msgname) == false && empty($load)) {",
"    echo \"<HTML><HEAD><TITLE>creating : $msgname</TITLE></HEAD><BODY><PRE>\";",
"    echo \"now executing:\\n\";",
"    echo \"<HR>\";",
"    echo \"$cmd\";",
"    echo \"<HR>\";",
"",
"    echo \"please be patient ... \\n<BR>\\n\";",
"    echo \"(THIS DOCUMENT IS  N O T  DONE YET. \";",
"    echo \"You will get informed if the download fails.)\\n\";",
"    echo \"<meta http-equiv=\\\"refresh\\\" content=\\\"0;\";",
"    echo \" URL=$SCRIPT_NAME?msg=$msg&load=1\\\">\\n\";",
"    echo \"</PRE></BODY></HTML>\";",
"    exit();",
"  }",
"} else {",
"  if (file_exists($msgname) == false) {",
"    $load = 1;",
"  }",
"}",
"",
"if (!empty($load)) {",
"  if ($load > 1 || ($load == 1 && file_exists($msgname) == false)) {",
"    set_time_limit(3600);",
"    $fd = popen($cmd,\"r\");",
"    while (!feof($fd)) {",
"      $buffer .= fgets($fd, 4096);",
"    }",
"    $ret = pclose($fd);",
"    sleep(1);",
"  }",
"}",
"",
"if (file_exists($msgname) == false) {",
"  echo \"<HTML><HEAD><TITLE>FAILURE</TITLE></HEAD><BODY><PRE>\";",
"  echo \"not created : $msgname\\n\";",
"  echo \"<HR>\";",
"  echo \"$cmd\\n\";",
"  echo \"returned $ret\\n\";",
"  echo \"<HR>\";",
"  echo $buffer;",
"  echo \"</PRE></BODY></HTML>\";",
"} else {",
"  readfile($msgname);",
"}",
"",
"?>",
"@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
 };
 char line[120+4];
 int i,ret;

 if(Sregex_string(reply,"",0)<=0)
   {ret= -1; goto ex;};
 for(i=0;strcmp(text[i],
             "@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@")!=0;i++){
   sprintf(line,"%s\n",text[i]);
   if(Sregex_string(reply,line,1)<=0)
     {ret= -1; goto ex;};
 }
 ret= 1;
ex:;
 return(ret);
}


int Nntpfiler_php3code_list_interval(reply,flag)
char **reply;
int flag;
{
 static char text[][120] = {
"<?php",
"",
"# automatically created and prone to overwriting by snntpbatch -subscribe",
"",
"# This script must be in a snntpbatch group directory below a regular server",
"# directory address (i.e.  $servername . \".port\" . $portnumber )",
"# In the base directory above the server directory, there has to be an",
"# executable snntpbatch with the same semantics as the stic program.",
"",
"# Parameters : none",
"",
"require \"snntp_doorman.php3\";",
"doorman(3,false,true);",
"",
"$cmd  = \"$basedir/snntpbatch \";",
"$cmd .= \"-basedir $basedir \";",
"$cmd .= \"-html php \" ;",
"$cmd .= \"-list_interval new end \" ;",
"$cmd .= \" 2>&1\" ;",
"",
"# execute snntpbatch command",
"set_time_limit(3600);",
"$fd = popen($cmd,\"r\");",
"while (!feof($fd)) {",
"  $buffer .= fgets($fd, 4096);",
"}",
"$ret = pclose($fd);",
"if ($ret != 0) {",
"  echo \"<HTML><HEAD><TITLE>FAILURE</TITLE>\";",
"  echo \"</HEAD><BODY><PRE>\";",
"  echo \"$cmd\";",
"  echo \"<HR>\";",
"  echo $buffer;",
"  echo \"</PRE></BODY></HTML>\";",
"  exit();",
"}",
"sleep(1);",
"",
"echo \"<HTML><HEAD><TITLE>SUCCESS: index updated</HEAD></TITLE><BODY><PRE>\";",
"echo \"The group index has been updated.\\n\";",
"echo \"<HR>\";",
"echo \"<A HREF=\\\"list_index.html\\\">to group index (list_index.html)</A>\\n\";",
"echo \"</PRE></BODY></HTML>\";",
"?>",
"@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
 };
 char line[120+4];
 int i,ret;

 if(Sregex_string(reply,"",0)<=0)
   {ret= -1; goto ex;};
 for(i=0;strcmp(text[i],
             "@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@")!=0;i++){
   sprintf(line,"%s\n",text[i]);
   if(Sregex_string(reply,line,1)<=0)
     {ret= -1; goto ex;};
 }
 ret= 1;
ex:;
 return(ret);
}


int Nntpfiler_php3code_remove_old(reply,flag)
char **reply;
int flag;
{
 static char text[][120] = {
"<?php",
"",
"# automatically created and prone to overwriting by snntpbatch -subscribe",
"",
"# This script must be in a snntpbatch group directory below a regular server ",
"# directory address (i.e.  $servername . \".port\" . $portnumber )",
"# In the base directory above the server directory, there has to be an",
"# executable snntpbatch with the same semantics as the stic program.",
"",
"# Parameters : none",
"",
"require \"snntp_doorman.php3\";",
"doorman(3,false,true);",
"",
"# set the minimum age of a file to be cleaned away (no less than -3d)",
"$age = \"-3d\";",
"",
"# stay away from the .snntpbatchrc files in and above the group directory",
"chdir(\"/\");",
"",
"$cmd  = \"$basedir/snntpbatch \";",
"$cmd .= \"-basedir $basedir \" ;",
"$cmd .= \"-html php \" ;",
"$cmd .= \"-remove_old_files $server $port $group $age \" ;",
"$cmd .= \"-remove_old_ids $age \" ;",
"$cmd .= \" 2>&1\" ;",
"",
"# execute snntpbatch command",
"set_time_limit(3600);",
"$fd = popen($cmd,\"r\");",
"while (!feof($fd)) {",
"  $buffer .= fgets($fd, 4096);",
"}",
"pclose($fd);",
"chdir($wd);",
"sleep(1);",
"",
"if (file_exists(\"maintenance_lock\") == true) {",
"  unlink(\"maintenance_lock\");",
"  echo \"<HTML><HEAD><TITLE>FAILED: cleaning and adjusting group</TITLE>\";",
"  echo \"</HEAD><BODY><PRE>\";",
"  echo \"$cmd\";",
"  echo \"<HR>\";",
"  echo $buffer;",
"  echo \"</PRE></BODY></HTML>\";",
"} else {",
"  echo \"<HTML><HEAD><TITLE>DONE: cleaning and adjusting group</TITLE>\";",
"  echo \"</HEAD><BODY><PRE>\";",
"  echo \"Group $group cleaned ($age) and adjusted.\\n<HR>\";",
"  echo \"Size now (KBytes) :<BR>\\n\";",
"  system(\"du -k .\");",
"  echo \"<HR>\";",
"  echo \"<A HREF=\\\"list_index.html\\\">to group index (list_index.html)</A>\\n\";",
"  echo \"</PRE></BODY></HTML>\";",
"}",
"?>",
"@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
 };
 char line[120+4];
 int i,ret;

 if(Sregex_string(reply,"",0)<=0)
   {ret= -1; goto ex;};
 for(i=0;strcmp(text[i],
             "@END_OF_TEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@")!=0;i++){
   sprintf(line,"%s\n",text[i]);
   if(Sregex_string(reply,line,1)<=0)
     {ret= -1; goto ex;};
 }
 ret= 1;
ex:;
 return(ret);
}


int Nntpfiler_write_coop_idx(filer,coop_idx_file,idx,flag)
struct NntpfileR *filer;
char *coop_idx_file;
int idx;
int flag;
{
 int ret;
 char temp_adr[SfileadrL],final_adr[SfileadrL],idx_text[81],hostname[257];

 Nntpfiler_std_dir(filer,final_adr,0);
 sprintf(final_adr+strlen(final_adr),"/%s",coop_idx_file);
 strcpy(temp_adr,final_adr);
 ret= gethostname(hostname,256);
 if(ret==-1)
   strcpy(hostname,"unknownhostname");
 sprintf(temp_adr+strlen(temp_adr),"_%ld+at+%s",getpid(),hostname);
 sprintf(idx_text,"%ld",idx);
 ret= Sfile_string_to_file(temp_adr,idx_text,1);
 if(ret<=0)
   goto ex;
 rename(temp_adr,final_adr);
 ret= 1;
ex:;
 return(ret);
}


int Nntpfiler_read_coop_idx(filer,coop_idx_file,idx,flag)
struct NntpfileR *filer;
char *coop_idx_file;
int *idx;
int flag;
{
 int ret;
 char adr[SfileadrL];

 *idx= -1;
 Nntpfiler_std_dir(filer,adr,0);
 sprintf(adr+strlen(adr),"/%s",coop_idx_file);
 ret= Sfile_file_to_int(adr,idx,2);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


