
/*
  cc -g -c nntpfilter.c
*/

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* Macro for allocation of object arrays */
#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));


#include "../s_tools/sregex.h"
#include "../s_tools/sfile.h"

#include "nntpconn.h"
#include "nntphead.h"
#include "nntpfilter.h"



/* ----------------------- Nttpftest ----------------------- */


int Nntpftest_new(ftest,flag)
struct NntpftesT **ftest;
int flag;
{
 struct NntpftesT *f;

 *ftest= f= TSOB_FELD(struct NntpftesT,1);
 if(f==NULL)
   return(-1);
 f->code= NULL;
 f->expr_text= NULL;
 f->re= NULL;
 f->invert= 0;
 f->head= 0;
 f->is_filetest= 0;
 f->filetest_type= 0;
 f->filename[0]= 0;
 f->re_filename= NULL;
 f->sfrec_dirlistpt= NULL;
 f->sfrec_namept= NULL;
 return(1);
}


int Nntpftest_destroy(ftest,flag)
struct NntpftesT **ftest;
int flag;
{
 struct NntpftesT *f;
 
 f= *ftest;
 if(f==NULL)
   return(0);

 Sregex_string(&(f->code),NULL,0);
 Sregex_string(&(f->expr_text),NULL,0);
 if(f->re!=NULL)
   Sregex_destroy(&(f->re),0);
 if(f->re_filename!=NULL)
   Sregex_destroy(&(f->re_filename),0);

 free((char *) f);
 *ftest= NULL;
 return(1);
}


int Nntpftest_set_regex(ftest,code,re_text,flag)
struct NntpftesT *ftest;
char *code;
char *re_text;
int flag;
/*
 bit0= do not compile regular expressions (used for comparison samples)
*/
{
 int ret= 1,is_regex= 0;

 if(Sregex_string(&(ftest->code),code,0)<=0)
   return(-1);
 if(Sregex_string(&(ftest->expr_text),re_text,0)<=0)
   return(-1);

 ftest->is_filetest= 0;
 if(strcmp(code,"from:")==0
    || strcmp(code,"date:")==0
    || strcmp(code,"subject:")==0
    || strcmp(code,"subjectfile:")==0
    || strcmp(code,"msgno:")==0
    || strcmp(code,"msgid:")==0
    || strcmp(code,"references:")==0
    || strcmp(code,"bytes:")==0
    || strcmp(code,"lines:")==0
   ){
   is_regex= 1;
 } else if(strcmp(code,"filename:")==0){
   ftest->is_filetest= 1;
   ftest->filetest_type= 0;
   is_regex= 1;
 } else if(strcmp(code,"subjectfilefound:")==0) {
   ftest->is_filetest= 1;
   ftest->filetest_type= 2;
 } else if(strcmp(code,"subjectfilerec:")==0) {
   ftest->is_filetest= 1;
   ftest->filetest_type= 3;
 } else if(strcmp(code,"smaller:")==0
          || strcmp(code,"larger:")==0
          || strcmp(code,"return:")==0
          || strcmp(code,"print:")==0
          ) {
   ;
 } else
   ret= 0;
 if(ret>0 && is_regex) {
   if(!(flag&1))
     ret= Sregex_replace(&(ftest->re),re_text,0);
 }

 return(ret);
}


int Nntpftest_set_sample(ftest,code,re_text,flag)
struct NntpftesT *ftest;
char *code;
char *re_text;
int flag;
{
 int ret= 1;
 ret= Nntpftest_set_regex(ftest,code,re_text,1); /* without regex compiling */
 return(ret);
}


int Nntpftest_set_head(ftest,head,flag)
struct NntpftesT *ftest;
struct NntpheaD *head;
int flag;
{
 ftest->head= head;
 return(1);
}


int Nntpftest_link_sfrec(ftest,dirlistpt,namept,flag)
struct NntpftesT *ftest;
char **dirlistpt,**namept;
int flag;
{
 ftest->sfrec_dirlistpt= dirlistpt;
 ftest->sfrec_namept= namept;
 return(1);
}


int Nntpftest_get_property(ftest,code,prop_text,prop_size,flag)
struct NntpftesT *ftest;
char *code;
char *prop_text;
int prop_size;
int flag;
{
 int ret;

 prop_text[0]= 0;

 if(   strcmp(code,"from:")==0 
    || strcmp(code,"date:")==0
    || strcmp(code,"subject:")==0
    || strcmp(code,"subjectfile:")==0
    || strcmp(code,"smaller:")==0
    || strcmp(code,"larger:")==0
    || strcmp(code,"msgno:")==0
    || strcmp(code,"msgid:")==0
    || strcmp(code,"references:")==0
    || strcmp(code,"bytes:")==0
    || strcmp(code,"lines:")==0
    ){
   ret= Nntphead_get_property(ftest->head,code,prop_text,prop_size,0);
   return(ret);
 } else if(strcmp(code,"return:")==0
    || strcmp(code,"print:")==0
    ){
   return(1);

 } else if(strcmp(code,"filename:")==0){
   if(ftest->filename[0]!=0)
     strncpy(prop_text,ftest->filename,prop_size-1);
 } else if(strcmp(code,"dirlist:")==0){ /* an alias for subjectfilefound: */
   if(ftest->expr_text!=NULL)
     if(ftest->expr_text[0]!=0)
       strncpy(prop_text,ftest->expr_text,prop_size-1);
 } else
   return(0);
 prop_text[prop_size-1]= 0; /* just in case that strncpy truncated */
 return(1);
}


int Nntpftest_prop_match(ftest,code,re,re_text,reason,flag)
struct NntpftesT *ftest;
char *code;
struct SregeX *re;
char *re_text;
char *reason;
int flag;
/*
 bit0= match also if code is not found in header
 bit1= this is a non regex test
*/
{
 int ret,size,limit,i;
 char line[Nntphead_strL];

 if(!(flag&2))
   if(re==NULL) /* empty expression is always a match */
     return(1);

 ret= Nntpftest_get_property(ftest,code,line,sizeof(line),0);
 if(ret<0) {
   return(ret);
 } else if (ret==0) {
   if(!(flag&1)){
     sprintf(reason,"Cannot get header property '%s'",code);
     return(0);
   }
 } else {
   if(flag&2) {
     size= 0;
     sscanf(line,"%ld",&size);
     limit= 0;
     if(strcmp(code,"larger:")==0) {
       sscanf(re_text,"%ld",&limit);
       if(size<=limit)
         goto failed;
     } else if(strcmp(code,"smaller:")==0) {
       sscanf(re_text,"%ld",&limit);
       if(size>=limit)
         goto failed;
     } else if(strcmp(code,"return:")==0) {
       if(strcmp(re_text,"1")==0 
          || strcmp(re_text,"true")==0
          || strcmp(re_text,"match")==0) {
         return(3);                                          /* forced match */
       } else {
         sprintf(reason,"Unconditional mismatch :\n(-filter) %s '%s'",
                        code,re_text); 
         return(2);                                       /* forced mismatch */
       }
     } else if(strcmp(code,"print:")==0) {
       char sub_code[Nntpfilter_strL],sub_prop[Nntphead_strL],*dpt,*rpt,*ept;

       printf("print: ");

#ifndef NIX

       /* text substitution by embedded properties  $name: */
       rpt= re_text;
       while(*rpt!=0) {
         /* scan for embedded property name start */
         dpt= strchr(rpt,'$');
         if(dpt==NULL) {
           printf("%s",rpt);
           rpt= rpt+strlen(rpt);
       break;
         }
         /* print up to found name */
         for(;rpt<dpt;rpt++)
           printf("%c",*rpt);
         if(*dpt!=0 && *(dpt+1)=='$') {
           printf("%c",'$');
           rpt= dpt+2;
      continue;
         }
         /* scan for property name end */
         ept= strchr(dpt,':');
         if(ept==NULL) {
           printf("%s",dpt);
           rpt= rpt+strlen(rpt);
       break;
         }
         i= 0;
         for(rpt= dpt+1; rpt<=ept && i<sizeof(sub_code)-1; rpt++)
           sub_code[i++]= *rpt;
         sub_code[i]= 0;
         /* get and print property text */
         Nntpftest_get_property(ftest,sub_code,sub_prop,sizeof(sub_prop),0);
         printf("%s",sub_prop);
         rpt= ept+1;
       }
       printf("\n");

#else /* NIX */

       printf("%s\n",re_text);

#endif /* NIX */

     }
   } else {
     if(!Sregex_is_match(re,line,0)) {
failed:;
       sprintf(reason,"Property does not match :\n%s\n(-filter) %s '%s'",
                      line,code,re_text); 
       return(0);
     }
   }
 }
 return(1); 
}


int Nntpftest_add_sfrec(ftest,dirlist,name,flag)
struct NntpftesT *ftest;
char *dirlist,*name;
int flag;
{
 int ret;
 
 if(ftest->sfrec_dirlistpt==NULL || ftest->sfrec_namept==NULL) 
   return(2);
 if((*ftest->sfrec_dirlistpt)!=NULL) {
   ret=  Sregex_set_dyn_string(ftest->sfrec_dirlistpt," ",1);
   if(ret<0)
     goto ex;
 }
 ret=  Sregex_set_dyn_string(ftest->sfrec_dirlistpt,dirlist,1);
 if(ret<0)
   goto ex;
 ret=  Sregex_set_dyn_string(ftest->sfrec_namept,name,0);
 if(ret<0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Nntpftest_match(ftest,sample,reason,flag)
struct NntpftesT *ftest;
struct NntpftesT *sample;
char *reason;
int flag;
/*
flag:
 bit0-7= testmode: 0=head , 1=filename
return:
 <0 = error
  0 = does not match
  1 = does match
  2 = immediate decision : does not match
  3 = immediate decision : does match
*/
{
 int value=0,mode,ret;
 static int debug= 1;

 if(ftest==NULL)
   return(1);
 mode= flag&255;
 if(mode==0){
   if(ftest->is_filetest==0) {
     ret= Nntpftest_prop_match(sample,ftest->code,ftest->re,ftest->expr_text,
                               reason,(ftest->re==NULL)<<1);
     if(ret<=0)
       goto ex;
     if(ret>1) {
       value= ret;
       goto ex;
     }
   } else if(ftest->filetest_type==1) {
     if(!Nntpftest_prop_match(sample,"subjectfile:",ftest->re,ftest->expr_text,
                              reason,0))
       goto ex;
   } else if(ftest->filetest_type==2) { /* subjectfilefound: */
     char dirlist[SfileadrL*4],name[SfileadrL];
     if(debug>2) 
       fprintf(stderr,"\ntesting listed directories for name : \n");
     ret= Nntpftest_get_property(sample,"subjectfile:",name,sizeof(name),0);
     if(ret<0 || name[0]==0) {
       if(debug>2)
         fprintf(stderr,"Nntpftest_get_property(): ret= %ld, name= '%s'\n",
                 ret,name);
       goto ex;
     }
     if(debug>2)
       fprintf(stderr,"       %s\n",name);
     else if(debug>1)
       fprintf(stderr,"\ntesting listed directories for name : %s\n",name);
     Sfile_esc_filename(name,sizeof(name),8);
     ret= Nntpftest_get_property(ftest,"dirlist:",
                                 dirlist,sizeof(dirlist),0);
     if(ret<0) 
       goto ex;
     if(!Sfile_search_file(dirlist,name,0)) {
       sprintf(reason,"file %s not found in listed directories",name);
       if(debug>1)
         fprintf(stderr,"%s (len=%ld)\n",reason,strlen(dirlist));
       goto ex;
     }
     if(debug>0)
       fprintf(stderr,"file %s found in listed directories\n",name);

   } else if(ftest->filetest_type==3) { /* subjectfilerec: */
     char dirlist[SfileadrL*4],name[SfileadrL];
     int no_name= 0;
      
     ret= Nntpftest_get_property(sample,"subjectfile:",name,sizeof(name),0);
     if(ret<0)
       goto ex;
     if(ret==0 || name[0]==0)
       no_name= 1;
     Sfile_esc_filename(name,sizeof(name),8);
     ret= Nntpftest_get_property(ftest,"dirlist:",
                                 dirlist,sizeof(dirlist),0);
     if(ret<0)
       goto ex;
     ret= Nntpftest_add_sfrec(ftest,dirlist,name,0);
     if(ret<0)
       goto ex;
     ret= Sfile_string_to_hashdir(dirlist,name,"",1|4|((!no_name)<<4));
     value= (ret>0);
     goto ex;
   }
 }else if(mode==1){
   if(ftest->filetest_type==0) {
     if(!Nntpftest_prop_match(sample,"filename:",ftest->re_filename,
                                    sample->filename,reason,0))
       goto ex;
   }
 }
 value= 1;
ex:;
 if(ftest->invert && value<=1 && value>=0)
   value= !value;
 return(value);
}


/* ----------------------- Nttpfnode ----------------------- */


int Nntpfnode_new(fnode,filter,up,origin,flag)
struct NntpfnodE **fnode;
struct NntpfilteR *filter;
struct NntpfnodE *up;
char *origin;
int flag;
/*
 bit0= set invert-property
 bit1= set use_shortcuts
*/
{
 struct NntpfnodE *n;
 int ret,i;

 *fnode= n= TSOB_FELD(struct NntpfnodE,1);
 if(n==NULL)
   return(-1);
 for(i=0;i<sizeof(n->origin);i++)
   n->origin[i]= 0;
 strncpy(n->origin,origin,sizeof(n->origin)-1);
 n->up= up;
 n->invert= (flag&1);
 n->assoc= 0;
 n->use_shortcuts= !!(flag&2);
 n->left= NULL;
 n->left_op= -1;
 n->right= NULL;
 n->right_op= -1;
 n->sub= NULL;
 n->test= NULL;
 n->is_if_then_else= 0;
 n->true_branch= NULL;
 n->false_branch= NULL;
 ret= Nntpftest_new(&(n->test),0);
 if(ret<=0){
   Nntpfnode_destroy(fnode,0);
   return(-1);
 }
 Nntpftest_link_sfrec(n->test,
                      &(filter->sfrec_dirlist),&(filter->sfrec_name),0);
 n->own_value= -1;
 n->composed_value= -1;
 return(1);
}


int Nntpfnode_destroy(fnode,flag)
struct NntpfnodE **fnode;
int flag;
{
 if(*fnode==NULL)
   return(0);
 Nntpfnode_destroy(&((*fnode)->right),0);
 Nntpfnode_destroy(&((*fnode)->sub),0);
 Nntpfnode_destroy(&((*fnode)->sub),0);
 Nntpfnode_destroy(&((*fnode)->true_branch),0);
 Nntpfnode_destroy(&((*fnode)->false_branch),0);
 Nntpftest_destroy(&((*fnode)->test),0);
 free((char *) *fnode);
 *fnode= NULL;
 return(1);
}


int Nntpfnode_set_is_if(fnode,value,flag)
struct NntpfnodE *fnode;
int value;
int flag;
{
 fnode->is_if_then_else= value;
 return(1);
}


int Nntpfnode_is_if(fnode,flag)
struct NntpfnodE *fnode;
int flag;
{
  return(fnode->is_if_then_else);
}


int Nntpfnode_set_branch(fnode,target,flag)
struct NntpfnodE *fnode;
struct NntpfnodE *target;
int flag;
/*
 bit0= false_branch (else true_branch)
*/
{
 struct NntpfnodE **branch;

 if(flag&1) 
   branch= &(fnode->false_branch);
 else
   branch= &(fnode->true_branch);
 Nntpfnode_destroy(branch,0);
 (*branch)= target;
 return(1);
}


int Nntpfnode_get_branch(fnode,branch,flag)
struct NntpfnodE *fnode;
struct NntpfnodE **branch;
int flag;
/*
 bit0= false_branch (else true_branch)
*/
{
 if(flag&1) 
   (*branch)= fnode->false_branch;
 else
   (*branch)= fnode->true_branch;
 return(1);
}


int Nntpfnode_is_defined(fnode,flag)
struct NntpfnodE *fnode;
int flag;
{
 struct NntpftesT *ftest;

 if(fnode==NULL)
   return(0);
 if(fnode->sub!=NULL)
   return(1);
 ftest= fnode->test;
 if(ftest==NULL)
   return(0);
 if(ftest->expr_text!=NULL)
   return(1);
 if(ftest->re_filename!=NULL)
   return(1);
 if(ftest->filetest_type>=1)
   return(1);
 return(0);
}


int Nntpfnode_own_value(fnode,sample,reason,flag)
struct NntpfnodE *fnode;
struct NntpftesT *sample;
char *reason;
int flag;
/*
flag:
 bit0-7= testmode: 0=head , 1=filename
return: (also from Nntpftest_match() and Nntpfnode_tree_value() )
 <0 = error
  0 = does not match
  1 = does match
  2 = immediate decision : does not match
  3 = immediate decision : does match
*/
{
 int ret;

 if(fnode==NULL)
   return(1);
 if(fnode->sub!=NULL) {
   ret= Nntpfnode_tree_value(fnode->sub,-1,sample,reason,flag&255);
 } else {
   ret= Nntpftest_match(fnode->test,sample,reason,flag&255);
 }
 if(ret<0)
   return(ret);
 if(ret>1)
   return(ret);

#ifdef Nntpfilter_wronG  
 /* <<< now in  Nntpfnode_tree_value() */
 if(fnode->is_if_then_else) {
   struct NntpfnodE *branch;
   if(ret>0) 
     branch= fnode->true_branch;
   else 
     branch= fnode->false_branch;
   if(branch!=NULL) {
     ret= Nntpfnode_tree_value(branch,-1,sample,reason,flag&255);
     if(ret<0)
       return(ret);
     if(ret>1)
       return(ret);
   }
 }
#endif /* Nntpfilter_wronG */

 if(fnode->invert)
   ret= !ret;
 return(ret);
}


int Nntpfnode_op(value1,value2,op,flag)
int value1,value2,op,flag;
{
 int ret;

 if(op==0)
   ret= value1 || value2 ;
 else
   ret= value1 && value2 ;
 return(ret);
}


int Nntpfnode_tree_value(fnode,left_value,sample,reason,flag)
struct NntpfnodE *fnode;
int left_value;
struct NntpftesT *sample;
char *reason;
int flag;
/*
 bit0-7= testmode: 0=head , 1=filename
return: (also from Nntpftest_match() and Nntpfnode_own_value() )
 <0 = error
  0 = does not match
  1 = does match
  2 = immediate decision : does not match
  3 = immediate decision : does match
*/
{
 int value= 1,ret;

 if(fnode==NULL)
   return(1);
 if(!Nntpfnode_is_defined(fnode,0))
   return(1);

 if(fnode->use_shortcuts && fnode->left!=NULL){
   fnode->composed_value= left_value;
   if(fnode->left_op==0) {/* OR */
     if(left_value!=0) 
       goto ex;
   } else {                /* AND */
     if(left_value==0)
       goto ex;
   }
 }
 fnode->composed_value= fnode->own_value= 
     Nntpfnode_own_value(fnode,sample,reason,flag&255);
 if(fnode->own_value<0||fnode->own_value>1)
   return(fnode->own_value);

 if(fnode->assoc==0){ /* left associative */
   if(fnode->left!=NULL && left_value>=0)
     fnode->composed_value= 
       Nntpfnode_op(left_value,fnode->own_value,fnode->left_op,0);
   /* compute right value */
   /* is the right value relevant ? */
   if(fnode->right!=NULL){
     if(fnode->use_shortcuts){
       if(fnode->right_op==0) {/* OR */
         if(fnode->composed_value!=0)
           goto ex;
       } else {                /* AND */
         if(fnode->composed_value==0)
           goto ex;
       }  
     }
     value= Nntpfnode_tree_value(fnode->right,fnode->composed_value,
                                 sample,reason,flag&255);
     if(value<0 || value>1)
       return(value);
     fnode->composed_value= value;
   }
 }else{ /* right associative */
   if(fnode->right!=NULL){
     /* is the right value relevant ? */
     if(fnode->use_shortcuts){
       if(fnode->right_op==0) {/* OR */
         if(fnode->composed_value!=0)
           goto ex;
       } else {                /* AND */
         if(fnode->composed_value==0)
           goto ex;
       }
     }
     value= Nntpfnode_tree_value(fnode->right,fnode->own_value,
                                 sample,reason,flag&255);
     if(value<0||value>1)
       return(value);
   } else
     value= fnode->own_value;
   fnode->composed_value= value;
   if(fnode->left!=NULL && left_value>=0)
     fnode->composed_value=
       Nntpfnode_op(left_value,fnode->composed_value,fnode->left_op,0);
 }
ex:
 ret= fnode->composed_value;
 if(fnode->is_if_then_else) {
   /* The if-condition is evaluated. Now follow the chosen branch */
   struct NntpfnodE *branch;
   if(ret>0) 
     branch= fnode->true_branch;
   else 
     branch= fnode->false_branch;
   if(branch!=NULL) {
     ret= Nntpfnode_tree_value(branch,-1,sample,reason,flag&255);
     if(ret<0)
       return(ret);
     if(ret>1)
       return(ret);
   }
   fnode->composed_value= ret;
 }
 return(fnode->composed_value);
}


/* --------------------- Nntpfilter -------------------- */


int Nntpfilter_new(filter,msgno_start,msgno_end,flag)
struct NntpfilteR **filter;
int msgno_start;
int msgno_end;
int flag;
/*
 bit0= do not create and attach a new node
 bit1= set use_shortcuts
*/
{
 struct NntpfilteR *f;
 int ret;

 *filter= f= TSOB_FELD(struct NntpfilteR,1);
 if(f==NULL)
   return(-1);
 f->msgno_start= msgno_start;
 f->msgno_end= msgno_end;
 f->msgno_idx= 0;
 f->msgno_prev[0]= f->msgno_prev[1]= 0;
 f->root= NULL;
 f->cursor= NULL;
 f->invert= 0;
 f->use_shortcuts= !!(flag&2);
 f->sfrec_dirlist= NULL;
 f->sfrec_name= NULL;
 f->sfrec_name_loaded= 0;
 f->errmsg[0]= 0;
 f->errn= 0;

 if(!(flag&1)){
   ret= Nntpfnode_new(&(f->root),f,NULL,"-filter",(f->use_shortcuts)<<1);
   if(ret<=0){
     Nntpfilter_destroy(filter,0);
     return(-1);
   }
   f->cursor= f->root;
 }
 return(1);
}


int Nntpfilter_destroy(filter,flag)
struct NntpfilteR **filter;
int flag;
{
 if(*filter==NULL)
   return(0);
 if((*filter)->root)
   Nntpfnode_destroy(&((*filter)->root),0); 
 Sregex_set_dyn_string(&((*filter)->sfrec_dirlist),NULL,0);
 Sregex_set_dyn_string(&((*filter)->sfrec_name),NULL,0);
 free((char *) *filter);
 *filter= NULL;
 return(1);
}


int Nntpfilter_set_sfn_loaded(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 filter->sfrec_name_loaded= 1;
 return(1);
}


int Nntpfilter_cursor_complete(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 int ret;
 struct NntpfnodE *fnode;
 struct NntpftesT *ftest;

 if(filter==NULL)
   return(0);
 ret= Nntpfnode_is_defined(filter->cursor,0);
 return(ret);
}


int Nntpfilter_is_restrictive(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 if(filter==NULL)
   return(0);
 if(filter->root==NULL)
   return(0);
 if(!Nntpfnode_is_defined(filter->root,0))
   return(0);
 return(1);
}


int Nntpfilter_head_match(filter,head,reason,flag)
struct NntpfilteR *filter;
struct NntpheaD *head;
char *reason;
int flag;
{
 int ret;
 struct NntpftesT *sample=NULL;
 char line[Nntphead_strL];

 if(filter==NULL)
   return(1);
 Sregex_set_dyn_string(&(filter->sfrec_dirlist),NULL,0);
 Sregex_set_dyn_string(&(filter->sfrec_name),NULL,0);
 filter->sfrec_name_loaded= 0;
 ret= Nntpftest_new(&sample,0);
 if(ret<=0)
   return(0);

 Nntpftest_set_head(sample,head,0);

 ret= Nntpfnode_tree_value(filter->root,-1,sample,reason,0);
 if(ret<0) 
   ret= 0;
 else if(ret==2)
   ret= 0;
 else if(ret==3)
   ret= 1;
ex:;
 Nntpftest_destroy(&sample,0);
 return(ret);
}


int Nntpfilter_filename_match(filter,filename,reason,flag)
struct NntpfilteR *filter;
char *filename;
char *reason;
int flag;
{
 int ret;
 struct NntpftesT *sample;

 return(1); /* >>> out of work */

 if(filter==NULL)
   return(1);
 ret= Nntpftest_new(&sample,0);
 if(ret<=0)
   return(0);
 Nntpftest_set_sample(sample,"filename:",filename,0);
 ret= Nntpfnode_tree_value(filter->root,-1,sample,reason,1);
 if(ret<0) 
   ret= 0;
 return(ret);
}


int Nntpfilter_set_regex(filter,code,re_text,flag)
struct NntpfilteR *filter;
char *code;
char *re_text;
int flag;
{
 struct NntpftesT *ftest;
 int ret;

 ftest= filter->cursor->test;
 ret= Nntpftest_set_regex(ftest,code,re_text,0);
 return(ret);
}


int Nntpfilter_new_node(filter,fnode,origin,flag)
struct NntpfilteR *filter;
struct NntpfnodE **fnode;
char *origin;
int flag;
/*
 bit0= open new branch 
 bit1= with bit1 : do not register as sub-node of filter->cursor
*/
{
 int ret;
 struct NntpfnodE *f;

 ret= Nntpfnode_new(fnode,filter,NULL,origin,
                    filter->invert|((filter->use_shortcuts)<<1));
 if(ret<=0)
   return(ret);
 f= *fnode;
 if(flag&1) {
   f->up= filter->cursor;
   if(filter->cursor!=NULL && !(flag&2)) {
     if(filter->cursor->sub!=NULL) {
       /* This would become a memory leak */
       filter->errn= -2;
       sprintf(filter->errmsg,
               "Program error while parsing -filter : sub branch overwrite");
       return(0);
     } else
       filter->cursor->sub= f;
   }
 } else {
   f->up= filter->cursor->up;
   f->left= filter->cursor;
   if(filter->cursor!=NULL)
     filter->cursor->right= f;
 }
 filter->invert= 0;
 return(1);
}


int Nntpfilter_open_bracket(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 int ret,make_new= 0;
 struct NntpfnodE *fnode;

 if(Nntpfilter_cursor_complete(filter,0)) {
   filter->errn= -2;
   sprintf(filter->errmsg,"Binary operator expected, opening-bracket found");
   return(0);
 } 
 ret= Nntpfilter_new_node(filter,&fnode,"-sub",1);
 if(ret<=0)
   return(ret);
 filter->cursor= fnode;
 return(1);
}


int Nntpfilter_close_bracket(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 int ret;

 if(!Nntpfilter_cursor_complete(filter,0)) {
   filter->errn= -3;
   sprintf(filter->errmsg,
     "Unary operator or expression expected, closing-bracket found");
   return(0);
 }

 if(filter->cursor->up==NULL){
   filter->errn= -1;
   sprintf(filter->errmsg,
           "No bracket open when encountering closing bracket.");
   return(0);
 }
 filter->cursor= filter->cursor->up;
 return(1);
}


int Nntpfilter_not(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 int ret;

 if(Nntpfilter_cursor_complete(filter,0)) {
   filter->errn= -2;
   sprintf(filter->errmsg,"Binary operator expected, unary operator found");
   return(0);
 } 
 filter->cursor->invert= !filter->cursor->invert;
 return(1);
}


int Nntpfilter_and(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 int ret;
 struct NntpfnodE *fnode;

 if(!Nntpfilter_cursor_complete(filter,0)) {
   filter->errn= -3;
   sprintf(filter->errmsg,
    "Unary operator or expression expected, binary operator found");
   return(0);
 }

 ret= Nntpfilter_new_node(filter,&fnode,"-and",0);
 if(ret<=0)
   return(ret);
 filter->cursor->right_op= 1;
 filter->cursor->assoc= 1;     /* compute right side first */
 fnode->left_op= 1;
 fnode->assoc= 0;              /* compute left side first */
 filter->cursor= fnode;
 return(1);
}


int Nntpfilter_or(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 int ret;
 struct NntpfnodE *fnode;

 if(!Nntpfilter_cursor_complete(filter,0)) {
   filter->errn= -3;
   sprintf(filter->errmsg,
    "Unary operator or expression expected, binary operator found");
   return(0);
 }

 ret= Nntpfilter_new_node(filter,&fnode,"-or",0);
 if(ret<=0)
   return(ret);
 filter->cursor->right= fnode;
 filter->cursor->right_op= 0;
                                    /* if existing : compute left side first */
 filter->cursor->assoc= (filter->cursor->left==NULL);
 fnode->left= filter->cursor;
 fnode->left_op= 0;
 fnode->assoc= 0;            /* no right side yet : compute left side first */
 filter->cursor= fnode;
 return(1);
}


int Nntpfilter_if(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 int ret;
 struct NntpfnodE *fnode;

 if(Nntpfilter_cursor_complete(filter,0)) {
   filter->errn= -2;
   sprintf(filter->errmsg,"Binary operator expected, -if-mark found");
   return(0);
 } 
 ret= Nntpfilter_new_node(filter,&fnode,"-if",1);
 if(ret<=0)
   return(ret);
 Nntpfnode_set_is_if(fnode,1,0);
 filter->cursor= fnode;
 return(1);
}


int Nntpfilter_then(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 int ret;
 struct NntpfnodE *fnode,*branch= NULL;

 if(! Nntpfilter_cursor_complete(filter,0)) {
   filter->errn= -3;
   sprintf(filter->errmsg,
           "Unary operator or expression expected, -then-operator found");
   return(0);
 } 
 /* Finding the -if that matches this -then
    Do not go up one node but look for the leftmost one.
    If everything is right we are at level of the -if node */
 while(filter->cursor->left!=NULL)
   filter->cursor= filter->cursor->left;
 Nntpfnode_get_branch(filter->cursor,&branch,0);
 if(!Nntpfnode_is_if(filter->cursor,0) || branch!=NULL) {
improper_range:;
   filter->errn= -5;
   sprintf(filter->errmsg,"-then-operator found outside its proper range.");
   return(0);
 }
 ret= Nntpfilter_new_node(filter,&fnode,"-then",1|2);
 if(ret<=0)
   return(ret);
 Nntpfnode_set_branch(filter->cursor,fnode,0);
 filter->cursor= fnode;
 return(1);
}


int Nntpfilter_else(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 int ret;
 struct NntpfnodE *fnode,*true_branch,*false_branch;

 if(! Nntpfilter_cursor_complete(filter,0)) {
   filter->errn= -3;
   sprintf(filter->errmsg,
           "Unary operator or expression expected, -else-operator found");
   return(0);
 } 
 if(filter->cursor->up==NULL)
   goto improper_range;
 filter->cursor= filter->cursor->up;
 Nntpfnode_get_branch(filter->cursor,&true_branch,0);
 Nntpfnode_get_branch(filter->cursor,&false_branch,1);
 if(!Nntpfnode_is_if(filter->cursor,0) || 
    true_branch==NULL || false_branch!=NULL) {
improper_range:;
   filter->errn= -5;
   sprintf(filter->errmsg,"-else-operator found outside its proper range.");
   return(0);
 }
 ret= Nntpfilter_new_node(filter,&fnode,"-else",1|2);
 if(ret<=0)
   return(ret);
 Nntpfnode_set_branch(filter->cursor,fnode,1);
 filter->cursor= fnode;
 return(1);
}


int Nntpfilter_elseif(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 int ret;
 struct NntpfnodE *fnode,*true_branch,*false_branch;

 if(!Nntpfilter_cursor_complete(filter,0)) {
   filter->errn= -3;
   sprintf(filter->errmsg,
           "Unary operator or expression expected, -elseif-operator found");
   return(0);
 } 
 if(filter->cursor->up==NULL)
   goto improper_range;
 filter->cursor= filter->cursor->up;
 Nntpfnode_get_branch(filter->cursor,&true_branch,0);
 Nntpfnode_get_branch(filter->cursor,&false_branch,1);
 if(!Nntpfnode_is_if(filter->cursor,0) || 
    true_branch==NULL || false_branch!=NULL) {
improper_range:;
   filter->errn= -5;
   sprintf(filter->errmsg,
           "-elseif-operator found outside its proper range.");
   return(0);
 }
 filter->cursor= filter->cursor->up;
 /* -elseif is equivalent to the three-step sequence :  -endif -or -if
    ( -endif has already been performed by following filter->cursor->up ) */
 ret= Nntpfilter_or(filter,0);
 if(ret<=0)
   return(0);
 ret= Nntpfilter_if(filter,0);
 if(ret<=0)
   return(0);
 return(1);
}


int Nntpfilter_endif(filter,flag)
struct NntpfilteR *filter;
int flag;
{
 int ret;
 struct NntpfnodE *fnode,*true_branch;

 if(!Nntpfilter_cursor_complete(filter,0)) {
   filter->errn= -3;
   sprintf(filter->errmsg,
           "Unary operator or expression expected, -endif found");
   return(0);
 }
 if(filter->cursor->up==NULL)
   goto improper_range;
 /* test wether parent node is -if */
 filter->cursor= filter->cursor->up;
 Nntpfnode_get_branch(filter->cursor,&true_branch,0);
 if(!Nntpfnode_is_if(filter->cursor,0) || true_branch==NULL) {
improper_range:;
   filter->errn= -5;
   sprintf(filter->errmsg,"-endif-mark found outside its proper range.");
   return(0);
 }
 /* go to grand parent node */
 filter->cursor= filter->cursor->up;
 return(1);
}


