
/*
  cc -g -c nntpload.c
*/

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>


#include "../s_tools/smem.h"


#include "nntpclient.h"
#include "nntpload.h"

#include "nntpmerge.h"



/* ----------------- Nntpmerge ------------------ */


int Nntpmerge_new(merger,boss,filev,filec,flag)
struct NntpmergE **merger;
struct NntpclienT *boss;
char **filev;
int filec;
int flag;
{
 struct NntpmergE *m;

 *merger= m= TSOB_FELD(struct NntpmergE, 1);
 if(m==NULL)
   return(-1);

 m->boss= boss;
 m->filev= filev;
 m->filec= filec;
 return(1);
}


int Nntpmerge_destroy(merger,flag)
struct NntpmergE **merger;
int flag;
/*
 bit0= also destroy filev (which was provided externally)
*/
{
 struct NntpmergE *m;

 m= *merger;
 if(m==NULL)
   return(0);

 if(flag&1)
   Sfile_destroy_argv(&(m->filec),&(m->filev),0);
 m->filev= NULL; 
 

 Smem_freE((char *) m);
 *merger= NULL;
 return(1);
}


int Nntpmerge__system(cmd,flag)
char *cmd;
int flag;
{
 int sysret= 0;

 printf("executing: %s\n",cmd);
 if(!Nntpmerge_dummy_modE)
   sysret= system(cmd);
 return(sysret==0);
}


int Nntpmerge__classify(line,filename,encodingtext,flag)
char *line;
char filename[SfileadrL];
char encodingtext[81];
int flag;
/*
 return: 0= no encoding header , 1= uuencoded , 2= yencoded , 3= raw_article
*/
{
 static int initialized= 0,begin_len;
 static struct SregeX *re_filename1,*re_filename2,*re_yenc;
 static int filename_start_char= 'b',yenc_start_char='=';
 int ret,is_yenc= 0;
 char *npt,*cpt;

 if(!initialized){
   if(Sregex_new(&re_filename1,"^begin [0-9][0-9][0-9] .*",0)<=0)
     return(-1);
   if(Sregex_new(&re_filename2,"^begin 0[0-9][0-9][0-9] .*",0)<=0)
     return(-1);
   begin_len= strlen("begin ??? ");
   filename_start_char= 'b';
   if(Sregex_new(&re_yenc,"^=ybegin.* name=.*",0)<=0)
     return(-1);
   yenc_start_char= '=';
   initialized= 1;
 }

 if(line[0]==filename_start_char &&
    (Sregex_is_match(re_filename1,line,0) ||
     Sregex_is_match(re_filename2,line,0))) {
   strcpy(encodingtext,"uuencoded");
   npt= line+begin_len;
 } else if(line[0]==yenc_start_char && Sregex_is_match(re_yenc,line,0)) {
   /* yEnc (nearly) 8-Bit encoding */;
   is_yenc= 1;
   strcpy(encodingtext,"yencoded");
   npt= strstr(line," name=");
   if(npt!=NULL)
     npt+= 6;
   else
     npt= line;
 } else if(Sregex_complines("X-Snntpbatch-Dl-Msgno: ",line,1)) {
   strcpy(encodingtext,"raw_article");
   sscanf(line+23,"%s",filename);
   ret= 3; 
   goto ex;
 } else {
   encodingtext[0]= 0;
   ret= 0;
   goto ex;
 }

 cpt= strrchr(npt,'/');
 if(cpt==NULL || cpt<npt)
   cpt= npt;
 else
   cpt++;
 strcpy(filename,cpt);
 Sregex_trimline(filename,8);
 if(filename[0]==0)
   strcpy(filename,"_");

 ret= 1+is_yenc; 
ex:;
 return(ret);
}


int Nntpmerge_uuencoded(m,parts,partcount,binname,flag)
struct NntpmergE *m;
char **parts;
int partcount;
char *binname;
int flag;
/*
 bit0= move parts to names that begin with X
*/
{
 int ret,i,sysret,len,errn;
 char *cmd= NULL,errmsg[4*SfileadrL];
 
 len= 100+strlen(binname);
 for(i=0;i<partcount;i++)
   len+= strlen(parts[i])+1;
 cmd= TSOB_FELD(char,len);
 if(cmd==NULL)
   return(-1);

 strcpy(cmd,"cat ");
 for(i=0;i<partcount;i++) {
   strcat(cmd,parts[i]);
   strcat(cmd," ");
 }
 strcat(cmd," | uudecode -o ");
 strcat(cmd,binname);
 ret= Nntpmerge__system(cmd,0);
 if(ret<=0)
   goto ex;

 if((flag&1)&&!Nntpmerge_dummy_modE) {
   for(i=0;i<partcount;i++) {
     sprintf(cmd,"X%s",parts[i]);
     ret= Sfile_move_file(parts[i],cmd,&errn,errmsg,0);
     if(ret<=0)
       goto ex;
   }
 }

 ret= 1;
ex:;
 if(cmd!=NULL)
   Smem_freE(cmd);
 return(ret);
}


int Nntpmerge_yencoded(m,parts,partcount,binname,flag)
struct NntpmergE *m;
char **parts;
int partcount;
char *binname;
int flag;
/*
 bit0= move parts to names that begin with X
*/
{
 int ret,i,sysret,errn;
 char cmd[2*SfileadrL+100],errmsg[4*SfileadrL];
 
 for(i=0;i<partcount;i++) {
   strcpy(cmd,"cat ");
   strcat(cmd,parts[i]);
   strcat(cmd," ");
   strcat(cmd," | syenc_decode -o ");
   strcat(cmd,binname);
   ret= Nntpmerge__system(cmd,0);
   if(ret<=0)
     goto ex;
 }
 if((flag&1)&&!Nntpmerge_dummy_modE) {
   for(i=0;i<partcount;i++) {
     sprintf(cmd,"X%s",parts[i]);
     ret= Sfile_move_file(parts[i],cmd,&errn,errmsg,2);
     if(ret<=0)
       goto ex;
   }
 }

 ret= 1;
ex:;
 return(ret);
}


int Nntpmerge_raw_message(m,parts,partcount,msgnotext,flag)
struct NntpmergE *m;
char **parts;
int partcount;
char *msgnotext; /* this is the article number as read from .tee file */
int flag;
/*
 bit0= move parts to names that begin with X
*/
{
 int ret,i,multi_bin=0,msgno= -1,errn;
 char filename[SfileadrL],fileadr[SfileadrL],wd[SfileadrL];
 char errmsg[4*SfileadrL];
 struct NntpconN *conn= NULL;
 struct NntploaD *loadjob= NULL;
 struct NntpclienT *client;


 static int dummy= 0;
 if(dummy)
   return(1);


 if(partcount<=0)
   return(1);

 getcwd(wd,sizeof(wd));
 client= m->boss;

 ret= Nntpconn_new(&conn,client->servername,client->port,1|2|8);
 if(ret<=0)
   goto ex;
 Nntpconn_set_filer(conn,client->filer,0);
 printf("\nmessage merge : \n");
 for(i=0;i<partcount;i++) {
   printf("  %s\n",parts[i]);
   sprintf(filename,"%s/%s",wd,parts[i]);
   ret= Nntpconn_add_source(conn,filename,0);
   if(ret<=0)
     goto ex;
 }

 sscanf(msgnotext,"%ld",&msgno);
 ret= Nntpclient_msg_target(client,filename,fileadr,msgno,1|16);
 if(ret<=0)
   goto ex;
 ret= Nntpload_new(&loadjob,conn,NULL,client->download_dir,filename,
                   (client->short_form<<1)|((client->allow_yenc&3)<<4));
 if(ret<=0)
   goto ex;
 Nntpload_set_html(loadjob,client->html,0);
 Nntpload_set_boss(loadjob,client,0);
 Nntphead_set_msgadr(loadjob->head,msgno,msgnotext,0);

 ret= Nntpconn_open(conn,0);
 if(ret<=0)
   goto ex;
 ret= Nntphead_read(loadjob->head,1);
 if(ret<=0)
   return(ret);
 ret= Nntpconn_seek0(conn,0);
 if(ret<=0)
   goto ex;

 ret= Nntpload_article(loadjob,4);
 if(ret<=0)
   goto ex;

 Nntpfiler_chdir_primitive(wd,&errn,&errmsg,1);
 if((flag&1)&&!Nntpmerge_dummy_modE) {
   for(i=0;i<partcount;i++) {
     sprintf(filename,"X%s",parts[i]);
     ret= Sfile_move_file(parts[i],filename,&errn,errmsg,2);
     if(ret<=0)
       goto ex;
   }
 }
 ret= 1;
ex:;
 Nntpfiler_chdir_primitive(wd,&errn,&errmsg,1);
 Nntpload_destroy(&loadjob,0);
 Nntpconn_destroy(&conn,0);
 return(ret);
}


int Nntpmerge_one(m,idx,marker,countpos,flag)
struct NntpmergE *m;
int idx;
char *marker;
int countpos;
int flag;
/*
 bit0-2 = verbosity
          0= silent , 2= report errors , 4= report rejections , 7= report all
*/
{
 int ret,i,partno,num_parts,verbosity,data_started= 0;
 char start[SfileadrL],*cpt,*npt,**partlist= NULL,filename[SfileadrL];
 char encodingtext[81],line[1024];
 FILE *fp= NULL;

 verbosity= flag&7;
 filename[0]= 0;

 /* ??? test name wether it looks like yenc.tee */;

 /* try to read number of parts */ 
 strcpy(start,m->filev[idx]);
 cpt= strstr(start,marker);
 if(cpt==NULL) {
   if(verbosity>=2)
     fprintf(stderr,
             "--- file : %s\n--- program error : marker '%s' not found.\n",
             m->filev[idx]);
   ret= -2; goto ex;
 }
 sscanf(cpt+strlen(marker),"%ld",&num_parts);
 if(num_parts<=0 || num_parts>Nntpmerge_max_partS) {
   if(verbosity>=2)
     fprintf(stderr,"--- file : %s\n--- no usable number of parts\n",
                    m->filev[idx]);
   ret= 0; goto ex;
 }
 partlist= TSOB_FELD(char *,num_parts);
 if(partlist==NULL) {
   if(verbosity>=2)
     fprintf(stderr,
             "--- file : %s\n--- system error : not enough memory available.\n",
             m->filev[idx]);
   ret= -1; goto ex;
 }
 for(i=0;i<num_parts;i++)
   partlist[i]= NULL;

 *(cpt+countpos)= 0;
 for(cpt=cpt+countpos+1;'0'<=(*cpt) && (*cpt)<='9';cpt++);
 for(i=0;i<m->filec;i++) {
   if(strstr(m->filev[i],start)!=m->filev[i])
 continue;
   partno= -1;
   sscanf(m->filev[i]+strlen(start),"%ld",&partno);
   if(partno<1||partno>num_parts)
 continue;
   for(npt=m->filev[i]+strlen(start);'0'<=(*npt) && (*npt)<='9';npt++);
   if(strcmp(npt,cpt)!=0)
 continue;
   partlist[partno-1]= m->filev[i];
 }
 for(i=0;i<num_parts;i++)
   if(partlist[i]==NULL)
 break;
 if(i<num_parts) {
   if(verbosity>=4) 
     fprintf(stderr,"--- file : %s\n--- not all parts found.\n",
                    m->filev[idx]);
   ret= 0; goto ex;
 }
 if(strcmp(partlist[0],m->filev[idx])!=0) {
   if(verbosity>=2) {
     fprintf(stderr,"--- file : %s\n--- not first name in ranklist.\n",
                    m->filev[idx]);
     fprintf(stderr,"--- first: %s\n",partlist[0]);
   }
   ret= 0; goto ex;
 }

 /* find out encoding */;
 fp= Sfile_fopen(partlist[0],"rb",0);
 if(fp==NULL) {
   if(verbosity>=2)
     fprintf(stderr,"--- file : %s\n--- cannot open part 1\n",m->filev[idx]);
   ret= 0; goto ex;
 }
 while(Sfile_fgets(line,sizeof(line)-1,fp)!=NULL) {
   ret= Nntpmerge__classify(line,filename,encodingtext,0);
   if(ret<0)
     goto ex;
   if(ret==0)
 continue;
   Nntpfiler_esc_filename(filename,sizeof(filename),1);
   if(ret==1) {
     if(verbosity>=7)
       fprintf(stderr,"trying to uudecode file %s\n",filename);
     ret= Nntpmerge_uuencoded(m,partlist,num_parts,filename,1);
   } else if(ret==2) {
     if(verbosity>=7)
       fprintf(stderr,"trying to yEnc decode file %s\n",filename);
     ret= Nntpmerge_yencoded(m,partlist,num_parts,filename,1);
   } else if(ret==3) {
     if(verbosity>=7)
       fprintf(stderr,"trying to decode file as raw message : %s\n",filename);
     ret= Nntpmerge_raw_message(m,partlist,num_parts,filename,1);
   } 
   goto ex;
 }
 if(verbosity>=2)
   fprintf(stderr,"--- file : %s\n--- cannot find start of data.\n",
                  m->filev[idx]);
 ret= 0;
ex:;
 if(partlist!=NULL)
   Smem_freE((char *) partlist);
 if(fp!=NULL)
   Sfile_fclose(fp,0);
 return(ret);
}


int Nntpmerge__cmp(pt1,pt2)
char **pt1,**pt2;
{
 return(strcmp(*pt1,*pt2));
}


int Nntpmerge_all(m,flag)
struct NntpmergE *m;
int flag;
{
 int i,partcount= 0,ret;

 /* ??? sort file list alphabetically */
 /* qsort(m->filev,m->filec,sizeof(char *),Nntpmerge__cmp); */

 for(i=0;i<m->filec;i++) {
   if(strstr(m->filev[i],"+281+2F")!=NULL) /*  "(1/"  */
     ret= Nntpmerge_one(m,i,"+281+2F",3,2);
   else if(strstr(m->filev[i],"+5B1+2F")!=NULL) /*  "[1/"  */
     ret= Nntpmerge_one(m,i,"+5B1+2F",3,2);
   else
     ret= 2;
   if(ret<0)
     goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


