#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* Macro for allocation of object arrays */
#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));

#include "../s_tools/sfile.h"

#include "nntpconn.h"
#include "nntphead.h"
#include "nntpfilter.h"
#include "nntpgroup.h"
#include "nntpnov.h"
#include "nntpstats.h"


int Nntpstats_new(stats,conn,group,head,flag)
struct NntpstatS **stats;
struct NntpconN *conn;
struct NntpgrouP *group;
struct NntpheaD *head;
int flag;
{
 struct NntpstatS *s;

 *stats= s= TSOB_FELD(struct NntpstatS,1);
 if(s==NULL)
   return(-1);
 s->reportfile[0]= 0;
 s->conn= conn;
 s->head= head;
 s->filter= NULL;
 Nntpstats_set_group(s,group,0);
 s->total_articles_limit= 0.0;
 s->total_bytes_limit= 0.0;
 s->largest_bytes_limit= 0.0;
 s->xover_maxlines= 10000.0;
 Nntpstats_accu_reset(s,0);
 return(1);
}


int Nntpstats_destroy(stats,flag)
struct NntpstatS **stats;
int flag;
{
 if((*stats)==NULL)
   return(0);
 free((char *) *stats);
 *stats= NULL;
 return(1);
}


int Nntpstats_accu_reset(stats,flag)
struct NntpstatS *stats;
int flag;
{
 stats->total_bytes= 0.0;
 stats->total_estim_bytes= 0.0;
 stats->total_lines= 0.0;
 stats->total_articles= 0.0;
 stats->largest_article= -1;
 stats->largest_bytes= 0.0;
 stats->largest_article_group[0]= 0;
 stats->over_largest_limit= 0;
 stats->oldest_article= -1;
 stats->oldest_time= 0;
 stats->youngest_article= -1;
 stats->youngest_time= 0;
 return(1);
}


int Nntpstats_set_group(stats,group,flag)
struct NntpstatS *stats;
struct NntpgrouP *group;
int flag;
{
 stats->group= group;
 if(group!=NULL) {
   stats->first_msgno= group->first_msgno;
   stats->last_msgno= group->last_msgno;
   if(stats->last_msgno-stats->first_msgno+1>stats->xover_maxlines)
     stats->first_msgno= stats->last_msgno-stats->xover_maxlines+1;
 } else {
   stats->first_msgno= 1;
   stats->last_msgno= 0;
 }
 return(1);
}


int Nntpstats_set_conn(stats,conn,flag)
struct NntpstatS *stats;
struct NntpconN *conn;
int flag;
{
 stats->conn= conn;
 return(1);
}


int Nntpstats_set_head(stats,head,flag)
struct NntpstatS *stats;
struct NntpheaD *head;
int flag;
{
 stats->head= head;
 return(1);
}


int Nntpstats_set_filter(stats,filter,flag)
struct NntpstatS *stats;
struct NntpfilteR *filter;
int flag;
{
 stats->filter= filter;
 return(1);
}


int Nntpstats_set_reportfile(stats,name,flag)
struct NntpstatS *stats;
char *name;
int flag;
{
 strcpy(stats->reportfile,name);
 return(1);
}


int Nntpstats_set_range(stats,first_msgno,last_msgno,flag)
struct NntpstatS *stats;
int first_msgno,last_msgno;
int flag;
{
 stats->first_msgno= first_msgno;
 stats->last_msgno= last_msgno;
 return(1);
}


int Nntpstats_set_limits(stats,total_articles_limit,total_bytes_limit,
                         largest_bytes_limit,flag)
struct NntpstatS *stats;
double total_articles_limit;
double total_bytes_limit,largest_bytes_limit;
int flag;
{
 stats->total_articles_limit= total_articles_limit;
 stats->total_bytes_limit= total_bytes_limit;
 stats->largest_bytes_limit= largest_bytes_limit;
 return(1);
}


int Nntpstats_set_curbs(stats,xover_maxlines,flag)
struct NntpstatS *stats;
double xover_maxlines;
int flag;
{
 stats->xover_maxlines= xover_maxlines;
 return(1);
}


int Nntpstats_scan(stats,flag)
struct NntpstatS *stats;
int flag;
{
 struct NntpheaD *head;
 struct NntpnoV *nov;
 int i,ret,valid,use_filter= 0;
 time_t time_ret;
 char reason[4096];

 head= stats->head;

 if(stats->filter!=NULL)
   if(Nntpfilter_is_restrictive(stats->filter,0))
     use_filter= 1;
 ret= Nntpnov_new(&nov,stats->conn,stats->first_msgno,stats->last_msgno,0);
 if(ret<=0)
   return(ret);
 ret= Nntpnov_read(nov,!use_filter);   /* 0=XOVER , 1=XPAT(Date,Bytes,Lines) */
 if(ret<=0)
   goto ex;

 if(stats->last_msgno>0)
 for(i= stats->first_msgno;i<=stats->last_msgno;i++) {
   ret= Nntpnov_fill_head(nov,stats->head,i,0);
   if(ret<0)
     goto ex;
   if(ret==0)
 continue;

   if(use_filter) {
     ret= Nntpfilter_head_match(stats->filter,stats->head,reason,0);
     if(ret<0)
       goto ex;
     if(ret==0)
 continue;
   }

   if(head->byte_count>0){
     stats->total_bytes+= head->byte_count;
   } else {
     /* estimating uuencoded lines */
     stats->total_estim_bytes+= head->line_count*63;
   }
   stats->total_lines+= head->line_count;
   stats->total_articles++;
   if(stats->largest_bytes_limit<head->byte_count)
     stats->over_largest_limit++;
   if(stats->largest_bytes<head->byte_count){
     stats->largest_bytes= head->byte_count;
     stats->largest_article= i;
   }
   Sfile_decode_time(head->date,&valid,&time_ret,0);
   if(valid) {
     if(time_ret<stats->oldest_time || stats->oldest_article<0) {
       stats->oldest_time= time_ret;
       stats->oldest_article= i;
     }
     if(time_ret>stats->youngest_time || stats->youngest_article<0) {
       stats->youngest_time= time_ret;
       stats->youngest_article= i;
     }
   }
 }
 ret= 1;
ex:;
 Nntpnov_destroy(&nov,0);
 return(ret);
}


int Nntpstats_print(stats,line,flag)
struct NntpstatS *stats;
char *line;
int flag;
/*
 bit0= add newline
*/
{
 printf("%s",line);
 if(flag&1)
   printf("\n");
 if(stats->reportfile[0]!=0 && 
    (stats->reportfile[0]!='-' || stats->reportfile[1]!=0) &&
    (stats->reportfile[0]!='.' || stats->reportfile[1]!=0) )
   Sfile_string_to_file(stats->reportfile,line,2|((flag&1)<<2));
 return(1);
}


int Nntpstats_report(stats,flag)
struct NntpstatS *stats;
int flag;
/*
 bit0= use largest_article_group rather than group->name
 bit1= just test wether stats would be printed
 bit2= print over_largest_limit even if largest_bytes_limit<=0
*/
{
 int ret,young_day,old_day;
 char line[SfileadrL],date[81],old_date[81],young_date[81];

 if(stats->total_articles_limit>stats->total_articles
    ||stats->largest_bytes<stats->largest_bytes_limit
    || stats->total_bytes<stats->total_bytes_limit)
   return(2-(flag&2));
 if(flag&2)
   return(1);
 Nntpstats_print(stats,"\n-----------------------------------------------------------------------------\n",0);
 Sfile_date(date,1);
 if(flag&1){
   sprintf(line,"Group : %s\n",stats->largest_article_group);
 } else {
   sprintf(line,"Group : %-48s (%ld-%ld)\n",
                stats->group->name,stats->group->first_msgno,
                stats->group->last_msgno);
 }
 Nntpstats_print(stats,line,0);
 if(stats->oldest_article>=0 && stats->youngest_article>=0) {
   Sfile_rel_datestr(stats->oldest_time,old_date,1|4);
   Sfile_rel_datestr(stats->youngest_time,young_date,4);
   sprintf(line,"Age   :    %s to  %s\n",old_date,young_date);
   Nntpstats_print(stats,line,0);
 }
 sprintf(line,"Size  : %9.lf  bytes in %.lf articles with %.lf lines\n",
              stats->total_bytes,stats->total_articles,stats->total_lines);
 Nntpstats_print(stats,line,0);
 if(stats->total_estim_bytes>0){
   sprintf(line,"+ est.: %9.lf  bytes in inconsistently reported articles\n",
                stats->total_estim_bytes);
   Nntpstats_print(stats,line,0);
 }
 if(stats->total_articles>0){
   sprintf(line,"Avg.  : %9.lf  bytes per article\n",stats->total_bytes/stats->total_articles);
   Nntpstats_print(stats,line,0);
   sprintf(line,"Max.  : %9.lf  bytes in article #%ld",
                stats->largest_bytes,stats->largest_article);
   if(stats->over_largest_limit>1 && 
      ((flag&4) || stats->largest_bytes_limit>0))
     sprintf(line+strlen(line)," ( %.lf more over %.lf)",
             stats->over_largest_limit,stats->largest_bytes_limit);
   strcat(line,"\n");
   Nntpstats_print(stats,line,0);
 }
 sprintf(line,"Server: %s   Port: %ld   GMT: %s\n",stats->conn->servername,
                                                 stats->conn->portnumber,date);
 Nntpstats_print(stats,line,0);
 Nntpstats_print(stats,"-----------------------------------------------------------------------------\n\n",0);
 return(1); 
}


int Nntpstats_sum(sum_stats,stats,flag)
struct NntpstatS *sum_stats,*stats;
int flag;
/*
 bit0= always copy largest_article and dates
*/
{
 sum_stats->total_bytes+= stats->total_bytes; 
 sum_stats->total_estim_bytes+= stats->total_estim_bytes; 
 sum_stats->total_lines+= stats->total_lines; 
 sum_stats->total_articles += stats->total_articles;
 if((flag&1) || sum_stats->largest_bytes<stats->largest_bytes){
   sum_stats->largest_bytes= stats->largest_bytes;
   sum_stats->largest_article= stats->largest_article;
   strcpy(sum_stats->largest_article_group,stats->largest_article_group);
 }
 if((flag&1) || (( sum_stats->oldest_article<0 || 
    sum_stats->oldest_time>stats->oldest_time) && stats->oldest_time>0)) {
   sum_stats->oldest_time= stats->oldest_time;
   sum_stats->oldest_article= stats->oldest_article;
 }
 if((flag&1) || (( sum_stats->youngest_article<0 || 
   sum_stats->youngest_time<stats->youngest_time) && stats->youngest_time>0)){
   sum_stats->youngest_time= stats->youngest_time;
   sum_stats->youngest_article= stats->youngest_article;
 }
 sum_stats->over_largest_limit+= stats->over_largest_limit;
 return(1);
}


int Nntpstats_largest_group(largest_stats,stats,flag)
struct NntpstatS *largest_stats,*stats;
int flag;
{
 if(largest_stats->total_bytes>=stats->total_bytes)
   return(2);
 Nntpstats_accu_reset(largest_stats,0);
 Nntpstats_sum(largest_stats,stats,0);
 return(1);
}


int Nntpstats_largest_article(largest_stats,stats,flag)
struct NntpstatS *largest_stats,*stats;
int flag;
{
 if(largest_stats->largest_bytes>=stats->largest_bytes)
   return(2);
 Nntpstats_accu_reset(largest_stats,0);
 Nntpstats_sum(largest_stats,stats,0);
 return(1);
}


int Nntpstats_copy_limits(target,source,flag)
struct NntpstatS *target,*source;
int flag;
{
 target->total_articles_limit= source->total_articles_limit;
 target->total_bytes_limit= source->total_bytes_limit;
 target->largest_bytes_limit= source->largest_bytes_limit;
 return(1);
}
