
/*
  cc -g -o snntpbatch snntpbatch.o nntpclient.o nntpgroup.o nntpload.o nntpfilter.o nntphead.o nntpconn.o sregex.o
*/

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include "nntpconn.h"
#include "nntpclient.h"




/* <<< Cleanup is grabbed out of a totally different project */
/*     it provides an outer shell to catch most signals */

static char cleanup_msg[81];
static int cleanup_exiting= 0;

#define Cleanup_temp_sizE  4
static int cleanup_temp_max= Cleanup_temp_sizE;
static int cleanup_temp_count= 0;
static char *cleanup_temp_adr[Cleanup_temp_sizE];
static int cleanup_temp_flags[Cleanup_temp_sizE]; 
/* bit0= rename to automatic name rather than delete */


void Cleanup_handler_exit()
{
 int silent_end,i,first= 1,ret,errn;
 char msg[161],adr[SfileadrL],errmsg[SfileadrL*2];

 if(cleanup_exiting)
   goto do_exit;
 cleanup_exiting= 1;
 alarm(0);

 /* remove temporary file objects */;
 for(i=0;i<cleanup_temp_count;i++) {
   if(cleanup_temp_adr[i]!=NULL) {
     if(first) {
       fprintf(stderr,"\n");
       fprintf(stderr,
               "Cleanup_handler_exit : temporary file objects handled :\n");
     }
     first= 0;
     fprintf(stderr,"   %s",cleanup_temp_adr[i]);
     if(cleanup_temp_flags[i]&1) {
       /* perform automatic rename */;
       strcpy(adr,cleanup_temp_adr[i]);
       Sfile_unique_adr(adr,1);
       fprintf(stderr," -> %s\n",adr);
       ret= Sfile_move_file(cleanup_temp_adr[i],adr,&errn,errmsg,0);
       if(ret<=0)
         fprintf(stderr," * %s\n",errmsg);
     } else {
       unlink(cleanup_temp_adr[i]);
       fprintf(stderr," -DELETED-\n");
     }
   }
 }
 if(first==0)
   fprintf(stderr,"\n");

do_exit:;
 if(! silent_end)
   exit(4);
 exit(0);
}

int Cleanup_handler_temp_file(adr,index,flag)
char *adr;
int index;
int flag;
/*
 bit0= delete entry
 bit1= use index-1 rather than adr to find entry
 bit2= set rename flag, meaning : do not delete but rename automatically
*/
{
 static int no_overflow_reported_yet= 1;
 int i= 0,upper_limit;

 i= 0;
 upper_limit= cleanup_temp_count;
 if(flag&2) {
   if(!(flag&1)) {
     fprintf(stderr,
             "Cleanup_handler_temp_file: will not allocate entry by index\n");
     return(-1);
   }
   if(index==0)
     return(2);
   if(index<0 || index>cleanup_temp_count) {
     fprintf(stderr,
             "Cleanup_handler_temp_file: invalid entry index : %ld\n", index);
     return(0);
   }
   upper_limit= index;
   i= index-1;
 }
 for(;i<upper_limit;i++) {
   if(cleanup_temp_adr[i]==NULL) {
     if(flag&1)
 continue;
     /* allocate entry */
     if(Sregex_string(&(cleanup_temp_adr[i]),adr,0)<=0)
       return(-1);
     cleanup_temp_flags[i]= !!(flag&4);
     return(i+1);
   } else {
     if(!(flag&1))
 continue;
     if(strcmp(cleanup_temp_adr[i],adr)!=0 && !(flag&2))
 continue;
     /* delete entry */;
     free(cleanup_temp_adr[i]);
     cleanup_temp_adr[i]= NULL;
     if(i==cleanup_temp_count-1) {
       while(cleanup_temp_count>0)
         if(cleanup_temp_adr[cleanup_temp_count-1]==NULL)
           cleanup_temp_count--;
         else
       break;
     }
     return(i+1);
   }
 }
 if(flag&1) {
   fprintf(stderr,"Cleanup_handler_temp_file: entry not found :\n   %s\n",adr);
   return(0);
 } else {
   if(cleanup_temp_count<cleanup_temp_max) {
     /* allocate entry */
     if(Sregex_string(&(cleanup_temp_adr[cleanup_temp_count]),adr,0)<=0)
       return(-1);
     cleanup_temp_flags[cleanup_temp_count]= !!(flag&4);
     cleanup_temp_count++;
     return(cleanup_temp_count);
   } else {
     fprintf(stderr,"Cleanup_handler_temp_file: overflow of temp list\n");
     fprintf(stderr," >>> %s\n",adr);
     if(no_overflow_reported_yet) {
       char line[80];
       for(i=0;i<upper_limit;i++) {
         fprintf(stderr," %ld : %s\n",i,cleanup_temp_adr[i]);
       }
       fprintf(stderr,"press Enter to continue : ");
       fgets(line,sizeof(line),stdin);
       no_overflow_reported_yet= 0;
     }
     return(-1);
   }
 }
 return(1);
}


void Cleanup_handler_hup() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGHUP   errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_int() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGINT   errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_quit() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGQUIT  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_ill() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGILL   errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_trap() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGTRAP  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_emt() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGEMT   errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_fpe() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGFPE   errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_bus() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGBUS   errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_segv() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGSEGV  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_sys() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGSYS   errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_pipe() 
{
 return;
}
void Cleanup_handler_urg() 
{
 return;
}

void Cleanup_handler_alrm()
{
 int silent_end= 0;

 alarm(0);
 if(! silent_end)
   sprintf(cleanup_msg,"TIMEOUT !");
 Cleanup_handler_exit();
}

void Cleanup_handler_term() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGTERM  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_xcpu() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGXCPU  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_xfsz() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGXFSZ  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_vtalrm() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGVTALRM  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_prof() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGPROF  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_usr1() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGUSR1  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_usr2() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGUSR2  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_addr() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGADDR  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_zero() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGZERO  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_chk() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGCHK   errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_over() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGOVER  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_priv() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGPRIV  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_pwr() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGPWR   errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_cld() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGCLD   errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_poll() 
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGPOLL  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_abrt()
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGABRT  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_tstp()
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGTSTP  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_ttin()
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGTTIN  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_ttou()
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGTTOU  errno= %ld",errno); 
 Cleanup_handler_exit();
}
void Cleanup_handler_lost()
{
 sprintf(cleanup_msg,"UNIX-SIGNAL:  SIGLOST  errno= %ld",errno); 
 Cleanup_handler_exit();
}


int Cleanup_set_handlers(flag)
int flag;
/*
 bit0= set to default handlers
*/
{ 
 if(flag&1)
   goto set_default;

 cleanup_msg[0]= 0;
 signal(SIGHUP,Cleanup_handler_hup);
 signal(SIGINT,Cleanup_handler_int);
 signal(SIGQUIT,Cleanup_handler_quit);
 signal(SIGILL,Cleanup_handler_ill);
 signal(SIGTRAP,Cleanup_handler_trap);
 signal(SIGFPE,Cleanup_handler_fpe);
 signal(SIGBUS,Cleanup_handler_bus);
 signal(SIGSEGV,Cleanup_handler_segv);
 signal(SIGPIPE,Cleanup_handler_pipe);
 signal(SIGALRM,Cleanup_handler_alrm);
 signal(SIGTERM,Cleanup_handler_term);
 signal(SIGUSR1,Cleanup_handler_usr1);
 signal(SIGUSR2,Cleanup_handler_usr2);
 signal(SIGURG,Cleanup_handler_urg);
 signal(SIGXCPU,Cleanup_handler_xcpu);
 signal(SIGXFSZ,Cleanup_handler_xfsz);
 signal(SIGVTALRM,Cleanup_handler_vtalrm);
 signal(SIGPROF,Cleanup_handler_prof);
 return(1);
set_default:;
 signal(SIGHUP,SIG_DFL);
 signal(SIGINT,SIG_DFL);
 signal(SIGQUIT,SIG_DFL);
 signal(SIGILL,SIG_DFL);
 signal(SIGTRAP,SIG_DFL);
 signal(SIGFPE,SIG_DFL);
 signal(SIGBUS,SIG_DFL);
 signal(SIGSEGV,SIG_DFL);
 signal(SIGPIPE,SIG_DFL);
 signal(SIGALRM,SIG_DFL);
 signal(SIGTERM,SIG_DFL);
 signal(SIGUSR1,SIG_DFL);
 signal(SIGUSR2,SIG_DFL);
 signal(SIGURG,SIG_DFL);
 signal(SIGXCPU,SIG_DFL);
 signal(SIGXFSZ,SIG_DFL);
 signal(SIGVTALRM,SIG_DFL);
 signal(SIGPROF,SIG_DFL);
 return(1);
}


main(argc,argv)
int argc;
char **argv;
{
 int i,ret,reading_rc,ret_code;
 struct NntpclienT *client;
 char *cpt;
 static int rc_flags[3]= {1,4,2};

 Cleanup_set_handlers(0);

 ret= Nntpclient_new(&client,"",1); /* no connection to server yet */
 if(ret<=0)
   goto err_ex;
 Nntpclient_set_timeout(client,240,0);

 for(reading_rc= 1;reading_rc<=3;reading_rc++) {
   ret= Nntpclient_rc(client,argv[0],rc_flags[reading_rc-1]); 
   if(ret<=0) 
     goto err_ex;
 }
 reading_rc= 0;

 i= 1;
 while(i<argc){
   /* is able to reconnect after timeouts */
   ret= Nntpclient_outer_fence(client,argc,argv,&i,0); 
   if(ret<=0)
     goto err_ex;
   while(++i<argc)
    if(argv[i][0]=='-')
      break;
 }
 exit(0);

err_ex:;
 ret_code= ret;
 if(reading_rc){
   cpt= strrchr(argv[0],'/');
   if(cpt==NULL)
     cpt= argv[0];
   else
     cpt++;
   fprintf(stderr,"\n++++++++++++++++++++++++++++++++++++++++++\n");
   fprintf(stderr,"+++ while reading  ");
   if(reading_rc==1)
     fprintf(stderr,"$HOME/");
   else if(reading_rc==2) 
     fprintf(stderr,"../");
   else if(reading_rc==3) 
     ;
   fprintf(stderr,".%src  :",cpt);
 }
 fprintf(stderr,"\n++++++++++++++++++++++++++++++++++++++++++\n");
 fprintf(stderr,"return code = %ld\n",ret_code);
 if(client!=NULL)
   if(client->conn!=0){
     fprintf(stderr,"error code = %ld\n",client->conn->errn);
     fprintf(stderr,"%s\n",client->conn->errmsg);
   }
 fprintf(stderr,"+++++++++ premature end of program +++++++\n");
 if(!reading_rc){
   fprintf(stderr,"%ld unprocessed arguments: \n",argc-i);
   for(;i<argc;i++){
     fprintf(stderr,"%s ",argv[i]);
   }
 }
 fprintf(stderr,"\n");
 exit(1);
}

