
#ifndef Sblowfish_includeD
#define Sblowfish_includeD


#ifndef ORDER_ABCD
#ifndef ORDER_DCBA
#ifndef ORDER_BADC
/* default is Linux on AMD/Intel */ 
#define ORDER_DCBA
#endif
#endif
#endif


/*********************blowfish.h********************/

/* derived from: $Id: blowfish.h,v 1.3 1995/01/23 12:38:02 pr Exp pr $*/

#define MAXKEYBYTES 56          /* 448 bits */
#define bf_N             16
#define noErr            0
#define DATAERROR         -1
#define KEYBYTES         8


#define UWORD_32bits  unsigned long
#define UWORD_16bits  unsigned short
#define UBYTE_08bits  unsigned char

/* choose a byte order for your hardware */
/* ABCD - big endian - motorola */
#ifdef ORDER_ABCD
union aword {
  UWORD_32bits word;
  UBYTE_08bits byte [4];
  struct {
    unsigned int byte0:8;
    unsigned int byte1:8;
    unsigned int byte2:8;
    unsigned int byte3:8;
  } w;
};
#endif  /* ORDER_ABCD */

/* DCBA - little endian - intel */
#ifdef ORDER_DCBA
union aword {
  UWORD_32bits word;
  UBYTE_08bits byte [4];
  struct {
    unsigned int byte3:8;
    unsigned int byte2:8;
    unsigned int byte1:8;
    unsigned int byte0:8;
  } w;
};
#endif  /* ORDER_DCBA */

/* BADC - vax */
#ifdef ORDER_BADC
union aword {
  UWORD_32bits word;
  UBYTE_08bits byte [4];
  struct {
    unsigned int byte1:8;
    unsigned int byte0:8;
    unsigned int byte3:8;
    unsigned int byte2:8;
  } w;
};
#endif  /* ORDER_BADC */


typedef UWORD_32bits Blowfish_spT[256];



/*
 A device to read strong random numbers from. 
 With Linux, /dev/random blocks until enough entropy is available 
                         entropy comes from hardware events (disk,keyboard,...)
             /dev/urandom does not block but might return pseudo-random data
 Undefine if you got no usable random device at all or if you need its 
 prescious entropy for other purposes. (/dev/random can be very slow if
 the entropy pool gets drained by too much reading of random numbers.
 /dev/urandom provides weakened security and also drains the pool.)
*/
#define Sblowfish_random_device_adR "/dev/urandom"

/*
 A keystring used as part of the weak seed for random numbers
 If it is unknown to an attacker this string obscures the random generation
 mechanism. It may consist of 16 to 56 characters in steps of 8 .
 If no /dev/random is available then one should try to make this key secret.
 (I.e. change the string and from then on keep others from seeing this file
  and all the binary files which were derived from it.) 
*/
#define Sblowfish_random_seeD "0123456789ABCDEF"


#ifdef Sblowfish_got_a_bicyclE

The global random generator function Sblowfish__global_random() produces
random numbers according to a preset quality. The quality is achieved by
reading trusted random bytes from a device like /dev/random.
See Sblowfish_random_device_adR (or Sblowfish_random_seeD).

Since such trusted bytes may be very rare they are read only once when quality
gets increased. Their randomness is then propagated by using them for
encrypting a sequence of floating point numbers. Therefore the resulting data
bytes are pseudo-random - but as hard to break as the weaker one of /dev/random
and the encryption algorithm.

Random numbers from Sblowfish__global_random() also depend on process id,
the memory address of a static object in Sblowfish__global_random() and
the microsecond of its creation (clock might be more coarse). Therefore
it is wise to use Sfile_microsleep to wait a few thousand random microseconds
after a quality increase.

A quality number of about 4 to 56 /dev/random bytes should be set by an 
application which intends to act as a receiving peer. Be aware that a large
number may need several seconds to come out of /dev/random and also may
slow down other consumers of /dev/random bytes.

Default quality is 0. An application eventually increases quality by parameter
quality_bytes of Sblowfish__global_random() . Here :  Sagent_random_seed_leN 

 ret= Sblowfish__global_random(NULL,0,Sagent_random_seed_leN,0);
 if(ret<=0) {
   /* insert failure handling here */;
 }

#endif /* Sblowfish_got_a_bicyclE */


struct SblowfisH {
  UWORD_32bits bf_P[bf_N + 2];
  UWORD_32bits bf_S[4][256];

  unsigned char last_inblock[8];
  unsigned char last_outblock[8];

  unsigned char iv[8];
};


/* make a binary key of fixed length from a text key with abritrary length */
char *Sblowfish_make_key();

#endif /* ! Sblowfish_includeD */

