
/*
 cc -g -c sdb_mysql.c
 link with :  -lmysqlclient 
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include <mysql/mysql.h>
#include <mysql/errmsg.h>

#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "sdb_mysql.h"


/* ------------------------- SDB wrapper for MySQL ----------------- */


int Sdb_mysql_new(sdb,hostname,username,password,dbname,message_pt,flag)
struct Sdb_mysqL **sdb;
char *hostname;
char *username;
char *password;
char *dbname;
char *message_pt;
int flag;
{
 struct Sdb_mysqL *m;

 m= *sdb= TSOB_FELD(struct Sdb_mysqL,1);
 if(m==NULL)
   return(-1);

 m->hostname= NULL;
 m->username= NULL;
 m->password= NULL;
 m->dbname= NULL;
 m->db= NULL;
 m->opened= 0;
 m->fetch_mode= 0;
 m->query_result= NULL;
 m->message= message_pt;

 if(Sregex_string(&(m->hostname),hostname,0)<=0)
   goto failed;
 if(Sregex_string(&(m->username),username,0)<=0)
   goto failed;
 if(Sregex_string(&(m->password),password,0)<=0)
   goto failed;
 if(Sregex_string(&(m->dbname),dbname,0)<=0)
   goto failed;
 return(1);

failed:;
 Sdb_mysql_destroy(sdb,0);
 return(-1);
}


int Sdb_mysql_destroy(sdb,flag)
struct Sdb_mysqL **sdb;
int flag;
{
 struct Sdb_mysqL *m;

 m= *sdb;
 if(m==NULL)
   return(0);
 Sdb_mysql_close(m,0);
 Sregex_string(&(m->dbname),NULL,0);
 Sregex_string(&(m->password),NULL,0);
 Sregex_string(&(m->username),NULL,0);
 Sregex_string(&(m->hostname),NULL,0);
 Smem_freE((char *) m);
 *sdb= NULL;
 return(1);
}


int Sdb_mysql_get_query_result(sdb,value,flag)
struct Sdb_mysqL *sdb;
MYSQL_RES **value;
int flag;
{
 *value= sdb->query_result;
 return(1);
}


int Sdb_mysql_set_query_result(sdb,value,flag)
struct Sdb_mysqL *sdb;
MYSQL_RES *value;
int flag;
{
 sdb->query_result= value;
 return(1);
}


int Sdb_mysql_init(sdb,flag)
struct Sdb_mysqL *sdb;
int flag;
{
 sdb->db= mysql_init(NULL);
 if(sdb->db==NULL) {
   strcpy(sdb->message,"mysql_init() failed. Probably out of memory.");
   return(-1);
 }
 return(1);
}


int Sdb_mysql_handle_error(sdb,flag)
struct Sdb_mysqL *sdb;
int flag;
/*
 bit2= copy mysql_error() to sdb->message
*/
{
 int ret,siz;
 char *msg;
 unsigned int merrno;

 merrno= mysql_errno(sdb->db);
 if(flag&2) {
   msg= mysql_error(sdb->db);
   siz= Sdb_mysql_min_msg_sizE;
   if(strlen(msg)>=siz) {
     strncpy(sdb->message,msg,siz-1);
     sdb->message[siz-1]= 0;
   } else
     strcpy(sdb->message,msg);
 }

 if(merrno==CR_SERVER_GONE_ERROR || merrno==CR_SERVER_LOST) {

   fprintf(stderr,"%s\n... trying to reconnect ...\n",mysql_error(sdb->db));

   /* shut down */;
   Sdb_mysql_free_result(sdb,0);
   Sdb_mysql_close(sdb,0);
   /* bring up again */;
   ret= Sdb_mysql_init(sdb,0);
   if(ret<=0)
     return(ret);
   /* try to reconnect */;
   ret= Sdb_mysql_open(sdb,1);
   if(ret<=0)
     return(ret);
   return(1);
 }

 return(0);
}


int Sdb_mysql_open(sdb,flag)
struct Sdb_mysqL *sdb;
int flag;
/*
 bit0= no Sdb_mysql_handle_error() in case of error
*/
{
 int retry_count= 0,ret;
 MYSQL *myret;

retry:;
 myret= mysql_real_connect(sdb->db,sdb->hostname,
                           sdb->username,sdb->password,
                           sdb->dbname,0,NULL,0);
 if(myret==NULL) {
   if(!flag&1) {
     ret= Sdb_mysql_handle_error(sdb,0);
     if(ret==1) {
       if(retry_count++<=Sdb_mysql_reconnect_maxtrY)
         goto retry;
     }     
   }     
   sprintf(sdb->message,"connection to MySQL failed : %s %s %s \n %s",
                              sdb->hostname,sdb->username,sdb->dbname,
                              mysql_error(sdb->db));
   return(-1);
 }
 sdb->opened= 1;
 return(1);
}


int Sdb_mysql_close(sdb,flag)
struct Sdb_mysqL *sdb;
int flag;
{
 Sdb_mysql_free_result(sdb,0);
 mysql_close(sdb->db);
 sdb->db= NULL;
 return(1);
}


int Sdb_mysql_query(sdb,query_text,query_length,flag)
struct Sdb_mysqL *sdb;
char *query_text;
int query_length;
int flag;
{
 int retry_count= 0,ret;
 int qret;

 if(sdb->opened==0) {
   ret= Sdb_mysql_open(sdb,0);
   if(ret<=0)
     return(ret);
 }
retry:;
 qret= mysql_real_query(sdb->db,query_text,query_length);
 if(qret!=0) {
   ret= Sdb_mysql_handle_error(sdb,2);
   if(ret==1) {
     if(retry_count++<=Sdb_mysql_reconnect_maxtrY)
       goto retry;
   }     
   return(0);
 }
 return(1);
}


int Sdb_mysql_request_result(sdb,flag)
struct Sdb_mysqL *sdb;
int flag;
/*
 bit4= buffer whole reply so query can be left open while intermediate query
*/
{
 int retry_count= 0,ret;

retry:;
 if(sdb->fetch_mode==0 || (flag&16)) {
   sdb->query_result= mysql_store_result(sdb->db);
 } else {
   sdb->query_result= mysql_use_result(sdb->db);
 }
 if(sdb->query_result==NULL) {
   Sdb_mysql_handle_error(sdb,2);
   return(0);
 }
 return(1);
}


int Sdb_mysql_update_worked(sdb,flag)
struct Sdb_mysqL *sdb;
int flag;
{
 int ret;
 static char head[]= {"Rows matched:"};
 char *text;

 text= mysql_info(sdb->db);
 if(text==NULL)
   return(0);
 if(strlen(text)<=strlen(head))
   return(0);
 sscanf(text+strlen(head),"%ld",&ret);
 return(ret);
}


int Sdb_mysql_esc_string(sdb,target,source,source_length,flag)
struct Sdb_mysqL *sdb;
char *target;
char *source;
int source_length;
int flag;
{
 int l;

 l= mysql_escape_string(target,source,source_length);
 return(l);
}


int Sdb_mysql_end_query(sdb,flag)
struct Sdb_mysqL *sdb;
int flag;
{
 int ret;
 MYSQL_ROW row;

 if(sdb->query_result==NULL)
   return(2);
 do {
   row= mysql_fetch_row(sdb->query_result);
 } while(row!=NULL);
 if(mysql_errno(sdb->db)!=0) {
   ret= Sdb_mysql_handle_error(sdb,2);
   if(ret!=1)
     return(0);
 }
 return(1);
}


int Sdb_mysql_free_result(sdb,flag)
struct Sdb_mysqL *sdb;
int flag;
{
 if(sdb->query_result==NULL)
   return(0);
 mysql_free_result(sdb->query_result);
 sdb->query_result= NULL;
 return(1);
}


int Sdb_mysql_unique_seed(sdb,fieldname,seed,flag)
/* produces a unique id or a seed value which is converted to a unique id
   by the RDB server */
struct Sdb_mysqL *sdb;
char *fieldname;
unsigned long *seed;
int flag;
{
 *seed= 0;
 return(1);
}


int Sdb_mysql_get_unique_id(sdb,fieldname,id,flag)
/* reads the actual unique id after the RDB server processed the seed */
struct Sdb_mysqL *sdb;
char *fieldname;
unsigned long *id;
int flag;
{
 /* >>> this only works up to 32 bit, the MySQl type is 64 bit as it seems */
 /* >>> how to use 64 bit in a portable way ? */
 *id= mysql_insert_id(sdb->db);
 return(1);
}


int Sdb_mysql_get_fieldcount(sdb,flag)
struct Sdb_mysqL *sdb;
int flag;
{
 int fieldcount;

 fieldcount= mysql_num_fields(sdb->query_result);
 return(fieldcount);
}


int Sdb_mysql_get_row(sdb,fields,lengths,flag)
struct Sdb_mysqL *sdb;
char ***fields;
unsigned long **lengths;
int flag;
{
 int retry_count= 0,ret;
 unsigned int *fieldlengths;
 MYSQL_ROW row;

retry:;
 row= mysql_fetch_row(sdb->query_result);
 if(row==NULL) {
   if(mysql_errno(sdb->db)!=0) {
     Sdb_mysql_handle_error(sdb,2);
     return(-1);
   }
   return(0);
 }
 *lengths= mysql_fetch_lengths(sdb->query_result);
 *fields= (char **) row;
 return(1);
}



#ifdef Sdb_mysql_gen_apI

/* ------------------ general SDB API (if desired) ---------------- */


int Sdb_new(sdb,hostname,username,password,dbname,message_pt,flag)
struct Sdb_generiC **sdb;
char *hostname;
char *username;
char *password;
char *dbname;
char *message_pt;
int flag;
{
 int ret;
 ret= Sdb_mysql_new(sdb,hostname,username,password,dbname,message_pt,flag);
 return(ret);
}


int Sdb_destroy(sdb,flag)
struct Sdb_generiC **sdb;
int flag;
{
 int ret;
 ret= Sdb_mysql_destroy(sdb,flag);
 return(ret);
}


int Sdb_get_query_result(sdb,value,flag)
struct Sdb_generiC *sdb;
Sdb_res_generiC **value;
int flag;
{
 int ret;
 ret= Sdb_mysql_get_query_result(sdb,value,flag);
 return(ret);
}


int Sdb_set_query_result(sdb,value,flag)
struct Sdb_generiC *sdb;
Sdb_res_generiC *value;
int flag;
{
 int ret;
 ret= Sdb_mysql_set_query_result(sdb,value,flag);
 return(ret);
}


int Sdb_init(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret; 
 ret= Sdb_mysql_init(sdb,flag);
 return(ret);
}


int Sdb_handle_error(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
/*
 bit2= copy mysql_error() to sdb->message
*/
{
 int ret;
 ret= Sdb_mysql_handle_error(sdb,flag);
 return(ret);
}


int Sdb_open(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
/*
 bit0= no Sdb_mysql_handle_error() in case of error
*/
{
 int ret;
 ret= Sdb_mysql_open(sdb,flag);
 return(ret);
}


int Sdb_close(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret;
 ret= Sdb_mysql_close(sdb,flag);
 return(ret);
}


int Sdb_query(sdb,query_text,query_length,flag)
struct Sdb_generiC *sdb;
char *query_text;
int query_length;
int flag;
{
 int ret;
 ret= Sdb_mysql_query(sdb,query_text,query_length,flag);
 return(ret);
}


int Sdb_request_result(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
/*
 bit4= buffer whole reply so query can be left open while intermediate query
*/
{
 int ret;
 ret= Sdb_mysql_request_result(sdb,flag);
 return(ret);
}


int Sdb_esc_string(sdb,target,source,source_length,flag)
struct Sdb_generiC *sdb;
char *target;
char *source;
int source_length;
int flag;
{
 int ret;
 ret= Sdb_mysql_esc_string(sdb,target,source,source_length,flag);
 return(ret);
}


int Sdb_end_query(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret;
 ret= Sdb_mysql_end_query(sdb,flag);
 return(ret); 
} 


int Sdb_free_result(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret;
 ret= Sdb_mysql_free_result(sdb,flag); 
 return(ret);  
}  


int Sdb_unique_seed(sdb,fieldname,seed,flag)
/* produces a unique id or a seed value which is converted to a unique id
   by the RDB server */
struct Sdb_generiC *sdb;
char *fieldname;
unsigned long *seed;
int flag;
{
 int ret; 
 ret= Sdb_mysql_unique_seed(sdb,fieldname,seed,flag);  
 return(ret);   
}   


int Sdb_get_unique_id(sdb,fieldname,id,flag)
/* reads the actual unique id after the RDB server processed the seed */
struct Sdb_generiC *sdb;
char *fieldname;
unsigned long *id;
int flag;
{
 int ret;  
 ret= Sdb_mysql_get_unique_id(sdb,fieldname,id,flag);
 return(ret);    
}    


int Sdb_get_fieldcount(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret;   
 ret= Sdb_mysql_get_fieldcount(sdb,flag); 
 return(ret);     
}     
 

int Sdb_get_row(sdb,fields,lengths,flag)
struct Sdb_generiC *sdb;
char ***fields;
unsigned long **lengths;
int flag;
{
 int ret;    
 ret= Sdb_mysql_get_row(sdb,fields,lengths,flag);  
 return(ret);      
}      


int Sdb_update_worked(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret;

 ret= Sdb_mysql_update_worked(sdb,flag);
 return(ret);
}


char *Sdb__get_typelist_pt(flag)
int flag;
{
 return(Sdb_mysql_typelisT);
}

  
#endif /* Sdb_mysql_gen_apI */
