
/*
 cc -g -c sdb_oci.c
 link with :  libclntsh
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"


/* this also includes "oci.h"  */
#include "sdb_oci.h"


/* --------------------- SDB wrapper for Oracle 8 via OCI --------------- */


int Sdb_oci_new(sdb,hostname,username,password,dbname,message_pt,flag)
struct Sdb_ocI **sdb;
char *hostname;
char *username;
char *password;
char *dbname;
char *message_pt;
int flag;
{
 struct Sdb_ocI *m;

 m= *sdb= TSOB_FELD(struct Sdb_ocI,1);
 if(m==NULL)
   return(-1);

 m->hostname= NULL;
 m->username= NULL;
 m->password= NULL;
 m->dbname= NULL;
 m->initialized= 0;
 m->opened= 0;
 m->auto_commit= 1;
 m->fetch_mode= 0;
 m->stq= NULL;
 m->query_token= NULL;
 m->search_where= NULL;
 m->search_stq= NULL;
 m->search_token= NULL;
 m->message= message_pt;

 if(Sregex_string(&(m->hostname),hostname,0)<=0)
   goto failed;
 if(Sregex_string(&(m->username),username,0)<=0)
   goto failed;
 if(Sregex_string(&(m->password),password,0)<=0)
   goto failed;
 if(Sregex_string(&(m->dbname),dbname,0)<=0)
   goto failed;
 return(1);

failed:;
 Sdb_oci_destroy(sdb,0);
 return(-1);
}


int Sdb_oci_destroy(sdb,flag)
struct Sdb_ocI **sdb;
int flag;
{
 struct Sdb_ocI *m;

 m= *sdb;
 if(m==NULL)
   return(0);
 Sdb_oci_close(m,0);
 Sregex_string(&(m->dbname),NULL,0);
 Sregex_string(&(m->password),NULL,0);
 Sregex_string(&(m->username),NULL,0);
 Sregex_string(&(m->hostname),NULL,0);
 Smem_freE((char *) m);
 *sdb= NULL;
 return(1);
}


int Sdb_oci_checkerr(sdb,in_oret,errmsg,flag)
struct Sdb_ocI *sdb;
sword in_oret;
char errmsg[Sdb_oci_min_msg_sizE];
int flag;
{
 int ret= 1;
 sb4 errcode;
 
 switch (in_oret) {
 case OCI_SUCCESS:
   ret= 2;                                             /* indicates no error */
   break;
 case OCI_SUCCESS_WITH_INFO:
   strcpy(errmsg,"ErrorOCI_SUCCESS_WITH_INFO");
   break;
 case OCI_NEED_DATA:
   strcpy(errmsg,"ErrorOCI_NEED_DATA");
   break;
 case OCI_NO_DATA:
   strcpy(errmsg,"ErrorOCI_NO_DATA");
   break;
 case OCI_ERROR:
   OCIErrorGet ((dvoid *) sdb->errhp, (ub4) 1, (text *) NULL, &errcode,
                 errmsg, (ub4) Sdb_oci_min_msg_sizE, (ub4) OCI_HTYPE_ERROR);
   break;
 case OCI_INVALID_HANDLE:
   strcpy(errmsg,"ErrorOCI_INVALID_HANDLE\n");
   break;
 case OCI_STILL_EXECUTING:
   strcpy(errmsg,"ErrorOCI_STILL_EXECUTE\n");
   break;
 case OCI_CONTINUE:
   strcpy(errmsg,"ErrorOCI_CONTINUE\n");
   break;
 default:
   sprintf(errmsg,"Unknown Error #%lu\n",in_oret);
   ret= 0;                                       /* indicates unknown error */
   break;
 }
 return(ret);
}


int Sdb_oci_init(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 sword oret;
 dvoid *dummy;

 sdb->initialized= 1;
 OCIInitialize((ub4) OCI_DEFAULT, (dvoid *)0,
               (dvoid * (*)()) 0, (dvoid * (*)()) 0,  (void (*)()) 0 );
 oret= OCIEnvInit(&(sdb->db),OCI_DEFAULT,0,&dummy);
 if(oret != OCI_SUCCESS) {
   Sdb_oci_checkerr(sdb,oret,sdb->message,0);
   return(0);
 } 
 oret= OCIHandleAlloc(sdb->db,(dvoid *) &(sdb->errhp),OCI_HTYPE_ERROR,0,&dummy);
 if(oret != OCI_SUCCESS) {
   Sdb_oci_checkerr(sdb,oret,sdb->message,0);
   return(0);
 } 
 return(1);
}


int Sdb_oci_open(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
/*
 bit0= no Sdb_oci_handle_error() in case of error
*/
{
 sword oret;
 int retry_count= 0,ret;
 char errmsg[Sdb_oci_min_msg_sizE];

retry:;
 oret= OCILogon(sdb->db,sdb->errhp,&(sdb->svchp),
                sdb->username,strlen(sdb->username),
                sdb->password,strlen(sdb->password),
                sdb->dbname,strlen(sdb->dbname));
 if(oret != OCI_SUCCESS) {
   if(!flag&1) {
     ret= Sdb_oci_handle_error(sdb,0);
     if(ret==1) {
       if(retry_count++<=Sdb_oci_reconnect_maxtrY)
         goto retry;
     }     
   }     
   return(0);
 }
 sdb->opened= 1;
 return(1);
}


int Sdb_oci_close(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 Sdb_oci_free_result(sdb,0);
 OCILogoff(sdb->svchp,sdb->errhp);
 OCIHandleFree(sdb->errhp,OCI_HTYPE_ERROR);
 OCIHandleFree(sdb->db,OCI_HTYPE_ENV);
 sdb->db= NULL;
 return(1);
}


int Sdb_oci_handle_error(sdb,oret,flag)
struct Sdb_ocI *sdb;
sword oret;
int flag;
/*
 bit2= copy oci_error() to sdb->message
*/
{
 int ret,siz;
 char errmsg[Sdb_oci_min_msg_sizE];
 unsigned int merrno= 0;

 Sdb_oci_checkerr(sdb,oret,errmsg,0);
 if(flag&2) 
   strcpy(sdb->message,errmsg);

 if(merrno==123456789) {                                      /* out of duty */

   fprintf(stderr,"%s\n... trying to reconnect ...\n",errmsg);

   /* shut down */;
   Sdb_oci_free_result(sdb,0);
   Sdb_oci_close(sdb,0);
   /* bring up again */;
   ret= Sdb_oci_init(sdb,0);
   if(ret<=0)
     return(ret);
   /* try to reconnect */;
   ret= Sdb_oci_open(sdb,1);
   if(ret<=0)
     return(ret);
   return(1);
 }

 return(0);
}


int Sdb_oci_query(sdb,query_text,query_length,flag)
struct Sdb_ocI *sdb;
char *query_text;
int query_length;
int flag;
/*
 bit0= do send COMMIT WORK by temporary statement
       (usually this is controlled by sdb->auto_commit)
 bit1= do not send query_text (but only COMMIT WORK if bit0 is set)
*/
{
 sword oret;
 char *dummy;
 int retry_count= 0,ret,send_commit= 0;
 struct Sdb_oci_stQ *commit_stq= NULL;

 if(sdb->opened==0) {
   ret= Sdb_oci_open(sdb,0);
   if(ret<=0)
     goto ex;
 }

 if(!(flag&2)) {
   Sdb_oci_stq_destroy(&(sdb->stq),0);
   ret= Sdb_oci_stq_new(&(sdb->stq),query_text,query_length,0);
   if(ret<=0) 
     goto ex;
   ret= Sdb_oci_stq_prepare(sdb->stq,0);
   if(ret<=0) 
     goto ex;
   if(sdb->auto_commit) {
     ret= Sdb_oci_stq_is_commitable(sdb->stq,0);
     if(ret<0)
       goto ex;
     if(ret>0)
       send_commit= 1;
   }
retry:;
   ret= Sdb_oci_stq_execute(sdb->stq,0);
   if(ret<=0) {
     ret= Sdb_oci_handle_error(sdb,2);
     if(ret==1) {
       if(retry_count++<=Sdb_oci_reconnect_maxtrY)
         goto retry;
     }     
     ret= 0; goto ex;
   }
 }
 if(flag&1)
   send_commit= 1;
 if(send_commit) {
   ret= Sdb_oci_stq_new(&commit_stq,"COMMIT WORK",11,0);
   if(ret<=0) 
     goto ex;
retry_commit:;
   ret= Sdb_oci_stq_execute(commit_stq,0);
   if(ret<=0) {
     ret= Sdb_oci_handle_error(sdb,2);
     if(ret==1) {
       if(retry_count++<=Sdb_oci_reconnect_maxtrY)
         goto retry_commit;
     }
     ret= 0; goto ex;
   }
 }
 ret= 1;
ex:;
 Sdb_oci_stq_destroy(&(commit_stq),0);
 return(ret);
}


int Sdb_oci_free_result(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 int ret;

 ret= Sdb_oci_stq_destroy(&(sdb->stq),0);
 return(ret);
}


int Sdb_oci_request_result(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
/*
 bit4= buffer whole reply so query can be left open while intermediate query
*/
{
 int retry_count= 0,ret;
 sword oret;
 unsigned long rowcount= 0,rowsize= 0,row_offsets[256],colcount= 0;
 OCIParam *pard;
 ub2 dtype;
 ub4 dsize;

 ret= Sdb_oci_stq_rowform(sdb->stq,0);
 if(ret<=0)
   return(ret);
 ret= Sdb_oci_stq_resultbuf(sdb->stq,(flag&16));
 if(ret<=0)
   return(ret);
 return(1);
}


int Sdb_oci_get_fieldcount(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 int fieldcount;

 fieldcount= Sdb_oci_stq_column_count(sdb->stq,0);
 return(fieldcount);
}


int Sdb_oci_get_row(sdb,fields,lengths,flag)
struct Sdb_ocI *sdb;
char ***fields;
unsigned long **lengths;
int flag;
{
 int retry_count= 0,ret;
 unsigned int *fieldlengths;

retry:;
 ret= Sdb_oci_stq_next_row(sdb->stq,fields,lengths,0);
 return(ret);
}


int Sdb_oci_end_query(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 int ret;

 if(sdb->stq==NULL)
   return(2);
 ret= Sdb_oci_stq_end_query(sdb->stq,0);
 return(ret);
}



int Sdb_oci_update_worked(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 int ret;

 ret= Sdb_oci_stq_update_worked(sdb->stq,0);
 return(ret);
}


int Sdb_oci_esc_string(sdb,target,source,source_length,flag)
struct Sdb_ocI *sdb;
char *target;
char *source;
int source_length;
int flag;
{
 char inflator= '\'',*rpt,*wpt;

 wpt= target;
 for(rpt= source; *rpt!= 0; rpt++) {
   if((*rpt)==inflator)
     *(wpt++)= inflator;
   *(wpt++)= *rpt;
 } 
 *wpt= 0;
 return(1);
}


/* >>> BAUSTELLE */


int Sdb_oci_unique_seed(sdb,fieldname,seed,flag)
/* produces a unique id or a seed value which is converted to a unique id
   by the RDB server */
struct Sdb_ocI *sdb;
char *fieldname;
unsigned long *seed;
int flag;
{
 /* >>> */;
 *seed= 0;
 return(1);
}


int Sdb_oci_get_unique_id(sdb,fieldname,id,flag)
/* reads the actual unique id after the RDB server processed the seed */
struct Sdb_ocI *sdb;
char *fieldname;
unsigned long *id;
int flag;
{
 /* >>> this only works up to 32 bit, the MySQl type is 64 bit as it seems */
 /* >>> how to use 64 bit in a portable way ? */

 /* >>> *id= oci_insert_id(sdb->db); */
 return(1);
}


#ifdef Sdb_oci_gen_apI

/* ------------------ general SDB API (if desired) ---------------- */


int Sdb_new(sdb,hostname,username,password,dbname,message_pt,flag)
struct Sdb_ocI **sdb;
char *hostname;
char *username;
char *password;
char *dbname;
char *message_pt;
int flag;
{
 int ret;
 ret= Sdb_oci_new(sdb,hostname,username,password,dbname,message_pt,flag);
 return(ret);
}


int Sdb_destroy(sdb,flag)
struct Sdb_ocI **sdb;
int flag;
{
 int ret;
 ret= Sdb_oci_destroy(sdb,flag);
 return(ret);
}


int Sdb_init(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 int ret; 
 ret= Sdb_oci_init(sdb,flag);
 return(ret);
}


int Sdb_handle_error(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
/*
 bit2= copy oci_error() to sdb->message
*/
{
 int ret;
 ret= Sdb_oci_handle_error(sdb,flag);
 return(ret);
}


int Sdb_open(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
/*
 bit0= no Sdb_oci_handle_error() in case of error
*/
{
 int ret;
 ret= Sdb_oci_open(sdb,flag);
 return(ret);
}


int Sdb_close(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 int ret;
 ret= Sdb_oci_close(sdb,flag);
 return(ret);
}


int Sdb_query(sdb,query_text,query_length,flag)
struct Sdb_ocI *sdb;
char *query_text;
int query_length;
int flag;
{
 int ret;
 ret= Sdb_oci_query(sdb,query_text,query_length,flag);
 return(ret);
}


int Sdb_request_result(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
/*
 bit4= buffer whole reply so query can be left open while intermediate query
*/
{
 int ret;
 ret= Sdb_oci_request_result(sdb,flag);
 return(ret);
}


int Sdb_esc_string(sdb,target,source,source_length,flag)
struct Sdb_ocI *sdb;
char *target;
char *source;
int source_length;
int flag;
{
 int ret;
 ret= Sdb_oci_esc_string(sdb,target,source,source_length,flag);
 return(ret);
}


int Sdb_end_query(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 int ret;
 ret= Sdb_oci_end_query(sdb,flag);
 return(ret); 
} 


int Sdb_free_result(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 int ret;
 ret= Sdb_oci_free_result(sdb,flag); 
 return(ret);  
}  


int Sdb_unique_seed(sdb,fieldname,seed,flag)
/* produces a unique id or a seed value which is converted to a unique id
   by the RDB server */
struct Sdb_ocI *sdb;
char *fieldname;
unsigned long *seed;
int flag;
{
 int ret; 
 ret= Sdb_oci_unique_seed(sdb,fieldname,seed,flag);  
 return(ret);   
}   


int Sdb_get_unique_id(sdb,fieldname,id,flag)
/* reads the actual unique id after the RDB server processed the seed */
struct Sdb_ocI *sdb;
char *fieldname;
unsigned long *id;
int flag;
{
 int ret;  
 ret= Sdb_oci_get_unique_id(sdb,fieldname,id,flag);
 return(ret);    
}    


int Sdb_get_fieldcount(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 int ret;   
 ret= Sdb_oci_get_fieldcount(sdb,flag); 
 return(ret);     
}     
 

int Sdb_get_row(sdb,fields,lengths,flag)
struct Sdb_ocI *sdb;
char ***fields;
unsigned long **lengths;
int flag;
{
 int ret;    
 ret= Sdb_oci_get_row(sdb,fields,lengths,flag);  
 return(ret);      
}      


int Sdb_update_worked(sdb,flag)
struct Sdb_ocI *sdb;
int flag;
{
 int ret;

 ret= Sdb_oci_update_worked(sdb,flag);
 return(ret);
}


char *Sdb__get_typelist_pt(flag)
int flag;
{
 return(Sdb_oci_typelisT);
}


#endif /* Sdb_oci_gen_apI */
