
/*
 cc -g -c sdb_oci_stq.c
 link with :  libclntsh
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"


/* this also includes "oci.h"  */
#include "sdb_oci.h"


#include "sdb_oci_stq.h"


/* ---------------------- SDB wrapper for OCIStmt -------------------- */


int Sdb_oci_stq_new(stq,boss,stmt_text,stmt_length,flag)
struct Sdb_oci_stQ **stq;
struct Sdb_ocI *boss;
char *stmt_text;
int stmt_length;
int flag;
{
 struct Sdb_oci_stQ *m;

 m= *stq= TSOB_FELD(struct Sdb_oci_stQ,1);
 if(m==NULL)
   return(-1);

 m->boss= boss;
 m->stmt= NULL;
 m->stmt_text= NULL;
 m->stmt_length= 0;
 m->stmt_type= 0;
 m->stmt_no_data= 0;
 m->row_size;
 m->row_offsets= NULL;
 m->row_types= NULL;
 m->row_rlengths= NULL;
 m->row_rcodeps= NULL;
 m->field_adr= NULL;
 m->field_len= NULL;
 m->column_count= 0;
 m->row_buffer= NULL;
 m->receive_buffer= NULL;
 m->receive_cursor= NULL;
 m->result_buffered= 0;

 if(stmt_length<=0)
   goto failed;
 m->stmt_text= Smem_malloC(stmt_length);
 if(m->stmt_text==NULL)
   goto failed;
 memcpy(m->stmt_text,stmt_text,stmt_length);
 m->stmt_length= stmt_length;

 return(1);

failed:;
 Sdb_oci_stq_destroy(stq,0);
 return(-1);
}

int Sdb_oci_stq_free_result(stq,flag)
struct Sdb_oci_stQ *stq;
int flag;
{
 if(stq->row_types!=NULL)
   Smem_free((char *) stq->row_types);
 stq->row_types= NULL;
 if(stq->row_offsets!=NULL)
   Smem_free((char *) stq->row_offsets);
 stq->row_offsets= NULL;
 if(stq->row_rlengths!=NULL)
   Smem_free(stq->row_rlengths);
 stq->row_rlengths= NULL;
 if(stq->row_rcodeps!=NULL)
   Smem_free(stq->row_rcodeps);
 stq->row_rcodeps= NULL;
 if(stq->field_adr!=NULL)
   Smem_free((char *) stq->field_adr);
 stq->field_adr= NULL;
 if(stq->field_len!=NULL)
   Smem_free((char *) stq->field_len);
 stq->field_len= NULL;
 if(stq->row_buffer!=NULL)
   Smem_free((char *) stq->row_buffer);
 stq->row_buffer= NULL;
 return(1);
}


int Sdb_oci_stq_destroy(stq,flag)
struct Sdb_oci_stQ **stq;
int flag;
{
 struct Sdb_oci_stQ *m;

 m= *stq;
 if(m==NULL)
   return(0);
 if(m->stmt!=NULL)
   OCIHandleFree((dvoid *) m->stmt, OCI_HTYPE_STMT);
 if(m->stmt_text!=NULL)
   Smem_free(m->stmt_text);
 Sdb_oci_stq_free_result(m,0);
 Lstring_destroy_all(&(m->receive_buffer),0);
 Smem_freE((char *) m);
 *stq= NULL;
 return(1);
}


int Sdb_oci_stq_prepare(stq,flag)
struct Sdb_oci_stQ *stq;
int flag;
{
 sword oret;
 char *dummy;
 int ret;
 struct Sdb_ocI *sdb;

 sdb= stq->boss;

 oret= OCIHandleAlloc(sdb->db,(dvoid *) &(stq->stmt),OCI_HTYPE_STMT,0,
                      (dvoid **) &dummy);
 if(oret != OCI_SUCCESS) {
   Sdb_oci_checkerr(sdb,oret,sdb->message,0);
   return(0);
 } 
 oret= OCIStmtPrepare(stq->stmt,sdb->errhp,stq->stmt_text,stq->stmt_length,
                      OCI_V8_SYNTAX,OCI_DEFAULT);
 if(oret != OCI_SUCCESS) {
   Sdb_oci_checkerr(sdb,oret,sdb->message,0);
   return(0);
 }
 oret= OCIAttrGet(stq->stmt,OCI_HTYPE_STMT,&(stq->stmt_type),NULL,
                  OCI_ATTR_STMT_TYPE,sdb->errhp);
 stq->stmt_no_data= 0;
 if(oret == OCI_NO_DATA) {
   stq->stmt_no_data= 1;
 } else if(oret != OCI_SUCCESS) {
   Sdb_oci_checkerr(sdb,oret,sdb->message,0);
   return(0);
 }
 return(1);
}


int Sdb_oci_stq_execute(stq,flag)
struct Sdb_oci_stQ *stq;
int flag;
{
 struct Sdb_ocI *sdb;
 sword oret;

 sdb= stq->boss;
 oret= OCIStmtExecute(sdb->svchp,stq->stmt,sdb->errhp,
                      stq->stmt_type!=OCI_STMT_SELECT,0,NULL,NULL,OCI_DEFAULT); 
 if(oret != OCI_SUCCESS) {
   Sdb_oci_checkerr(sdb,oret,sdb->message,0);
   return(0);
 }
 return(1);
}


int Sdb_oci_stq_is_commitable(stq,flag)
struct Sdb_oci_stQ *stq;
int flag;
{
 struct Sdb_ocI *sdb;
 sword oret;
 ub2 stmt_type;

 sdb= stq->boss;
 oret= OCIAttrGet((void*) stq->stmt, (ub4) OCI_HTYPE_STMT, (dvoid*) &stmt_type,
                  (ub4 *) NULL, (ub4) OCI_ATTR_STMT_TYPE, sdb->errhp);
 if(oret != OCI_SUCCESS) {
   Sdb_oci_checkerr(sdb,oret,sdb->message,0);
   return(-1);
 }
 if(stmt_type==OCI_STMT_UPDATE || stmt_type==OCI_STMT_INSERT ||
    stmt_type==OCI_STMT_DELETE)
   return(1);
 return(0);
}


int Sdb_oci_stq_rowform(stq,flag)
struct Sdb_oci_stQ *stq;
int flag;
{
 int ret,i;
 sword oret;
 unsigned long rowsize= 0,row_offsets[256],colcount= 0;
 OCIParam *pard;
 ub2 dtypes[256];
 ub4 dsize;
 struct Sdb_ocI *sdb;

 sdb= stq->boss;

 /* compute pointer offsets and total rowsize */
 oret= OCI_SUCCESS;
 row_offsets[0]= 0;
 for(i= 0;i<sizeof(row_offsets)/sizeof(int);i++) {
   oret= OCIParamGet(stq->stmt,OCI_HTYPE_STMT,sdb->errhp,(dvoid **) &pard,
                     (ub4) i+1);
   if(oret!=OCI_SUCCESS) 
 break;   
   oret= OCIAttrGet((dvoid*) pard,(ub4) OCI_DTYPE_PARAM,(dvoid*) &(dtypes[i]),
                    (ub4 *) NULL, (ub4) OCI_ATTR_DATA_TYPE,sdb->errhp);
   if(oret != OCI_SUCCESS) {
     Sdb_oci_checkerr(sdb,oret,sdb->message,0);
     return(0);
   }
   oret= OCIAttrGet((dvoid*) pard,(ub4) OCI_DTYPE_PARAM,(dvoid*) &dsize,
                    (ub4 *) NULL, (ub4) OCI_ATTR_DATA_SIZE,sdb->errhp);
   if(oret != OCI_SUCCESS) {
     Sdb_oci_checkerr(sdb,oret,sdb->message,0);
     return(0);
   }
   rowsize+= dsize;
   rowsize+= 4-(rowsize%4);
   row_offsets[i+1]= rowsize;
   colcount= i;
 }

 /* attach row description and define receiving memory */;
 Sdb_oci_stq_free_result(stq,0);
 stq->row_size= rowsize;
 stq->column_count= 0;
 if(colcount<=0) 
   return(1);

 stq->row_offsets= TSOB_FELD(unsigned long,colcount+1);
 if(stq->row_offsets==NULL) {
no_mem:;
   sprintf(sdb->message,"Not enough memory to receive %ld columns.",
                        colcount);
   return(-1);
 }
 stq->row_types= TSOB_FELD(ub2,colcount);
 if(stq->row_types==NULL)
   goto no_mem;
 stq->row_rlengths= TSOB_FELD(ub2,colcount);
 if(stq->row_rlengths==NULL)
   goto no_mem;
 stq->row_rcodeps= TSOB_FELD(ub2,colcount);
 if(stq->row_rcodeps==NULL)
   goto no_mem;
 stq->field_adr= TSOB_FELD(char *,colcount);
 if(stq->field_adr==NULL)
   goto no_mem;
 stq->field_len= TSOB_FELD(unsigned long,colcount);
 if(stq->field_len==NULL)
   goto no_mem;
 stq->row_buffer= TSOB_FELD(char,rowsize);
 if(stq->row_buffer==NULL)
   goto no_mem;
 stq->column_count= colcount;
 stq->row_offsets[0]= 0;
 for(i=0;i<colcount;i++) {
   stq->row_offsets[i+1]= row_offsets[i+1];
   stq->row_types[i]= dtypes[i];
   stq->field_adr[i]= stq->row_buffer+stq->row_offsets[i];

   /* Define the receiving memory areas */;
   oret= OCIDefineByPos(stq->stmt,(OCIDefine **) NULL, sdb->errhp,i+1,
                        (dvoid *) stq->field_adr[i],
                        stq->row_offsets[i+1]-stq->row_offsets[i],
                        stq->row_types[i], (dvoid *) NULL,
                        (dvoid *) &(stq->row_rlengths[i]),
                        (dvoid *) &(stq->row_rcodeps[i]),
                        OCI_DEFAULT);
   if(oret != OCI_SUCCESS) {
     Sdb_oci_checkerr(sdb,oret,sdb->message,0);
     return(0);
   }
 }
 return(1);
}


int Sdb_oci_stq_resultbuf(stq,flag)
struct Sdb_oci_stQ *stq;
int flag;
/*
 bit4= buffer whole reply so query can be left open while intermediate query
*/
{
 int ret,bsize;
 sword oret;
 struct Sdb_ocI *sdb;
 struct LstrinG *lpt;
 char *wpt;

 sdb= stq->boss;
 
 bsize= stq->row_size+2*(stq->column_count*sizeof(ub2));
 stq->result_buffered= 0;
 if( sdb->fetch_mode==0 || (flag&16) ) {
   /* read all rows into a list buffer */
   Lstring_destroy_all(&(stq->receive_buffer),0);
   lpt= NULL;
   while(1) {
     ret= Sdb_oci_stq_fetch_row(stq,0);
     if(ret<0)
       return(ret);  
     if(ret==0)
   break;
     ret= Lstring_new_binary(&(lpt),NULL,stq->row_size,lpt,2);
     if(ret<=0)
       goto no_mem;
     if(stq->receive_buffer==NULL)
       stq->receive_buffer= lpt;
     wpt= lpt->text;
     memcpy(wpt,stq->row_buffer,stq->row_size);
     wpt+= stq->row_size;
     memcpy(wpt,stq->row_rlengths,stq->column_count*sizeof(ub2));
     wpt+= stq->column_count*sizeof(ub2);
     memcpy(wpt,stq->row_rcodeps,stq->column_count*sizeof(ub2));
     wpt+= stq->column_count*sizeof(ub2);
   }
   stq->result_buffered= 1;
 } 
 stq->receive_cursor= stq->receive_buffer; 
 return(1);

no_mem:;
 sprintf(sdb->message,"Not enough memory to allocate receive buffer.");
 return(-1);
}


int Sdb_oci_stq_column_count(stq,flag)
struct Sdb_oci_stQ *stq;
int flag;
{
 return(stq->column_count);
}


int Sdb_oci_stq_next_row(stq,fields,lengths,flag)
struct Sdb_oci_stQ *stq;
char ***fields;
unsigned long **lengths;
int flag;
{
 int retry_count= 0,ret,i;
 unsigned int *fieldlengths;
 struct Sdb_ocI *sdb;
 char *rpt;

 sdb= stq->boss;

 if(stq->result_buffered) {
   if(stq->receive_cursor==NULL)
     return(0);
   rpt= stq->receive_cursor->text;
   memcpy(stq->row_buffer,rpt,stq->row_size);
   rpt+= stq->row_size;
   memcpy(stq->row_rlengths,rpt,stq->column_count*sizeof(ub2));
   rpt+= stq->column_count*sizeof(ub2);
   memcpy(stq->row_rcodeps,rpt,stq->column_count*sizeof(ub2));
   rpt+= stq->column_count*sizeof(ub2);
   stq->receive_cursor= stq->receive_cursor->next;
   for(i=0;i<stq->column_count;i++) 
     stq->field_len[i]= stq->row_rlengths[i];
 } else {
   ret= Sdb_oci_stq_fetch_row(stq,0);
   if(ret<=0)
     return(ret);
 }
 *fields= stq->field_adr;
 *lengths= stq->field_len;
 return(1);
}


int Sdb_oci_stq_fetch_row(stq,flag)
struct Sdb_oci_stQ *stq;
int flag;
{
 sword oret;
 struct Sdb_ocI *sdb;
 int i;

 sdb= stq->boss;

 /* fetch row data into row_buffer */
 oret= OCIStmtFetch(stq->stmt,sdb->errhp,1,OCI_FETCH_NEXT,OCI_DEFAULT);
 if(oret == OCI_NO_DATA) {
   return(0);
 } else if(oret != OCI_SUCCESS) {
   Sdb_oci_checkerr(sdb,oret,sdb->message,0);
   return(-1);
 }
 for(i=0;i<stq->column_count;i++) 
   stq->field_len[i]= stq->row_rlengths[i];

 return(1);
}


int Sdb_oci_stq_end_query(stq,flag)
struct Sdb_oci_stQ *stq;
int flag;
{
 int ret;

 if(stq->result_buffered)
   return(2);

 /* >>> ??? is there a fast way to cancel a query ? */

 do {
   ret= Sdb_oci_stq_fetch_row(stq,0);
 } while(ret>0);
 if(ret<0)
   return(ret);
 return(1);
}


/* >>> BAUSTELLE */


int Sdb_oci_stq_update_worked(stq,flag)
struct Sdb_oci_stQ *stq;
int flag;
{
 int ret;

 /* >>> find out wether the update clause matched at least one row */
 ret= 1;

 return(ret);
}


