
/* 
  cc -g -c slist.c
*/

#include <stdio.h>


#include "smem.h"

#include "slist.h"


int Slist_new(list,data,Data_refer,Data_destroy,link,flag)
struct SlisT **list;
char *data;
int (*Data_refer) ();              /* (char *data, char **referer, int flag) */
                                                      /* NULL = refer by "=" */
int (*Data_destroy) ();
struct SlisT *link;
int flag;
/*
 bit0= insert before link rather than after it
*/
{
 int ret;
 struct SlisT *s;

 (*list)= s= TSOB_FELD(struct SlisT,1);
 if(s==NULL)
   return(-1);
 s->data= NULL;
 s->next= s->prev= NULL;
 if(Data_refer!=NULL && data!=NULL)
   Data_refer(data,&(s->data),0);
 else
   s->data= data;
 s->Data_destroy= Data_destroy;
 Slist_attach(s,link,flag&1);
 return(1);
}


/* use this as Data_destroy() for plain Smem_malloc or TSOB_FELD data */
int Slist__free(data,flag)
char **data;
int flag;
{
 if(*data==NULL)
   return(0);
 Smem_freE(*data);
 *data= NULL;
 return(1);
}


int Slist_destroy_data(s,flag)
struct SlisT *s;
int flag;
{
 if(s->data==NULL) 
   return(0);
 if(s->Data_destroy!=NULL)
   s->Data_destroy(&(s->data),0);
 s->data= NULL;
 return(1);
}


int Slist_destroy(list,flag)
struct SlisT **list;
int flag;
{
 struct SlisT *s;

 s= *list;
 if(s==NULL)
   return(0);
 if(s->prev!=NULL)
   s->prev->next= s->next;
 if(s->next!=NULL)
   s->next->prev= s->prev;
 Slist_destroy_data(s,0);
 Smem_freE((char *) s);
 (*list)= NULL;
 return;
}


int Slist_destroy_all(list,flag)
struct SlisT **list;
int flag;
{
 struct SlisT *s,*next;

 if((*list)==NULL)
   return(0);
 for(s= *list; s->prev!=NULL; s= s->prev);
 for(;s!=NULL;s= next){
   next= s->next;
   Slist_destroy(&s,0);
 }
 *list= NULL;
 return(1);
}


int Slist_get_data(list,data,flag)
struct SlisT *list;
char **data;
int flag;
{
 if(list==NULL)
   { *data= NULL; return(0); }
 *data= list->data;
 return(1);
}


char *Slist_return_data(list,flag)
struct SlisT *list;
int flag;
{
 if(list==NULL)
   return(NULL);
 return(list->data);
}


int Slist_replace_data(list,data,Data_destroy,flag)
struct SlisT *list;
char *data;
int (*Data_destroy) ();
int flag;
/*
 bit0= do not Slist_destroy_data()
*/
{
 if(list==NULL)
   return(0);
 if(!(flag&1))
   Slist_destroy_data(list,0);
 list->data= data;
 list->Data_destroy= Data_destroy;
 return(1);
}


int Slist_get_next(list,next,flag)
struct SlisT *list;
struct SlisT **next;
int flag;
{
 if(list==NULL)
   { *next= NULL; return(0); }
 *next= list->next;
 return((*next)!=NULL);
}


int Slist_get_prev(list,prev,flag)
struct SlisT *list;
struct SlisT **prev;
int flag;
{
 if(list==NULL)
   { *prev= NULL; return(0); }
 *prev= list->prev;
 return((*prev)!=NULL);
}


int Slist_return_count(list,flag)
struct SlisT *list;
int flag;
{
 struct SlisT *l;
 int count= 0;

 if(list==NULL)
   return(0);
 for(l= list; l!=NULL; Slist_get_next(l,&l,0))
   count++;
 return(count);
}


int Slist_make_vector(list,vector,num,flag)
struct SlisT *list;
char ***vector;
int *num;
int flag;
/*
 bit0= revert sequence
 bit1= *num is actual count
*/
{
 struct SlisT *l;
 char **v; 
 int w;

 *vector= NULL;
 if(!(flag&2))
   *num= Slist_return_count(list,0);
 if(list==NULL || *num==0)
   return(1);
 (*vector)= v= TSOB_FELD(char *,*num);
 if(v==NULL)
   return(1);
 if(flag&1) {
   w= (*num)-1;
   for(l= list;w>=0;Slist_get_next(l,&l,0)) {
     Slist_get_data(l,&(v[w]),0);
     w--;
   }
 } else {
   w= 0;
   for(l= list;w<*num;Slist_get_next(l,&l,0)) {
     Slist_get_data(l,&(v[w]),0);
     w++;
   }
 }
 return(1);
}


int Slist_detach(list,flag)
struct SlisT *list;
int flag;
{
 if(list==NULL)
   return(0);
 if(list->next!=NULL)
   list->next->prev= list->prev;
 if(list->prev!=NULL)
   list->prev->next= list->next;
 list->next= NULL;
 list->prev= NULL;
 return(1);
}


int Slist_attach(list,link,flag)
struct SlisT *list;
struct SlisT *link;
int flag;
/*
 bit0= insert before link rather than after it
*/
{
 Slist_detach(list,0);
 if(link==NULL) {
   ;
 } else if(flag&1) {
   list->next= link;
   list->prev= link->prev;
   if(link->prev!=NULL)
     link->prev->next= list;
   link->prev= list;
 } else {
   list->prev= link;
   list->next= link->next;
   if(link->next!=NULL)
     link->next->prev= list;
   link->next= list;
 }
 return(1);
}


int Slist_shuffle(start,end,flag)
struct SlisT **start,**end;
int flag;
/*
 bit0= shuffle in prev direction
*/
{
 struct SlisT *new_start,*new_end;

 if((*start)==(*end) || (*start)==NULL || (*end)==NULL)
   return(0);
 if((*start)->next==NULL || (*end)->prev==NULL)
   return(0);
 if(flag&1) {
   new_start= (*start)->next;
   new_end= *start;
   Slist_attach(*start,*end,0);
 } else {
   new_start= *end;
   new_end= (*end)->prev;
   Slist_attach(*end,*start,1);  
 }
 *start= new_start;
 *end= new_end;
 return(1);
}


