

/*
 cc -g -c stcp.c
*/

#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <fcntl.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define closesocket close
/* for debugging : #define closesocket Stcp_closesocket */

#include "smem.h"
#include "sfile.h"
#include "stcp.h"


/* --------------------------------- Stcp ------------------------------ */
/* some raw TCP/IP functionality, used by higher level connection models */


int Stcp_clientfd(fd,servername,portnumber,flag)
int *fd;
char *servername;
int portnumber;
int flag;
{
 int ret,sock,protocol_no;
 struct sockaddr_in server;
 struct hostent *hp;
 struct hostent *gethostbyname();
 struct protoent *pe;

 pe = getprotobyname("tcp");
 if(pe==NULL)
   protocol_no= 0;
 else
   protocol_no= pe->p_proto;
 sock= socket(AF_INET,SOCK_STREAM,protocol_no);
 if(sock<0) {
   fprintf(stderr,"Stcp_clientfd : socket() failed. errno=%ld  (%s)\n",
                  errno,strerror(errno));
   return(-1);
 }
 server.sin_family= AF_INET;
 hp= gethostbyname(servername);
 if(hp==NULL) {
   fprintf(stderr,
           "Stcp_clientfd : gethostbyname(\"%s\") failed. errno=%ld  (%s)\n",
           servername,errno,strerror(errno));
   return(0);
 }
 memcpy((char *) &server.sin_addr,(char *) hp->h_addr,hp->h_length);
 server.sin_port = htons((unsigned short) portnumber);
 if(connect(sock,(struct sockaddr *) &server,sizeof(server))<0) {
   fprintf(stderr,"Stcp_clientfd : connect() failed. errno=%ld  (%s)\n",
                  errno,strerror(errno));
   return(0);
 }
 *fd= sock;
 return(1);
}


int Stcp_client_get_hostadr(sock,adr,flag)
int sock;
char adr[81];
int flag;
{
 int ret,client_len,x,i;
 struct sockaddr_in client;

 client_len= sizeof(client);
 ret= getsockname(sock,(struct sockaddr *) &client,&client_len);
 if(ret==-1) {
   fprintf(stderr,
           "Stcp_client_get_hostadr : getsockname() failed. errno=%ld  (%s)\n",
           errno,strerror(errno));
   return(0);
 }
 adr[0]= 0;
 for(i=0;i<4;i++) {
   x= (client.sin_addr.s_addr>>(i*8))&255;
   sprintf(adr+strlen(adr),"%u",x);
   if(i<3)
     strcat(adr,".");
 }
 return(1);
}

/* ------------------------------- StcpserV --------------------------- */


int Stcpserv_new(serv,servername,portnumber,external_s_mask,flag)
struct StcpserV **serv;
char *servername;
int portnumber;
int external_s_mask;
int flag;
{
 struct StcpserV *s;
 int ret;

 *serv= s= TSOB_FELD(struct StcpserV,1);
 if(s==NULL)
   return(-1);
 strcpy(s->servername,servername);
 s->portnumber= portnumber;
 s->is_ready= 0;
 s->server_socket= -1;
 s->server_socket_fp= NULL;
 s->external_s_addr= 0x0;
 s->external_s_mask= external_s_mask;
 s->published_file= NULL;
 ret= Stcpserv_start(s,0);
 if(ret<=0)
   Stcpserv_destroy(serv,0);
 return(ret);
}


int Stcpserv_destroy(serv,flag)
struct StcpserV **serv;
int flag;
{
 if(*serv==NULL)
   return(0);
 if((*serv)->is_ready)
   Stcpserv_stop(*serv,0);
 Smem_freE((char *) *serv);
 *serv= NULL;
 return(1);
}


int Stcpserv_start(serv,flag)
struct StcpserV *serv;
int flag;
{
 int sock,length,protocol_no;
 struct sockaddr_in server;
 struct hostent *hp;
 struct protoent *pe;

 if(serv==NULL)
   return(0);

 hp = gethostbyname(serv->servername) ;
 if(hp == 0) {
   fprintf(stderr,"Stcpserv_start : gethostbyname() failed. errno=%ld  (%s)\n",
                  errno,strerror(errno));
   return(-1);
 }
 serv->external_s_addr= ntohl(*((unsigned long *) hp->h_addr));
 
 pe = getprotobyname("tcp");
 if(pe==NULL)
   protocol_no= 0;
 else
   protocol_no= pe->p_proto;
 sock = socket(AF_INET, SOCK_STREAM, protocol_no);
 if(sock<0){
   fprintf(stderr,"Stcpserv_start : socket() failed. errno=%ld  (%s)\n",
                  errno,strerror(errno));
   return(-1);
 }
 server.sin_family= AF_INET;
 server.sin_addr.s_addr= INADDR_ANY;
 server.sin_port = htons(serv->portnumber) ;
 if(bind(sock,(struct sockaddr *)&server,sizeof server)<0) {
   fprintf(stderr,"Stcpserv_start : bind() failed. errno=%ld  (%s)\n",
                  errno,strerror(errno));
   goto ex; 
 }
 length= sizeof(server);
 if(getsockname(sock,(struct sockaddr *)&server,&length)<0){
   fprintf(stderr,"Stcpserv_start : getsockname() failed. errno=%ld  (%s)\n",
                  errno,strerror(errno));
   goto ex; 
 }
 if(fcntl(sock,F_SETFL,O_NONBLOCK)==-1){
   fprintf(stderr,
       "Stcpserv_start : socket refuses non-blocking mode. errno=%ld  (%s)\n",
           errno,strerror(errno));
   goto ex;
 }

 serv->portnumber= ntohs(server.sin_port);
 serv->server_socket= sock;
 serv->is_ready= 1;
 listen(sock,5);
 return(1);

ex:;
 closesocket(sock);
 return(0);
}


int Stcpserv_stop(serv,flag)
struct StcpserV *serv;
int flag;
{
 if(serv==NULL)
   return(0);
 Stcpserv_publish_adress(serv,NULL,0);
 if(serv->server_socket>=0)
   closesocket(serv->server_socket);
 serv->server_socket= -1;
 serv->is_ready= 0;
 return(1);
}


int Stcpserv_get_server_socket(serv,flag)
struct StcpserV *serv;
int flag;
{
 return(serv->server_socket);
}


int Stcpserv_get_portnumber(serv,flag)
struct StcpserV *serv;
int flag;
{
 return(serv->portnumber);
}


int Stcpserv_publish_adress(serv,fileadr,flag)
struct StcpserV *serv;
char *fileadr;
int flag;
/*
 bit0= only revoke current publication (same happens if fileadr==NULL)
*/
{
 int ret;
 char line[SfileadrL];

 if(serv->published_file!=NULL) {
   unlink(serv->published_file);
   Sregex_string(&(serv->published_file),NULL,0);
 }
 if((flag&1) || fileadr==NULL)
   return(1);
 Sregex_string(&(serv->published_file),fileadr,0);
 if(serv->is_ready<=0 || serv->server_socket<0)
   return(0);
 sprintf(line,"%s %ld",serv->servername,serv->portnumber);
 ret= Sfile_string_to_file(fileadr,line,1);
 return(ret);
}


int Stcpserv_accept_fd(serv,acc_sock,flag)
struct StcpserV *serv;
int *acc_sock;
int flag;
/*
 bit0= do not set accepted fd to non blocking mode
 bit1= report host address of caller
*/
{
 int ret;
 char hostadr[80];

 ret= Stcpserv_accept_fd_adr(serv,acc_sock,hostadr,flag&3);
 return(ret);
}


int Stcpserv_accept_fd_adr(serv,acc_sock,hostadr,flag)
struct StcpserV *serv;
int *acc_sock;
unsigned char hostadr[80];              /* oversized for future enhancements */
int flag;
/*
 bit0= do not set accepted fd to non blocking mode
 bit1= report host address of caller
*/
{
 int accepted_l,s_addr,i;
 unsigned char *cpt;
 struct sockaddr_in accepted;

 accepted_l= sizeof(accepted);

 if(!Stcpfd_is_ready(serv->server_socket,0,1|4)) {
   return(0); /* no connection request pending */
 }
 *acc_sock= accept(serv->server_socket,
                   (struct sockaddr *) &accepted,&accepted_l);
 if(*acc_sock==-1){
   if(errno==EAGAIN)
     return(0); /* no connection request pending */
   fprintf(stderr,
           "Stcpserv_accept : accept() failed. errno=%ld  (%s)\n",
           errno,strerror(errno));
   return(-1);
 }

 s_addr= ntohl(accepted.sin_addr.s_addr);
 cpt= (char *) &accepted.sin_addr.s_addr;
 for(i=0;i<4;i++)
   hostadr[i]= cpt[i];
 hostadr[4]= 0;
#ifdef Stcp_localhost_onlY
 if(s_addr!=0x7F000001) {
#else /* Stcp_localhost_onlY */
 if(s_addr!=0x7F000001 &&
    (s_addr&serv->external_s_mask)!=
    (serv->external_s_addr&serv->external_s_mask)) {
#endif /* ! Stcp_localhost_onlY */
   fprintf(stderr,
           "Stcpserv_accept : accept() of external connection refused.");
   fprintf(stderr," origin= %lu.%lu.%lu.%lu\n",cpt[0],cpt[1],cpt[2],cpt[3]);
   closesocket(*acc_sock);
   *acc_sock= -1;
   return(0);
 }
 if(flag&2) {
   fprintf(stderr," origin= %lu.%lu.%lu.%lu\n",cpt[0],cpt[1],cpt[2],cpt[3]);
 }
 if((flag&1)) 
   return(1);
 if(fcntl(*acc_sock,F_SETFL,O_NONBLOCK)==-1){
   fprintf(stderr,
     "Stcpserv_accept_fd : socket refuses non-blocking mode. errno=%ld  (%s)\n",
     errno,strerror(errno));
   return(0);
 }

 return(1);
}


int Stcpserv_accept_conn(serv,conn,flag)
struct StcpserV *serv;
struct StcpconN **conn;
int flag;
{
 int fd= -1,ret;
 unsigned char adr[80];
 char foreign_hostname[320];

 ret= Stcpserv_accept_fd_adr(serv,&fd,adr,0);
 if(ret<=0)
   return(ret);
 sprintf(foreign_hostname,"%lu.%lu.%lu.%lu",adr[0],adr[1],adr[2],adr[3]);
 ret= Stcpconn_new(conn,serv,fd,foreign_hostname,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 if(ret<=0){
   if(fd>=0)
     closesocket(fd);
 }
 return(ret);
}


/* ------------------------------ StcpconN --------------------------------*/


int Stcpconn_new(conn,serv,fd,foreign_hostname,flag)
struct StcpconN **conn;
struct StcpserV *serv;
int fd;
char *foreign_hostname;
int flag;
{
 struct StcpconN *c;
 int ret;

 *conn= c= TSOB_FELD(struct StcpconN,1);
 if(c==NULL)
   return(-1);
 c->fd= fd;
 c->is_open= (fd>=0);
 c->server= serv;
 c->foreign_hostname= NULL;
 c->poll_microsecs= Stcpconn_microsleeP;

 if(Sregex_string(&(c->foreign_hostname),foreign_hostname,0)<=0)
   goto failed;

 return(1);
failed:
 Stcpconn_destroy(conn,0);
 return(-1);
}


int Stcpconn_destroy(conn,flag)
struct StcpconN **conn;
int flag;
{
 struct StcpconN *c;

 c= *conn;
 if(c==NULL)
   return(0);
 if(c->is_open)
   closesocket(c->fd);
 Sregex_string(&(c->foreign_hostname),NULL,0);
 Smem_freE((char *) c);
 *conn= NULL;
 return(1);
}


int Stcpconn_fd(conn,flag)
struct StcpconN *conn;
int flag;
{
 if(!conn->is_open)
   return(-1);
 return(conn->fd);
}


int Stcpconn_get_foreign_hostname(conn,foreign_hostname,flag)
struct StcpconN *conn;
char **foreign_hostname;
int flag;
/*
 bit0= point to "" rather than contain NULL
*/
{
 *foreign_hostname= conn->foreign_hostname;
 if(flag&1)
   if((*foreign_hostname) == NULL)
     (*foreign_hostname)= "";
 return(1);
}


int Stcpconn_set_fd(conn,value,flag)
struct StcpconN *conn;
int value;
int flag;
{
 conn->fd= value;
 return(1);
}


int Stcpconn_set_is_open(conn,value,flag)
struct StcpconN *conn;
int value;
int flag;
{
 conn->is_open= !!value;
 return(1);
}


int Stcpconn_read(conn,buf,bufsize,count,flag)
struct StcpconN *conn;
char *buf;
int bufsize;
int *count;
int flag;
{
 int ret;

 ret= Stcpfd_read(conn->fd,buf,bufsize,count,0);
 return(ret);
}


int Stcpconn_write(conn,buf,count,flag)
struct StcpconN *conn;
char *buf;
int *count;
int flag;
/*
 bit0= if would-block : return(2) rather than wait
*/
{
 int ret;

 ret= Stcpfd_write(conn->fd,buf,count,conn->poll_microsecs,(flag&1));
 return(ret);
}


int Stcpconn_is_ready(conn,flag)
struct StcpconN *conn;
int flag;
/*
 bit0= check for reading
 bit1= check for writing
 bit2= check for exception
return-bits are set according to matching flag-bits
*/
{
 int ret;

 ret= Stcpfd_is_ready(conn->fd,0,flag&7);
 return(ret);
}



/* -------------------------- Stcpfd --------------------------- */
/*  A layer between TCP/IP filedescriptors and standard UNIX I/O */
/*  There are systems where write+read do not work with TCP/IP   */
/*  Best is to avoid them, second best is to implement a         */
/*  workaround here. */


int Stcpfd_read(fd,buf,bufsize,count,flag)
int fd;
char *buf;
int bufsize;
int *count;
int flag;
/*
 return:
  <0= error
   0= EOF (connection closed)
  >0= number of bytes read
*/
{
 int ret;

 ret= Sfile_read(fd,buf,bufsize,count,0);
 return(ret);
}


int Stcpfd_write(fd,buf,count,microsec,flag)
int fd;
char *buf;
int *count;
unsigned int microsec;
int flag;
/*
 bit0= if would-block : return(2) rather than wait
*/
{
 int ret;

 ret= Sfile_write(fd,buf,count,microsec,(flag&1));
 return(ret);
}


int Stcpfd_is_ready(fd,microsec,flag)
int fd;
unsigned int microsec;
int flag;
/*
 bit0= check for reading
 bit1= check for writing
 bit2= check for exception
return-bits are set according to matching flag-bits
*/
{
 int ret;

 ret= Sfile_fd_is_ready(fd,microsec,flag&7);
 return(ret);
}


int Stcpfd_close(fd,flag)
int fd;
int flag;
{
 closesocket(fd);
 return(1);
}


/* ----------------------- for debugging purposes ---------------------- */

int Stcp_closesocket(fd)
int fd;
{
 return(close(fd));
}


