

/*
  cc -g -c stunnel.c
*/

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>



#ifdef Stunnel_use_smeM

#include "smem.h"

#else /* Stunnel_use_smeM */

#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));
#define Smem_malloC malloc
#define Smem_freE free

#endif /* ! Stunnel_use_smeM */



#include "sfile.h"
#include "sregex.h"
#include "sblowfish.h"

#include "stunnel.h"


/* -------------------------------- StunneL ------------------------------ */


int Stunnel_new(tunnel,user,idno,keydir,random_key,flag)
struct StunneL **tunnel;
char *user;
int idno;
char *keydir;
char *random_key; /* NULL or length matching kl of protocol version */
int flag;
/*
 flag:
  bit0= do not make personal fish
  bit4-11= protocol version :
        0= use default version (currently 0.2)
        1= 0.0 kl=16 cl=8 ml=1024
        2= 0.2 kl=32 cl=16 ml=1024
 return:
  1 = ok
  0 = non final error
 -1 = final error
 -2 = unknown user
*/
{
 struct StunneL *t;
 char key[Stunnel_max_keyleN+1];
 int i,ret,protocol_version;
 double m;

 protocol_version= (flag>>4)&155;
 if(strlen(user)>Stunnel_userleN)
  return(-2);
 *tunnel= t= TSOB_FELD(struct StunneL,1);
 if(t==NULL)
   return(-1);
 if(protocol_version==0) /* default is version 0.2 */
   protocol_version= 2;
 if(protocol_version==2) {
   t->version_major= 0;
   t->version_minor= 2;
   t->keylen= Stunnel_v0_2_keyleN;
   t->challen= Stunnel_v0_2_challeN;
   t->headlen= Stunnel_v0_2_headleN;
 } else {
   t->version_major= 0;
   t->version_minor= 0;
   t->keylen= Stunnel_v0_0_keyleN;
   t->challen= Stunnel_v0_0_challeN;
   t->headlen= Stunnel_v0_0_headleN;
 }
 t->user[0]= 0;
 t->authenticated= 0;
 t->keydir= NULL;
 t->idno= idno;
 t->internal_fish= NULL;
 t->random_fish= NULL;
 t->random_count= 0;
 t->personal_fish= NULL;
 t->session_fish_in= NULL;
 t->session_fish_out= NULL;
 for(i=0;i<=Stunnel_max_challeN;i++)
   t->challenge[i]= t->reply[i]= t->in_reply[i]= t->in_challenge[i]= 0;
 t->in_status= 0;
 for(i=0;i<Stunnel_max_headleN;i++)
   t->in_head[i]= 0;
 t->head_count= 0;
 t->data_len= 0;
 t->data_count= 0;
 t->data_buffer= NULL;
 t->data_crc= 0;
 for(i=0;i<sizeof(t->data_sha1);i++)
   t->data_sha1[i]= 0;
 t->sha1_ctx= NULL;
 t->in_seqno= 0;
 t->out_seqno= 0;
 t->out_seqno_start= 0;
 t->errmsg[0]= 0;
 t->next= t->prev= NULL;

 ret= Stunnel_set_user(t,user,0);
 if(ret<=0)
   goto failed;
 if(Sregex_string(&(t->keydir),keydir,0)<=0)
   {ret= -1; goto failed;}
 ret= Sblowfish_new(&(t->internal_fish),
                    Sblowfish_make_key(Stunnel_internal_keY,t->keylen,
                                       key,0),
                    t->keylen,0);
 if(ret<=0)
   goto failed;

 if(random_key==NULL) {
   Sblowfish__global_random(key,t->keylen,-1,0);
 } else {
   for(i=0;i<t->keylen;i++)
     key[i]= random_key[i];
 }
 ret= Sblowfish_new(&(t->random_fish),key,t->keylen,0);
 if(ret<=0)
   goto failed;

 if(!(flag&1)){
   ret= Stunnel_make_personal(t,0);
   if(ret<0)
     goto failed;
   if(ret==0)
     {ret= -2; goto failed;}
 }
 return(1);
failed:
 Stunnel_destroy(tunnel,0);
 return(ret);
}


int Stunnel_destroy(tunnel,flag)
struct StunneL **tunnel;
int flag;
{
 struct StunneL *t;

 if(*tunnel==NULL)
   return(0);
 t= *tunnel;
 Sregex_string(&(t->keydir),NULL,0);
 Sblowfish_destroy(&(t->internal_fish),0);
 Sblowfish_destroy(&(t->random_fish),0);
 Sblowfish_destroy(&(t->personal_fish),0);
 if(t->session_fish_in!=t->session_fish_out)
   Sblowfish_destroy(&(t->session_fish_in),0);
 else
   t->session_fish_in= NULL;
 Sblowfish_destroy(&(t->session_fish_out),0);
 if(t->sha1_ctx!=NULL)
   Shash_sha1(&(t->sha1_ctx),NULL,0,NULL,1<<15);
   
 Smem_freE((char *) *tunnel);
 *tunnel= NULL;
 return(1);
}


int Stunnel_set_user(tunnel,username,flag)
struct StunneL *tunnel;
char *username;
int flag;
{
 char userfile[3*Stunnel_userleN+4];

 if(strlen(username)>=Stunnel_userleN)
   return(0);
 strcpy(userfile,username);
 if(strcmp(username,"internal_owner")==0)
   goto illegal_username;
 Sfile_esc_filename(userfile,sizeof(userfile)-1,1|4);
 if(strcmp(userfile,username)!=0 || strchr(username,':')!=NULL) {
illegal_username:;
   sprintf(tunnel->errmsg,"illegal username '%s'",userfile);
   return(0);
 }
 strcpy(tunnel->user,username);
 return(1);
}


int Stunnel_get_username(tunnel,username,flag)
struct StunneL *tunnel;
char **username;
int flag;
{
 if(Sregex_string(username,tunnel->user,0)<=0)
   return(-1);
 if(*username==NULL)
   return(0);
 return(1);
}


int Stunnel__protocol_number(text,flag)
char *text;
int flag;
{
 if(strcmp(text,"all")==0 || strcmp(text,"any")==0 || strcmp(text,"default")==0)   return(0);
 if(strcmp(text,"0.0")==0)
   return(1);
 if(strcmp(text,"0.2")==0)
   return(2);
 return(-1);
}

int Stunnel_make_personal(tunnel,flag)
struct StunneL *tunnel;
int flag;
{
 int ret,i;
 char key[Stunnel_max_keyleN+1];
 char filename[SfileadrL],userfile[3*Stunnel_userleN+4];

 Sblowfish_destroy(&(tunnel->personal_fish),0);

 strcpy(userfile,tunnel->user);
 Sfile_esc_filename(userfile,sizeof(userfile)-1,1|4);
 if(strcmp(userfile,tunnel->user)!=0 || strchr(userfile,':')!=NULL) {
   sprintf(tunnel->errmsg,"illegal username '%s'",userfile);
   return(0);
 }
 sprintf(filename,"%s/%s.tnl",tunnel->keydir,userfile);
 for(i=0;i<tunnel->keylen;i++)
   key[i]= 0;
 ret= Sfile_file_to_bin(filename,key,tunnel->keylen,4);
 if(ret<=0)
   return(ret);
 if(ret<tunnel->keylen) {
   if(ret<Stunnel_v0_0_keyleN) {
     sprintf(tunnel->errmsg,"insufficient number of key bytes in '%s'",
                            filename);
     return(0);
   }
   /* >>> ??? warn of non sufficient key length ? */;
   for(i=ret;i<tunnel->keylen;i++)
     key[i]= key[(i-ret)%ret];
 }
#ifndef Stunnel_no_fishinG
 ret= Sblowfish_transform(tunnel->internal_fish,key,tunnel->keylen,1|2);
#endif /* ! Stunnel_no_fishinG */
 if(ret<=0)
   return(ret);
 ret= Sblowfish_new(&(tunnel->personal_fish),key,tunnel->keylen,0);
 if(ret<=0)
   return(-1);
 return(1);
}


int Stunnel_make_challenge(tunnel,flag)
struct StunneL *tunnel;
int flag;
{
 Sblowfish_make_random(tunnel->random_fish,&(tunnel->random_count),
                       tunnel->challenge,tunnel->challen,4);
 Sblowfish_make_random(tunnel->random_fish,&(tunnel->random_count),
                       tunnel->iv_seed,Stunnel_iv_seedleN,4);
 return(1);
}


int Stunnel_make_reply(tunnel,challenge,reply,flag)
struct StunneL *tunnel;
char *challenge,*reply;
int flag;
{
 int j,i,c;
 
 /* just see data as 8-byte integers and increment each of them */
 for(j=0;j<tunnel->challen;j+=8) {
   for(i=0;i<8;i++) {
     c= (((unsigned char *) challenge)[j+i]+1)&0xff;
     reply[j+i]= c;
     if(c!=0)
   break;
   }
   for(i++;i<8;i++)
     reply[j+i]= challenge[j+i];
 }
 return(1);
}


int Stunnel_new_session_fish(tunnel,key,flag)
struct StunneL *tunnel;
char key[];
int flag;
/*
 bit0= in-fish (else out-fish)
*/
{
 int ret;
 struct SblowfisH **fishpt;

 if(flag&1)
   fishpt= &(tunnel->session_fish_in);
 else
   fishpt= &(tunnel->session_fish_out);
 if((*fishpt)!=NULL && tunnel->session_fish_in!=tunnel->session_fish_out)
   Sblowfish_destroy(fishpt,0);
 else
   *fishpt= NULL;
 ret= Sblowfish_new(fishpt,key,tunnel->keylen,0);
 if(ret<=0)
   goto ex;
 if(tunnel->version_major==0 && tunnel->version_minor==2) {
   Sblowfish_set_iv(*fishpt,tunnel->iv,0);
   Sblowfish_reset_chain(*fishpt,3);
 }
 ret= 1;
ex:;
 return(ret);
}


int Stunnel_format_token_v0_0(tunnel,message,messagelen,type,data,len,flag)
struct StunneL *tunnel;
char **message;
int *messagelen;
int type;
char *data;
int len;
int flag;
/*
 bit0= use personal_fish rather than session_fish
*/
{
 int ret,i,l=0,buffersize,result_len,transmit_len;
 unsigned long crc;
 char *buffer,typechars[8],seqnochars[16],rndchars[16];

 result_len= tunnel->headlen+len+8*(type!=0);
 result_len= result_len+7-((result_len+7)%8);
 buffersize= result_len+8;
 buffer= Smem_malloC(buffersize);
 if(flag&1)
   Sblowfish_reset_chain(tunnel->personal_fish,2);
 else
   Sblowfish_reset_chain(tunnel->session_fish_out,2);

 tunnel->out_seqno++;
 for(i=0;i<4;i++)
   seqnochars[3-i]= ((tunnel->out_seqno)>>(8*i))&0xff;
 /* some random bytes are added to obscure the header (especially the
    sequence numbers) and for padding :
      0-3:header  4-7:control header pad  8-15:data pad  */
 Sblowfish_make_random(tunnel->random_fish,&(tunnel->random_count),rndchars,16,
                       4);
 if(buffer==NULL)
   return(-1);
 for(i=0;i<4;i++)
    seqnochars[4+i]= rndchars[i];
 Sregex_crc32(seqnochars,8,&crc,1);
 if(type) {
   for(i=0;i<4;i++)
     typechars[3-i]= ((type)>>(8*i))&0xff;
   for(i=4;i<8;i++)
     typechars[i]= rndchars[i];
   Sregex_crc32(typechars,8,&crc,0);
 } 
 Sregex_crc32(data,len,&crc,0);

 for(i=0;i<4;i++)
   buffer[l+3-i]= (crc>>(8*i))&0xff;
 l+=4;
 if(type)
   transmit_len= -len;
 else
   transmit_len= len;
 for(i=0;i<4;i++)
   buffer[l+3-i]= (transmit_len>>(8*i))&0xff;
 l+= 4;
 for(i=0;i<8;i++)
   buffer[l+i]= seqnochars[i];
 l+= 8;
 if(type){
   for(i=0;i<8;i++)
     buffer[l+i]= typechars[i];
   l+= 8;
 }
 for(i=0;i<len;i++)
   buffer[l+i]= data[i];
 l+= len;
 for(i=l;i<buffersize;i++)
   buffer[i]= rndchars[8+(i%8)];
#ifndef Stunnel_no_fishinG
 if(flag&1)
   Sblowfish_transform(tunnel->personal_fish,buffer,result_len,0);
 else
   Sblowfish_transform(tunnel->session_fish_out,buffer,result_len,0);
#endif /* ! Stunnel_no_fishinG */

 if(*message!=NULL)
   Smem_freE((char *) *message);
 *message= buffer;

#ifdef Stunnel_no_formaT

 memcpy(buffer,data,len);
 *messagelen= len;
 return(1);

#endif /* Stunnel_no_formaT */

 *messagelen= result_len;
 return(1);
}


int Stunnel_format_token_v0_2(tunnel,message,messagelen,type,data,len,flag)
struct StunneL *tunnel;
char **message;
int *messagelen;
int type;
char *data;
int len;
int flag;
/*
 bit0= use personal_fish rather than session_fish
*/
{
 int ret,i,l=0,buffersize,result_len,transmit_len;
 unsigned long crc;
 char *buffer,*ctx= NULL;
 char sha1_chars[20],typechars[8],seqnochars[16],rndchars[16],lenchars[8];

 result_len= tunnel->headlen+len+8*(type!=0);
 result_len= result_len+7-((result_len+7)%8);
 buffersize= result_len+8;
 buffer= Smem_malloC(buffersize);
 if(buffer==NULL)
   return(-1);
 ret= Shash_sha1(&ctx,NULL,0,sha1_chars,1);
 if(ret<=0)
   goto ex;
 /* some random bytes are added to obscure the header (especially the
    sequence numbers) and for padding :
      0-3:header  4-7:control header pad  8-15:data pad  */
 Sblowfish_make_random(tunnel->random_fish,&(tunnel->random_count),rndchars,16,
                       4);

 if(type)
   transmit_len= -len;
 else
   transmit_len= len;
 for(i=0;i<4;i++)
   lenchars[i]= rndchars[i];
 for(i=0;i<4;i++)
   lenchars[7-i]= (transmit_len>>(8*i))&0xff;
 Shash_sha1(&ctx,lenchars,8,NULL,0);
 tunnel->out_seqno++;
 for(i=0;i<4;i++)
   seqnochars[3-i]= ((tunnel->out_seqno)>>(8*i))&0xff;
 Shash_sha1(&ctx,seqnochars,4,NULL,0);
 if(type) {
   for(i=0;i<4;i++)
     typechars[3-i]= ((type)>>(8*i))&0xff;
   for(i=4;i<8;i++)
     typechars[i]= rndchars[i];
   Shash_sha1(&ctx,typechars,8,NULL,0);
 } 
 if(len>0)
   Shash_sha1(&ctx,data,len,NULL,0);
 Shash_sha1(&ctx,NULL,0,sha1_chars,2); /* get result in sha1_chars */

 l= 0;
 memcpy(buffer+l,lenchars,8);
 l+=8;
 memcpy(buffer+l,seqnochars,4);
 l+= 4;
 memcpy(buffer+l,sha1_chars,20);
 l+= 20;
 if(type){
   memcpy(buffer+l,typechars,8);
   l+= 8;
 }
 if(len>0)
   memcpy(buffer+l,data,len);
 l+= len;
 for(i=l;i<result_len;i++)
   buffer[i]= rndchars[8+(i%8)];
#ifndef Stunnel_no_fishinG
 if(flag&1)
   Sblowfish_transform(tunnel->personal_fish,buffer,result_len,0);
 else
   Sblowfish_transform(tunnel->session_fish_out,buffer,result_len,0);
#endif /* ! Stunnel_no_fishinG */

 if(*message!=NULL)
   Smem_freE((char *) *message);
 *message= buffer;

#ifdef Stunnel_no_formaT

 memcpy(buffer,data,len);
 *messagelen= len;
 {ret= 1; goto ex;}

#endif /* Stunnel_no_formaT */

 *messagelen= result_len;
 ret= 1;
ex:;
 Shash_sha1(&ctx,NULL,0,NULL,1<<15); /* release memory */
 return(ret);
}


int Stunnel_format(tunnel,message,messagelen,type,data,len,flag)
struct StunneL *tunnel;
char **message;
int *messagelen;
int type;
char *data;
int len;
int flag;
/*
 bit0= use personal_fish rather than session_fish
*/
{
 int ret,i,tunneled_len,l,v0_2= 0;
 char cpt,*tunneled=NULL,key[Stunnel_max_keyleN];

 if(*message!=NULL)
   Smem_freE((char *) *message);
 *message= NULL;
 *messagelen= 0;

 if(tunnel->version_major==0 && tunnel->version_minor==2)
   v0_2= 1;

 if(v0_2 && (!(flag&1)) &&
    tunnel->out_seqno-tunnel->out_seqno_start>Stunnel_session_key_lifetimE) {
   /* Renew session key */
   Sblowfish_make_random(tunnel->random_fish,&(tunnel->random_count),
                         key,tunnel->keylen,4);
   /* prepend new session key token */
   ret= Stunnel_format_token_v0_2(tunnel,message,messagelen,
                                  2,key,tunnel->keylen,0);
   /* replace session_fish_out by fish with new key */;
   ret= Stunnel_new_session_fish(tunnel,key,0);
   if(ret<=0)
     goto ex;
   tunnel->out_seqno_start= tunnel->out_seqno;
 }

 if(*message==NULL && len<=Stunnel_max_tokenleN){ /* simple case */
   if(v0_2)
     ret= Stunnel_format_token_v0_2(tunnel,message,messagelen,
                                    type,data,len,flag);
   else
     ret= Stunnel_format_token_v0_0(tunnel,message,messagelen,
                                    type,data,len,flag);
   goto ex;
 }
 /* general case */
 for(i=0;i<len;i+=Stunnel_max_tokenleN){
   l= len-i;
   if(l>Stunnel_max_tokenleN)
     l= Stunnel_max_tokenleN;
   if(v0_2)
     ret= Stunnel_format_token_v0_2(tunnel,&tunneled,&tunneled_len,
                                    type,data+i,l,0);
   else
     ret= Stunnel_format_token_v0_0(tunnel,&tunneled,&tunneled_len,
                                    type,data+i,l,0);
   if(ret<=0)
     return(ret);
   if(Sregex_concat_bin(message,messagelen,tunneled,tunneled_len,0)<=0)
     {ret= -1; goto ex;}
 }
 ret= 1;
ex:;
 if(tunneled!=NULL)
   Smem_freE(tunneled);
 return(ret);
}


int Stunnel_retrieve_token_v0_0(tunnel,in_message,messagelen,endpt,
                           type,data,len,flag)
struct StunneL *tunnel;
char *in_message;
int messagelen;
char **endpt;
int *type;
char **data;
int *len;
int flag;
/*
 bit0= use personal_fish rather than session_fish
 bit1= do not check sequence number

return:
 <0 = error
  0 = authentication failed
  1 = data valid
  2 = no data available yet
*/
{
 int ret,i,l=0,buffersize,datastart,bufferend;
 unsigned char *buffer,*head,*message;
 unsigned long crc,seqno=0;


 message= in_message;  /* to make me feel better if  endpt == &in_message  */
 datastart= 0;
 *len= 0;
 if(*data!=NULL) 
   Smem_freE(*data);
 *data= NULL;
 *endpt= NULL;

 /* init token */
 if(tunnel->in_status==0) {
   tunnel->head_count= 0;
   tunnel->in_status= 1;
   tunnel->data_len= 0;
   tunnel->data_count= 0;
   if(flag&1)
     Sblowfish_reset_chain(tunnel->personal_fish,1);
   else
     Sblowfish_reset_chain(tunnel->session_fish_in,1);
 }

 /* assemble and decode head */
 if(tunnel->in_status==1) {
   if(messagelen+tunnel->head_count<tunnel->headlen) {
     for(i=0;i<messagelen;i++)
       tunnel->in_head[tunnel->head_count+i]= message[i];
     tunnel->head_count+= messagelen;
     return(2);
   }
   for(i=0;i<tunnel->headlen-tunnel->head_count;i++)
     tunnel->in_head[tunnel->head_count+i]= message[i];
   datastart= i;
   tunnel->head_count= tunnel->headlen;
#ifndef Stunnel_no_fishinG
   if(flag&1)
     Sblowfish_transform(tunnel->personal_fish,
                         tunnel->in_head,tunnel->headlen,1);
   else
     Sblowfish_transform(tunnel->session_fish_in,
                         tunnel->in_head,tunnel->headlen,1);
#endif /* ! Stunnel_no_fishinG */
   head= tunnel->in_head;
   l= 0;
   tunnel->data_crc= (head[l+0]<<24)|(head[l+1]<<16)|(head[l+2]<<8)|head[l+3];
   l= 4;
   tunnel->data_len= (head[l+0]<<24)|(head[l+1]<<16)|(head[l+2]<<8)|head[l+3];
   tunnel->in_status= 2;
   if(tunnel->data_len<-Stunnel_max_tokenleN
      || tunnel->data_len>Stunnel_max_tokenleN ) {
     sprintf(tunnel->errmsg,
             "--- tunnel %ld : received oversized token #%lu : %ld > %ld\n",
             tunnel->idno,tunnel->in_seqno+1,
             (tunnel->data_len>0?tunnel->data_len:-tunnel->data_len),
             Stunnel_max_tokenleN);
     return(0);
   }
   if(tunnel->data_len<0) {
     tunnel->in_status= 3;
     tunnel->data_len= 8-tunnel->data_len;
   }
   l= 8;
   seqno= (head[l+0]<<24)|(head[l+1]<<16)|(head[l+2]<<8)|head[l+3];
   if(!(flag&2))
     if(tunnel->in_seqno+1!=seqno) {
       sprintf(tunnel->errmsg,
             "--- tunnel %ld : sequence number mismatch : %lu should be %lu\n",
               tunnel->idno,seqno,tunnel->in_seqno+1);
       return(0);
     }
   tunnel->in_seqno= seqno;

   if(tunnel->data_len==0) {
     tunnel->in_status= 0;
     return(1);
   }
   buffersize= tunnel->data_len+7-((tunnel->data_len+7)%8)+8;
   buffer= Smem_malloC(buffersize);
   if(buffer==NULL)
     return(-1);
   if(tunnel->data_buffer!=NULL)
     Smem_freE((char *) tunnel->data_buffer);
   tunnel->data_buffer= buffer;

 } 
 if(datastart>=messagelen)
   return(2);

 /* read payload data */
 buffer= tunnel->data_buffer;
 if(tunnel->in_status==2 || tunnel->in_status==3) {
   *type= 0;

   for(i=datastart;i<messagelen;i++) {
     if(tunnel->data_count>=tunnel->data_len && (tunnel->data_count%8)==0)
   break;
     buffer[tunnel->data_count++]= message[i];
   }
   if(i<messagelen) 
     *endpt= message+i;
   if(!(tunnel->data_count>=tunnel->data_len && (tunnel->data_count%8)==0))
     return(2);
#ifndef Stunnel_no_fishinG
   if(flag&1)
     Sblowfish_transform(tunnel->personal_fish,buffer,tunnel->data_count,1);
   else
     Sblowfish_transform(tunnel->session_fish_in,buffer,tunnel->data_count,1);
#endif /* ! Stunnel_no_fishinG */

   Sregex_crc32(tunnel->in_head+8,8,&crc,1);
   Sregex_crc32(buffer,tunnel->data_len,&crc,0);
   if(tunnel->data_crc!=crc) {
     sprintf(tunnel->errmsg,
             "--- tunnel %ld : crc mismatch in token #%lu : %-8.8X != %-8.8X\n",
             tunnel->idno,tunnel->in_seqno, crc,tunnel->data_crc);
     return(0);
   }

   /* read control type */
   if(tunnel->in_status==3) {
     l= 0;
     *type= (buffer[l+0]<<24)|(buffer[l+1]<<16)|(buffer[l+2]<<8)|buffer[l+3];
     for(i=8;i<tunnel->data_len;i++)
       buffer[i-8]= buffer[i];
     tunnel->data_len-= 8;
     if(tunnel->data_len<0)
       tunnel->data_len= 0;
   }

   /* transfer data buffer from tunnel to reply parameter */
   *data= (char *) tunnel->data_buffer;
   *len= tunnel->data_len;
   tunnel->data_buffer= NULL;
   tunnel->data_len= tunnel->data_count= 0;
   tunnel->in_status= 0;
   if(!(flag&2))
     tunnel->authenticated= 1;
   return(1);
 }

 return(2-(tunnel->in_status==0));
}


int Stunnel_sha1_check(tunnel,flag)
struct StunneL *tunnel;
int flag;
{
 char sha1_local[20];
 int i;

 if(tunnel->sha1_ctx==NULL)
   return(-1);
 Shash_sha1(&(tunnel->sha1_ctx),NULL,0,sha1_local,2|(1<<14));
 for(i=0;i<sizeof(tunnel->data_sha1);i++)
   if(sha1_local[i]!=tunnel->data_sha1[i])
 break;
 if(i<sizeof(tunnel->data_sha1)) {
   sprintf(tunnel->errmsg,
           "--- tunnel %ld : SHA1 mismatch in token #%lu\n",
           tunnel->idno,tunnel->in_seqno);
   return(0);
 }
 return(1);
}


int Stunnel_retrieve_token_v0_2(tunnel,in_message,messagelen,endpt,
                           type,data,len,flag)
struct StunneL *tunnel;
char *in_message;
int messagelen;
char **endpt;
int *type;
char **data;
int *len;
int flag;
/*
 bit0= use personal_fish rather than session_fish
 bit1= do not check sequence number

return:
 <0 = error
  0 = authentication failed
  1 = data valid
  2 = no data available yet
*/
{
 int ret,i,l=0,buffersize,datastart,bufferend;
 unsigned char *buffer,*head,*message;
 unsigned long crc,seqno=0;

 message= in_message;  /* to make me feel better if  endpt == &in_message  */
 datastart= 0;
 *len= 0;
 if(*data!=NULL) 
   Smem_freE(*data);
 *data= NULL;
 *endpt= NULL;

 /* init token */
 if(tunnel->in_status==0) {
   tunnel->head_count= 0;
   tunnel->in_status= 1;
   tunnel->data_len= 0;
   tunnel->data_count= 0;
   ret= Shash_sha1(&(tunnel->sha1_ctx),NULL,0,NULL,1);
   if(ret<=0)
     return(ret);
 }

 /* assemble and decode head */
 if(tunnel->in_status==1) {
   if(messagelen+tunnel->head_count<tunnel->headlen) {
     for(i=0;i<messagelen;i++)
       tunnel->in_head[tunnel->head_count+i]= message[i];
     tunnel->head_count+= messagelen;
     return(2);
   }
   for(i=0;i<tunnel->headlen-tunnel->head_count;i++)
     tunnel->in_head[tunnel->head_count+i]= message[i];
   datastart= i;
   tunnel->head_count= tunnel->headlen;
#ifndef Stunnel_no_fishinG
   if(flag&1)
     Sblowfish_transform(tunnel->personal_fish,
                         tunnel->in_head,tunnel->headlen,1);
   else
     Sblowfish_transform(tunnel->session_fish_in,
                         tunnel->in_head,tunnel->headlen,1);
#endif /* ! Stunnel_no_fishinG */
   head= tunnel->in_head;
   tunnel->in_status= 2;
   l= 4;
   tunnel->data_len= (head[l+0]<<24)|(head[l+1]<<16)|(head[l+2]<<8)|head[l+3];
   l+= 4;
   seqno= (head[l+0]<<24)|(head[l+1]<<16)|(head[l+2]<<8)|head[l+3];
   l+= 4;
   memcpy(tunnel->data_sha1,head+l,20);

   if(tunnel->data_len<-Stunnel_max_tokenleN
      || tunnel->data_len>Stunnel_max_tokenleN ) {
     sprintf(tunnel->errmsg,
             "--- tunnel %ld : received oversized token #%lu : %ld > %ld\n",
             tunnel->idno,tunnel->in_seqno+1,
             (tunnel->data_len>0?tunnel->data_len:-tunnel->data_len),
             Stunnel_max_tokenleN);
     return(0);
   }
   if(!(flag&2))
     if(tunnel->in_seqno+1!=seqno) {
       sprintf(tunnel->errmsg,
       "--- tunnel %ld : sequence number mismatch : %lu should be %lu\n",
               tunnel->idno,seqno,tunnel->in_seqno+1);
       return(0);
     }
   if(tunnel->data_len<0) {
     tunnel->in_status= 3;
     tunnel->data_len= 8-tunnel->data_len;
   }
   tunnel->in_seqno= seqno;

   Shash_sha1(&(tunnel->sha1_ctx),head,8,NULL,0);
   Shash_sha1(&(tunnel->sha1_ctx),head+8,4,NULL,0);

   if(tunnel->data_len==0) {
     tunnel->in_status= 0;
     if(Stunnel_sha1_check(tunnel,0)<=0)
       return(0); 
     return(1);
   }
   buffersize= tunnel->data_len+7-((tunnel->data_len+7)%8)+8;
   buffer= Smem_malloC(buffersize);
   if(buffer==NULL)
     return(-1);
   if(tunnel->data_buffer!=NULL)
     Smem_freE((char *) tunnel->data_buffer);
   tunnel->data_buffer= buffer;

 } 
 if(datastart>=messagelen)
   return(2);

 /* read payload data */
 buffer= tunnel->data_buffer;
 if(tunnel->in_status==2 || tunnel->in_status==3) {
   *type= 0;

   for(i=datastart;i<messagelen;i++) {
     if(tunnel->data_count>=tunnel->data_len && (tunnel->data_count%8)==0)
   break;
     buffer[tunnel->data_count++]= message[i];
   }
   if(i<messagelen) 
     *endpt= message+i;
   if(!(tunnel->data_count>=tunnel->data_len && (tunnel->data_count%8)==0))
     return(2); /* data not complete yet */

   /* data are completely read now */

#ifndef Stunnel_no_fishinG
   if(flag&1)
     Sblowfish_transform(tunnel->personal_fish,buffer,tunnel->data_count,1);
   else
     Sblowfish_transform(tunnel->session_fish_in,buffer,tunnel->data_count,1);
#endif /* ! Stunnel_no_fishinG */

   /* read control type */
   if(tunnel->in_status==3) {
     l= 0;
     *type= (buffer[l+0]<<24)|(buffer[l+1]<<16)|(buffer[l+2]<<8)|buffer[l+3];
     Shash_sha1(&(tunnel->sha1_ctx),buffer,8,NULL,0);
     for(i=8;i<tunnel->data_len;i++)
       buffer[i-8]= buffer[i];
     tunnel->data_len-= 8;
     if(tunnel->data_len<0)
       tunnel->data_len= 0;
   }
   Shash_sha1(&(tunnel->sha1_ctx),buffer,tunnel->data_len,NULL,0);
   if(Stunnel_sha1_check(tunnel,0)<=0)
     return(0); 

   /* transfer data buffer from tunnel to reply parameter */
   *data= (char *) tunnel->data_buffer;
   *len= tunnel->data_len;
   tunnel->data_buffer= NULL;
   tunnel->data_len= tunnel->data_count= 0;
   tunnel->in_status= 0;
   if(!(flag&2))
     tunnel->authenticated= 1;
   return(1);
 }

 return(2-(tunnel->in_status==0));
}


int Stunnel_retrieve_token(tunnel,in_message,messagelen,endpt,
                           type,data,len,flag)
struct StunneL *tunnel;
char *in_message;
int messagelen;
char **endpt;
int *type;
char **data;
int *len;
int flag;
/*
 bit0= use personal_fish rather than session_fish
 bit1= do not check sequence number
 bit15= abort of incomplete token : reset token buffer and exit

return:
 <0 = error
  0 = authentication failed
  1 = data valid
  2 = no data available yet
*/
{
 int ret;

 if(flag&(1<<15)) {
   if(tunnel->data_buffer!=NULL)
     Smem_freE(tunnel->data_buffer);
   tunnel->data_buffer= NULL;
   tunnel->data_len= tunnel->data_count= 0;
   tunnel->in_status= 0;
   return(1);
 }
 if(tunnel->version_major==0 && tunnel->version_minor==2)
   ret= Stunnel_retrieve_token_v0_2(tunnel,in_message,messagelen,
                                    endpt,type,data,len,flag);
 else
   ret= Stunnel_retrieve_token_v0_0(tunnel,in_message,messagelen,
                                    endpt,type,data,len,flag);
 return(ret);
}


int Stunnel_retrieve(tunnel,in_message,messagelen,data,len,flag)
struct StunneL *tunnel;
char *in_message;
int messagelen;
char **data;
int *len;
int flag;
/*
 bit0= use personal_fish rather than session_fish
 bit1= do not check sequence number

return:
 <0 = error
  0 = authentication failed
  1 = data valid
  2 = no data available yet
*/
{
 int ret,untunneled_len=0,type;
 char *untunneled= NULL,*rpt;

 rpt= in_message;
 if(*data!=NULL)
   Smem_freE(*data);
 *data= NULL;
 *len= 0;
 while(rpt!=NULL && messagelen-(rpt-in_message)>0) {
   ret= Stunnel_retrieve_token(tunnel,rpt,messagelen-(rpt-in_message),&rpt,
                               &type,&untunneled,&untunneled_len,flag&3);
   if(ret<=0)
     goto ex;

   if(ret==1 && type!=0) {
     /* handle any control messages after handshake is completed
        (handshake reply and first session key are handled elsewhere) */
     if(type==2) {
       /*  new session key for inbound stream */
       if(untunneled_len!=tunnel->keylen) {
         /* >>> key not of suitable length */;
         ret= -1; goto ex;
       }
       ret= Stunnel_new_session_fish(tunnel,untunneled,1);
       if(ret<=0)
         {ret= -1; goto ex;}
     }
     if(type==3) {
       /* authentication opportunity message : ignore */;
     }
 continue;
   }
   if(untunneled_len>0)
     if(Sregex_concat_bin(data,len,untunneled,untunneled_len,0)<=0)
       {ret= -1; goto ex;}
 }
 if(*len>0)
   ret= 1;
 else
   ret= 2;
ex:;
 if(untunneled!=NULL)
   Smem_freE(untunneled);
 return(ret);
}


int Stunnel_handshake_init(tunnel,message,message_len,backchannel,flag)
struct StunneL *tunnel;
char **message;
int *message_len;
char *backchannel;
int flag;
{
 int ret,i,v0_2= 0,line_len;
 char line[2*SfileadrL],dibmmfohf[Stunnel_max_challeN+1];
 char encrypted_line[2*SfileadrL],iv[Stunnel_iv_seedleN];
 char *ctx= NULL,sha_result[20];

 if(tunnel->version_major==0 && tunnel->version_minor==2)
   v0_2= 1;
 Stunnel_make_challenge(tunnel,0);
 sprintf(line,"-handshake:%ld.%ld kl=%ld cl=%ld ml=%ld:%s:",
              tunnel->version_major,tunnel->version_minor,tunnel->keylen,
              tunnel->challen,Stunnel_max_tokenleN,tunnel->user);
 for(i=0;i<tunnel->challen;i++)
   dibmmfohf[i]= tunnel->challenge[i];

#ifndef Stunnel_no_fishinG
 if(v0_2) {
   for(i=0;i<8;i++)
     sha_result[i]= 0;
   Sblowfish_set_iv(tunnel->personal_fish,sha_result,0);
   Sblowfish_reset_chain(tunnel->personal_fish,2);
   for(i=0;i<Stunnel_iv_seedleN;i++)
     iv[i]= tunnel->iv_seed[i];
   Sblowfish_transform(tunnel->personal_fish,iv,Stunnel_iv_seedleN,0);
   Shash_sha1(&ctx,iv,Stunnel_iv_seedleN,sha_result,1|2|(1<<14));
   for(i=0;i<8;i++)
     tunnel->iv[i]= sha_result[i];
   Sblowfish_set_iv(tunnel->personal_fish,tunnel->iv,0);
 }
 Sblowfish_reset_chain(tunnel->personal_fish,2); /* v0_2 approved */
 Sblowfish_transform(tunnel->personal_fish,dibmmfohf,tunnel->challen,0);
#endif /* ! Stunnel_no_fishinG */

 for(i=0;i<tunnel->challen;i++)
   sprintf(line+strlen(line),"%-2.2X",dibmmfohf[i]&0xff);
 sprintf(line+strlen(line),":%s",backchannel);
 if(v0_2) {
   sprintf(line+strlen(line),":");
   for(i=0;i<Stunnel_iv_seedleN;i++)
     sprintf(line+strlen(line),"%-2.2X",tunnel->iv_seed[i]&0xff);
   line_len= strlen(line);
   for(i=line_len;i>=0;i--)
     encrypted_line[i]= line[i];
   if(line_len%8) {
     for(i=line_len;i<line_len+8-(line_len%8);i++)
       encrypted_line[i]= 0;
     line_len= line_len+8-(line_len%8);
   }
#ifndef Stunnel_no_fishinG
   Sblowfish_reset_chain(tunnel->personal_fish,2);
   Sblowfish_transform(tunnel->personal_fish,encrypted_line,line_len,0);
#endif /* ! Stunnel_no_fishinG */
   Shash_sha1(&ctx,encrypted_line,line_len,sha_result,1|2|(1<<14));
   sprintf(line+strlen(line),":");
   for(i=0;i<20;i++)
     sprintf(line+strlen(line),"%-2.2X",sha_result[i]&0xff);
 }
 sprintf(line+strlen(line),"\n");
 if(Sregex_string(message,line,0)<=0)
   return(-1);
 *message_len= strlen(*message);
 return(1);
}


int Stunnel_handshake_reply(tunnel,errmsg,message,message_len,backchannel,
                            idno,keydir,flag)
struct StunneL **tunnel;
char errmsg[SfileadrL*2];            
char **message;                     /* must point to NULL or freeable memory */
int *message_len;
char **backchannel;
int idno;
char *keydir;
int flag;
/*
  bit4-11= protocol version :
        0= allow all known protocols
        1= 0.0 kl=16 cl=8 ml=1024
        2= 0.2 kl=32 cl=16 ml=1024
*/
{
 int ret,i,c,reply_len= 0,skey_reply_len= 0,protocol_version,challen,line_len;
 int allowed_protocol;
 char line[SfileadrL],challenge[Stunnel_max_challeN+1];
 char iv_seed[Stunnel_iv_seedleN+1],sha1_of_encrypted[20];
 char key[Stunnel_max_keyleN+1],*ctx= NULL,sha_result[20];
 char *cpt,*gotstr= NULL,*user= NULL,rev[81],*reply= NULL,*skey_reply= NULL;
 char *challenge_reply;
 static char general_error[]= 
       {"--- malformed handshake attempt (eventually shown escaped here) :\n"};
 struct StunneL *t=NULL;

 allowed_protocol= (flag>>4)&0xff;
 cpt= *message;
 errmsg[0]= 0;
 if(strlen(*message)>SfileadrL-8)
   {ret= 0;goto ex;}

 /* verify protocol */
 ret= Sregex_cut_neck(cpt,':',&gotstr,&cpt,0);
 if(ret!=1)
   {ret= 0; goto ex;}
 if(strcmp(gotstr,"-handshake")!=0)
   {ret= 0; goto ex;}
 ret= Sregex_cut_neck(cpt,':',&gotstr,&cpt,0);
 if(ret!=1)
   {ret= 0; goto ex;}
 /* for now, the protocol range is quite small :) */
 if(allowed_protocol==0 || allowed_protocol==1) {
   sprintf(rev,"0.0 kl=%ld cl=%ld ml=%ld",
               Stunnel_v0_0_keyleN,Stunnel_v0_0_challeN,Stunnel_max_tokenleN);
    if(strcmp(gotstr,rev)==0) {
      protocol_version= 1; 
      challen= 8;
      goto allowed_protocol_found;
    }
 }
 if(allowed_protocol==0 || allowed_protocol==2) {
   sprintf(rev,"0.2 kl=%ld cl=%ld ml=%ld",
               Stunnel_v0_2_keyleN,Stunnel_v0_2_challeN,Stunnel_max_tokenleN);
   if(strcmp(gotstr,rev)==0) {
     protocol_version= 2; 
     challen= 32;
     goto allowed_protocol_found;
   }
 }
 /* not an allowed protocol */
 {ret= 0; goto ex;}

allowed_protocol_found:;
 /* read user */
 ret= Sregex_cut_neck(cpt,':',&user,&cpt,0);
 if(ret!=1)
   {ret= 0; goto ex;}

 /* read in_challenge */
 if(strlen(cpt)<2*challen)
   {ret= 0; goto ex;}
 for(i=0;i<challen;i++)
   challenge[i]= 0;
 ret= Sregex_hex_string(cpt,challenge,2*challen,0);
 if(ret<=0)
   goto ex;
 cpt+= 2*challen;
 if(*cpt!=':')
   {ret= 0; goto ex;}
 cpt++;

 /* read back channel (and eventually more) */
 if(protocol_version==1) {
   ret= Sregex_string(backchannel,cpt,0);
   if(ret!=1)
     {ret= 0; goto ex;}
 } else if(protocol_version==2) {
   ret= Sregex_cut_neck(cpt,':',backchannel,&cpt,0);
   if(ret!=1)
     {ret= 0; goto ex;}
   if(strlen(cpt)<2*Stunnel_iv_seedleN+1+40)
     {ret= 0; goto ex;}
   /* read iv_seed */
   ret= Sregex_hex_string(cpt,iv_seed,2*Stunnel_iv_seedleN,0);
   if(ret<=0)
     goto ex;
   cpt+= 2*Stunnel_iv_seedleN;
   if(*cpt!=':')
     {ret= 0; goto ex;}
   cpt++;
   line_len= (cpt- *message)-1;
   /* read sha1_of_encrypted */
   ret= Sregex_hex_string(cpt,sha1_of_encrypted,40,0);
   if(ret<=0)
     goto ex;
   cpt+= 40;
 } 

 ret= Stunnel_new(tunnel,user,idno,keydir,NULL,protocol_version<<4);
 if(ret<=0) {
   if(ret==-2){
     sprintf(errmsg,"--- tunnel %ld : unknown user : %s",idno,user);
     ret= 0;
   }
   goto ex;
 }
 t= *tunnel;
 if(protocol_version==2) {
   /* install new initial vector */
   for(i=0;i<8;i++)
     t->iv[i]= 0;
   Sblowfish_set_iv(t->personal_fish,t->iv,0);
   Sblowfish_reset_chain(t->personal_fish,2);
#ifndef Stunnel_no_fishinG
   Sblowfish_transform(t->personal_fish,iv_seed,Stunnel_iv_seedleN,0);
#endif /* ! Stunnel_no_fishinG */
   Shash_sha1(&ctx,iv_seed,Stunnel_iv_seedleN,sha_result,1|2|(1<<14));
   for(i=0;i<8;i++)
     t->iv[i]= sha_result[i];
   Sblowfish_set_iv(t->personal_fish,t->iv,0);

   /* check wether sha1_of_encrypted matches line */
   for(i=0;i<line_len;i++)
     line[i]= (*message)[i];
   if(line_len%8) {
     for(i=line_len;i<line_len+8-(line_len%8);i++)
       line[i]= 0;
     line_len= line_len+8-(line_len%8);
   }
   Sblowfish_reset_chain(t->personal_fish,2);
#ifndef Stunnel_no_fishinG
   Sblowfish_transform(t->personal_fish,line,line_len,0);
   Sblowfish_reset_chain(t->personal_fish,2);
#endif /* ! Stunnel_no_fishinG */
   Shash_sha1(&ctx,line,line_len,sha_result,1|2|(1<<14));
   for(i=0;i<20;i++) {
     if(sha_result[i]!=sha1_of_encrypted[i]) {
       strcpy(errmsg,"SHA1 mismatch of encrypted handshake string");
       ret= 0; goto ex;
     }
   }

 }

 /* decipher challenge and make reply */
 Sblowfish_reset_chain(t->personal_fish,1);
#ifndef Stunnel_no_fishinG
 Sblowfish_transform(t->personal_fish,challenge,t->challen,1);
#endif /* ! Stunnel_no_fishinG */
 for(i=0;i<t->challen;i++)
   t->in_challenge[i]= challenge[i];
 Stunnel_make_reply(t,t->in_challenge,t->reply,0);

 Sblowfish_reset_chain(t->personal_fish,3);
 ret= Stunnel_format(t,&reply,&reply_len,1,t->reply,t->challen,1);
 if(ret<=0)
   goto ex;

 /* make session key and format it encrypted by personal_fish */
 Sblowfish_make_random(t->random_fish,&(t->random_count),key,t->keylen,4);
 ret= Stunnel_new_session_fish(t,key,0);
 if(ret<=0)
   goto ex;
 ret= Stunnel_new_session_fish(t,key,1);
 if(ret<=0)
   goto ex;
 ret= Stunnel_format(t,&skey_reply,&skey_reply_len,2,key,t->keylen,1);
 if(ret<=0)
   goto ex;
 if(protocol_version==2) {
   challenge_reply= reply;
   reply= TSOB_FELD(char,8+reply_len+skey_reply_len+1);
   if(reply==NULL) {
     Smem_freE(challenge_reply);
     ret= -1; goto ex;
   }
   strcpy(reply,"OK v0.2\n");
   memcpy(reply+8,challenge_reply,reply_len);
   Smem_freE(challenge_reply);
   memcpy(reply+8+reply_len,skey_reply,skey_reply_len);
   reply_len= 8+reply_len+skey_reply_len;
 } else {
   ret= Sregex_concat_bin(&reply,&reply_len,skey_reply,skey_reply_len,0);
   if(ret<=0)
     goto ex;
 }

 ret= 1;
ex:;
 if(ret<=0) {
   if(errmsg[0]==0) 
     if(*message_len<SfileadrL/2-1-strlen(general_error)) {
       strcpy(errmsg,general_error);
       cpt= errmsg+strlen(errmsg);
       sprintf(cpt,"%s",*message);
       Sfile_esc_filename(cpt,SfileadrL*2-1-(cpt-errmsg),1|16|32|128);
     }
   if(t!=NULL)
     Stunnel_destroy(tunnel,0);
   Sregex_string(&reply,NULL,0);
 } else {
   if(*message!=NULL)
     Smem_freE((char *) *message);
   *message= reply;
   *message_len= reply_len;
 }
 Sregex_string(&gotstr,NULL,0);
 Sregex_string(&user,NULL,0);
 Sregex_string(&skey_reply,NULL,0);
 return(ret);
}


int Stunnel_handshake_accept(tunnel,message,message_len,endpt,flag)
struct StunneL *tunnel;
char *message;
int message_len;
char **endpt;  /* points to unconsumed message data for Stunnel_retrieve () */
int flag;
/*
return:
 <0 = error
  0 = authentication failed
  1 = handshake accepted
  2 = no decision yet, append more data and submit again
*/
{
 int ret,i,type,untunneled_len= 0,v0_2= 0,l,size;
 char *cpt,*untunneled= NULL,expected_reply[Stunnel_max_challeN+1];
 char new_random_key[Stunnel_max_keyleN];
 struct SblowfisH *temp_fish= NULL;

 if(tunnel->version_major==0 && tunnel->version_minor==2)
   v0_2= 1;
 cpt= message;
 if(v0_2) {
   if(message_len<8) {
     if(strncmp(cpt,"OK v0.2\n",message_len)==0)
       {ret= 2; goto ex;}
     else
       goto failed;
   }
   if(strncmp(cpt,"OK v0.2\n",8)!=0) {
failed:;
     strcpy(tunnel->errmsg,"-handshake v0.2 failed. Server replies (eventually shown escaped here) :\n");
     l= strlen(tunnel->errmsg);
     size= sizeof(tunnel->errmsg)/4-l-1;
     cpt= tunnel->errmsg+l;
     if(message_len>size) {
       memcpy(cpt,message,size);
       cpt[size]= 0;
     } else {
       memcpy(cpt,message,message_len);
       cpt[message_len]= 0;
     }
     Sfile_esc_filename(cpt,sizeof(tunnel->errmsg)-1-l,1|16|32|128);
     ret= 0;
     goto ex;
   }
   cpt+= 8; 
 }
 /* decode message using personal_fish , accept any sequence number */
 Sblowfish_reset_chain(tunnel->personal_fish,3);
 ret= Stunnel_retrieve_token(tunnel,cpt,message_len-(cpt-message),&cpt,
                            &type,&untunneled,&untunneled_len,1|2);
 if(ret<=0)
   goto ex;
 if(ret==1 && cpt==NULL) /* the message ends exactly after first token */
   ret= 2;
 if(ret==2) {
   Stunnel_retrieve_token(tunnel,cpt,0,&cpt,&type,&untunneled,&untunneled_len,
                          1<<15); /* abort token */
   goto ex;
 }

#ifdef Stunnel_no_formaT
 type= 1;
 untunneled_len= tunnel->challen;
 cpt= message+tunnel->challen;
#endif Stunnel_no_formaT

 /* compare with expected reply */
 if(untunneled_len!=tunnel->challen || type!=1)
   {ret=0; goto ex;}
 Stunnel_make_reply(tunnel,tunnel->challenge,expected_reply,0);
 for(i=0;i<tunnel->challen;i++)
   if(expected_reply[i]!=untunneled[i])
     break;
 if(i<tunnel->challen) {
   sprintf(tunnel->errmsg,
           "--- tunnel %ld : handshake reply does not match challenge.\n",
           tunnel->idno);
   ret=0;
   goto ex;
 }

 /* decode message and obtain session key */
 ret= Stunnel_retrieve_token(tunnel,cpt,message_len-(cpt-message),&cpt,
                             &type,&untunneled,&untunneled_len,1);
 if(ret<=0)
   goto ex;
 if(ret==2) {
   Stunnel_retrieve_token(tunnel,cpt,0,&cpt,&type,&untunneled,&untunneled_len,
                          1<<15); /* abort token */
   goto ex;
 }

#ifdef Stunnel_no_formaT
 type= 2;
 untunneled_len= tunnel->keylen;
 cpt= message+tunnel->challen+tunnel->keylen;
#endif /* Stunnel_no_formaT */

 if(untunneled_len!=tunnel->keylen || type!=2)
   {ret=0; goto ex;}

 /* create session_fish */
 ret= Stunnel_new_session_fish(tunnel,untunneled,0);
 if(ret<=0)
   goto ex;
 ret= Stunnel_new_session_fish(tunnel,untunneled,1);
 if(ret<=0)
   goto ex;

 /* re-create random fish with (hopefully) better random key */
 ret= Sblowfish_new(&temp_fish,untunneled,tunnel->keylen,0);
 if(ret<=0)
   goto ex;
 Sblowfish_make_random(temp_fish,&(tunnel->random_count),
                       new_random_key,tunnel->keylen,4);
 Sblowfish_destroy(&(tunnel->random_fish),0);
 ret= Sblowfish_new(&(tunnel->random_fish),new_random_key,tunnel->keylen,0);
 if(ret<=0)
   goto ex;


 ret= 1;
ex:;
 *endpt= cpt;
 if(untunneled!=NULL) {
   for(cpt= untunneled; cpt-untunneled<untunneled_len; cpt++)
     *cpt= 0;
   Smem_freE(untunneled);
 }
 if(temp_fish!=NULL)
   Sblowfish_destroy(&temp_fish,0);
 return(ret);
}


int Stunnel_reauthenticate(tunnel,message,messagelen,flag)
struct StunneL * tunnel;
char **message;
int *messagelen;
int flag;
{
 int ret;
 char msg[16];
 
 Sblowfish_make_random(tunnel->random_fish,&(tunnel->random_count),msg,16,4);
 ret= Stunnel_format(tunnel,message,messagelen,3,msg,16,0);
 return(ret);
}


int Stunnel_is_authenticated(tunnel,flag)
struct StunneL * tunnel;
int flag;
{
 if(tunnel==NULL)
   return(0);
 return(tunnel->authenticated);
}


int Stunnel_has_bytes(tunnel,flag)
struct StunneL * tunnel;
int flag;
{
 if(tunnel==NULL)
   return(0);
 if(tunnel->in_status==0)
   return(0);
 if(tunnel->in_status==1)
   return(tunnel->head_count);
 if(tunnel->in_status==2 || tunnel->in_status==3)
   return(tunnel->head_count+tunnel->data_count);
 return(0);
}


int Stunnel_make_userkey(tunnel,flag)
struct StunneL * tunnel;
int flag;
/*
 bit0= do not write SHA-1 part (i.e. old 128 bit format)
*/
{
 char *adr= NULL,buffer[64],*ctx= NULL;
 unsigned char key[Stunnel_max_keyleN+20],userfile[3*Stunnel_userleN+4];
 int ret,i,j,key_len;

 unsigned long crc[Stunnel_max_keyleN/4];

 key_len= Stunnel_v0_0_keyleN+16*!(flag&1);
 strcpy(userfile,tunnel->user);
 if(strcmp(userfile,"internal_owner")==0)
   goto illegal_username;
 Sfile_esc_filename(userfile,sizeof(userfile)-1,1|4);
 if(strcmp(userfile,tunnel->user)!=0 || strchr(userfile,':')!=NULL) {
illegal_username:;
   sprintf(tunnel->errmsg,"illegal username '%s'",userfile);
   return(0);
 }
 if(Sregex_string(&adr,tunnel->keydir,0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&adr,"/",1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&adr,userfile,1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&adr,".tnl",1)<=0)
   {ret= -1; goto ex;}
 ret= Sfile_file_to_bin(adr,buffer,64,2);
 if(ret<=0) {
   sprintf(tunnel->errmsg,"cannot read %ld bytes from file  %s",64,adr);
   goto ex;
 }  

 /* compute key */
 for(i=0;i<Stunnel_v0_0_keyleN/4;i++)
   crc[i]= 0;

 /* first (128) bits from old key generator */
 for(i=0;i<64;i++)
   Sregex_crc32(buffer+i,1,&(crc[i%(Stunnel_v0_0_keyleN/4)]),1);
 for(i=0;i<Stunnel_v0_0_keyleN/4;i++)
   for(j=0;j<4;j++)
    key[4*i+j]= (crc[i]>>(8*j))*0xff;

 /* then add 128 SHA-1 bits (throw away last 32 bits) */
 if(!(flag&1))
   Shash_sha1(&ctx,buffer,64,key+Stunnel_v0_0_keyleN,1|2|(1<<14));

#ifndef Stunnel_no_fishinG
 ret= Sblowfish_transform(tunnel->internal_fish,key,key_len,1|2);
 if(ret<=0)
   return(ret);
#endif /* ! Stunnel_no_fishinG */
 ret= Sfile_bin_to_file(adr,key,key_len,1);
 if(ret<=0){
cannot_write:;
   sprintf(tunnel->errmsg,"cannot write key back to file  %s",adr);
   goto ex;
 }
 chmod(adr,S_IRUSR|S_IWUSR);
 ret= Stunnel_make_personal(tunnel,0);
 if(ret<=0)
   return(ret);
 ret= 1;
ex:;
 Sregex_string(&adr,NULL,0);
 return(ret);
}
