

#ifndef Stunnel_includeD
#define Stunnel_includeD


/* You may personalize your stunnel binary here by changing this string 
   recompiling and keeping the resulting binaries secret */
#define Stunnel_internal_keY "stunnel Default Configuration"


/* debugging only , not intended for applications */
/* #define Stunnel_no_fishinG Yes */
/* #define Stunnel_no_formaT Yes */


/* not more than 56 , should be multiples of 8 */
#define Stunnel_max_keyleN 32 
#define Stunnel_v0_0_keyleN 16
#define Stunnel_v0_2_keyleN 32

/* must be multiples of 8 */
#define Stunnel_max_challeN 32
#define Stunnel_v0_0_challeN 8
#define Stunnel_v0_2_challeN 32

#define Stunnel_iv_seedleN 32

#define Stunnel_userleN 256

#define Stunnel_max_headleN 32
#define Stunnel_v0_0_headleN 16
#define Stunnel_v0_2_headleN 32

/* Maximum payload length of a token (added: Stunnel_*_headleN) */
/* larger tokens have to be divided automatically */
/* has to be larger than challen+Stunnel_userleN+keylen */
/* but not more than 65536 */
#define Stunnel_max_tokenleN 1024

/* v0.2 :  "OK v0.2\n",Challenge-reply-token,Sessionkey-token, */
#define Stunnel_max_handshake_reply_leN \
           (8+2*(Stunnel_max_headleN+8)+Stunnel_max_challeN+Stunnel_max_keyleN)

/* v0.2 : defines how many tokens a session key shall last before renewal */
#define Stunnel_session_key_lifetimE 100000 



struct StunneL {

 /* fundamental parameters */
 int version_major;
 int version_minor;
 int keylen;
 int challen;
 int headlen;

 /* wether there has been received any valid token */
 int authenticated;

 /* name of user identity */
 char user[Stunnel_userleN+1];
 char *keydir;

 /* opaque handle for owner of this object (filedescriptor of connection) */
 int idno;

 /* used to read and write keylists */
 struct SblowfisH *internal_fish;

 /* used to generate "random" numbers  */
 struct SblowfisH *random_fish;
 double random_count;

 /* used to authenticate client and server while session handshake */
 struct SblowfisH *personal_fish;

 /* choosen by server. used to encrypt data transmission */
 /* (after a session key change they might use differing keys) */ 
 struct SblowfisH *session_fish_in;
 struct SblowfisH *session_fish_out;


 /* last challenge sent */
 char challenge[Stunnel_max_challeN+1];
 char iv_seed[Stunnel_iv_seedleN];
 char iv[8];

 /* reply to be sent */
 char reply[Stunnel_max_challeN+1];

 /* last reply received */
 char in_reply[Stunnel_max_challeN+1];
 
 /* last challenge received */
 char in_challenge[Stunnel_max_challeN+1];


 /* status of receiving buffers :
      0=closed , 1=waiting for head , 2=waiting for payload data, 
      3=waiting for control type */
 int in_status;

 /* to assemble an incoming token header */
 char in_head[Stunnel_max_headleN];
 int head_count;

 /* to assemble a complete incoming token */
 char *data_buffer;
 int data_len;
 int data_count;
 unsigned long data_crc;
 char data_sha1[20]; /* received alleged SHA1 */
 char *sha1_ctx;     /* context for computing SHA1 of received token */


 /* sequence numbers to make tokens individual */
 unsigned long in_seqno;
 unsigned long out_seqno;
 unsigned long out_seqno_start;

 char errmsg[SfileadrL*2];

 struct StunneL *next,*prev;

};


/*

A StunneL object is one half of a quite secure virtual communications
connection.
Security in this case means, that a third party may not easily read or alter
the data transmitted. A faulty data transmission will appear as such an
outside offender to the owner of StunneL.

Currently security is achieved by the blowfish encryption algorithm using
256-bit keys and a 160-bit SHA-1 seal. Replay of recorded transmissions is
prevented by a 32-bit sequence number which has to rise in single steps.
(The outdated protocol 0.0 uses 128 bit-keys and 32-bit CRC seals.)


Sending:

Users of a StunneL can put cleartext into Stunnel_format() and get encrypted
data in reply, which can be sent over insecure connections.
The result of such a conversion consists of one or more tokens assembled in
a single character array. To prevent giant buffers there is a rather small
maximum size for a single token (Stunnel_max_tokenleN).

A token consists of its head, an optional control type and a data part
of variable length.

In protocol 0.0 a token's head consists of :

  crc[4],len[4],seqno[4],random[4]
  seen as integers represented with high bytes first, len is signed
  each token is padded up with random numbers to the next multiple of eight
  if len%8!=0

In protocol 0.2 a token's head consists of :

  random[4],len[4],seqno[4],sha1[20]
  The 32 bit checksum has been replaced by 160 bits of SHA-1. 

In both protocols :

  len<0 indicates a control message : type[4],pad[4],data[-len]
    type= 0 : payload data (will hardly occur as control message)
          1 : reply to challenge of Stunnel_handshake_reply()
          2 : new session_key for all following token
          3 : authentication opportunity, senseless data just to check crc 
    pad=  random characters (for now)
  len>= indicates a data message ; data[len]


Receiving:

The receiver of StunneL tokens may hand them over to a corresponding
StunneL object by Stunnel_retrieve() and get the original cleartext.
The decoder of StunneL will buffer data as long as a token is incomplete. The
content of any token completed by the input to Stunnel_retrieve() will be
concatenated and form the reply.
Be aware that Stunnel_retrieve() may have no data at all available and
return 2. If Stunnel_has_bytes()>0 , then more data is expected. Else
there has been a control message which was consumed by the StunneL object. 


Establishing a StunneL Pair:
 
Despite it is possible to create two matching StunneL objects without any
communication between their owners, usually the parameters are negotiated
at the beginning of communication. The key is depicted by a username which
leads to a file containing the key text in a locally encrypted format.
Two participants are needed for this negociation: the initiating peer and the
receiving peer. Both are using StunneL and at start are able to communicate
at least one-way from initiating peer to receiving peer. 


Initiating Peer:

Stunnel_handshake_init() produces a string of printable characters which can
be sent to the receiving peer. This string contains the username, a challenge
encrypted by the user's key and the name of a backchannel to be used for reply.
The backchannel may be necessary if the communication relies on oneway
connections (e.g pipes). Seen from StunneL the backchannel name is simply a
text to be transported. Seen from the using applications, it may be important
or not.

Handshake string format of protocol 0.0 :
 -handshake:0.0 kl=16 cl=8 ml=1024:username:challenge:backchannel\n

The challenge size of protocol 0.0 is 8 bytes which are transmitted as
16 hex characters.

Handshake string format of protocol 0.2 :
 -handshake:0.2 kl=32 cl=32 ml=1024:username:challenge:backchannel:iv_seed:sha1_of_encrypted\n

With protocol 0.2 the challenge is 32 bytes long. Additionally 32 random bytes
of iv_seed are generated. Both peers will encrypt these 32 bytes by the
user's personal key with zero initial vector and condense the result to a
SHA-1 value of which the first 8 bytes are used as initial vector for
sblowfish's further chaining of the session key and the user's personal key.

The whole message up to iv_seed gets encrypted by the user's personal key.
This encrypted data is condensed to a SHA-1 and appended as sha1_of_encrypted
By that the initiating peer already authentifies with the handshake string.

The (unencrypted) handshake string has to be transmitted to the receiving peer
now.

The reply coming back from the peer is handed over to Stunnel_handshake_accept()
which will return 0 if the authentication failed, 1 if the StunneL pair is
ready for transmissions or 2 if more reply data are needed. Excess reply
data is marked for use by the application. So if the application protocol
allows the peer to send immediately, be sure to check that pointer.

See Sconn_try_tunnel() for an example.


Receiving Peer:

The receiving peer should call Stunnel_handshake_reply() which tries to obtain
the user's key from file. If the key is available, the challenge gets its reply
and a key for the current session is choosen. Both results are formatted as
tokens by the user's key. The session key will be used for all further
transmissions.
Also the backchannel name is retrieved from the handshake string. If the
application (e.g. SconN) does not have established a sending channel yet, it
can use this name for doing so.

With protocol 0.2 the first 8 bytes of the handshake reply are "OK v0.2\n"
if the handshake attempt was accepted. Any other reply is a clear text error
message and will be followed by connection termination.
Protocol 0.0 does not send such a confirmation text.

The two encrypted reply token have to be sent to the initiating peer now.

See Sconn_expect_tunnel() for an example.


After Handshake:

Now the handshake is complete. The receiving peer has authenticated via the
64-bit reply to the 64-bit challenge. The initiating peer is authenticated
if protocol 0.2 was used. 

With protocol 0.0 the initiating peer will authenticate by having understood
the session key. Without that key, it will not be able to format a valid token.

So if the receiving peer has to be sure about the authenticity of the initiating
peer, the application protocol should demand some initiating message before 
normal communications start. Such a message can be obtained by a call to
Stunnel_reauthenticate(). The resulting token will be a good authentication
test which unpacks to zero bytes in Stunnel_retrieve().
Stunnel_is_authenticated() tells wether a valid token already was received.
If Stunnel_reauthenticate() is used, one has to be aware that Stunnel_retrieve()
might return 2 without having an unfinished token. Use Stunnel_has_bytes()
to find out wether an unfished token is being processed if this information
is of interest for your application. 

If the receiving peer is a kind of server which only replies on request, then
a special initiating message is not necessary. The first request will do.
(Note again : this only applies to protocol 0.0)


Several Keys:

A StunneL has up to five SblowfisH encryption objects using different keys.
Also there is a global random generator which owns a static SblowfisH. 

The internal key is hardcoded into the program. It is used to read and write
keyfiles. 
This key is set by the macro Stunnel_internal_keY and should be altered if
it is possible to keep secret the programs that use StunneL. This adds a
great deal of security, if possible.

The random key is generated by Sblowfish__global_random() and used to generate
further random numbers. Such numbers are needed to generate keys and
challenges. Also they help to obscure sequence numbers with their high
redundancy (which is the enemy of all encryption).
See sblowfish.h for details on  Sblowfish__global_random() . Especially the
quality considerations.

The user's personal key is read from a file where it is stored encrypted by
the programs internal key. It is used for a short while during the handshake
until the session key is transmitted. Therefore it does not wear off quickly.

The session key is generated at random (hopefully) during handshake. As soon
as it is transmitted, all encryption is done with this key.

At the end of the handshake the initiating peer recomputes its random key
by use of the session key. This is to take advantage of the possibly better
random seed of the receiving peer.

With protocol version 0.2 there are two independent blowfishes with possibly
different keys for the two channel directions.
After about  Stunnel_session_key_lifetimE  packets the key of a direction 
gets changed by the sender.

*/

#endif /* ! Stunnel_includeD */

